// KRATOS  __  __ _____ ____  _   _ ___ _   _  ____
//        |  \/  | ____/ ___|| | | |_ _| \ | |/ ___|
//        | |\/| |  _| \___ \| |_| || ||  \| | |  _
//        | |  | | |___ ___) |  _  || || |\  | |_| |
//        |_|  |_|_____|____/|_| |_|___|_| \_|\____| APPLICATION
//
//  License:         BSD License
//                   Kratos default license: kratos/license.txt
//
//  Main authors:    Miguel Maso Sotomayor
//


// System includes


// External includes


// Project includes
#include "custom_processes/multiscale_refining_process.h"
#include "geometries/point.h"
#include "processes/fast_transfer_between_model_parts_process.h"
#include "utilities/assign_unique_model_part_collection_tag_utility.h"
#include "custom_includes/meshing_flags.h"

namespace Kratos
{

MultiscaleRefiningProcess::MultiscaleRefiningProcess(
    ModelPart& rThisCoarseModelPart,
    ModelPart& rThisRefinedModelPart,
    ModelPart& rThisVisualizationModelPart,
    Parameters ThisParameters)
    : mrCoarseModelPart(rThisCoarseModelPart)
    , mrRefinedModelPart(rThisRefinedModelPart)
    , mrVisualizationModelPart(rThisVisualizationModelPart)
    , mParameters(ThisParameters)
    , mUniformRefinement(mrRefinedModelPart)
{
    mParameters.ValidateAndAssignDefaults(GetDefaultParameters());

    mDivisionsAtSubscale = mParameters["number_of_divisions_at_subscale"].GetInt();
    mEchoLevel = mParameters["echo_level"].GetInt();

    std::string interface_base_name = mParameters["subscale_interface_base_name"].GetString();
    mRefinedInterfaceName = interface_base_name + "_" + std::to_string(mrCoarseModelPart.GetValue(SUBSCALE_INDEX) + 1);
    mInterfaceConditionName = mParameters["subscale_boundary_condition"].GetString();

    KRATOS_INFO_IF("MultiscaleRefiningProcess", mEchoLevel > 1) << "Parameters\n" << mParameters.Info() << std::endl;

    mStepDataSize = mrCoarseModelPart.GetNodalSolutionStepDataSize();

    Check();

    // Initialize the coarse model part
    InitializeCoarseModelPartInterface();

    // Initialize the refined model part
    InitializeRefinedModelPartInterface();

    // Copy all the entities to the visualization model part
    // InitializeVisualizationModelPart(sub_model_parts_names);
}


int MultiscaleRefiningProcess::Check()
{
    KRATOS_TRY

    KRATOS_CHECK(KratosComponents<Condition>::Has(mInterfaceConditionName));

    KRATOS_CHECK_NOT_EQUAL(mDivisionsAtSubscale, 0);

    KRATOS_CHECK_EQUAL(mStepDataSize, mrRefinedModelPart.GetNodalSolutionStepDataSize());
    KRATOS_CHECK_EQUAL(mStepDataSize, mrVisualizationModelPart.GetNodalSolutionStepDataSize());

    return 0;

    KRATOS_CATCH("")
}


void MultiscaleRefiningProcess::ExecuteRefinement()
{
    // Initialize the maps
    IndexIndexMapType node_tag, elem_tag, cond_tag;
    AssignUniqueModelPartCollectionTagUtility model_part_collection(mrCoarseModelPart);
    model_part_collection.ComputeTags(node_tag, cond_tag, elem_tag, mCollections);

    // Get the Id's
    IndexType node_id, elem_id, cond_id;
    GetLastId(node_id, elem_id, cond_id);

    // Clone the nodes and set the nodal flags
    CloneNodesToRefine(node_id);

    // Set the elements and conditions flags
    MarkElementsFromNodalFlag();
    MarkConditionsFromNodalFlag();

    // Create the auxiliary entities
    CreateElementsToRefine(elem_id, elem_tag);
    CreateConditionsToRefine(cond_id, cond_tag);

    // Check and prepare the interface
    IdentifyCurrentInterface();

    // Execute the refinement
    int divisions = mrRefinedModelPart.GetValue(SUBSCALE_INDEX) * mDivisionsAtSubscale;
    mUniformRefinement.SetCustomIds(node_id, elem_id, cond_id);
    mUniformRefinement.Refine(divisions);
    mUniformRefinement.GetLastCreatedIds(node_id, elem_id, cond_id);

    // Update the refined interface after creation
    UpdateRefinedInterface();

    // Update the visualization model part
    UpdateVisualizationAfterRefinement();

    // Reset the flags
    FinalizeRefinement();
}


void MultiscaleRefiningProcess::ExecuteCoarsening()
{
    IdentifyParentNodesToCoarsen();
    IdentifyElementsToErase();
    IdentifyConditionsToErase();
    IdentifyRefinedNodesToErase();

    mUniformRefinement.RemoveRefinedEntities(TO_ERASE);

    // Update the visualization model part
    IdentifyCurrentInterface();
    UpdateVisualizationAfterCoarsening();

    FinalizeCoarsening();
}


void MultiscaleRefiningProcess::InitializeVisualizationModelPart(ModelPart& rReferenceModelPart, ModelPart& rNewModelPart)
{
    // Create the sub model parts and add properties and tables
    InitializeNewModelPart(rReferenceModelPart, rNewModelPart);

    // Adding the variables
    VariablesList& variables_list = rNewModelPart.GetNodalSolutionStepVariablesList();
    variables_list = rReferenceModelPart.GetNodalSolutionStepVariablesList();

    // Add the entities to the root model part
    FastTransferBetweenModelPartsProcess(rNewModelPart, rReferenceModelPart)();

    // Get the model part hierarchy
    StringVectorType sub_model_parts_names;
    sub_model_parts_names = rReferenceModelPart.GetSubModelPartNames();

    // Add the entities to the submodel parts
    for (auto name : sub_model_parts_names)
    {
        ModelPart& destination = rNewModelPart.GetSubModelPart(name);
        ModelPart& origin = rReferenceModelPart.GetSubModelPart(name);
        FastTransferBetweenModelPartsProcess(destination, origin)();
    }
}


void MultiscaleRefiningProcess::InitializeRefinedModelPart(ModelPart& rReferenceModelPart, ModelPart& rNewModelPart)
{
    // Create the sub model parts and add properties and tables
    InitializeNewModelPart(rReferenceModelPart, rNewModelPart);

    // Increase the refinement level
    int subscale_index = rReferenceModelPart.GetValue(SUBSCALE_INDEX);
    rNewModelPart.SetValue(SUBSCALE_INDEX, ++subscale_index);
}


void MultiscaleRefiningProcess::InitializeNewModelPart(ModelPart& rReferenceModelPart, ModelPart& rNewModelPart)
{
    // Copy all the tables and properties
    AddAllTablesToModelPart(rReferenceModelPart, rNewModelPart);
    AddAllPropertiesToModelPart(rReferenceModelPart, rNewModelPart);

    // Copy the ProcessInfo
    ProcessInfo& process_info = rNewModelPart.GetProcessInfo();
    process_info = rReferenceModelPart.GetProcessInfo();

    // Get the model part hierarchy
    StringVectorType sub_model_parts_names;
    sub_model_parts_names = rReferenceModelPart.GetSubModelPartNames();

    // Copy the hierarchy to the refined model part
    for (auto name : sub_model_parts_names)
    {
        ModelPart& sub_model_part = rNewModelPart.CreateSubModelPart(name);

        ModelPart& origin_model_part = rReferenceModelPart.GetSubModelPart(name);

        // Copy all the tables and properties
        AddAllTablesToModelPart(origin_model_part, sub_model_part);
        AddAllPropertiesToModelPart(origin_model_part, sub_model_part);
    }
}


void MultiscaleRefiningProcess::InitializeCoarseModelPartInterface()
{
    // Create a model part to store the interface boundary conditions
    if (mrCoarseModelPart.HasSubModelPart(mRefinedInterfaceName))
    {
        mrCoarseModelPart.GetSubModelPart(mRefinedInterfaceName).RemoveNodesFromAllLevels();
        mrCoarseModelPart.GetSubModelPart(mRefinedInterfaceName).RemoveElementsFromAllLevels();
        mrCoarseModelPart.GetSubModelPart(mRefinedInterfaceName).RemoveConditionsFromAllLevels();
    }
    else
        mrCoarseModelPart.CreateSubModelPart(mRefinedInterfaceName);
}


void MultiscaleRefiningProcess::InitializeRefinedModelPartInterface()
{
    // Create a model part to store the interface boundary conditions
    if (mrRefinedModelPart.HasSubModelPart(mRefinedInterfaceName))
    {
        mrRefinedModelPart.GetSubModelPart(mRefinedInterfaceName).RemoveNodesFromAllLevels();
        mrRefinedModelPart.GetSubModelPart(mRefinedInterfaceName).RemoveElementsFromAllLevels();
        mrRefinedModelPart.GetSubModelPart(mRefinedInterfaceName).RemoveConditionsFromAllLevels();
    }
    else
        mrRefinedModelPart.CreateSubModelPart(mRefinedInterfaceName);
}


void MultiscaleRefiningProcess::UpdateVisualizationAfterRefinement()
{
    // Remove the refined elements and conditions to substitute them by the refined ones
    mrVisualizationModelPart.RemoveElementsFromAllLevels(MeshingFlags::REFINED);
    mrVisualizationModelPart.RemoveConditionsFromAllLevels(MeshingFlags::REFINED);

    // Remove the refined nodes which are not interface
    ModelPart::NodesContainerType::iterator nodes_begin = mrCoarseModelPart.NodesBegin();
    #pragma omp parallel for
    for (int i = 0; i < static_cast<int>(mrCoarseModelPart.Nodes().size()); i++)
    {
        auto coarse_node = nodes_begin + i;
        if (coarse_node->Is((MeshingFlags::REFINED)) && (coarse_node->IsNot(INTERFACE)))
            coarse_node->Set(INSIDE, true);
        else
            coarse_node->Set(INSIDE, false);
    }
    // mrVisualizationModelPart.RemoveNodesFromAllLevels(INSIDE); // TODO: for some reason it is modifying the refined model part...

    // Add the new entities which are refined
    FastTransferBetweenModelPartsProcess(mrVisualizationModelPart, mrRefinedModelPart,
        FastTransferBetweenModelPartsProcess::EntityTransfered::ALL, NEW_ENTITY)();
}


void MultiscaleRefiningProcess::UpdateVisualizationAfterCoarsening()
{
    // Remove the coarsened entities
    mrVisualizationModelPart.RemoveNodesFromAllLevels(TO_ERASE);
    mrVisualizationModelPart.RemoveElementsFromAllLevels(TO_ERASE);
    mrVisualizationModelPart.RemoveConditionsFromAllLevels(TO_ERASE);

    // Add the origin entities which are coarsened
    FastTransferBetweenModelPartsProcess(mrVisualizationModelPart, mrCoarseModelPart,
        FastTransferBetweenModelPartsProcess::EntityTransfered::ALL, MeshingFlags::TO_COARSEN)();
    FastTransferBetweenModelPartsProcess(mrVisualizationModelPart, mrCoarseModelPart,
        FastTransferBetweenModelPartsProcess::EntityTransfered::NODES, INTERFACE)();
}


void MultiscaleRefiningProcess::AddAllPropertiesToModelPart(ModelPart& rOriginModelPart, ModelPart& rDestinationModelPart)
{
    const IndexType nprop = rOriginModelPart.NumberOfProperties();
    ModelPart::PropertiesContainerType::iterator prop_begin = rOriginModelPart.PropertiesBegin();

    for (IndexType i = 0; i < nprop; i++)
    {
        auto prop = prop_begin + i;
        rDestinationModelPart.AddProperties(*prop.base());
    }
}


void MultiscaleRefiningProcess::AddAllTablesToModelPart(ModelPart& rOriginModelPart, ModelPart& rDestinationModelPart)
{
    const IndexType ntables = rOriginModelPart.NumberOfTables();
    ModelPart::TablesContainerType::iterator table_begin = rOriginModelPart.TablesBegin();

    for (IndexType i = 0; i < ntables; i++)
    {
        auto table = table_begin + i;
        rDestinationModelPart.AddTable(table.base()->first, table.base()->second);
    }
}


void MultiscaleRefiningProcess::MarkElementsFromNodalFlag()
{
    const int nelems = static_cast<int>(mrCoarseModelPart.Elements().size());
    ModelPart::ElementsContainerType::iterator elem_begin = mrCoarseModelPart.ElementsBegin();

    // We assume all the elements have the same number of nodes
    const IndexType number_of_nodes = elem_begin->GetGeometry().size();

    // We will refine the elements which:
    // 1. all the nodes are to refine
    // 2. at least, one node is marked as new entity
    // NEW_ENTITY flag is used to avoid duplication of refined entities on previous steps
    #pragma omp parallel for
    for (int i = 0; i < nelems; i++)
    {
        auto elem = elem_begin + i;
        bool to_refine = true;
        bool new_entity = false;
        for (IndexType node = 0; node < number_of_nodes; node++)
        {
            if (elem->GetGeometry()[node].IsNot(TO_REFINE))
                to_refine = false;

            if (elem->GetGeometry()[node].Is(NEW_ENTITY))
                new_entity = true;
        }
        elem->Set(TO_REFINE, (to_refine && new_entity));
    }
}


void MultiscaleRefiningProcess::MarkConditionsFromNodalFlag()
{
    const int nconds = static_cast<int>(mrCoarseModelPart.Conditions().size());
    ModelPart::ConditionsContainerType::iterator cond_begin = mrCoarseModelPart.ConditionsBegin();

    // We assume all the conditions have the same number of nodes
    const IndexType number_of_nodes = cond_begin->GetGeometry().size();

    // We will refine the conditions which:
    // 1. all the nodes are to refine
    // 2. at least, one node is marked as new entity
    // NEW_ENTITY flag is used to avoid duplication of refined entities on previous steps
    #pragma omp parallel for
    for (int i = 0; i < nconds; i++)
    {
        auto cond = cond_begin + i;
        bool to_refine = true;
        bool new_entity = false;
        for (IndexType node = 0; node < number_of_nodes; node++)
        {
            if (cond->GetGeometry()[node].IsNot(TO_REFINE))
                to_refine = false;

            if (cond->GetGeometry()[node].Is(NEW_ENTITY))
                new_entity = true;
        }
        cond->Set(TO_REFINE, (to_refine && new_entity));
    }
}


void MultiscaleRefiningProcess::CloneNodesToRefine(IndexType& rNodeId)
{
    const int nnodes = static_cast<int>(mrCoarseModelPart.Nodes().size());
    ModelPart::NodesContainerType::iterator nodes_begin = mrCoarseModelPart.NodesBegin();

    // Adding the nodes to the refined model part
    for (int i = 0; i < nnodes; i++)
    {
        auto coarse_node = nodes_begin + i;
        if (coarse_node->Is(TO_REFINE))
        {
            if (coarse_node->IsNot(MeshingFlags::REFINED))
            {
                coarse_node->Set(NEW_ENTITY, true);
                NodeType::Pointer new_node = coarse_node->Clone();
                new_node->SetId(++rNodeId);
                mrRefinedModelPart.AddNode(new_node);
                new_node->Set(TO_REFINE, false);
                new_node->GetValue(FATHER_NODES).resize(0);
                new_node->GetValue(FATHER_NODES).push_back( NodeType::WeakPointer(*coarse_node.base()) );
                new_node->GetValue(FATHER_NODES_WEIGHTS).resize(0);
                new_node->GetValue(FATHER_NODES_WEIGHTS).push_back(1.0);
                coarse_node->SetValue(SLAVE_NODE, new_node);
                coarse_node->Set(MeshingFlags::REFINED, true);
            }
        }
    }

    // Adding the nodes to the refined sub model parts
    StringVectorType sub_model_part_names = mrCoarseModelPart.GetSubModelPartNames();
    for (auto name : sub_model_part_names)
    {
        ModelPart& coarse_sub_model_part = mrCoarseModelPart.GetSubModelPart(name);
        ModelPart& refined_sub_model_part = mrRefinedModelPart.GetSubModelPart(name);

        const int nnodes = static_cast<int>(coarse_sub_model_part.Nodes().size());
        ModelPart::NodesContainerType::iterator nodes_begin = coarse_sub_model_part.NodesBegin();

        for (int i = 0; i < nnodes; i++)
        {
            auto coarse_node = nodes_begin + i;
            if (coarse_node->Is(NEW_ENTITY))
                refined_sub_model_part.AddNode(coarse_node->GetValue(SLAVE_NODE));
        }
    }
}


void MultiscaleRefiningProcess::IdentifyParentNodesToCoarsen()
{
    const int nnodes = static_cast<int>(mrCoarseModelPart.Nodes().size());
    ModelPart::NodesContainerType::iterator nodes_begin = mrCoarseModelPart.NodesBegin();

    // Identify the nodes to remove
    for (int i = 0; i < nnodes; i++)
    {
        auto coarse_node = nodes_begin + i;
        if (coarse_node->IsNot(TO_REFINE))
        {
            if (coarse_node->Is(MeshingFlags::REFINED))
            {
                // We need to ensure the refined mesh does not has dependencies
                if (coarse_node->GetValue(SLAVE_NODE)->IsNot(MeshingFlags::REFINED))
                {
                    coarse_node->Set(MeshingFlags::TO_COARSEN, true);
                    coarse_node->Set(MeshingFlags::REFINED, false);
                    coarse_node->SetValue(SLAVE_NODE, nullptr);
                }
            }
        }
    }
}


void MultiscaleRefiningProcess::IdentifyElementsToErase()
{
    // Identify the parent elements to coarse
    const int nelems_coarse = static_cast<int>(mrCoarseModelPart.Elements().size());
    ModelPart::ElementsContainerType::iterator coarse_elem_begin = mrCoarseModelPart.ElementsBegin();

    // The number of nodes of the elements
    const IndexType element_nodes = coarse_elem_begin->GetGeometry().size();

    #pragma omp parallel for
    for (int i = 0; i < nelems_coarse; i++)
    {
        auto coarse_elem = coarse_elem_begin + i;
        if (coarse_elem->Is(MeshingFlags::REFINED))
        {
            bool to_coarsen = false;
            for (IndexType inode = 0; inode < element_nodes; inode++)
            {
                if (coarse_elem->GetGeometry()[inode].Is(MeshingFlags::TO_COARSEN))
                    to_coarsen = true;
            }
            coarse_elem->Set(MeshingFlags::TO_COARSEN, to_coarsen);
            if (to_coarsen)
                coarse_elem->Set(MeshingFlags::REFINED, false);
        }
    }

    // Identify the refined elements to remove
    const int nelems_ref = static_cast<int>(mrRefinedModelPart.Elements().size());
    ModelPart::ElementsContainerType::iterator refined_elem_begin = mrRefinedModelPart.ElementsBegin();

    #pragma omp parallel for
    for (int i = 0; i < nelems_ref; i++)
    {
        auto refined_elem = refined_elem_begin + i;
        if ((refined_elem->GetValue(FATHER_ELEMENT))->Is(MeshingFlags::TO_COARSEN))
            refined_elem->Set(TO_ERASE, true);
    }
}


void MultiscaleRefiningProcess::IdentifyConditionsToErase()
{
    // Identify the parent conditions  to coarse
    const int nconds_coarse = static_cast<int>(mrCoarseModelPart.Conditions().size());
    ModelPart::ConditionsContainerType::iterator coarse_cond_begin = mrCoarseModelPart.ConditionsBegin();

    // The number of nodes of the conditions
    const IndexType condition_nodes = coarse_cond_begin->GetGeometry().size();

    #pragma omp parallel for
    for (int i = 0; i < nconds_coarse; i++)
    {
        auto coarse_cond = coarse_cond_begin + i;
        if (coarse_cond->Is(MeshingFlags::REFINED))
        {
            bool to_coarsen = false;
            for (IndexType inode = 0; inode < condition_nodes; inode++)
            {
                if (coarse_cond->GetGeometry()[inode].Is(MeshingFlags::TO_COARSEN))
                    to_coarsen = true;
            }
            coarse_cond->Set(MeshingFlags::TO_COARSEN, to_coarsen);
            if (to_coarsen)
                coarse_cond->Set(MeshingFlags::REFINED, false);
        }
    }

    // Identify the refined conditions to remove
    const int nconds_ref = static_cast<int>(mrRefinedModelPart.Conditions().size());
    ModelPart::ConditionsContainerType::iterator refined_cond_begin = mrRefinedModelPart.ConditionsBegin();

    #pragma omp parallel for
    for (int i = 0; i < nconds_ref; i++)
    {
        auto refined_cond = refined_cond_begin + i;
        if ((refined_cond->GetValue(FATHER_CONDITION))->Is(MeshingFlags::TO_COARSEN))
            refined_cond->Set(TO_ERASE, true);
    }
}


void MultiscaleRefiningProcess::IdentifyRefinedNodesToErase()
{
    const IndexType nelems = mrRefinedModelPart.Elements().size();
    if (nelems != 0) // just avoiding segfault in case of an empty coarse model part
    {
        ModelPart::NodeIterator nodes_begin = mrRefinedModelPart.NodesBegin();
        #pragma omp parallel for
        for (int i = 0; i < static_cast<int>(mrRefinedModelPart.Nodes().size()); i++)
        {
            auto node = nodes_begin + i;
            node->Set(TO_ERASE, true);
        }

        ModelPart::ElementIterator elements_begin = mrRefinedModelPart.ElementsBegin();
        const IndexType element_nodes = elements_begin->GetGeometry().size();

        for (IndexType i = 0; i < nelems; i++)
        {
            auto elem = elements_begin + i;
            if (elem->IsNot(TO_ERASE))
            {
                for (IndexType inode = 0; inode < element_nodes; inode++)
                {
                    (elem->GetGeometry()[inode]).Set(TO_ERASE, false);
                }
            }
        }
    }
}


void MultiscaleRefiningProcess::CreateElementsToRefine(IndexType& rElemId, IndexIndexMapType& rElemTag)
{
    const int nelems = static_cast<int>(mrCoarseModelPart.Elements().size());
    ModelPart::ElementsContainerType::iterator elements_begin = mrCoarseModelPart.ElementsBegin();

    // We assume all the elements have the same number of nodes
    const IndexType number_of_nodes = elements_begin->GetGeometry().size();

    // The map to add the elements to the sub model parts
    IndexVectorMapType tag_elems_map;

    for (int i = 0; i < nelems; i++)
    {
        auto coarse_elem = elements_begin + i;
        if (coarse_elem->Is(TO_REFINE))
        {
            Geometry<NodeType>::PointsArrayType p_elem_nodes;
            for (IndexType node = 0; node < number_of_nodes; node++)
                p_elem_nodes.push_back(coarse_elem->GetGeometry()[node].GetValue(SLAVE_NODE));

            Element::Pointer aux_elem = coarse_elem->Clone(++rElemId, p_elem_nodes);
            mrRefinedModelPart.AddElement(aux_elem);

            aux_elem->SetValue(FATHER_ELEMENT, *coarse_elem.base());
            aux_elem->Set(NEW_ENTITY, true);

            IndexType tag = rElemTag[coarse_elem->Id()];
            tag_elems_map[tag].push_back(rElemId);

            coarse_elem->Set(MeshingFlags::REFINED, true);
        }
    }

    // Loop the sub model parts and add the new elements to it
    for (auto& collection : mCollections)
    {
        const auto tag = collection.first;
        if (tag != 0)
        {
            for (auto name : collection.second)
            {
                ModelPart& sub_model_part = mrRefinedModelPart.GetSubModelPart(name);
                sub_model_part.AddElements(tag_elems_map[tag]);
            }
        }
    }
}


void MultiscaleRefiningProcess::CreateConditionsToRefine(IndexType& rCondId, IndexIndexMapType& rCondTag)
{
    const int nconds = static_cast<int>(mrCoarseModelPart.Conditions().size());
    ModelPart::ConditionsContainerType::iterator conditions_begin = mrCoarseModelPart.ConditionsBegin();

    // We assume all the conditions have the same number of nodes
    const IndexType number_of_nodes = conditions_begin->GetGeometry().size();

    // The map to add the conditions to the sub model parts
    IndexVectorMapType tag_conds_map;

    // #pragma omp parallel for
    for (int i = 0; i < nconds; i++)
    {
        auto coarse_cond = conditions_begin + i;
        if (coarse_cond->Is(TO_REFINE))
        {
            Geometry<NodeType>::PointsArrayType p_cond_nodes;
            for (IndexType node = 0; node < number_of_nodes; node++)
                p_cond_nodes.push_back(coarse_cond->GetGeometry()[node].GetValue(SLAVE_NODE));

            Condition::Pointer aux_cond = coarse_cond->Clone(++rCondId, p_cond_nodes);
            mrRefinedModelPart.AddCondition(aux_cond);

            aux_cond->SetValue(FATHER_CONDITION, *coarse_cond.base());
            aux_cond->Set(NEW_ENTITY, true);

            IndexType tag = rCondTag[coarse_cond->Id()];
            tag_conds_map[tag].push_back(rCondId);

            coarse_cond->Set(MeshingFlags::REFINED, true);
        }
    }

    // Loop the sub model parts and add the new conditions to it
    for (auto& collection : mCollections)
    {
        const auto tag = collection.first;
        if (tag != 0)
        {
            for (auto name : collection.second)
            {
                ModelPart& sub_model_part = mrRefinedModelPart.GetSubModelPart(name);
                sub_model_part.AddConditions(tag_conds_map[tag]);
            }
        }
    }
}


void MultiscaleRefiningProcess::FinalizeRefinement()
{
    /// Coarse model part
    // Resetting the nodes flags
    ModelPart::NodeIterator coarse_nodes_begin = mrCoarseModelPart.NodesBegin();
    #pragma omp parallel for
    for (int i = 0; i < static_cast<int>(mrCoarseModelPart.Nodes().size()); i++)
    {
        auto node = coarse_nodes_begin + i;
        node->Set(NEW_ENTITY, false);
    }

    /// Refined model part
    // Resetting the nodes flags
    ModelPart::NodeIterator refined_nodes_begin = mrRefinedModelPart.NodesBegin();
    #pragma omp parallel for
    for (int i = 0; i < static_cast<int>(mrRefinedModelPart.Nodes().size()); i++)
    {
        auto node = refined_nodes_begin + i;
        node->Set(NEW_ENTITY, false);
    }

    // Resetting the elements flags
    ModelPart::ElementIterator refined_elem_begin = mrRefinedModelPart.ElementsBegin();
    #pragma omp parallel for
    for (int i = 0; i < static_cast<int>(mrRefinedModelPart.Elements().size()); i++)
    {
        auto elem = refined_elem_begin + i;
        elem->Set(NEW_ENTITY, false);
    }

    // Resetting the conditions flags
    ModelPart::ConditionIterator refined_cond_begin = mrRefinedModelPart.ConditionsBegin();
    #pragma omp parallel for
    for (int i = 0; i < static_cast<int>(mrRefinedModelPart.Conditions().size()); i++)
    {
        auto cond = refined_cond_begin + i;
        cond->Set(NEW_ENTITY, false);
    }
}

void MultiscaleRefiningProcess::FinalizeCoarsening()
{
    /// Coarse model part
    // Resetting the nodes flags
    ModelPart::NodeIterator nodes_begin = mrCoarseModelPart.NodesBegin();
    #pragma omp parallel for
    for (int i = 0; i < static_cast<int>(mrCoarseModelPart.Nodes().size()); i++)
    {
        auto node = nodes_begin + i;
        node->Set(MeshingFlags::TO_COARSEN, false);
    }

    // Resetting the elements flags
    ModelPart::ElementIterator elements_begin = mrCoarseModelPart.ElementsBegin();
    #pragma omp parallel for
    for (int i = 0; i < static_cast<int>(mrCoarseModelPart.Elements().size()); i++)
    {
        auto elem = elements_begin + i;
        elem->Set(MeshingFlags::TO_COARSEN, false);
    }

    // Resetting the conditions flags
    ModelPart::ConditionIterator conditions_begin = mrCoarseModelPart.ConditionsBegin();
    #pragma omp parallel for
    for (int i = 0; i < static_cast<int>(mrCoarseModelPart.Conditions().size()); i++)
    {
        auto cond = conditions_begin + i;
        cond->Set(MeshingFlags::TO_COARSEN, false);
    }
}


void MultiscaleRefiningProcess::IdentifyCurrentInterface()
{
    // Resetting the flag
    int nnodes = static_cast<int>(mrCoarseModelPart.Nodes().size());
    ModelPart::NodesContainerType::iterator nodes_begin = mrCoarseModelPart.NodesBegin();

    #pragma omp parallel for
    for (int i = 0; i < nnodes; i++)
    {
        auto node = nodes_begin + i;
        node->Set(INTERFACE, false);
    }

    // The number of nodes of the elements
    ModelPart::ElementIterator elem_begin = mrCoarseModelPart.ElementsBegin();
    const IndexType element_nodes = elem_begin->GetGeometry().size();

    // Identify the current interface:
    // Look for the elements which are not to refine and have some nodes to refine
    for (int i = 0; i < static_cast<int>(mrCoarseModelPart.Elements().size()); i++)
    {
        auto elem = elem_begin + i;
        if (elem->IsNot(MeshingFlags::REFINED))
        {
            // set the nodal flags
            for (IndexType node = 0; node < element_nodes; node++)
            {
                if (elem->GetGeometry()[node].Is(MeshingFlags::REFINED))
                    elem->GetGeometry()[node].Set(INTERFACE, true);
            }
        }
    }
}


void MultiscaleRefiningProcess::UpdateRefinedInterface()
{
    mRefinedInterfaceContainer.clear();

    ModelPart::NodeIterator refined_begin = mrRefinedModelPart.NodesBegin();
    for (int i = 0; i < static_cast<int>(mrRefinedModelPart.Nodes().size()); i++)
    {
        auto node = refined_begin + i;
        bool is_refined_interface = true;
        GlobalPointersVector<NodeType>& father_nodes = node->GetValue(FATHER_NODES);
        for (auto father_node = father_nodes.begin(); father_node < father_nodes.end(); father_node++)
        {
            if (father_node->IsNot(INTERFACE))
                is_refined_interface = false;
        }

        if (is_refined_interface)
            mRefinedInterfaceContainer.push_back(Kratos::intrusive_ptr<NodeType>(*node.base()));
    }
}


void MultiscaleRefiningProcess::GetLastId(
    IndexType& rNodesId,
    IndexType& rElemsId,
    IndexType& rCondsId)
{
    // Initialize the output
    rNodesId = 0;
    rElemsId = 0;
    rCondsId = 0;

    // Get the absolute root model part
    ModelPart& root_model_part = mrVisualizationModelPart.GetRootModelPart();

    // Get the maximum node id
    const IndexType nnodes = root_model_part.Nodes().size();
    ModelPart::NodesContainerType::iterator nodes_begin = root_model_part.NodesBegin();
    for (IndexType i = 0; i < nnodes; i++)
    {
        auto inode = nodes_begin + i;
        if (rNodesId < inode->Id())
            rNodesId = inode->Id();
    }

    // Get the maximum element id
    const IndexType nelems = root_model_part.Elements().size();
    ModelPart::ElementsContainerType::iterator elements_begin = root_model_part.ElementsBegin();
    for (IndexType i = 0; i < nelems; i++)
    {
        auto elem = elements_begin + i;
        if (rElemsId < elem->Id())
            rElemsId = elem->Id();
    }

    // Get the maximum condition id
    const IndexType nconds = root_model_part.Conditions().size();
    ModelPart::ConditionsContainerType::iterator conditions_begin = root_model_part.ConditionsBegin();
    for (IndexType i = 0; i < nconds; i++)
    {
        auto cond = conditions_begin + i;
        if (rCondsId < cond->Id())
            rCondsId = cond->Id();
    }
}

const Parameters MultiscaleRefiningProcess::GetDefaultParameters() const
{
    const Parameters default_parameters = Parameters(R"(
    {
        "number_of_divisions_at_subscale"     : 2,
        "echo_level"                          : 0,
        "subscale_interface_base_name"        : "refined_interface",
        "subscale_boundary_condition"         : "LineCondition2D2N"
    })" );

    return default_parameters;
}

} // namespace Kratos
