//
//   Project Name:        KratosPfemApplication    $
//   Created by:          $Author:     JMCarbonell $
//   Last modified by:    $Co-Author:              $
//   Date:                $Date:          May 2018 $
//   Revision:            $Revision:           0.0 $
//
//

#if !defined(KRATOS_FLUID_ELEMENT_H_INCLUDED)
#define  KRATOS_FLUID_ELEMENT_H_INCLUDED

// System includes

// External includes

// Project includes
#include "includes/checks.h"
#include "includes/element.h"

namespace Kratos
{
///@name Kratos Globals
///@{
///@}
///@name Type Definitions
///@{
///@}
///@name  Enum's
///@{
///@}
///@name  Functions
///@{
///@}
///@name Kratos Classes
///@{

/// Large Displacement Lagrangian Element for 3D and 2D geometries. (base class)

/**
 * Implements a Large Displacement Lagrangian definition for structural analysis.
 * This works for arbitrary geometries in 3D and 2D (base class)
 */

class KRATOS_API(PFEM_APPLICATION) FluidElement
    : public Element
{
public:

    ///@name Type Definitions
    ///@{
    ///Reference type definition for constitutive laws
    typedef ConstitutiveLaw ConstitutiveLawType;
    ///Pointer type for constitutive laws
    typedef ConstitutiveLawType::Pointer ConstitutiveLawPointerType;
    ///StressMeasure from constitutive laws
    typedef ConstitutiveLawType::StressMeasure StressMeasureType;
    ///Type definition for integration methods
    typedef GeometryData::IntegrationMethod IntegrationMethod;
    ///Type for size
    typedef GeometryData::SizeType SizeType;

    /// Counted pointer of FluidElement
    KRATOS_CLASS_INTRUSIVE_POINTER_DEFINITION( FluidElement );
    ///@}

protected:

    /**
     * Flags related to the element computation
     */

    KRATOS_DEFINE_LOCAL_FLAG( COMPUTE_RHS_VECTOR );
    KRATOS_DEFINE_LOCAL_FLAG( COMPUTE_LHS_MATRIX );
    KRATOS_DEFINE_LOCAL_FLAG( FINALIZED_STEP );

    /**
     * Parameters to be used in the Element as they are. Direct interface to Parameters Struct
     */

    struct ElementData
    {
      private:

        //variables including all integration points
        const GeometryType::ShapeFunctionsGradientsType* pDN_De;
        const Matrix* pNcontainer;
        const ProcessInfo* pProcessInfo;

      public:

        StressMeasureType StressMeasure;

        double  Alpha;
        double  Tau;
        double  IntegrationWeight;

        //general variables for large displacement use
        double  detF;
        double  detJ;

        Vector  StrainVector;
        Vector  StressVector;
        Vector  N;

        Matrix  B;
        Matrix  L;    //Velocity Gradient
        Matrix  F;    //Incremental Deformation Gradient (n to n+1)

        Matrix  DN_DX;
        Matrix  ConstitutiveMatrix;

        //variables including all integration points
        GeometryType::JacobiansType J;
        GeometryType::JacobiansType j;
        Matrix  DeltaPosition;


        /**
         * sets the value of a specified pointer variable
         */
        void SetShapeFunctionsGradients(const GeometryType::ShapeFunctionsGradientsType &rDN_De)
        {
            pDN_De=&rDN_De;
        };

        void SetShapeFunctions(const Matrix& rNcontainer)
        {
            pNcontainer=&rNcontainer;
        };

        void SetProcessInfo(const ProcessInfo& rProcessInfo)
        {
            pProcessInfo=&rProcessInfo;
        };


        /**
         * returns the value of a specified pointer variable
         */
        const GeometryType::ShapeFunctionsGradientsType& GetShapeFunctionsGradients()
        {
            return *pDN_De;
        };

        const Matrix& GetShapeFunctions()
        {
            return *pNcontainer;
        };

        const ProcessInfo& GetProcessInfo()
        {
            return *pProcessInfo;
        };

        void Initialize( const unsigned int& voigt_size,
			 const unsigned int& dimension,
			 const unsigned int& number_of_nodes )
        {
	  StressMeasure = ConstitutiveLaw::StressMeasure_PK2;

          //integration
          Alpha = 1.0;

          //stabilization
          Tau = 0;

          //time step
          IntegrationWeight = 1;

	  //jacobians
	  detF  = 1;
	  detJ  = 1;

	  //vectors
	  StrainVector.resize(voigt_size,false);
          StressVector.resize(voigt_size,false);
	  N.resize(number_of_nodes,false);

	  //matrices
	  B.resize(voigt_size,dimension*number_of_nodes,false);
	  L.resize(dimension,dimension,false);
	  F.resize(dimension,dimension,false);
	  DN_DX.resize(number_of_nodes,dimension,false);
	  ConstitutiveMatrix.resize(voigt_size,voigt_size,false);
	  DeltaPosition.resize(number_of_nodes,dimension,false);

          //others
	  J.resize(1,false);
	  j.resize(1,false);
	  J[0].resize(dimension,dimension,false);
	  j[0].resize(dimension,dimension,false);

          //pointers
	  pDN_De = NULL;
	  pNcontainer = NULL;
          pProcessInfo = NULL;

          //fill containters
          // noalias(StrainVector) = ZeroVector(voigt_size);
	  // noalias(StressVector) = ZeroVector(voigt_size);
	  // noalias(N) = ZeroVector(number_of_nodes);

	  // noalias(B)  = ZeroMatrix(voigt_size, dimension*number_of_nodes);
	  // noalias(L)  = ZeroMatrix(dimension,dimension);
	  // noalias(F)  = IdentityMatrix(dimension,dimension);

	  // noalias(DN_DX) = ZeroMatrix(number_of_nodes, dimension);
	  // noalias(ConstitutiveMatrix) = ZeroMatrix(voigt_size, voigt_size);
	  // noalias(DeltaPosition) = ZeroMatrix(number_of_nodes, dimension);

          // noalias(J[0]) = ZeroMatrix(dimension,dimension);
	  // noalias(j[0]) = ZeroMatrix(dimension,dimension);

	}

    };


    /**
     * This struct is used in the component wise calculation only
     * is defined here and is used to declare a member variable in the component wise elements
     * private pointers can only be accessed by means of set and get functions
     * this allows to set and not copy the local system variables
     */

    struct LocalSystemComponents
    {
    private:

      //for calculation local system with compacted LHS and RHS
      MatrixType *mpLeftHandSideMatrix;
      VectorType *mpRightHandSideVector;

    public:

      //calculation flags
      Flags  CalculationFlags;

      /**
       * sets the value of a specified pointer variable
       */
      void SetLeftHandSideMatrix( MatrixType& rLeftHandSideMatrix ) { mpLeftHandSideMatrix = &rLeftHandSideMatrix; };

      void SetRightHandSideVector( VectorType& rRightHandSideVector ) { mpRightHandSideVector = &rRightHandSideVector; };


      /**
       * returns the value of a specified pointer variable
       */
      MatrixType& GetLeftHandSideMatrix() { return *mpLeftHandSideMatrix; };

      VectorType& GetRightHandSideVector() { return *mpRightHandSideVector; };

    };


public:

    ///Type for element variables
    typedef ElementData ElementDataType;

    ///@name Life Cycle
    ///@{

    /// Empty constructor needed for serialization
    FluidElement();

    /// Default constructors
    FluidElement(IndexType NewId, GeometryType::Pointer pGeometry);

    FluidElement(IndexType NewId, GeometryType::Pointer pGeometry, PropertiesType::Pointer pProperties);

    ///Copy constructor
    FluidElement(FluidElement const& rOther);

    /// Destructor.
    virtual ~FluidElement();

    ///@}
    ///@name Operators
    ///@{

    /// Assignment operator.
    FluidElement& operator=(FluidElement const& rOther);

    ///@}
    ///@name Operations
    ///@{

    /**
     * creates a new element pointer
     * @param NewId: the ID of the new element
     * @param ThisNodes: the nodes of the new element
     * @param pProperties: the properties assigned to the new element
     * @return a Pointer to the new element
     */
    Element::Pointer Create(IndexType NewId, NodesArrayType const& ThisNodes, PropertiesType::Pointer pProperties) const override;

    /**
     * clones the selected element variables, creating a new one
     * @param NewId: the ID of the new element
     * @param ThisNodes: the nodes of the new element
     * @param pProperties: the properties assigned to the new element
     * @return a Pointer to the new element
     */
    Element::Pointer Clone(IndexType NewId, NodesArrayType const& ThisNodes) const override;


    //************* GETTING METHODS

    /**
     * Returns the currently selected integration method
     * @return current integration method selected
     */
    IntegrationMethod GetIntegrationMethod() const override;

    /**
     * Sets on rElementalDofList the degrees of freedom of the considered element geometry
     */
    void GetDofList(DofsVectorType& rElementalDofList, ProcessInfo& rCurrentProcessInfo) override;

    /**
     * Sets on rResult the ID's of the element degrees of freedom
     */
    void EquationIdVector(EquationIdVectorType& rResult, ProcessInfo& rCurrentProcessInfo) override;

    /**
     * Sets on rValues the nodal displacements
     */
    void GetValuesVector(Vector& rValues, int Step = 0) override;

    /**
     * Sets on rValues the nodal velocities
     */
    void GetFirstDerivativesVector(Vector& rValues, int Step = 0) override;

    /**
     * Sets on rValues the nodal accelerations
     */
    void GetSecondDerivativesVector(Vector& rValues, int Step = 0) override;



    //on integration points:
    /**
     * Access for variables on Integration points.
     * This gives access to variables stored in the constitutive law on each integration point.
     * Specialisations of element.h (e.g. the TotalLagrangian) must specify the actual
     * interface to the constitutive law!
     * Note, that these functions expect a std::vector of values for the
     * specified variable type that contains a value for each integration point!
     * SetValuesOnIntegrationPoints: set the values for given Variable.
     * GetValueOnIntegrationPoints: get the values for given Variable.
     */

    //SET
    /**
     * Set a double  Value on the Element Constitutive Law
     */
    virtual void SetValuesOnIntegrationPoints(const Variable<double>& rVariable, std::vector<double>& rValues, const ProcessInfo& rCurrentProcessInfo) override;

    /**
     * Set a Vector Value on the Element Constitutive Law
     */
    void SetValuesOnIntegrationPoints(const Variable<Vector>& rVariable, std::vector<Vector>& rValues, const ProcessInfo& rCurrentProcessInfo) override;

    /**
     * Set a Matrix Value on the Element Constitutive Law
     */
    void SetValuesOnIntegrationPoints(const Variable<Matrix>& rVariable, std::vector<Matrix>& rValues, const ProcessInfo& rCurrentProcessInfo) override;

    /**
    * Set a Constitutive Law Value
    */
    void SetValuesOnIntegrationPoints( const Variable<ConstitutiveLaw::Pointer>& rVariable,
                                      std::vector<ConstitutiveLaw::Pointer>& rValues,
                                      const ProcessInfo& rCurrentProcessInfo ) override;


    //GET:
    /**
     * Get on rVariable a double Value from the Element Constitutive Law
     */
    virtual void GetValueOnIntegrationPoints(const Variable<double>& rVariable, std::vector<double>& rValues, const ProcessInfo& rCurrentProcessInfo) override;

    /**
     * Get on rVariable a Vector Value from the Element Constitutive Law
     */
    virtual void GetValueOnIntegrationPoints(const Variable<Vector>& rVariable, std::vector<Vector>& rValues, const ProcessInfo& rCurrentProcessInfo) override;

    /**
     * Get on rVariable a Matrix Value from the Element Constitutive Law
     */
    virtual void GetValueOnIntegrationPoints(const Variable<Matrix>& rVariable, std::vector<Matrix>& rValues, const ProcessInfo& rCurrentProcessInfo) override;

    /**
     * Get a Constitutive Law Value
     */
    void GetValueOnIntegrationPoints( const Variable<ConstitutiveLaw::Pointer>& rVariable,
                                      std::vector<ConstitutiveLaw::Pointer>& rValues,
                                      const ProcessInfo& rCurrentProcessInfo ) override;



    //************* STARTING - ENDING  METHODS

    /**
      * Called to initialize the element.
      * Must be called before any calculation is done
      */
    virtual void Initialize() override;

    /**
     * Called at the beginning of each solution step
     */
    void InitializeSolutionStep(ProcessInfo& rCurrentProcessInfo) override;

    /**
     * this is called for non-linear analysis at the beginning of the iteration process
     */
    void InitializeNonLinearIteration(ProcessInfo& rCurrentProcessInfo) override;

    /**
     * this is called for non-linear analysis at the beginning of the iteration process
     */
    void FinalizeNonLinearIteration(ProcessInfo& rCurrentProcessInfo) override;

    /**
     * Called at the end of eahc solution step
     */
    void FinalizeSolutionStep(ProcessInfo& rCurrentProcessInfo) override;


    //************* COMPUTING  METHODS


    /**
     * this is called during the assembling process in order
     * to calculate all elemental contributions to the global system
     * matrix and the right hand side
     * @param rLeftHandSideMatrix: the elemental left hand side matrix
     * @param rRightHandSideVector: the elemental right hand side
     * @param rCurrentProcessInfo: the current process info instance
     */

    void CalculateLocalSystem(MatrixType& rLeftHandSideMatrix,
			      VectorType& rRightHandSideVector,
			      ProcessInfo& rCurrentProcessInfo) override;


    /**
      * this is called during the assembling process in order
      * to calculate the elemental right hand side vector only
      * @param rRightHandSideVector: the elemental right hand side vector
      * @param rCurrentProcessInfo: the current process info instance
      */
    void CalculateRightHandSide(VectorType& rRightHandSideVector,
				ProcessInfo& rCurrentProcessInfo) override;


    /**
     * this is called during the assembling process in order
     * to calculate the elemental left hand side vector only
     * @param rLeftHandSideVector: the elemental left hand side vector
     * @param rCurrentProcessInfo: the current process info instance
     */
    void CalculateLeftHandSide (MatrixType& rLeftHandSideMatrix,
				ProcessInfo& rCurrentProcessInfo) override;

    /**
     * this is called during the assembling process in order
     * to calculate the first derivatives contributions for the LHS and RHS
     * @param rLeftHandSideMatrix: the elemental left hand side matrix
     * @param rRightHandSideVector: the elemental right hand side
     * @param rCurrentProcessInfo: the current process info instance
     */
    void CalculateFirstDerivativesContributions(MatrixType& rLeftHandSideMatrix,
						VectorType& rRightHandSideVector,
						ProcessInfo& rCurrentProcessInfo) override;

    /**
     * this is called during the assembling process in order
     * to calculate the second derivatives contributions for the LHS and RHS
     * @param rLeftHandSideMatrix: the elemental left hand side matrix
     * @param rRightHandSideVector: the elemental right hand side
     * @param rCurrentProcessInfo: the current process info instance
     */
    void CalculateSecondDerivativesContributions(MatrixType& rLeftHandSideMatrix,
						VectorType& rRightHandSideVector,
						ProcessInfo& rCurrentProcessInfo) override;

    /**
     * this is called during the assembling process in order
     * to calculate the elemental left hand side matrix for the second derivatives constributions
     * @param rLeftHandSideMatrix: the elemental left hand side matrix
     * @param rCurrentProcessInfo: the current process info instance
     */
    void CalculateSecondDerivativesLHS(MatrixType& rLeftHandSideMatrix,
				       ProcessInfo& rCurrentProcessInfo) override;


    /**
     * this is called during the assembling process in order
     * to calculate the elemental right hand side vector for the second derivatives constributions
     * @param rRightHandSideVector: the elemental right hand side vector
     * @param rCurrentProcessInfo: the current process info instance
     */
    void CalculateSecondDerivativesRHS(VectorType& rRightHandSideVector,
				       ProcessInfo& rCurrentProcessInfo) override;

    /**
      * this is called during the assembling process in order
      * to calculate the elemental mass matrix
      * @param rMassMatrix: the elemental mass matrix
      * @param rCurrentProcessInfo: the current process info instance
      */
    void CalculateMassMatrix(MatrixType& rMassMatrix,
		    ProcessInfo& rCurrentProcessInfo) override;

    /**
      * this is called during the assembling process in order
      * to calculate the elemental damping matrix
      * @param rDampingMatrix: the elemental damping matrix
      * @param rCurrentProcessInfo: the current process info instance
      */
    void CalculateDampingMatrix(MatrixType& rDampingMatrix,
		    ProcessInfo& rCurrentProcessInfo) override;


    /**
     * this function is designed to make the element to assemble an rRHS vector
     * identified by a variable rRHSVariable by assembling it to the nodes on the variable
     * rDestinationVariable.
     * @param rRHSVector: input variable containing the RHS vector to be assembled
     * @param rRHSVariable: variable describing the type of the RHS vector to be assembled
     * @param rDestinationVariable: variable in the database to which the rRHSvector will be assembled
      * @param rCurrentProcessInfo: the current process info instance
     */
    virtual void AddExplicitContribution(const VectorType& rRHSVector,
					 const Variable<VectorType>& rRHSVariable,
					 Variable<array_1d<double,3> >& rDestinationVariable,
					 const ProcessInfo& rCurrentProcessInfo) override;

    //on integration points:
    /**
     * Calculate a double Variable on the Element Constitutive Law
     */
    void CalculateOnIntegrationPoints(const Variable<double>& rVariable, std::vector<double>& rOutput, const ProcessInfo& rCurrentProcessInfo) override;

    /**
     * Calculate a Vector Variable on the Element Constitutive Law
     */
    void CalculateOnIntegrationPoints(const Variable<Vector>& rVariable, std::vector<Vector>& rOutput, const ProcessInfo& rCurrentProcessInfo) override;

    /**
     * Calculate a Matrix Variable on the Element Constitutive Law
     */
    void CalculateOnIntegrationPoints(const Variable<Matrix >& rVariable, std::vector< Matrix >& rOutput, const ProcessInfo& rCurrentProcessInfo) override;


    //************************************************************************************
    //************************************************************************************
    /**
     * This function provides the place to perform checks on the completeness of the input.
     * It is designed to be called only once (or anyway, not often) typically at the beginning
     * of the calculations, so to verify that nothing is missing from the input
     * or that no common error is found.
     * @param rCurrentProcessInfo
     */
    int Check(const ProcessInfo& rCurrentProcessInfo) override;

    ///@}
    ///@name Access
    ///@{

    ///@}
    ///@name Inquiry
    ///@{
    ///@}
    ///@name Input and output
    ///@{
    /// Turn back information as a string.
    std::string Info() const override
    {
        std::stringstream buffer;
        buffer << "Fluid Element #" << Id();
        return buffer.str();
    }

    /// Print information about this object.
    void PrintInfo(std::ostream& rOStream) const override
    {
        rOStream << "Fluid Element #" << Id();
    }

    /// Print object's data.
    void PrintData(std::ostream& rOStream) const override
    {
      GetGeometry().PrintData(rOStream);
    }
    ///@}
    ///@name Friends
    ///@{
    ///@}

protected:
    ///@name Protected static Member Variables
    ///@{
    ///@}
    ///@name Protected member Variables
    ///@{

    /**
     * Currently selected integration methods
     */
    IntegrationMethod mThisIntegrationMethod;

    /**
     * Container for constitutive law instances on each integration point
     */
    std::vector<ConstitutiveLaw::Pointer> mConstitutiveLawVector;

    ///@}
    ///@name Protected Operators
    ///@{

    ///@}
    ///@name Protected Operations
    ///@{

    /**
     * Sets process information to set member variables
     */
    virtual void SetProcessInformation(const ProcessInfo& rCurrentProcessInfo);

    /**
     * Increases the integration method in the "increment" order
     */
    void IncreaseIntegrationMethod(IntegrationMethod& rThisIntegrationMethod,
				   unsigned int increment) const;

    /**
     * Calculates the elemental contributions
     */
    virtual void CalculateElementalSystem(LocalSystemComponents& rLocalSystem,
                                          ProcessInfo& rCurrentProcessInfo);

    /**
     * Calculates the elemental dynamic contributions
     */
    virtual void CalculateDynamicSystem(LocalSystemComponents& rLocalSystem,
					ProcessInfo& rCurrentProcessInfo);

    /**
     * Prints element information for each gauss point (debugging purposes)
     */
    void PrintElementCalculation(LocalSystemComponents& rLocalSystem, ElementDataType& rVariables);


    /**
     * Calculation of the tangent via perturbation of the dofs variables : testing purposes
     */
    void CalculatePerturbedLeftHandSide (MatrixType& rLeftHandSideMatrix,
					 ProcessInfo& rCurrentProcessInfo);

    /**
     * Calculation and addition of the matrices of the LHS
     */

    virtual void CalculateAndAddLHS(LocalSystemComponents& rLocalSystem,
                                    ElementDataType& rVariables);

    /**
     * Calculation and addition of the vectors of the RHS
     */

    virtual void CalculateAndAddRHS(LocalSystemComponents& rLocalSystem,
                                    ElementDataType& rVariables);


    /**
     * Calculation and addition of the matrices of the LHS
     */

    virtual void CalculateAndAddDynamicLHS(MatrixType& rLeftHandSideMatrix,
					   ElementDataType& rVariables);

    /**
     * Calculation and addition of the vectors of the RHS
     */
    virtual void CalculateAndAddDynamicRHS(VectorType& rRightHandSideVector,
					   ElementDataType& rVariables);

    /**
     * Calculation of the Material Stiffness Matrix. Kvvm = BT * C * B
     */
    virtual void CalculateAndAddKvvm(MatrixType& rLeftHandSideMatrix,
                                     ElementDataType& rVariables);

    /**
     * Calculation of the Geometric Stiffness Matrix. Kvvg = BT * S
     */
    virtual void CalculateAndAddKvvg(MatrixType& rLeftHandSideMatrix,
                                     ElementDataType& rVariables);


    /**
     * Calculation of the External Forces Vector. Fe = N * t + N * b
     */
    virtual void CalculateAndAddExternalForces(VectorType& rRightHandSideVector,
					       ElementDataType& rVariables);


    /**
      * Calculation of the Internal Forces Vector. Fi = B * sigma
      */
    virtual void CalculateAndAddInternalForces(VectorType& rRightHandSideVector,
					       ElementDataType & rVariables);


    /**
     * Set Variables of the Element to the Parameters of the Constitutive Law
     */
    virtual void SetElementData(ElementDataType& rVariables,
                                ConstitutiveLaw::Parameters& rValues,
                                const int & rPointNumber);


    /**
     * Set Parameters for the Constitutive Law and Calculate Material Response
     */
    virtual void CalculateMaterialResponse(ElementDataType& rVariables,
                                           ConstitutiveLaw::Parameters& rValues,
                                           const int & rPointNumber);

    /**
     * Get element size from the dofs
     */
    virtual unsigned int GetDofsSize();


    /**
     * Check if the element is labeled as sliver
     */
    bool IsSliver();

    /**
     * Initialize System Matrices
     */
    virtual void InitializeSystemMatrices(MatrixType& rLeftHandSideMatrix,
                                          VectorType& rRightHandSideVector,
                                          Flags& rCalculationFlags);

    /**
     * Initialize Material Properties on the Constitutive Law
     */
    void InitializeConstitutiveLaw();


    /**
     * Reset the Constitutive Law Parameters
     */
    void ResetConstitutiveLaw() override;


    /**
     * Clear Nodal Forces
     */
    void InitializeExplicitContributions();


    /**
     * Calculate Element Kinematics
     */
    virtual void CalculateKinematics(ElementDataType& rVariables,
                                     const double& rPointNumber);

    /**
     * Calculate Element Jacobian
     */
    virtual void CalculateKinetics(ElementDataType& rVariables,
				   const double& rPointNumber);

    /**
     * Initialize Element General Variables
     */
    virtual void InitializeElementData(ElementDataType & rVariables,
					    const ProcessInfo& rCurrentProcessInfo);

    /**
     * Transform Element General Variables
     */
    virtual void TransformElementData(ElementDataType & rVariables,
					   const double& rPointNumber);

    /**
     * Finalize Element Internal Variables
     */
    virtual void FinalizeStepVariables(ElementDataType & rVariables,
				       const double& rPointNumber);

    /**
     * Calculation of the velocity gradient
     */
    void CalculateVelocityGradient(Matrix& rL,
                                   const Matrix& rDN_DX,
                                   unsigned int step = 0);

    /**
     * Calculation of the velocity gradient
     */
    void CalculateVelocityGradientVector(Vector& rVector,
                                         const Matrix& rL,
                                         const Matrix& rDN_DX,
                                         unsigned int step = 0);
    /**
     * Calculation of the velocity gradient
     */
    void CalculateVelocityGradientVector(Vector& rVector,
                                         const Matrix& rDN_DX,
                                         unsigned int step = 0);

    /**
     * Calculation of the symmetric velocity gradient Vector
     */
    void CalculateSymmetricVelocityGradient(const Matrix& rL,
                                            Vector& rStrainVector);

    /**
     * Calculation of the skew symmetric velocity gradient Vector
     */
    void CalculateSkewSymmetricVelocityGradient(const Matrix& rL,
                                                Vector& rStrainVector);

    /**
     * Calculation of the Integration Weight
     */
    virtual double& CalculateIntegrationWeight(double& rIntegrationWeight);


    /**
     * Calculation of the Total Mass of the Element
     */
    virtual double& CalculateTotalMass(double& rTotalMass, const ProcessInfo& rCurrentProcessInfo);


    /**
     * Calculation of the Volume Change of the Element
     */
    virtual double& CalculateVolumeChange(double& rVolumeChange, ElementDataType& rVariables);

    /**
     * Calculation of the Volume Force of the Element
     */
    virtual Vector& CalculateVolumeForce(Vector& rVolumeForce, ElementDataType& rVariables);


    ///@}
    ///@name Protected  Access
    ///@{
    ///@}
    ///@name Protected Inquiry
    ///@{
    ///@}
    ///@name Protected LifeCycle
    ///@{
    ///@}

private:

    ///@name Static Member Variables
    ///@{
    ///@}
    ///@name Member Variables
    ///@{


    ///@}
    ///@name Private Operators
    ///@{


    ///@}
    ///@name Private Operations
    ///@{


    ///@}
    ///@name Private  Access
    ///@{
    ///@}

    ///@}
    ///@name Serialization
    ///@{
    friend class Serializer;

    // A private default constructor necessary for serialization

    virtual void save(Serializer& rSerializer) const override;

    virtual void load(Serializer& rSerializer) override;


    ///@name Private Inquiry
    ///@{
    ///@}
    ///@name Un accessible methods
    ///@{
    ///@}

}; // Class FluidElement

///@}
///@name Type Definitions
///@{
///@}
///@name Input and output
///@{
///@}

} // namespace Kratos.
#endif // KRATOS_FLUID_ELEMENT_H_INCLUDED  defined
