//    |  /           |
//    ' /   __| _` | __|  _ \   __|
//    . \  |   (   | |   (   |\__ `
//   _|\_\_|  \__,_|\__|\___/ ____/
//                   Multi-Physics
//
//  License:         BSD License
//                     Kratos default license: kratos/license.txt
//
//  Main authors:    Riccardo Rossi
//
//
#if !defined(KRATOS_NODAL_RESIDUAL_BASED_BLOCK_BUILDER_AND_SOLVER)
#define KRATOS_NODAL_RESIDUAL_BASED_BLOCK_BUILDER_AND_SOLVER

/* System includes */
#include <set>
#ifdef _OPENMP
#include <omp.h>
#endif

/* #include <unordered_set> */

/* #ifdef USE_GOOGLE_HASH */
/* #include "sparsehash/dense_hash_set" //included in external libraries */
/* #endif */

#ifdef USE_GOOGLE_HASH
#include "sparsehash/dense_hash_set" //included in external libraries
#else
#include <unordered_set>
#endif

/* Project includes */
#include "utilities/timer.h"
#include "includes/define.h"
#include "includes/key_hash.h"
#include "solving_strategies/builder_and_solvers/builder_and_solver.h"
#include "includes/model_part.h"

#include "utilities/openmp_utils.h"
#include "includes/kratos_flags.h"

namespace Kratos
{

///@name Kratos Globals
///@{

///@}
///@name Type Definitions
///@{

///@}
///@name  Enum's
///@{

///@}
///@name  Functions
///@{

///@}
///@name Kratos Classes
///@{

/**
   * @class NodalResidualBasedBlockBuilderAndSolver
   * @ingroup KratosCore
   * @brief Current class provides an implementation for standard builder and solving operations.
   * @details The RHS is constituted by the unbalanced loads (residual)
   * Degrees of freedom are reordered putting the restrained degrees of freedom at
   * the end of the system ordered in reverse order with respect to the DofSet.
   * Imposition of the dirichlet conditions is naturally dealt with as the residual already contains
   * this information.
   * Calculation of the reactions involves a cost very similiar to the calculation of the total residual
   * @author Riccardo Rossi
   */
template <class TSparseSpace,
          class TDenseSpace,  //= DenseSpace<double>,
          class TLinearSolver //= LinearSolver<TSparseSpace,TDenseSpace>
          >
class NodalResidualBasedBlockBuilderAndSolver
    : public BuilderAndSolver<TSparseSpace, TDenseSpace, TLinearSolver>
{
public:
  ///@name Type Definitions
  ///@{
  KRATOS_CLASS_POINTER_DEFINITION(NodalResidualBasedBlockBuilderAndSolver);

  typedef BuilderAndSolver<TSparseSpace, TDenseSpace, TLinearSolver> BaseType;

  typedef typename BaseType::TSchemeType TSchemeType;

  typedef typename BaseType::TDataType TDataType;

  typedef typename BaseType::DofsArrayType DofsArrayType;

  typedef typename BaseType::TSystemMatrixType TSystemMatrixType;

  typedef typename BaseType::TSystemVectorType TSystemVectorType;

  typedef typename BaseType::LocalSystemVectorType LocalSystemVectorType;

  typedef typename BaseType::LocalSystemMatrixType LocalSystemMatrixType;

  typedef typename BaseType::TSystemMatrixPointerType TSystemMatrixPointerType;
  typedef typename BaseType::TSystemVectorPointerType TSystemVectorPointerType;

  typedef Node<3> NodeType;

  typedef typename BaseType::NodesArrayType NodesArrayType;
  typedef typename BaseType::ElementsArrayType ElementsArrayType;
  typedef typename BaseType::ConditionsArrayType ConditionsArrayType;

  typedef typename BaseType::ElementsContainerType ElementsContainerType;

  typedef Vector VectorType;

  ///@}
  ///@name Life Cycle
  ///@{

  /** Constructor.
       */
  NodalResidualBasedBlockBuilderAndSolver(
      typename TLinearSolver::Pointer pNewLinearSystemSolver)
      : BuilderAndSolver<TSparseSpace, TDenseSpace, TLinearSolver>(pNewLinearSystemSolver)
  {
  }

  /** Destructor.
       */
  ~NodalResidualBasedBlockBuilderAndSolver() override
  {
  }

  ///@}
  ///@name Operators
  ///@{

  ///@}
  ///@name Operations
  ///@{

  /**
       * @brief Function to perform the build of the RHS. The vector could be sized as the total number
       * of dofs or as the number of unrestrained ones
       * @param pScheme The integration scheme considered
       * @param rModelPart The model part of the problem to solve
       * @param A The LHS matrix
       * @param b The RHS vector
       */
  void Build(
      typename TSchemeType::Pointer pScheme,
      ModelPart &rModelPart,
      TSystemMatrixType &A,
      TSystemVectorType &b) override
  {
    KRATOS_TRY

    KRATOS_ERROR_IF(!pScheme) << "No scheme provided!" << std::endl;

    // Getting the elements from the model
    const int nelements = static_cast<int>(rModelPart.Elements().size());

    // Getting the array of the conditions
    const int nconditions = static_cast<int>(rModelPart.Conditions().size());

    ProcessInfo &CurrentProcessInfo = rModelPart.GetProcessInfo();
    ModelPart::ElementsContainerType::iterator el_begin = rModelPart.ElementsBegin();
    ModelPart::ConditionsContainerType::iterator cond_begin = rModelPart.ConditionsBegin();

    //contributions to the system
    LocalSystemMatrixType LHS_Contribution = LocalSystemMatrixType(0, 0);
    LocalSystemVectorType RHS_Contribution = LocalSystemVectorType(0);

    //vector containing the localization in the system of the different
    //terms
    Element::EquationIdVectorType EquationId;

    // assemble all elements
    double start_build = OpenMPUtils::GetCurrentTime();

#pragma omp parallel firstprivate(nelements, nconditions, LHS_Contribution, RHS_Contribution, EquationId)
    {
#pragma omp for schedule(guided, 512) nowait
      for (int k = 0; k < nelements; k++)
      {
        ModelPart::ElementsContainerType::iterator it = el_begin + k;

        //detect if the element is active or not. If the user did not make any choice the element
        //is active by default
        bool element_is_active = true;
        if ((it)->IsDefined(ACTIVE))
          element_is_active = (it)->Is(ACTIVE);

        if (element_is_active)
        {
          //calculate elemental contribution
          pScheme->CalculateSystemContributions(*(it.base()), LHS_Contribution, RHS_Contribution, EquationId, CurrentProcessInfo);

          //assemble the elemental contribution
#ifdef USE_LOCKS_IN_ASSEMBLY
          Assemble(A, b, LHS_Contribution, RHS_Contribution, EquationId, mlock_array);
#else
          Assemble(A, b, LHS_Contribution, RHS_Contribution, EquationId);
#endif
          // clean local elemental memory
          pScheme->CleanMemory(*(it.base()));
        }
      }

      //#pragma omp parallel for firstprivate(nconditions, LHS_Contribution, RHS_Contribution, EquationId ) schedule(dynamic, 1024)
#pragma omp for schedule(guided, 512)
      for (int k = 0; k < nconditions; k++)
      {
        ModelPart::ConditionsContainerType::iterator it = cond_begin + k;

        //detect if the element is active or not. If the user did not make any choice the element
        //is active by default
        bool condition_is_active = true;
        if ((it)->IsDefined(ACTIVE))
          condition_is_active = (it)->Is(ACTIVE);

        if (condition_is_active)
        {
          //calculate elemental contribution
          pScheme->Condition_CalculateSystemContributions(*(it.base()), LHS_Contribution, RHS_Contribution, EquationId, CurrentProcessInfo);

          //assemble the elemental contribution
#ifdef USE_LOCKS_IN_ASSEMBLY
          Assemble(A, b, LHS_Contribution, RHS_Contribution, EquationId, mlock_array);
#else
          Assemble(A, b, LHS_Contribution, RHS_Contribution, EquationId);
#endif

          // clean local elemental memory
          pScheme->CleanMemory(*(it.base()));
        }
      }
    }

    const double stop_build = OpenMPUtils::GetCurrentTime();
    KRATOS_INFO_IF("NodalResidualBasedBlockBuilderAndSolver", (this->GetEchoLevel() >= 1 && rModelPart.GetCommunicator().MyPID() == 0)) << "Build time: " << stop_build - start_build << std::endl;

    //for (int i = 0; i < A_size; i++)
    //    omp_destroy_lock(&lock_array[i]);
    KRATOS_INFO_IF("NodalResidualBasedBlockBuilderAndSolver", (this->GetEchoLevel() > 2 && rModelPart.GetCommunicator().MyPID() == 0)) << "Finished parallel building" << std::endl;

    KRATOS_CATCH("")
  }

  void BuildNodally(
      typename TSchemeType::Pointer pScheme,
      ModelPart &rModelPart,
      TSystemMatrixType &A,
      TSystemVectorType &b)
  {
    KRATOS_TRY

    KRATOS_ERROR_IF(!pScheme) << "No scheme provided!" << std::endl;

    std::cout << "Build Nodally Continuity Equation" << std::endl;

    //contributions to the system
    LocalSystemMatrixType LHS_Contribution = LocalSystemMatrixType(0, 0);
    LocalSystemVectorType RHS_Contribution = LocalSystemVectorType(0);

    //vector containing the localization in the system of the different terms
    Element::EquationIdVectorType EquationId;

    ProcessInfo &CurrentProcessInfo = rModelPart.GetProcessInfo();

    /* const double start_build = OpenMPUtils::GetCurrentTime(); */

    /* #pragma omp parallel */
    {
      ModelPart::NodeIterator NodesBegin;
      ModelPart::NodeIterator NodesEnd;
      OpenMPUtils::PartitionedIterators(rModelPart.Nodes(), NodesBegin, NodesEnd);

      for (ModelPart::NodeIterator itNode = NodesBegin; itNode != NodesEnd; ++itNode)
      {

        VectorType nodalSFDneighboursId = itNode->FastGetSolutionStepValue(NODAL_SFD_NEIGHBOURS_ORDER);
        const unsigned int neighSize = nodalSFDneighboursId.size();

        if (neighSize > 1)
        {

          const unsigned int dimension = rModelPart.ElementsBegin()->GetGeometry().WorkingSpaceDimension();
          const double nodalVolume = itNode->FastGetSolutionStepValue(NODAL_VOLUME);
          const double timeInterval = CurrentProcessInfo[DELTA_TIME];

          LHS_Contribution = ZeroMatrix(neighSize, neighSize);
          RHS_Contribution = ZeroVector(neighSize);

          if (EquationId.size() != neighSize)
            EquationId.resize(neighSize, false);

          double deviatoricCoeff = itNode->FastGetSolutionStepValue(DEVIATORIC_COEFFICIENT);
          double volumetricCoeff = itNode->FastGetSolutionStepValue(VOLUMETRIC_COEFFICIENT) + 2.0 * deviatoricCoeff / 3.0;

          const unsigned int xpos = itNode->GetDofPosition(VELOCITY_X);

          double deltaPressure = itNode->FastGetSolutionStepValue(PRESSURE, 0) - itNode->FastGetSolutionStepValue(PRESSURE, 1);
          double volumetricDefRate = itNode->GetSolutionStepValue(NODAL_VOLUMETRIC_DEF_RATE);

          LHS_Contribution(0, 0) += nodalVolume / volumetricCoeff;
          RHS_Contribution[0] += (-deltaPressure / volumetricCoeff + volumetricDefRate) * nodalVolume;

          bool stabilizationNeeded = false;
          if ((itNode->Is(FLUID) || (itNode->Is(SOLID) && itNode->FastGetSolutionStepValue(POISSON_RATIO) > 0.49999)))
          {
            stabilizationNeeded = true;
          }
          else
          {
            for (unsigned int i = 0; i < neighSize; i++)
            {
              unsigned int idNode = nodalSFDneighboursId[i];
              EquationId[i] = rModelPart.Nodes()[idNode].GetDof(PRESSURE, xpos).EquationId();
            }
          }

          if (stabilizationNeeded == true)
          {

            /* Vector& rNodalSFDneigh = itNode->FastGetSolutionStepValue(NODAL_SFD_NEIGHBOURS); */

            unsigned int firstRow = 0;
            unsigned int firstCol = 0;
            double meanMeshSize = itNode->FastGetSolutionStepValue(NODAL_MEAN_MESH_SIZE);
            double characteristicLength = 2.0 * meanMeshSize;
            /* double nodalFreesurfaceArea=itNode->FastGetSolutionStepValue(NODAL_FREESURFACE_AREA); */
            double density = itNode->FastGetSolutionStepValue(DENSITY);

            /* double tauStab=1.0/(8.0*deviatoricCoeff/(meanMeshSize*meanMeshSize)+2.0*density/timeInterval); */

            double nodalVelocity = 0;
            if (dimension == 2)
            {
              nodalVelocity = sqrt(itNode->FastGetSolutionStepValue(VELOCITY_X) * itNode->FastGetSolutionStepValue(VELOCITY_X) +
                                   itNode->FastGetSolutionStepValue(VELOCITY_Y) * itNode->FastGetSolutionStepValue(VELOCITY_Y));
            }
            else if (dimension == 3)
            {
              nodalVelocity = sqrt(itNode->FastGetSolutionStepValue(VELOCITY_X) * itNode->FastGetSolutionStepValue(VELOCITY_X) +
                                   itNode->FastGetSolutionStepValue(VELOCITY_Y) * itNode->FastGetSolutionStepValue(VELOCITY_Y) +
                                   itNode->FastGetSolutionStepValue(VELOCITY_Z) * itNode->FastGetSolutionStepValue(VELOCITY_Z));
            }

            double tauStab = 1.0 * (characteristicLength * characteristicLength * timeInterval) / (density * nodalVelocity * timeInterval * characteristicLength + density * characteristicLength * characteristicLength + 8.0 * deviatoricCoeff * timeInterval);
            /* tauStab*=10.0; */
            /* tauStab=0.0000001; */
            /* tauStab=100.0; */
            LHS_Contribution(0, 0) += +nodalVolume * tauStab * density / (volumetricCoeff * timeInterval);
            RHS_Contribution[0] += -nodalVolume * tauStab * density / (volumetricCoeff * timeInterval) * (deltaPressure - itNode->FastGetSolutionStepValue(PRESSURE_VELOCITY, 0) * timeInterval);

            if (itNode->Is(FREE_SURFACE))
            {

              /* LHS_Contribution(0,0) += + 2.0 * tauStab * nodalFreesurfaceArea / meanMeshSize; */
              /* RHS_Contribution[0]   += - 2.0 * tauStab * nodalFreesurfaceArea / meanMeshSize * itNode->FastGetSolutionStepValue(PRESSURE,0); */

              /* double boundLHScontribution=4.0 * tauStab * nodalVolume /(meanMeshSize*meanMeshSize); */
              /* std::cout<<"boundLHScontribution  "<<boundLHScontribution<<std::endl; */
              /* if(itNode->IsNot(RIGID)){ */
              LHS_Contribution(0, 0) += +4.0 * 2.0 * tauStab * nodalVolume / (meanMeshSize * meanMeshSize);
              RHS_Contribution[0] += -4.0 * 2.0 * tauStab * nodalVolume / (meanMeshSize * meanMeshSize) * itNode->FastGetSolutionStepValue(PRESSURE, 0);
              /* } */
              /* else { */
              /*   LHS_Contribution(0,0) += + 4.0/3.0 * tauStab * nodalVolume /(meanMeshSize*meanMeshSize); */
              /*   RHS_Contribution[0]   += - 4.0/3.0 * tauStab * nodalVolume /(meanMeshSize*meanMeshSize) * itNode->FastGetSolutionStepValue(PRESSURE,0); */
              /* } */
              const array_1d<double, 3> &Normal = itNode->FastGetSolutionStepValue(NORMAL);
              Vector &SpatialDefRate = itNode->FastGetSolutionStepValue(NODAL_SPATIAL_DEF_RATE);
              array_1d<double, 3> nodalAcceleration = 0.5 * (itNode->FastGetSolutionStepValue(VELOCITY, 0) - itNode->FastGetSolutionStepValue(VELOCITY, 1)) / timeInterval - itNode->FastGetSolutionStepValue(ACCELERATION, 1);
              /* nodalAcceleration=  (itNode->FastGetSolutionStepValue(VELOCITY,0)-itNode->FastGetSolutionStepValue(VELOCITY,1))/timeInterval; */

              double nodalNormalAcceleration = 0;
              double nodalNormalProjDefRate = 0;
              if (dimension == 2)
              {
                nodalNormalProjDefRate = Normal[0] * SpatialDefRate[0] * Normal[0] + Normal[1] * SpatialDefRate[1] * Normal[1] + 2 * Normal[0] * SpatialDefRate[2] * Normal[1];
                /* nodalNormalAcceleration=Normal[0]*itNode->FastGetSolutionStepValue(ACCELERATION_X,1) + Normal[1]*itNode->FastGetSolutionStepValue(ACCELERATION_Y,1); */
                /* nodalNormalAcceleration=(itNode->FastGetSolutionStepValue(VELOCITY_X,0)-itNode->FastGetSolutionStepValue(VELOCITY_X,1))*Normal[0]/timeInterval + */
                /* 	(itNode->FastGetSolutionStepValue(VELOCITY_Y,0)-itNode->FastGetSolutionStepValue(VELOCITY_Y,1))*Normal[1]/timeInterval; */
                nodalNormalAcceleration = Normal[0] * nodalAcceleration[0] + Normal[1] * nodalAcceleration[1];
              }
              else if (dimension == 3)
              {
                nodalNormalProjDefRate = Normal[0] * SpatialDefRate[0] * Normal[0] + Normal[1] * SpatialDefRate[1] * Normal[1] + Normal[2] * SpatialDefRate[2] * Normal[2] +
                                         2 * Normal[0] * SpatialDefRate[3] * Normal[1] + 2 * Normal[0] * SpatialDefRate[4] * Normal[2] + 2 * Normal[1] * SpatialDefRate[5] * Normal[2];

                /* nodalNormalAcceleration=Normal[0]*itNode->FastGetSolutionStepValue(ACCELERATION_X) + Normal[1]*itNode->FastGetSolutionStepValue(ACCELERATION_Y) + Normal[2]*itNode->FastGetSolutionStepValue(ACCELERATION_Z); */
                /* nodalNormalAcceleration=Normal[0]*nodalAcceleration[0] + Normal[1]*nodalAcceleration[1] + Normal[2]*nodalAcceleration[2]; */
              }
              /* RHS_Contribution[0]  += tauStab * (density*nodalNormalAcceleration - 4.0*deviatoricCoeff*nodalNormalProjDefRate/meanMeshSize) * nodalFreesurfaceArea; */
              double accelerationContribution = 2.0 * density * nodalNormalAcceleration / meanMeshSize;

              double deviatoricContribution = 8.0 * deviatoricCoeff * nodalNormalProjDefRate / (meanMeshSize * meanMeshSize);
              /* std::cout<<"nodalNormalAcceleration= "<<nodalNormalAcceleration<<std::endl; */
              /* std::cout<<"nodalNormalProjDefRate= "<<nodalNormalProjDefRate<<std::endl; */
              /* std::cout<<"meanMeshSize "<<meanMeshSize<<std::endl; */

              /* accelerationContribution=0; */
              /* deviatoricContribution=0; */
              /* if(itNode->IsNot(RIGID)){ */
              RHS_Contribution[0] += 2.0 * tauStab * (accelerationContribution + deviatoricContribution) * nodalVolume;
              /* }else{ */
              /*   RHS_Contribution[0]  += 1.0/3.0* tauStab * (accelerationContribution - deviatoricContribution) * nodalVolume; */
              /* } */
            }

            for (unsigned int i = 0; i < neighSize; i++)
            {
              unsigned int idNode = nodalSFDneighboursId[i];
              EquationId[i] = rModelPart.Nodes()[idNode].GetDof(PRESSURE, xpos).EquationId();

              double Density = rModelPart.Nodes()[idNode].FastGetSolutionStepValue(DENSITY);
              array_1d<double, 3> &VolumeAcceleration = rModelPart.Nodes()[idNode].FastGetSolutionStepValue(VOLUME_ACCELERATION);

              double dNdXi = itNode->FastGetSolutionStepValue(NODAL_SFD_NEIGHBOURS)[firstCol];
              double dNdYi = itNode->FastGetSolutionStepValue(NODAL_SFD_NEIGHBOURS)[firstCol + 1];
              double dNdZi = 0;

              if (dimension == 2)
              {
                RHS_Contribution[i] += -tauStab * Density * (dNdXi * VolumeAcceleration[0] + dNdYi * VolumeAcceleration[1]) * nodalVolume;
              }
              else if (dimension == 3)
              {
                dNdZi = itNode->FastGetSolutionStepValue(NODAL_SFD_NEIGHBOURS)[firstCol + 2];
                RHS_Contribution[i] += -tauStab * Density * (dNdXi * VolumeAcceleration[0] + dNdYi * VolumeAcceleration[1] + dNdZi * VolumeAcceleration[2]) * nodalVolume;
              }

              firstRow = 0;

              for (unsigned int j = 0; j < neighSize; j++)
              {
                unsigned int idNodeJ = nodalSFDneighboursId[j];
                double dNdXj = itNode->FastGetSolutionStepValue(NODAL_SFD_NEIGHBOURS)[firstRow];
                double dNdYj = itNode->FastGetSolutionStepValue(NODAL_SFD_NEIGHBOURS)[firstRow + 1];

                if (dimension == 2)
                {
                  ////////////////// Laplacian term for LHS
                  LHS_Contribution(i, j) += +tauStab * (dNdXi * dNdXj + dNdYi * dNdYj) * nodalVolume;
                  ////////////////// Laplacian term L_ij*P_j for RHS
                  RHS_Contribution[i] += -tauStab * (dNdXi * dNdXj + dNdYi * dNdYj) * nodalVolume * rModelPart.Nodes()[idNodeJ].FastGetSolutionStepValue(PRESSURE, 0);
                }
                else if (dimension == 3)
                {
                  double dNdZj = itNode->FastGetSolutionStepValue(NODAL_SFD_NEIGHBOURS)[firstRow + 2];
                  ////////////////// Laplacian term for LHS
                  LHS_Contribution(i, j) += +tauStab * (dNdXi * dNdXj + dNdYi * dNdYj + dNdZi * dNdZj) * nodalVolume;
                  ////////////////// Laplacian term L_ij*P_j for RHS
                  RHS_Contribution[i] += -tauStab * (dNdXi * dNdXj + dNdYi * dNdYj + dNdZi * dNdZj) * nodalVolume * rModelPart.Nodes()[idNodeJ].FastGetSolutionStepValue(PRESSURE, 0);
                }
                /* std::cout << "dNdXi= " <<dNdXi<< "dNdYi= " <<dNdYi<< "dNdYj= " <<dNdYj<< "dNdXj= " <<dNdXj<< std::endl; */
                firstRow += dimension;
              }
              firstCol += dimension;
            }
          }

          //assemble the elemental contribution
#ifdef USE_LOCKS_IN_ASSEMBLY
          Assemble(A, b, LHS_Contribution, RHS_Contribution, EquationId, mlock_array);
#else
          Assemble(A, b, LHS_Contribution, RHS_Contribution, EquationId);
#endif

          /* AssembleLHS(A, LHS_Contribution, EquationId); */
          /* AssembleRHS(b, RHS_Contribution, EquationId); */
        }
      }
    }

    /* /\* std::cout<<".... Build Nodally Continuity Equation DONE!"<<std::endl; *\/ */

    /* const double stop_build = OpenMPUtils::GetCurrentTime(); */
    /* KRATOS_INFO_IF("NodalResidualBasedBlockBuilderAndSolver", (this->GetEchoLevel() >= 1 && rModelPart.GetCommunicator().MyPID() == 0)) << "Build time: " << stop_build - start_build << std::endl; */

    /* //for (int i = 0; i < A_size; i++) */
    /* //    omp_destroy_lock(&lock_array[i]); */
    /* KRATOS_INFO_IF("NodalResidualBasedBlockBuilderAndSolver", (this->GetEchoLevel() > 2 && rModelPart.GetCommunicator().MyPID() == 0)) << "Finished parallel building" << std::endl; */

    KRATOS_CATCH("")
  }

  /**
       * @brief Function to perform the building of the LHS
       * @details Depending on the implementation choosen the size of the matrix could
       * be equal to the total number of Dofs or to the number of unrestrained dofs
       * @param pScheme The integration scheme considered
       * @param rModelPart The model part of the problem to solve
       * @param A The LHS matrix
       */
  void BuildLHS(
      typename TSchemeType::Pointer pScheme,
      ModelPart &rModelPart,
      TSystemMatrixType &A) override
  {
    KRATOS_TRY

    TSystemVectorType tmp(A.size1(), 0.0);
    this->Build(pScheme, rModelPart, A, tmp);

    KRATOS_CATCH("")
  }

  /**
       * @brief Build a rectangular matrix of size n*N where "n" is the number of unrestrained degrees of freedom
       * and "N" is the total number of degrees of freedom involved.
       * @details This matrix is obtained by building the total matrix without the lines corresponding to the fixed
       * degrees of freedom (but keeping the columns!!)
       * @param pScheme The integration scheme considered
       * @param rModelPart The model part of the problem to solve
       * @param A The LHS matrix
       */
  void BuildLHS_CompleteOnFreeRows(
      typename TSchemeType::Pointer pScheme,
      ModelPart &rModelPart,
      TSystemMatrixType &A) override
  {
    KRATOS_TRY

    TSystemVectorType tmp(A.size1(), 0.0);
    this->Build(pScheme, rModelPart, A, tmp);

    KRATOS_CATCH("")
  }

  /**
       * @brief This is a call to the linear system solver
       * @param A The LHS matrix
       * @param Dx The Unknowns vector
       * @param b The RHS vector
       */
  void SystemSolve(
      TSystemMatrixType &A,
      TSystemVectorType &Dx,
      TSystemVectorType &b) override
  {
    KRATOS_TRY

    double norm_b;
    if (TSparseSpace::Size(b) != 0)
      norm_b = TSparseSpace::TwoNorm(b);
    else
      norm_b = 0.00;

    if (norm_b != 0.00)
    {
      //do solve
      BaseType::mpLinearSystemSolver->Solve(A, Dx, b);
    }
    else
      TSparseSpace::SetToZero(Dx);

    //prints informations about the current time
    KRATOS_INFO_IF("NodalResidualBasedBlockBuilderAndSolver", this->GetEchoLevel() > 1) << *(BaseType::mpLinearSystemSolver) << std::endl;

    KRATOS_CATCH("")
  }

  /**
       *@brief This is a call to the linear system solver (taking into account some physical particularities of the problem)
       * @param A The LHS matrix
       * @param Dx The Unknowns vector
       * @param b The RHS vector
       * @param rModelPart The model part of the problem to solve
       */
  void SystemSolveWithPhysics(
      TSystemMatrixType &A,
      TSystemVectorType &Dx,
      TSystemVectorType &b,
      ModelPart &rModelPart)
  {
    KRATOS_TRY

    double norm_b;
    if (TSparseSpace::Size(b) != 0)
      norm_b = TSparseSpace::TwoNorm(b);
    else
      norm_b = 0.00;

    if (norm_b != 0.00)
    {
      //provide physical data as needed
      if (BaseType::mpLinearSystemSolver->AdditionalPhysicalDataIsNeeded())
        BaseType::mpLinearSystemSolver->ProvideAdditionalData(A, Dx, b, BaseType::mDofSet, rModelPart);

      //do solve
      BaseType::mpLinearSystemSolver->Solve(A, Dx, b);
    }
    else
    {
      TSparseSpace::SetToZero(Dx);
      KRATOS_WARNING("NodalResidualBasedBlockBuilderAndSolver") << "ATTENTION! setting the RHS to zero!" << std::endl;
    }

    //prints informations about the current time
    KRATOS_INFO_IF("NodalResidualBasedBlockBuilderAndSolver", this->GetEchoLevel() > 1) << *(BaseType::mpLinearSystemSolver) << std::endl;

    KRATOS_CATCH("")
  }

  /**
       * @brief Function to perform the building and solving phase at the same time.
       * @details It is ideally the fastest and safer function to use when it is possible to solve
       * just after building
       * @param pScheme The integration scheme considered
       * @param rModelPart The model part of the problem to solve
       * @param A The LHS matrix
       * @param Dx The Unknowns vector
       * @param b The RHS vector
       */
  void BuildAndSolve(
      typename TSchemeType::Pointer pScheme,
      ModelPart &rModelPart,
      TSystemMatrixType &A,
      TSystemVectorType &Dx,
      TSystemVectorType &b) override
  {
    KRATOS_TRY

    std::cout << "CONTINUITY EQ: buildAndSolve " << std::endl;

    Timer::Start("Build");

    /* Build(pScheme, rModelPart, A, b); */

    //boost::timer build_time;
    BuildNodally(pScheme, rModelPart, A, b);
    //std::cout << "CONTINUITY EQ: build_time : " << build_time.elapsed() << std::endl;

    Timer::Stop("Build");

    ApplyDirichletConditions(pScheme, rModelPart, A, Dx, b);

    KRATOS_INFO_IF("NodalResidualBasedBlockBuilderAndSolver", (this->GetEchoLevel() == 3)) << "Before the solution of the system"
                                                                                           << "\nSystem Matrix = " << A << "\nUnknowns vector = " << Dx << "\nRHS vector = " << b << std::endl;

    const double start_solve = OpenMPUtils::GetCurrentTime();
    Timer::Start("Solve");

    //boost::timer solve_time;
    SystemSolveWithPhysics(A, Dx, b, rModelPart);
    //std::cout << "CONTINUITY EQ: solve_time : " << solve_time.elapsed() << std::endl;

    Timer::Stop("Solve");
    const double stop_solve = OpenMPUtils::GetCurrentTime();
    KRATOS_INFO_IF("NodalResidualBasedBlockBuilderAndSolver", (this->GetEchoLevel() >= 1 && rModelPart.GetCommunicator().MyPID() == 0)) << "System solve time: " << stop_solve - start_solve << std::endl;

    KRATOS_INFO_IF("NodalResidualBasedBlockBuilderAndSolver", (this->GetEchoLevel() == 3)) << "After the solution of the system"
                                                                                           << "\nSystem Matrix = " << A << "\nUnknowns vector = " << Dx << "\nRHS vector = " << b << std::endl;

    KRATOS_CATCH("")
  }

  /**
       * @brief Corresponds to the previews, but the System's matrix is considered already built and only the RHS is built again
       * @param pScheme The integration scheme considered
       * @param rModelPart The model part of the problem to solve
       * @param A The LHS matrix
       * @param Dx The Unknowns vector
       * @param b The RHS vector
       */
  void BuildRHSAndSolve(
      typename TSchemeType::Pointer pScheme,
      ModelPart &rModelPart,
      TSystemMatrixType &A,
      TSystemVectorType &Dx,
      TSystemVectorType &b) override
  {
    KRATOS_TRY

    BuildRHS(pScheme, rModelPart, b);
    SystemSolve(A, Dx, b);

    KRATOS_CATCH("")
  }

  /**
       * @brief Function to perform the build of the RHS.
       * @details The vector could be sized as the total number of dofs or as the number of unrestrained ones
       * @param pScheme The integration scheme considered
       * @param rModelPart The model part of the problem to solve
       */
  void BuildRHS(
      typename TSchemeType::Pointer pScheme,
      ModelPart &rModelPart,
      TSystemVectorType &b) override
  {
    KRATOS_TRY

    BuildRHSNoDirichlet(pScheme, rModelPart, b);

    const int ndofs = static_cast<int>(BaseType::mDofSet.size());

    //NOTE: dofs are assumed to be numbered consecutively in the BlockBuilderAndSolver
#pragma omp parallel for firstprivate(ndofs)
    for (int k = 0; k < ndofs; k++)
    {
      typename DofsArrayType::iterator dof_iterator = BaseType::mDofSet.begin() + k;
      const std::size_t i = dof_iterator->EquationId();

      if (dof_iterator->IsFixed())
        b[i] = 0.0f;
    }

    KRATOS_CATCH("")
  }

  /**
       * @brief Builds the list of the DofSets involved in the problem by "asking" to each element
       * and condition its Dofs.
       * @details The list of dofs is stores insde the BuilderAndSolver as it is closely connected to the
       * way the matrix and RHS are built
       * @param pScheme The integration scheme considered
       * @param rModelPart The model part of the problem to solve
       */
  void SetUpDofSet(
      typename TSchemeType::Pointer pScheme,
      ModelPart &rModelPart) override
  {
    KRATOS_TRY;

    KRATOS_INFO_IF("NodalResidualBasedBlockBuilderAndSolver", (this->GetEchoLevel() > 1 && rModelPart.GetCommunicator().MyPID() == 0)) << "Setting up the dofs" << std::endl;

    //Gets the array of elements from the modeler
    ElementsArrayType &pElements = rModelPart.Elements();
    const int nelements = static_cast<int>(pElements.size());

    Element::DofsVectorType ElementalDofList;

    ProcessInfo &CurrentProcessInfo = rModelPart.GetProcessInfo();

    unsigned int nthreads = OpenMPUtils::GetNumThreads();

    //        typedef boost::fast_pool_allocator< NodeType::DofType::Pointer > allocator_type;
    //         typedef std::unordered_set < NodeType::DofType::Pointer,
    //             DofPointerHasher,
    //             DofPointerComparor,
    //             allocator_type     >  set_type;

#ifdef USE_GOOGLE_HASH
    typedef google::dense_hash_set<NodeType::DofType::Pointer, DofPointerHasher> set_type;
#else
    typedef std::unordered_set<NodeType::DofType::Pointer, DofPointerHasher> set_type;
#endif
    //

    std::vector<set_type> dofs_aux_list(nthreads);
    //         std::vector<allocator_type> allocators(nthreads);

    KRATOS_INFO_IF("NodalResidualBasedBlockBuilderAndSolver", (this->GetEchoLevel() > 2)) << "Number of threads" << nthreads << "\n"
                                                                                          << std::endl;

    for (int i = 0; i < static_cast<int>(nthreads); i++)
    {
#ifdef USE_GOOGLE_HASH
      dofs_aux_list[i].set_empty_key(NodeType::DofType::Pointer());
#else
      //             dofs_aux_list[i] = set_type( allocators[i]);
      dofs_aux_list[i].reserve(nelements);
#endif
    }

    KRATOS_INFO_IF("NodalResidualBasedBlockBuilderAndSolver", (this->GetEchoLevel() > 2)) << "Initializing element loop" << std::endl;

#pragma omp parallel firstprivate(nelements, ElementalDofList)
    {
#pragma omp for schedule(guided, 512) nowait
      for (int i = 0; i < nelements; i++)
      {
        typename ElementsArrayType::iterator it = pElements.begin() + i;
        const unsigned int this_thread_id = OpenMPUtils::ThisThread();

        // gets list of Dof involved on every element
        pScheme->GetElementalDofList(*(it.base()), ElementalDofList, CurrentProcessInfo);

        dofs_aux_list[this_thread_id].insert(ElementalDofList.begin(), ElementalDofList.end());
      }

      KRATOS_INFO_IF("NodalResidualBasedBlockBuilderAndSolver", (this->GetEchoLevel() > 2)) << "Initializing condition loop" << std::endl;

      ConditionsArrayType &pConditions = rModelPart.Conditions();
      const int nconditions = static_cast<int>(pConditions.size());
#pragma omp for schedule(guided, 512)
      for (int i = 0; i < nconditions; i++)
      {
        typename ConditionsArrayType::iterator it = pConditions.begin() + i;
        const unsigned int this_thread_id = OpenMPUtils::ThisThread();

        // gets list of Dof involved on every element
        pScheme->GetConditionDofList(*(it.base()), ElementalDofList, CurrentProcessInfo);
        dofs_aux_list[this_thread_id].insert(ElementalDofList.begin(), ElementalDofList.end());
      }
    }

    KRATOS_INFO_IF("NodalResidualBasedBlockBuilderAndSolver", (this->GetEchoLevel() > 2)) << "Initializing tree reduction\n"
                                                                                          << std::endl;

    // Here we do a reduction in a tree so to have everything on thread 0
    unsigned int old_max = nthreads;
    unsigned int new_max = ceil(0.5 * static_cast<double>(old_max));
    while (new_max >= 1 && new_max != old_max)
    {
      if (this->GetEchoLevel() > 2)
      {
        //just for debugging
        std::cout << "old_max" << old_max << " new_max:" << new_max << std::endl;
        for (int i = 0; i < static_cast<int>(new_max); i++)
        {
          if (i + new_max < old_max)
          {
            std::cout << i << " - " << i + new_max << std::endl;
          }
        }
        std::cout << "********************" << std::endl;
      }

#pragma omp parallel for
      for (int i = 0; i < static_cast<int>(new_max); i++)
      {
        if (i + new_max < old_max)
        {
          dofs_aux_list[i].insert(dofs_aux_list[i + new_max].begin(), dofs_aux_list[i + new_max].end());
          dofs_aux_list[i + new_max].clear();
        }
      }

      old_max = new_max;
      new_max = ceil(0.5 * static_cast<double>(old_max));
    }

    KRATOS_INFO_IF("NodalResidualBasedBlockBuilderAndSolver", (this->GetEchoLevel() > 2)) << "Initializing ordered array filling\n"
                                                                                          << std::endl;

    DofsArrayType Doftemp;
    BaseType::mDofSet = DofsArrayType();

    Doftemp.reserve(dofs_aux_list[0].size());
    for (auto it = dofs_aux_list[0].begin(); it != dofs_aux_list[0].end(); it++)
    {
      Doftemp.push_back((*it));
    }
    Doftemp.Sort();

    BaseType::mDofSet = Doftemp;

    //Throws an exception if there are no Degrees Of Freedom involved in the analysis
    KRATOS_ERROR_IF(BaseType::mDofSet.size() == 0) << "No degrees of freedom!" << std::endl;

    KRATOS_INFO_IF("NodalResidualBasedBlockBuilderAndSolver", (this->GetEchoLevel() > 2)) << "Number of degrees of freedom:" << BaseType::mDofSet.size() << std::endl;

    BaseType::mDofSetIsInitialized = true;

    KRATOS_INFO_IF("NodalResidualBasedBlockBuilderAndSolver", (this->GetEchoLevel() > 2 && rModelPart.GetCommunicator().MyPID() == 0)) << "Finished setting up the dofs" << std::endl;

    KRATOS_INFO_IF("NodalResidualBasedBlockBuilderAndSolver", (this->GetEchoLevel() > 2)) << "Initializing lock array" << std::endl;

#ifdef _OPENMP
    if (mlock_array.size() != 0)
    {
      for (int i = 0; i < static_cast<int>(mlock_array.size()); i++)
      {
        omp_destroy_lock(&mlock_array[i]);
      }
    }
    mlock_array.resize(BaseType::mDofSet.size());

    for (int i = 0; i < static_cast<int>(mlock_array.size()); i++)
    {
      omp_init_lock(&mlock_array[i]);
    }
#endif

    KRATOS_INFO_IF("NodalResidualBasedBlockBuilderAndSolver", (this->GetEchoLevel() > 2)) << "End of setup dof set\n"
                                                                                          << std::endl;

    // If reactions are to be calculated, we check if all the dofs have reactions defined
    // This is tobe done only in debug mode

#ifdef KRATOS_DEBUG

    if (BaseType::GetCalculateReactionsFlag())
    {
      for (auto dof_iterator = BaseType::mDofSet.begin(); dof_iterator != BaseType::mDofSet.end(); ++dof_iterator)
      {
        KRATOS_ERROR_IF_NOT(dof_iterator->HasReaction()) << "Reaction variable not set for the following : " << std::endl
                                                         << "Node : " << dof_iterator->Id() << std::endl
                                                         << "Dof : " << (*dof_iterator) << std::endl
                                                         << "Not possible to calculate reactions." << std::endl;
      }
    }
#endif

    KRATOS_CATCH("");
  }

  /**
       * @brief Organises the dofset in order to speed up the building phase
       * @param rModelPart The model part of the problem to solve
       */
  void SetUpSystem(
      ModelPart &rModelPart) override
  {

    //int free_id = 0;
    BaseType::mEquationSystemSize = BaseType::mDofSet.size();
    int ndofs = static_cast<int>(BaseType::mDofSet.size());

#pragma omp parallel for firstprivate(ndofs)
    for (int i = 0; i < static_cast<int>(ndofs); i++)
    {
      typename DofsArrayType::iterator dof_iterator = BaseType::mDofSet.begin() + i;
      dof_iterator->SetEquationId(i);
    }

    //for (typename DofsArrayType::iterator dof_iterator = BaseType::mDofSet.begin(); dof_iterator != BaseType::mDofSet.end(); ++dof_iterator)
    //    dof_iterator->SetEquationId(free_id++);
  }

  //**************************************************************************
  //**************************************************************************

  void ResizeAndInitializeVectors(
      typename TSchemeType::Pointer pScheme,
      TSystemMatrixPointerType &pA,
      TSystemVectorPointerType &pDx,
      TSystemVectorPointerType &pb,
      ModelPart &rModelPart) override
  {
    KRATOS_TRY

    //boost::timer contruct_matrix;

    if (pA == NULL) //if the pointer is not initialized initialize it to an empty matrix
    {
      TSystemMatrixPointerType pNewA = TSystemMatrixPointerType(new TSystemMatrixType(0, 0));
      pA.swap(pNewA);
    }
    if (pDx == NULL) //if the pointer is not initialized initialize it to an empty matrix
    {
      TSystemVectorPointerType pNewDx = TSystemVectorPointerType(new TSystemVectorType(0));
      pDx.swap(pNewDx);
    }
    if (pb == NULL) //if the pointer is not initialized initialize it to an empty matrix
    {
      TSystemVectorPointerType pNewb = TSystemVectorPointerType(new TSystemVectorType(0));
      pb.swap(pNewb);
    }

    TSystemMatrixType &A = *pA;
    TSystemVectorType &Dx = *pDx;
    TSystemVectorType &b = *pb;

    //resizing the system vectors and matrix
    if (A.size1() == 0 || BaseType::GetReshapeMatrixFlag() == true) //if the matrix is not initialized
    {
      A.resize(BaseType::mEquationSystemSize, BaseType::mEquationSystemSize, false);
      ConstructMatrixStructure(pScheme, A, rModelPart);
    }
    else
    {
      if (A.size1() != BaseType::mEquationSystemSize || A.size2() != BaseType::mEquationSystemSize)
      {
        KRATOS_WATCH("   it should not come here!!!!!!!! ... this is SLOW");
        KRATOS_ERROR << "The equation system size has changed during the simulation. This is not permited." << std::endl;
        A.resize(BaseType::mEquationSystemSize, BaseType::mEquationSystemSize, true);
        ConstructMatrixStructure(pScheme, A, rModelPart);
      }
    }
    if (Dx.size() != BaseType::mEquationSystemSize)
      Dx.resize(BaseType::mEquationSystemSize, false);
    if (b.size() != BaseType::mEquationSystemSize)
      b.resize(BaseType::mEquationSystemSize, false);

    //std::cout << "CONTINUITY EQ: contruct_matrix : " << contruct_matrix.elapsed() << std::endl;

    KRATOS_CATCH("")
  }

  //**************************************************************************
  //**************************************************************************

  void InitializeSolutionStep(
      ModelPart &rModelPart,
      TSystemMatrixType &A,
      TSystemVectorType &Dx,
      TSystemVectorType &b) override
  {
    KRATOS_TRY
    std::cout << "Initialize Solution Step in nodal res based " << std::endl;

    KRATOS_CATCH("")
  }

  //**************************************************************************
  //**************************************************************************

  void FinalizeSolutionStep(
      ModelPart &rModelPart,
      TSystemMatrixType &A,
      TSystemVectorType &Dx,
      TSystemVectorType &b) override
  {
  }

  //**************************************************************************
  //**************************************************************************

  void CalculateReactions(
      typename TSchemeType::Pointer pScheme,
      ModelPart &rModelPart,
      TSystemMatrixType &A,
      TSystemVectorType &Dx,
      TSystemVectorType &b) override
  {
    TSparseSpace::SetToZero(b);

    //refresh RHS to have the correct reactions
    BuildRHSNoDirichlet(pScheme, rModelPart, b);

    const int ndofs = static_cast<int>(BaseType::mDofSet.size());

    //NOTE: dofs are assumed to be numbered consecutively in the BlockBuilderAndSolver
#pragma omp parallel for firstprivate(ndofs)
    for (int k = 0; k < ndofs; k++)
    {
      typename DofsArrayType::iterator dof_iterator = BaseType::mDofSet.begin() + k;

      const int i = (dof_iterator)->EquationId();
      (dof_iterator)->GetSolutionStepReactionValue() = -b[i];
    }

    //KRATOS_WATCH(__LINE__)
  }

  /**
       * @brief Applies the dirichlet conditions. This operation may be very heavy or completely
       * unexpensive depending on the implementation choosen and on how the System Matrix is built.
       * @details For explanation of how it works for a particular implementation the user
       * should refer to the particular Builder And Solver choosen
       * @param pScheme The integration scheme considered
       * @param rModelPart The model part of the problem to solve
       * @param A The LHS matrix
       * @param Dx The Unknowns vector
       * @param b The RHS vector
       */
  void ApplyDirichletConditions(
      typename TSchemeType::Pointer pScheme,
      ModelPart &rModelPart,
      TSystemMatrixType &A,
      TSystemVectorType &Dx,
      TSystemVectorType &b) override
  {
    std::size_t system_size = A.size1();
    std::vector<double> scaling_factors(system_size, 0.0f);

    const int ndofs = static_cast<int>(BaseType::mDofSet.size());

    //NOTE: dofs are assumed to be numbered consecutively in the BlockBuilderAndSolver
#pragma omp parallel for firstprivate(ndofs)
    for (int k = 0; k < ndofs; k++)
    {
      typename DofsArrayType::iterator dof_iterator = BaseType::mDofSet.begin() + k;
      if (dof_iterator->IsFixed())
        scaling_factors[k] = 0.0f;
      else
        scaling_factors[k] = 1.0f;
    }

    double *Avalues = A.value_data().begin();
    std::size_t *Arow_indices = A.index1_data().begin();
    std::size_t *Acol_indices = A.index2_data().begin();

    //detect if there is a line of all zeros and set the diagonal to a 1 if this happens
#pragma omp parallel for firstprivate(system_size)
    for (int k = 0; k < static_cast<int>(system_size); ++k)
    {
      std::size_t col_begin = Arow_indices[k];
      std::size_t col_end = Arow_indices[k + 1];
      bool empty = true;
      for (std::size_t j = col_begin; j < col_end; ++j)
      {
        if (Avalues[j] != 0.0)
        {
          empty = false;
          break;
        }
      }

      if (empty == true)
      {
        A(k, k) = 1.0;
        b[k] = 0.0;
      }
    }

#pragma omp parallel for
    for (int k = 0; k < static_cast<int>(system_size); ++k)
    {
      std::size_t col_begin = Arow_indices[k];
      std::size_t col_end = Arow_indices[k + 1];
      double k_factor = scaling_factors[k];
      if (k_factor == 0)
      {
        // zero out the whole row, except the diagonal
        for (std::size_t j = col_begin; j < col_end; ++j)
          if (static_cast<int>(Acol_indices[j]) != k)
            Avalues[j] = 0.0;

        // zero out the RHS
        b[k] = 0.0;
      }
      else
      {
        // zero out the column which is associated with the zero'ed row
        for (std::size_t j = col_begin; j < col_end; ++j)
          if (scaling_factors[Acol_indices[j]] == 0)
            Avalues[j] = 0.0;
      }
    }
  }

  /**
       * @brief This function is intended to be called at the end of the solution step to clean up memory storage not needed
       */
  void Clear() override
  {
#ifdef _OPENMP
    for (int i = 0; i < static_cast<int>(mlock_array.size()); i++)
      omp_destroy_lock(&mlock_array[i]);
    mlock_array.resize(0);
#endif
    BaseType::Clear();
  }

  /**
       * @brief This function is designed to be called once to perform all the checks needed
       * on the input provided. Checks can be "expensive" as the function is designed
       * to catch user's errors.
       * @param rModelPart The model part of the problem to solve
       * @return 0 all ok
       */
  int Check(ModelPart &rModelPart) override
  {
    KRATOS_TRY

    return 0;
    KRATOS_CATCH("");
  }

  ///@}
  ///@name Access
  ///@{

  ///@}
  ///@name Inquiry
  ///@{

  ///@}
  ///@name Friends
  ///@{

  ///@}

protected:
  ///@name Protected static Member Variables
  ///@{

  ///@}
  ///@name Protected member Variables
  ///@{
#ifdef _OPENMP
  std::vector<omp_lock_t> mlock_array;
#endif
  ///@}
  ///@name Protected Operators
  ///@{

  ///@}
  ///@name Protected Operations
  ///@{

  virtual void ConstructMatrixStructure(
      typename TSchemeType::Pointer pScheme,
      TSystemMatrixType &A,
      ModelPart &rModelPart)
  {
    std::cout << " ConstructMatrixStructure for Continuity equation " << std::endl;
    //filling with zero the matrix (creating the structure)
    Timer::Start("MatrixStructure");

    ProcessInfo &CurrentProcessInfo = rModelPart.GetProcessInfo();

    // Getting the array of the conditions
    const int nconditions = static_cast<int>(rModelPart.Conditions().size());
    ModelPart::ConditionsContainerType::iterator cond_begin = rModelPart.ConditionsBegin();

    const std::size_t equation_size = BaseType::mEquationSystemSize;

#ifdef USE_GOOGLE_HASH
    std::vector<google::dense_hash_set<std::size_t>> indices(equation_size);
    const std::size_t empty_key = 2 * equation_size + 10;
#else
    std::vector<std::unordered_set<std::size_t>> indices(equation_size);
#endif

#pragma omp parallel for firstprivate(equation_size)
    for (int iii = 0; iii < static_cast<int>(equation_size); iii++)
    {
#ifdef USE_GOOGLE_HASH
      indices[iii].set_empty_key(empty_key);
#else
      indices[iii].reserve(40);
#endif
    }

    Element::EquationIdVectorType EquationId;

    /* #pragma omp parallel */
    /* 	{ */
    ModelPart::NodeIterator NodesBegin;
    ModelPart::NodeIterator NodesEnd;
    OpenMPUtils::PartitionedIterators(rModelPart.Nodes(), NodesBegin, NodesEnd);

    for (ModelPart::NodeIterator itNode = NodesBegin; itNode != NodesEnd; ++itNode)
    {

      /* VectorType nodalSFDneighboursId=itNode->FastGetSolutionStepValue(NODAL_SFD_NEIGHBOURS_ORDER); */
      /* const unsigned int neighSize = nodalSFDneighboursId.size(); */

      const unsigned int neighSize = itNode->FastGetSolutionStepValue(NODAL_SFD_NEIGHBOURS_ORDER).size();
      if (EquationId.size() != neighSize)
        EquationId.resize(neighSize, false);

      /* const unsigned int xpos = itNode->GetDofPosition(VELOCITY_X); */
      const unsigned int xpos = itNode->GetDofPosition(PRESSURE);

      for (unsigned int i = 0; i < neighSize; i++)
      {
        /* unsigned int idNode=nodalSFDneighboursId[i]; */
        unsigned int idNode = itNode->FastGetSolutionStepValue(NODAL_SFD_NEIGHBOURS_ORDER)[i];
        EquationId[i] = rModelPart.Nodes()[idNode].GetDof(PRESSURE, xpos).EquationId();
      }

      for (std::size_t i = 0; i < EquationId.size(); i++)
      {
        if (EquationId[i] < BaseType::mEquationSystemSize)
        {
#ifdef _OPENMP
          omp_set_lock(&mlock_array[EquationId[i]]);
#endif
          auto &row_indices = indices[EquationId[i]];
          for (auto it = EquationId.begin(); it != EquationId.end(); it++)
          {
            if (*it < BaseType::mEquationSystemSize)
              row_indices.insert(*it);
          }
#ifdef _OPENMP
          omp_unset_lock(&mlock_array[EquationId[i]]);
#endif
        }
      }

      /* 	      for (std::size_t i = 0; i < EquationId.size(); i++) */
      /* 		{ */
      /* #ifdef _OPENMP */
      /* 		  omp_set_lock(&mlock_array[EquationId[i]]); */
      /* #endif */
      /* 		  auto& row_indices = indices[EquationId[i]]; */
      /* 		  row_indices.insert(EquationId.begin(), EquationId.end()); */

      /* #ifdef _OPENMP */
      /* 		  omp_unset_lock(&mlock_array[EquationId[i]]); */
      /* #endif */
      /* 		} */
    }

    /* } */

    Element::EquationIdVectorType ids(3, 0);

#pragma omp parallel for firstprivate(nconditions, ids)
    for (int iii = 0; iii < nconditions; iii++)
    {
      typename ConditionsArrayType::iterator i_condition = cond_begin + iii;
      pScheme->Condition_EquationId(*(i_condition.base()), ids, CurrentProcessInfo);
      for (std::size_t i = 0; i < ids.size(); i++)
      {
#ifdef _OPENMP
        omp_set_lock(&mlock_array[ids[i]]);
#endif
        auto &row_indices = indices[ids[i]];
        row_indices.insert(ids.begin(), ids.end());
#ifdef _OPENMP
        omp_unset_lock(&mlock_array[ids[i]]);
#endif
      }
    }

    //count the row sizes
    unsigned int nnz = 0;
    for (unsigned int i = 0; i < indices.size(); i++)
      nnz += indices[i].size();

    A = boost::numeric::ublas::compressed_matrix<double>(indices.size(), indices.size(), nnz);

    double *Avalues = A.value_data().begin();
    std::size_t *Arow_indices = A.index1_data().begin();
    std::size_t *Acol_indices = A.index2_data().begin();

    //filling the index1 vector - DO NOT MAKE PARALLEL THE FOLLOWING LOOP!
    Arow_indices[0] = 0;
    for (int i = 0; i < static_cast<int>(A.size1()); i++)
      Arow_indices[i + 1] = Arow_indices[i] + indices[i].size();

#pragma omp parallel for
    for (int i = 0; i < static_cast<int>(A.size1()); i++)
    {
      const unsigned int row_begin = Arow_indices[i];
      const unsigned int row_end = Arow_indices[i + 1];
      unsigned int k = row_begin;
      for (auto it = indices[i].begin(); it != indices[i].end(); it++)
      {
        Acol_indices[k] = *it;
        Avalues[k] = 0.0;
        k++;
      }

      indices[i].clear(); //deallocating the memory

      std::sort(&Acol_indices[row_begin], &Acol_indices[row_end]);
    }

    A.set_filled(indices.size() + 1, nnz);

    Timer::Stop("MatrixStructure");

    /* std::cout<<"..... ConstructMatrixStructure for Continuity equation DONE"<<std::endl; */
  }

  /*       virtual void ConstructMatrixStructure( */
  /* 					    typename TSchemeType::Pointer pScheme, */
  /* 					    TSystemMatrixType& A, */
  /* 					    ModelPart& rModelPart) */
  /*       { */
  /*         //filling with zero the matrix (creating the structure) */
  /*         Timer::Start("MatrixStructure"); */

  /*         // Getting the elements from the model */
  /*         const int nelements = static_cast<int>(rModelPart.Elements().size()); */

  /*         // Getting the array of the conditions */
  /*         const int nconditions = static_cast<int>(rModelPart.Conditions().size()); */

  /*         ProcessInfo& CurrentProcessInfo = rModelPart.GetProcessInfo(); */
  /*         ModelPart::ElementsContainerType::iterator el_begin = rModelPart.ElementsBegin(); */
  /*         ModelPart::ConditionsContainerType::iterator cond_begin = rModelPart.ConditionsBegin(); */

  /*         const std::size_t equation_size = BaseType::mEquationSystemSize; */

  /* #ifdef USE_GOOGLE_HASH */
  /*         std::vector<google::dense_hash_set<std::size_t> > indices(equation_size); */
  /*         const std::size_t empty_key = 2*equation_size + 10; */
  /* #else */
  /*         std::vector<std::unordered_set<std::size_t> > indices(equation_size); */
  /* #endif */

  /* #pragma omp parallel for firstprivate(equation_size) */
  /*         for (int iii = 0; iii < static_cast<int>(equation_size); iii++) */
  /* 	  { */
  /* #ifdef USE_GOOGLE_HASH */
  /* 	    indices[iii].set_empty_key(empty_key); */
  /* #else */
  /* 	    indices[iii].reserve(40); */
  /* #endif */
  /* 	  } */

  /*         Element::EquationIdVectorType ids(3, 0); */

  /* #pragma omp parallel for firstprivate(nelements, ids) */
  /*         for(int iii=0; iii<nelements; iii++) */
  /* 	  { */
  /*             typename ElementsContainerType::iterator i_element = el_begin + iii; */
  /*             pScheme->EquationId( *(i_element.base()) , ids, CurrentProcessInfo); */
  /*             for (std::size_t i = 0; i < ids.size(); i++) */
  /* 	      { */
  /* #ifdef _OPENMP */
  /*                 omp_set_lock(&mlock_array[ids[i]]); */
  /* #endif */
  /*                 auto& row_indices = indices[ids[i]]; */
  /*                 row_indices.insert(ids.begin(), ids.end()); */

  /* #ifdef _OPENMP */
  /*                 omp_unset_lock(&mlock_array[ids[i]]); */
  /* #endif */
  /* 	      } */

  /* 	  } */

  /* #pragma omp parallel for firstprivate(nconditions, ids) */
  /*         for (int iii = 0; iii<nconditions; iii++) */
  /* 	  { */
  /*             typename ConditionsArrayType::iterator i_condition = cond_begin + iii; */
  /*             pScheme->Condition_EquationId( *(i_condition.base()), ids, CurrentProcessInfo); */
  /*             for (std::size_t i = 0; i < ids.size(); i++) */
  /* 	      { */
  /* #ifdef _OPENMP */
  /*                 omp_set_lock(&mlock_array[ids[i]]); */
  /* #endif */
  /*                 auto& row_indices = indices[ids[i]]; */
  /*                 row_indices.insert(ids.begin(), ids.end()); */
  /* #ifdef _OPENMP */
  /*                 omp_unset_lock(&mlock_array[ids[i]]); */
  /* #endif */
  /* 	      } */
  /* 	  } */

  /*         //count the row sizes */
  /*         unsigned int nnz = 0; */
  /*         for (unsigned int i = 0; i < indices.size(); i++) */
  /* 	  nnz += indices[i].size(); */

  /*         A = boost::numeric::ublas::compressed_matrix<double>(indices.size(), indices.size(), nnz); */

  /*         double* Avalues = A.value_data().begin(); */
  /*         std::size_t* Arow_indices = A.index1_data().begin(); */
  /*         std::size_t* Acol_indices = A.index2_data().begin(); */

  /*         //filling the index1 vector - DO NOT MAKE PARALLEL THE FOLLOWING LOOP! */
  /*         Arow_indices[0] = 0; */
  /*         for (int i = 0; i < static_cast<int>(A.size1()); i++) */
  /* 	  Arow_indices[i+1] = Arow_indices[i] + indices[i].size(); */

  /* #pragma omp parallel for */
  /*         for (int i = 0; i < static_cast<int>(A.size1()); i++) */
  /* 	  { */
  /*             const unsigned int row_begin = Arow_indices[i]; */
  /*             const unsigned int row_end = Arow_indices[i+1]; */
  /*             unsigned int k = row_begin; */
  /*             for (auto it = indices[i].begin(); it != indices[i].end(); it++) */
  /* 	      { */
  /*                 Acol_indices[k] = *it; */
  /*                 Avalues[k] = 0.0; */
  /*                 k++; */
  /* 	      } */

  /*             indices[i].clear(); //deallocating the memory */

  /*             std::sort(&Acol_indices[row_begin], &Acol_indices[row_end]); */

  /* 	  } */

  /*         A.set_filled(indices.size()+1, nnz); */

  /*         Timer::Stop("MatrixStructure"); */
  /*       } */

  //**************************************************************************

  void AssembleLHS(
      TSystemMatrixType &A,
      LocalSystemMatrixType &LHS_Contribution,
      Element::EquationIdVectorType &EquationId)
  {
    unsigned int local_size = LHS_Contribution.size1();

    for (unsigned int i_local = 0; i_local < local_size; i_local++)
    {
      unsigned int i_global = EquationId[i_local];

      for (unsigned int j_local = 0; j_local < local_size; j_local++)
      {
        unsigned int j_global = EquationId[j_local];

        A(i_global, j_global) += LHS_Contribution(i_local, j_local);
      }
    }
  }

  void Assemble(
      TSystemMatrixType &A,
      TSystemVectorType &b,
      const LocalSystemMatrixType &LHS_Contribution,
      const LocalSystemVectorType &RHS_Contribution,
      Element::EquationIdVectorType &EquationId
#ifdef USE_LOCKS_IN_ASSEMBLY
      ,
      std::vector<omp_lock_t> &lock_array
#endif
  )
  {
    unsigned int local_size = LHS_Contribution.size1();

    for (unsigned int i_local = 0; i_local < local_size; i_local++)
    {
      unsigned int i_global = EquationId[i_local];

#ifdef USE_LOCKS_IN_ASSEMBLY
      omp_set_lock(&lock_array[i_global]);
      b[i_global] += RHS_Contribution(i_local);
#else
      double &r_a = b[i_global];
      const double &v_a = RHS_Contribution(i_local);
#pragma omp atomic
      r_a += v_a;
#endif

      AssembleRowContribution(A, LHS_Contribution, i_global, i_local, EquationId);

#ifdef USE_LOCKS_IN_ASSEMBLY
      omp_unset_lock(&lock_array[i_global]);
#endif

      //note that computation of reactions is not performed here!
    }
  }

  //**************************************************************************

  void AssembleRHS(
      TSystemVectorType &b,
      LocalSystemVectorType &RHS_Contribution,
      Element::EquationIdVectorType &EquationId)
  {
    unsigned int local_size = RHS_Contribution.size();

    for (unsigned int i_local = 0; i_local < local_size; i_local++)
    {
      unsigned int i_global = EquationId[i_local];

      // ASSEMBLING THE SYSTEM VECTOR
      double &b_value = b[i_global];
      const double &rhs_value = RHS_Contribution[i_local];

#pragma omp atomic
      b_value += rhs_value;
    }
  }

  ///@}
  ///@name Protected  Access
  ///@{

  ///@}
  ///@name Protected Inquiry
  ///@{

  ///@}
  ///@name Protected LifeCycle
  ///@{

  ///@}

private:
  ///@name Static Member Variables
  ///@{

  ///@}
  ///@name Member Variables
  ///@{

  ///@}
  ///@name Private Operators
  ///@{

  ///@}
  ///@name Private Operations
  ///@{

  inline void AddUnique(std::vector<std::size_t> &v, const std::size_t &candidate)
  {
    std::vector<std::size_t>::iterator i = v.begin();
    std::vector<std::size_t>::iterator endit = v.end();
    while (i != endit && (*i) != candidate)
    {
      i++;
    }
    if (i == endit)
    {
      v.push_back(candidate);
    }
  }

  void BuildRHSNoDirichlet(
      typename TSchemeType::Pointer pScheme,
      ModelPart &rModelPart,
      TSystemVectorType &b)
  {
    KRATOS_TRY

    //Getting the Elements
    ElementsArrayType &pElements = rModelPart.Elements();

    //getting the array of the conditions
    ConditionsArrayType &ConditionsArray = rModelPart.Conditions();

    ProcessInfo &CurrentProcessInfo = rModelPart.GetProcessInfo();

    //contributions to the system
    LocalSystemMatrixType LHS_Contribution = LocalSystemMatrixType(0, 0);
    LocalSystemVectorType RHS_Contribution = LocalSystemVectorType(0);

    //vector containing the localization in the system of the different
    //terms
    Element::EquationIdVectorType EquationId;

    // assemble all elements
    //for (typename ElementsArrayType::ptr_iterator it = pElements.ptr_begin(); it != pElements.ptr_end(); ++it)

    const int nelements = static_cast<int>(pElements.size());
#pragma omp parallel firstprivate(nelements, RHS_Contribution, EquationId)
    {
#pragma omp for schedule(guided, 512) nowait
      for (int i = 0; i < nelements; i++)
      {
        typename ElementsArrayType::iterator it = pElements.begin() + i;
        //detect if the element is active or not. If the user did not make any choice the element
        //is active by default
        bool element_is_active = true;
        if ((it)->IsDefined(ACTIVE))
          element_is_active = (it)->Is(ACTIVE);

        if (element_is_active)
        {
          //calculate elemental Right Hand Side Contribution
          pScheme->Calculate_RHS_Contribution(*(it.base()), RHS_Contribution, EquationId, CurrentProcessInfo);

          //assemble the elemental contribution
          AssembleRHS(b, RHS_Contribution, EquationId);
        }
      }

      LHS_Contribution.resize(0, 0, false);
      RHS_Contribution.resize(0, false);

      // assemble all conditions
      //for (typename ConditionsArrayType::ptr_iterator it = ConditionsArray.ptr_begin(); it != ConditionsArray.ptr_end(); ++it)
      const int nconditions = static_cast<int>(ConditionsArray.size());
      //#pragma omp parallel for firstprivate(nconditions, RHS_Contribution, EquationId) schedule(dynamic, 1024)
#pragma omp for schedule(guided, 512)
      for (int i = 0; i < nconditions; i++)
      {
        auto it = ConditionsArray.begin() + i;
        //detect if the element is active or not. If the user did not make any choice the element
        //is active by default
        bool condition_is_active = true;
        if ((it)->IsDefined(ACTIVE))
          condition_is_active = (it)->Is(ACTIVE);

        if (condition_is_active)
        {

          //calculate elemental contribution
          pScheme->Condition_Calculate_RHS_Contribution(*(it.base()), RHS_Contribution, EquationId, CurrentProcessInfo);

          //assemble the elemental contribution
          AssembleRHS(b, RHS_Contribution, EquationId);
        }
      }
    }

    KRATOS_CATCH("")
  }

  //******************************************************************************************
  //******************************************************************************************

  inline void CreatePartition(unsigned int number_of_threads, const int number_of_rows, vector<unsigned int> &partitions)
  {
    partitions.resize(number_of_threads + 1);
    int partition_size = number_of_rows / number_of_threads;
    partitions[0] = 0;
    partitions[number_of_threads] = number_of_rows;
    for (unsigned int i = 1; i < number_of_threads; i++)
      partitions[i] = partitions[i - 1] + partition_size;
  }

  inline void AssembleRowContribution(TSystemMatrixType &A, const Matrix &Alocal, const unsigned int i, const unsigned int i_local, Element::EquationIdVectorType &EquationId)
  {
    double *values_vector = A.value_data().begin();
    std::size_t *index1_vector = A.index1_data().begin();
    std::size_t *index2_vector = A.index2_data().begin();

    size_t left_limit = index1_vector[i];
    //    size_t right_limit = index1_vector[i+1];

    //find the first entry
    size_t last_pos = ForwardFind(EquationId[0], left_limit, index2_vector);
    size_t last_found = EquationId[0];

#ifndef USE_LOCKS_IN_ASSEMBLY
    double &r_a = values_vector[last_pos];
    const double &v_a = Alocal(i_local, 0);
#pragma omp atomic
    r_a += v_a;
#else
    values_vector[last_pos] += Alocal(i_local, 0);
#endif

    //now find all of the other entries
    size_t pos = 0;
    for (unsigned int j = 1; j < EquationId.size(); j++)
    {
      unsigned int id_to_find = EquationId[j];
      if (id_to_find > last_found)
        pos = ForwardFind(id_to_find, last_pos + 1, index2_vector);
      else
        pos = BackwardFind(id_to_find, last_pos - 1, index2_vector);

#ifndef USE_LOCKS_IN_ASSEMBLY
      double &r = values_vector[pos];
      const double &v = Alocal(i_local, j);
#pragma omp atomic
      r += v;
#else
      values_vector[pos] += Alocal(i_local, j);
#endif

      last_found = id_to_find;
      last_pos = pos;
    }
  }

  inline unsigned int ForwardFind(const unsigned int id_to_find,
                                  const unsigned int start,
                                  const size_t *index_vector)
  {
    unsigned int pos = start;
    while (id_to_find != index_vector[pos])
      pos++;
    return pos;
  }

  inline unsigned int BackwardFind(const unsigned int id_to_find,
                                   const unsigned int start,
                                   const size_t *index_vector)
  {
    unsigned int pos = start;
    while (id_to_find != index_vector[pos])
      pos--;
    return pos;
  }

  ///@}
  ///@name Private Operations
  ///@{

  ///@}
  ///@name Private  Access
  ///@{

  ///@}
  ///@name Private Inquiry
  ///@{

  ///@}
  ///@name Un accessible methods
  ///@{

  ///@}

}; /* Class NodalResidualBasedBlockBuilderAndSolver */

///@}

///@name Type Definitions
///@{

///@}

} /* namespace Kratos.*/

#endif /* KRATOS_NODAL_RESIDUAL_BASED_BLOCK_BUILDER_AND_SOLVER  defined */
