//    |  /           |
//    ' /   __| _` | __|  _ \   __|
//    . \  |   (   | |   (   |\__ `
//   _|\_\_|  \__,_|\__|\___/ ____/
//                   Multi-Physics
//
//  License:		 BSD License
//					 Kratos default license: kratos/license.txt
//
//  Main authors:    Elisa Magliozzi
//

#include "custom_elements/compressible_navier_stokes.h"

namespace Kratos {

template<>
void CompressibleNavierStokes<3>::EquationIdVector(EquationIdVectorType& rResult, ProcessInfo& rCurrentProcessInfo)
{
    KRATOS_TRY

    unsigned int Dim = 3;
    unsigned int BlockSize = Dim+2;
    unsigned int NumNodes = 4;
    unsigned int DofSize  = NumNodes*(BlockSize);

    if (rResult.size() != DofSize)
        rResult.resize(DofSize, false);

    for(unsigned int i=0; i<NumNodes; i++)
    {
        rResult[i*(BlockSize)  ]  =  this->GetGeometry()[i].GetDof(DENSITY).EquationId();
        rResult[i*(BlockSize)+1]  =  this->GetGeometry()[i].GetDof(MOMENTUM_X).EquationId();
        rResult[i*(BlockSize)+2]  =  this->GetGeometry()[i].GetDof(MOMENTUM_Y).EquationId();
        rResult[i*(BlockSize)+3]  =  this->GetGeometry()[i].GetDof(MOMENTUM_Z).EquationId(); 
        rResult[i*(BlockSize)+4]  =  this->GetGeometry()[i].GetDof(TOTAL_ENERGY).EquationId();
    }

    KRATOS_CATCH("")
}

template<>
void CompressibleNavierStokes<3>::GetDofList(DofsVectorType& ElementalDofList, ProcessInfo& rCurrentProcessInfo)
{
    KRATOS_TRY

    unsigned int Dim = 3;
    unsigned int BlockSize = Dim+2;
    unsigned int NumNodes = 4;
    unsigned int DofSize  = NumNodes*(BlockSize);

    if (ElementalDofList.size() != DofSize)
        ElementalDofList.resize(DofSize);

    for(unsigned int i=0; i<NumNodes; i++)
    {
        ElementalDofList[i*(BlockSize)  ]  =  this->GetGeometry()[i].pGetDof(DENSITY);
        ElementalDofList[i*(BlockSize)+1]  =  this->GetGeometry()[i].pGetDof(MOMENTUM_X);
        ElementalDofList[i*(BlockSize)+2]  =  this->GetGeometry()[i].pGetDof(MOMENTUM_Y);
        ElementalDofList[i*(BlockSize)+3]  =  this->GetGeometry()[i].pGetDof(MOMENTUM_Z);
        ElementalDofList[i*(BlockSize)+4]  =  this->GetGeometry()[i].pGetDof(TOTAL_ENERGY);
    }

    KRATOS_CATCH("");
}

template<>
void CompressibleNavierStokes<3>::ComputeGaussPointLHSContribution(BoundedMatrix<double,20,20>& lhs, const ElementDataStruct& data,double data_v_sc, double data_k_sc)
{
    const int nnodes = 4;
    const int dim = 3;
    const int BlockSize = dim+2;
    const double h = data.h; 
 
    const double& bdf0 = data.bdf0;
    const double& bdf1 = data.bdf1;
    const double& bdf2 = data.bdf2;
 
    const BoundedMatrix<double,nnodes,BlockSize>& U = data.U;
    const BoundedMatrix<double,nnodes,BlockSize>& Un = data.Un;
    const BoundedMatrix<double,nnodes,BlockSize>& Unn = data.Unn;
    const BoundedMatrix<double,nnodes,dim>& f_ext = data.f_ext;
    const array_1d<double,nnodes>& r = data.r;
    const double mu = data.mu;
    const double nu = data.nu;
    const double lambda = data.lambda;
    const double c_v = data.c_v;
    const double gamma = data.gamma;
    const double cp = c_v*gamma;
    const double v_sc = data_v_sc;
    const double k_sc = data_k_sc;
 
    // Get shape function values
    const array_1d<double,nnodes>& N = data.N;
    const BoundedMatrix<double,nnodes,dim>& DN = data.DN_DX;
    
    // Stabilization parameters
    const double stab_c1 = 4.0;
    const double stab_c2 = 2.0;
    
    const array_1d<double,BlockSize> U_gauss= prod(trans(U),N);
    
    double tmp = U_gauss(dim+1)/U_gauss(0);
    for(unsigned int ll=0; ll<dim; ll++)
        tmp -=(U_gauss(ll+1)*U_gauss(ll+1))/(2*U_gauss(0)*U_gauss(0));
    double c = sqrt(gamma*(gamma-1)*tmp);

    double tau1inv = 0.0;
    for(unsigned int ll=0; ll<dim; ll++)
        tau1inv += (U_gauss(ll+1)/U_gauss(0))*(U_gauss(ll+1)/U_gauss(0));
    tau1inv = (sqrt(tau1inv)+c)*stab_c2/h;
    double tau2inv = stab_c1*nu/(h*h)+tau1inv;
    double tau3inv = stab_c1*lambda/(U_gauss(0)*cp*h*h)+tau1inv;
        
    const double tau1 = 1/tau1inv;
    const double tau2 = 1/tau2inv;
    const double tau3 = 1/tau3inv;
    
    const double clhs0 =             pow(N[0], 2);
const double clhs1 =             -bdf0*clhs0;
const double clhs2 =             (1.0L/2.0L)*DN(0,0)*tau2;
const double clhs3 =             2*N[0];
const double clhs4 =             N[0]*f_ext(0,0) + N[1]*f_ext(1,0) + N[2]*f_ext(2,0) + N[3]*f_ext(3,0);
const double clhs5 =             clhs3*clhs4;
const double clhs6 =             N[0]*U(0,0) + N[1]*U(1,0) + N[2]*U(2,0) + N[3]*U(3,0);
const double clhs7 =             pow(clhs6, -2);
const double clhs8 =             N[0]*U(0,2) + N[1]*U(1,2) + N[2]*U(2,2) + N[3]*U(3,2);
const double clhs9 =             2*DN(0,1);
const double clhs10 =             N[0]*U(0,1) + N[1]*U(1,1) + N[2]*U(2,1) + N[3]*U(3,1);
const double clhs11 =             clhs10*clhs9;
const double clhs12 =             clhs11*clhs8;
const double clhs13 =             clhs12*clhs7;
const double clhs14 =             N[0]*U(0,3) + N[1]*U(1,3) + N[2]*U(2,3) + N[3]*U(3,3);
const double clhs15 =             2*DN(0,2);
const double clhs16 =             clhs10*clhs15;
const double clhs17 =             clhs14*clhs16;
const double clhs18 =             clhs17*clhs7;
const double clhs19 =             DN(0,1)*U(0,1);
const double clhs20 =             DN(1,1)*U(1,1);
const double clhs21 =             DN(2,1)*U(2,1);
const double clhs22 =             DN(3,1)*U(3,1);
const double clhs23 =             clhs19 + clhs20 + clhs21 + clhs22;
const double clhs24 =             2*N[0]*clhs8;
const double clhs25 =             clhs23*clhs24;
const double clhs26 =             clhs25*clhs7;
const double clhs27 =             DN(0,1)*U(0,2);
const double clhs28 =             DN(1,1)*U(1,2);
const double clhs29 =             DN(2,1)*U(2,2);
const double clhs30 =             DN(3,1)*U(3,2);
const double clhs31 =             clhs27 + clhs28 + clhs29 + clhs30;
const double clhs32 =             clhs3*clhs31;
const double clhs33 =             clhs10*clhs32;
const double clhs34 =             clhs33*clhs7;
const double clhs35 =             DN(0,2)*U(0,1);
const double clhs36 =             DN(1,2)*U(1,1);
const double clhs37 =             DN(2,2)*U(2,1);
const double clhs38 =             DN(3,2)*U(3,1);
const double clhs39 =             clhs35 + clhs36 + clhs37 + clhs38;
const double clhs40 =             2*N[0]*clhs14;
const double clhs41 =             clhs39*clhs40;
const double clhs42 =             clhs41*clhs7;
const double clhs43 =             DN(0,2)*U(0,3);
const double clhs44 =             DN(1,2)*U(1,3);
const double clhs45 =             DN(2,2)*U(2,3);
const double clhs46 =             DN(3,2)*U(3,3);
const double clhs47 =             clhs43 + clhs44 + clhs45 + clhs46;
const double clhs48 =             clhs3*clhs47;
const double clhs49 =             clhs10*clhs48;
const double clhs50 =             clhs49*clhs7;
const double clhs51 =             DN(0,0)*U(0,1);
const double clhs52 =             DN(1,0)*U(1,1);
const double clhs53 =             DN(2,0)*U(2,1);
const double clhs54 =             DN(3,0)*U(3,1);
const double clhs55 =             clhs51 + clhs52 + clhs53 + clhs54;
const double clhs56 =             clhs3*clhs55;
const double clhs57 =             gamma - 3;
const double clhs58 =             clhs10*clhs57;
const double clhs59 =             clhs56*clhs58;
const double clhs60 =             clhs59*clhs7;
const double clhs61 =             -clhs60;
const double clhs62 =             DN(0,0)*U(0,2);
const double clhs63 =             DN(1,0)*U(1,2);
const double clhs64 =             DN(2,0)*U(2,2);
const double clhs65 =             DN(3,0)*U(3,2);
const double clhs66 =             clhs62 + clhs63 + clhs64 + clhs65;
const double clhs67 =             gamma - 1;
const double clhs68 =             2*N[0]*clhs67*clhs8;
const double clhs69 =             clhs66*clhs68;
const double clhs70 =             clhs69*clhs7;
const double clhs71 =             -clhs70;
const double clhs72 =             DN(0,0)*U(0,3);
const double clhs73 =             DN(1,0)*U(1,3);
const double clhs74 =             DN(2,0)*U(2,3);
const double clhs75 =             DN(3,0)*U(3,3);
const double clhs76 =             clhs72 + clhs73 + clhs74 + clhs75;
const double clhs77 =             2*N[0]*clhs14*clhs67;
const double clhs78 =             clhs76*clhs77;
const double clhs79 =             clhs7*clhs78;
const double clhs80 =             -clhs79;
const double clhs81 =             DN(0,1)*U(0,0) + DN(1,1)*U(1,0) + DN(2,1)*U(2,0) + DN(3,1)*U(3,0);
const double clhs82 =             clhs10*clhs81;
const double clhs83 =             pow(clhs6, -3);
const double clhs84 =             4*N[0]*U(0,2) + 4*N[1]*U(1,2) + 4*N[2]*U(2,2) + 4*N[3]*U(3,2);
const double clhs85 =             N[0]*clhs83*clhs84;
const double clhs86 =             clhs82*clhs85;
const double clhs87 =             -clhs86;
const double clhs88 =             DN(0,2)*U(0,0) + DN(1,2)*U(1,0) + DN(2,2)*U(2,0) + DN(3,2)*U(3,0);
const double clhs89 =             clhs10*clhs88;
const double clhs90 =             4*N[0]*U(0,3) + 4*N[1]*U(1,3) + 4*N[2]*U(2,3) + 4*N[3]*U(3,3);
const double clhs91 =             N[0]*clhs83*clhs90;
const double clhs92 =             clhs89*clhs91;
const double clhs93 =             -clhs92;
const double clhs94 =             pow(clhs10, 2);
const double clhs95 =             2*clhs94;
const double clhs96 =             clhs67*clhs94;
const double clhs97 =             pow(clhs8, 2);
const double clhs98 =             clhs67*clhs97;
const double clhs99 =             pow(clhs14, 2);
const double clhs100 =             clhs67*clhs99;
const double clhs101 =             clhs100 + clhs96 + clhs98;
const double clhs102 =             clhs101 - clhs95;
const double clhs103 =             DN(0,0)*clhs102;
const double clhs104 =             clhs103*clhs7;
const double clhs105 =             -clhs104;
const double clhs106 =             DN(0,0)*U(0,0) + DN(1,0)*U(1,0) + DN(2,0)*U(2,0) + DN(3,0)*U(3,0);
const double clhs107 =             clhs106*clhs83;
const double clhs108 =             2*N[0]*clhs102;
const double clhs109 =             clhs107*clhs108;
const double clhs110 =             clhs105 + clhs109 + clhs13 + clhs18 + clhs26 + clhs34 + clhs42 + clhs50 + clhs61 + clhs71 + clhs80 + clhs87 + clhs93;
const double clhs111 =             clhs110 + clhs5;
const double clhs112 =             (1.0L/2.0L)*DN(0,1)*tau2;
const double clhs113 =             N[0]*f_ext(0,1) + N[1]*f_ext(1,1) + N[2]*f_ext(2,1) + N[3]*f_ext(3,1);
const double clhs114 =             clhs113*clhs3;
const double clhs115 =             2*DN(0,0);
const double clhs116 =             clhs10*clhs115;
const double clhs117 =             clhs116*clhs8;
const double clhs118 =             clhs117*clhs7;
const double clhs119 =             clhs15*clhs8;
const double clhs120 =             clhs119*clhs14;
const double clhs121 =             clhs120*clhs7;
const double clhs122 =             clhs56*clhs8;
const double clhs123 =             clhs122*clhs7;
const double clhs124 =             2*N[0]*clhs10;
const double clhs125 =             clhs124*clhs66;
const double clhs126 =             clhs125*clhs7;
const double clhs127 =             DN(0,2)*U(0,2);
const double clhs128 =             DN(1,2)*U(1,2);
const double clhs129 =             DN(2,2)*U(2,2);
const double clhs130 =             DN(3,2)*U(3,2);
const double clhs131 =             clhs127 + clhs128 + clhs129 + clhs130;
const double clhs132 =             clhs131*clhs40;
const double clhs133 =             clhs132*clhs7;
const double clhs134 =             clhs48*clhs8;
const double clhs135 =             clhs134*clhs7;
const double clhs136 =             clhs57*clhs8;
const double clhs137 =             clhs136*clhs32;
const double clhs138 =             -clhs137*clhs7;
const double clhs139 =             2*N[0]*clhs10*clhs67;
const double clhs140 =             clhs139*clhs23;
const double clhs141 =             clhs140*clhs7;
const double clhs142 =             -clhs141;
const double clhs143 =             DN(0,1)*U(0,3);
const double clhs144 =             DN(1,1)*U(1,3);
const double clhs145 =             DN(2,1)*U(2,3);
const double clhs146 =             DN(3,1)*U(3,3);
const double clhs147 =             clhs143 + clhs144 + clhs145 + clhs146;
const double clhs148 =             clhs147*clhs77;
const double clhs149 =             clhs148*clhs7;
const double clhs150 =             -clhs149;
const double clhs151 =             clhs106*clhs8;
const double clhs152 =             4*N[0]*clhs10*clhs83;
const double clhs153 =             clhs151*clhs152;
const double clhs154 =             -clhs153;
const double clhs155 =             clhs8*clhs88;
const double clhs156 =             -clhs155*clhs91;
const double clhs157 =             2*clhs97;
const double clhs158 =             clhs101 - clhs157;
const double clhs159 =             DN(0,1)*clhs158;
const double clhs160 =             -clhs159*clhs7;
const double clhs161 =             2*N[0]*clhs83;
const double clhs162 =             clhs158*clhs81;
const double clhs163 =             clhs118 + clhs121 + clhs123 + clhs126 + clhs133 + clhs135 + clhs138 + clhs142 + clhs150 + clhs154 + clhs156 + clhs160 + clhs161*clhs162;
const double clhs164 =             clhs114 + clhs163;
const double clhs165 =             (1.0L/2.0L)*DN(0,2)*tau2;
const double clhs166 =             N[0]*f_ext(0,2) + N[1]*f_ext(1,2) + N[2]*f_ext(2,2) + N[3]*f_ext(3,2);
const double clhs167 =             clhs166*clhs3;
const double clhs168 =             clhs116*clhs14;
const double clhs169 =             clhs168*clhs7;
const double clhs170 =             clhs8*clhs9;
const double clhs171 =             clhs14*clhs170;
const double clhs172 =             clhs171*clhs7;
const double clhs173 =             clhs14*clhs56;
const double clhs174 =             clhs173*clhs7;
const double clhs175 =             clhs124*clhs76;
const double clhs176 =             clhs175*clhs7;
const double clhs177 =             clhs14*clhs32;
const double clhs178 =             clhs177*clhs7;
const double clhs179 =             clhs147*clhs24;
const double clhs180 =             clhs179*clhs7;
const double clhs181 =             clhs14*clhs57;
const double clhs182 =             clhs181*clhs48;
const double clhs183 =             -clhs182*clhs7;
const double clhs184 =             clhs139*clhs39;
const double clhs185 =             clhs184*clhs7;
const double clhs186 =             -clhs185;
const double clhs187 =             clhs131*clhs68;
const double clhs188 =             clhs187*clhs7;
const double clhs189 =             -clhs188;
const double clhs190 =             clhs106*clhs14;
const double clhs191 =             clhs152*clhs190;
const double clhs192 =             -clhs191;
const double clhs193 =             clhs14*clhs81;
const double clhs194 =             -clhs193*clhs85;
const double clhs195 =             2*clhs99;
const double clhs196 =             clhs101 - clhs195;
const double clhs197 =             DN(0,2)*clhs196;
const double clhs198 =             -clhs197*clhs7;
const double clhs199 =             clhs196*clhs88;
const double clhs200 =             clhs161*clhs199 + clhs169 + clhs172 + clhs174 + clhs176 + clhs178 + clhs180 + clhs183 + clhs186 + clhs189 + clhs192 + clhs194 + clhs198;
const double clhs201 =             clhs167 + clhs200;
const double clhs202 =             -DN(0,0)*N[0];
const double clhs203 =             1.0/clhs6;
const double clhs204 =             DN(0,1)*clhs203*tau2;
const double clhs205 =             DN(0,0)*clhs8;
const double clhs206 =             N[0]*clhs66;
const double clhs207 =             DN(0,1)*clhs10;
const double clhs208 =             clhs207*clhs67;
const double clhs209 =             N[0]*clhs23;
const double clhs210 =             clhs209*clhs67;
const double clhs211 =             -clhs210;
const double clhs212 =             N[0]*clhs203;
const double clhs213 =             clhs151*clhs212;
const double clhs214 =             clhs212*clhs82;
const double clhs215 =             clhs205 + clhs206 - clhs208 + clhs211 - clhs213 + clhs214*clhs67;
const double clhs216 =             DN(0,2)*clhs203*tau2;
const double clhs217 =             DN(0,0)*clhs14;
const double clhs218 =             N[0]*clhs76;
const double clhs219 =             DN(0,2)*clhs10;
const double clhs220 =             clhs219*clhs67;
const double clhs221 =             N[0]*clhs39;
const double clhs222 =             clhs221*clhs67;
const double clhs223 =             -clhs222;
const double clhs224 =             clhs190*clhs212;
const double clhs225 =             clhs212*clhs89;
const double clhs226 =             clhs217 + clhs218 - clhs220 + clhs223 - clhs224 + clhs225*clhs67;
const double clhs227 =             DN(0,0)*tau2;
const double clhs228 =             DN(0,2)*clhs14;
const double clhs229 =             clhs203*clhs228;
const double clhs230 =             N[0]*clhs47;
const double clhs231 =             clhs203*clhs230;
const double clhs232 =             N[0]*clhs7;
const double clhs233 =             clhs14*clhs88;
const double clhs234 =             clhs232*clhs233;
const double clhs235 =             -clhs229 - clhs231 + clhs234;
const double clhs236 =             N[0]*bdf0;
const double clhs237 =             -clhs236;
const double clhs238 =             DN(0,1)*clhs8;
const double clhs239 =             clhs203*clhs238;
const double clhs240 =             -clhs239;
const double clhs241 =             N[0]*clhs31;
const double clhs242 =             clhs203*clhs241;
const double clhs243 =             -clhs242;
const double clhs244 =             DN(0,0)*clhs10;
const double clhs245 =             clhs244*clhs57;
const double clhs246 =             clhs203*clhs245;
const double clhs247 =             N[0]*clhs55;
const double clhs248 =             clhs247*clhs57;
const double clhs249 =             clhs203*clhs248;
const double clhs250 =             clhs8*clhs81;
const double clhs251 =             clhs232*clhs250;
const double clhs252 =             clhs10*clhs106*clhs57;
const double clhs253 =             clhs232*clhs252;
const double clhs254 =             clhs235 + clhs237 + clhs240 + clhs243 + clhs246 + clhs249 + clhs251 - clhs253;
const double clhs255 =             -DN(0,1)*N[0];
const double clhs256 =             DN(0,0)*clhs203*tau2;
const double clhs257 =             clhs206*clhs67;
const double clhs258 =             -clhs209 + clhs257;
const double clhs259 =             clhs205*clhs67 - clhs207 - clhs213*clhs67 + clhs214 + clhs258;
const double clhs260 =             DN(0,1)*clhs14;
const double clhs261 =             N[0]*clhs147;
const double clhs262 =             DN(0,2)*clhs8;
const double clhs263 =             clhs262*clhs67;
const double clhs264 =             N[0]*clhs131;
const double clhs265 =             clhs264*clhs67;
const double clhs266 =             -clhs265;
const double clhs267 =             clhs193*clhs212;
const double clhs268 =             clhs155*clhs212;
const double clhs269 =             clhs260 + clhs261 - clhs263 + clhs266 - clhs267 + clhs268*clhs67;
const double clhs270 =             DN(0,1)*tau2;
const double clhs271 =             clhs229 + clhs231 - clhs234;
const double clhs272 =             clhs203*clhs244;
const double clhs273 =             clhs203*clhs247;
const double clhs274 =             clhs238*clhs57;
const double clhs275 =             clhs203*clhs274;
const double clhs276 =             clhs241*clhs57;
const double clhs277 =             clhs203*clhs276;
const double clhs278 =             clhs10*clhs106;
const double clhs279 =             clhs232*clhs278;
const double clhs280 =             -clhs279;
const double clhs281 =             clhs57*clhs8*clhs81;
const double clhs282 =             clhs232*clhs281;
const double clhs283 =             clhs236 + clhs271 + clhs272 + clhs273 - clhs275 - clhs277 + clhs280 + clhs282;
const double clhs284 =             -DN(0,2)*N[0];
const double clhs285 =             clhs218*clhs67;
const double clhs286 =             -clhs221 + clhs285;
const double clhs287 =             clhs217*clhs67 - clhs219 - clhs224*clhs67 + clhs225 + clhs286;
const double clhs288 =             clhs261*clhs67;
const double clhs289 =             -clhs264 + clhs288;
const double clhs290 =             clhs260*clhs67 - clhs262 - clhs267*clhs67 + clhs268 + clhs289;
const double clhs291 =             DN(0,2)*tau2;
const double clhs292 =             -clhs251;
const double clhs293 =             clhs228*clhs57;
const double clhs294 =             clhs203*clhs293;
const double clhs295 =             clhs230*clhs57;
const double clhs296 =             clhs203*clhs295;
const double clhs297 =             clhs14*clhs57*clhs88;
const double clhs298 =             clhs232*clhs297;
const double clhs299 =             clhs236 + clhs239 + clhs242 + clhs272 + clhs273 + clhs280 + clhs292 - clhs294 - clhs296 + clhs298;
const double clhs300 =             clhs67*tau2;
const double clhs301 =             -N[1]*clhs236;
const double clhs302 =             2*N[1];
const double clhs303 =             clhs302*clhs4;
const double clhs304 =             2*DN(1,1);
const double clhs305 =             clhs10*clhs304;
const double clhs306 =             clhs305*clhs8;
const double clhs307 =             clhs306*clhs7;
const double clhs308 =             2*DN(1,2);
const double clhs309 =             clhs10*clhs308;
const double clhs310 =             clhs14*clhs309;
const double clhs311 =             clhs310*clhs7;
const double clhs312 =             2*N[1]*clhs8;
const double clhs313 =             clhs23*clhs312;
const double clhs314 =             clhs313*clhs7;
const double clhs315 =             clhs302*clhs31;
const double clhs316 =             clhs10*clhs315;
const double clhs317 =             clhs316*clhs7;
const double clhs318 =             2*N[1]*clhs14;
const double clhs319 =             clhs318*clhs39;
const double clhs320 =             clhs319*clhs7;
const double clhs321 =             clhs302*clhs47;
const double clhs322 =             clhs10*clhs321;
const double clhs323 =             clhs322*clhs7;
const double clhs324 =             clhs302*clhs55;
const double clhs325 =             clhs324*clhs58;
const double clhs326 =             clhs325*clhs7;
const double clhs327 =             -clhs326;
const double clhs328 =             2*N[1]*clhs67*clhs8;
const double clhs329 =             clhs328*clhs66;
const double clhs330 =             clhs329*clhs7;
const double clhs331 =             -clhs330;
const double clhs332 =             2*N[1]*clhs14*clhs67;
const double clhs333 =             clhs332*clhs76;
const double clhs334 =             clhs333*clhs7;
const double clhs335 =             -clhs334;
const double clhs336 =             N[1]*clhs83*clhs84;
const double clhs337 =             clhs336*clhs82;
const double clhs338 =             -clhs337;
const double clhs339 =             N[1]*clhs83*clhs90;
const double clhs340 =             clhs339*clhs89;
const double clhs341 =             -clhs340;
const double clhs342 =             DN(1,0)*clhs102;
const double clhs343 =             clhs342*clhs7;
const double clhs344 =             -clhs343;
const double clhs345 =             2*N[1]*clhs102;
const double clhs346 =             clhs107*clhs345;
const double clhs347 =             clhs307 + clhs311 + clhs314 + clhs317 + clhs320 + clhs323 + clhs327 + clhs331 + clhs335 + clhs338 + clhs341 + clhs344 + clhs346;
const double clhs348 =             clhs303 + clhs347;
const double clhs349 =             clhs113*clhs302;
const double clhs350 =             2*DN(1,0);
const double clhs351 =             clhs10*clhs350;
const double clhs352 =             clhs351*clhs8;
const double clhs353 =             clhs352*clhs7;
const double clhs354 =             clhs308*clhs8;
const double clhs355 =             clhs14*clhs354;
const double clhs356 =             clhs355*clhs7;
const double clhs357 =             clhs324*clhs8;
const double clhs358 =             clhs357*clhs7;
const double clhs359 =             2*N[1]*clhs10;
const double clhs360 =             clhs359*clhs66;
const double clhs361 =             clhs360*clhs7;
const double clhs362 =             clhs131*clhs318;
const double clhs363 =             clhs362*clhs7;
const double clhs364 =             clhs321*clhs8;
const double clhs365 =             clhs364*clhs7;
const double clhs366 =             clhs136*clhs315;
const double clhs367 =             -clhs366*clhs7;
const double clhs368 =             2*N[1]*clhs10*clhs67;
const double clhs369 =             clhs23*clhs368;
const double clhs370 =             clhs369*clhs7;
const double clhs371 =             -clhs370;
const double clhs372 =             clhs147*clhs332;
const double clhs373 =             clhs372*clhs7;
const double clhs374 =             -clhs373;
const double clhs375 =             4*N[0]*U(0,1) + 4*N[1]*U(1,1) + 4*N[2]*U(2,1) + 4*N[3]*U(3,1);
const double clhs376 =             N[1]*clhs375*clhs83;
const double clhs377 =             -clhs151*clhs376;
const double clhs378 =             -clhs155*clhs339;
const double clhs379 =             DN(1,1)*clhs158;
const double clhs380 =             -clhs379*clhs7;
const double clhs381 =             2*N[1]*clhs83;
const double clhs382 =             clhs162*clhs381 + clhs353 + clhs356 + clhs358 + clhs361 + clhs363 + clhs365 + clhs367 + clhs371 + clhs374 + clhs377 + clhs378 + clhs380;
const double clhs383 =             clhs349 + clhs382;
const double clhs384 =             clhs166*clhs302;
const double clhs385 =             clhs14*clhs351;
const double clhs386 =             clhs385*clhs7;
const double clhs387 =             clhs304*clhs8;
const double clhs388 =             clhs14*clhs387;
const double clhs389 =             clhs388*clhs7;
const double clhs390 =             clhs14*clhs324;
const double clhs391 =             clhs390*clhs7;
const double clhs392 =             clhs359*clhs76;
const double clhs393 =             clhs392*clhs7;
const double clhs394 =             clhs14*clhs315;
const double clhs395 =             clhs394*clhs7;
const double clhs396 =             clhs147*clhs312;
const double clhs397 =             clhs396*clhs7;
const double clhs398 =             clhs181*clhs321;
const double clhs399 =             -clhs398*clhs7;
const double clhs400 =             clhs368*clhs39;
const double clhs401 =             clhs400*clhs7;
const double clhs402 =             -clhs401;
const double clhs403 =             clhs131*clhs328;
const double clhs404 =             clhs403*clhs7;
const double clhs405 =             -clhs404;
const double clhs406 =             -clhs190*clhs376;
const double clhs407 =             -clhs193*clhs336;
const double clhs408 =             DN(1,2)*clhs196;
const double clhs409 =             -clhs408*clhs7;
const double clhs410 =             clhs199*clhs381 + clhs386 + clhs389 + clhs391 + clhs393 + clhs395 + clhs397 + clhs399 + clhs402 + clhs405 + clhs406 + clhs407 + clhs409;
const double clhs411 =             clhs384 + clhs410;
const double clhs412 =             DN(1,0)*N[0];
const double clhs413 =             -clhs412;
const double clhs414 =             DN(1,0)*clhs8;
const double clhs415 =             N[1]*clhs66;
const double clhs416 =             DN(1,1)*clhs10;
const double clhs417 =             clhs416*clhs67;
const double clhs418 =             N[1]*clhs23;
const double clhs419 =             clhs418*clhs67;
const double clhs420 =             -clhs419;
const double clhs421 =             N[1]*clhs203;
const double clhs422 =             clhs151*clhs421;
const double clhs423 =             clhs421*clhs82;
const double clhs424 =             clhs414 + clhs415 - clhs417 + clhs420 - clhs422 + clhs423*clhs67;
const double clhs425 =             DN(1,0)*clhs14;
const double clhs426 =             N[1]*clhs76;
const double clhs427 =             DN(1,2)*clhs10;
const double clhs428 =             clhs427*clhs67;
const double clhs429 =             N[1]*clhs39;
const double clhs430 =             clhs429*clhs67;
const double clhs431 =             -clhs430;
const double clhs432 =             clhs190*clhs421;
const double clhs433 =             clhs421*clhs89;
const double clhs434 =             clhs425 + clhs426 - clhs428 + clhs431 - clhs432 + clhs433*clhs67;
const double clhs435 =             DN(1,2)*clhs14;
const double clhs436 =             clhs203*clhs435;
const double clhs437 =             N[1]*clhs47;
const double clhs438 =             clhs203*clhs437;
const double clhs439 =             N[1]*clhs7;
const double clhs440 =             clhs233*clhs439;
const double clhs441 =             -clhs436 - clhs438 + clhs440;
const double clhs442 =             N[1]*bdf0;
const double clhs443 =             -clhs442;
const double clhs444 =             DN(1,1)*clhs8;
const double clhs445 =             clhs203*clhs444;
const double clhs446 =             -clhs445;
const double clhs447 =             N[1]*clhs31;
const double clhs448 =             clhs203*clhs447;
const double clhs449 =             -clhs448;
const double clhs450 =             DN(1,0)*clhs10;
const double clhs451 =             clhs450*clhs57;
const double clhs452 =             clhs203*clhs451;
const double clhs453 =             N[1]*clhs55;
const double clhs454 =             clhs453*clhs57;
const double clhs455 =             clhs203*clhs454;
const double clhs456 =             clhs250*clhs439;
const double clhs457 =             clhs252*clhs439;
const double clhs458 =             clhs441 + clhs443 + clhs446 + clhs449 + clhs452 + clhs455 + clhs456 - clhs457;
const double clhs459 =             DN(1,1)*N[0];
const double clhs460 =             -clhs459;
const double clhs461 =             clhs415*clhs67;
const double clhs462 =             -clhs418 + clhs461;
const double clhs463 =             clhs414*clhs67 - clhs416 - clhs422*clhs67 + clhs423 + clhs462;
const double clhs464 =             DN(1,1)*clhs14;
const double clhs465 =             N[1]*clhs147;
const double clhs466 =             DN(1,2)*clhs8;
const double clhs467 =             clhs466*clhs67;
const double clhs468 =             N[1]*clhs131;
const double clhs469 =             clhs468*clhs67;
const double clhs470 =             -clhs469;
const double clhs471 =             clhs193*clhs421;
const double clhs472 =             clhs155*clhs421;
const double clhs473 =             clhs464 + clhs465 - clhs467 + clhs470 - clhs471 + clhs472*clhs67;
const double clhs474 =             clhs436 + clhs438 - clhs440;
const double clhs475 =             clhs203*clhs450;
const double clhs476 =             clhs203*clhs453;
const double clhs477 =             clhs444*clhs57;
const double clhs478 =             clhs203*clhs477;
const double clhs479 =             clhs447*clhs57;
const double clhs480 =             clhs203*clhs479;
const double clhs481 =             clhs278*clhs439;
const double clhs482 =             -clhs481;
const double clhs483 =             clhs281*clhs439;
const double clhs484 =             clhs442 + clhs474 + clhs475 + clhs476 - clhs478 - clhs480 + clhs482 + clhs483;
const double clhs485 =             DN(1,2)*N[0];
const double clhs486 =             -clhs485;
const double clhs487 =             clhs426*clhs67;
const double clhs488 =             -clhs429 + clhs487;
const double clhs489 =             clhs425*clhs67 - clhs427 - clhs432*clhs67 + clhs433 + clhs488;
const double clhs490 =             clhs465*clhs67;
const double clhs491 =             -clhs468 + clhs490;
const double clhs492 =             clhs464*clhs67 - clhs466 - clhs471*clhs67 + clhs472 + clhs491;
const double clhs493 =             -clhs456;
const double clhs494 =             clhs435*clhs57;
const double clhs495 =             clhs203*clhs494;
const double clhs496 =             clhs437*clhs57;
const double clhs497 =             clhs203*clhs496;
const double clhs498 =             clhs297*clhs439;
const double clhs499 =             clhs442 + clhs445 + clhs448 + clhs475 + clhs476 + clhs482 + clhs493 - clhs495 - clhs497 + clhs498;
const double clhs500 =             -clhs300*(DN(0,0)*DN(1,0) + DN(0,1)*DN(1,1) + DN(0,2)*DN(1,2));
const double clhs501 =             -N[2]*clhs236;
const double clhs502 =             2*N[2];
const double clhs503 =             clhs4*clhs502;
const double clhs504 =             2*DN(2,1);
const double clhs505 =             clhs10*clhs504;
const double clhs506 =             clhs505*clhs8;
const double clhs507 =             clhs506*clhs7;
const double clhs508 =             2*DN(2,2);
const double clhs509 =             clhs10*clhs508;
const double clhs510 =             clhs14*clhs509;
const double clhs511 =             clhs510*clhs7;
const double clhs512 =             2*N[2]*clhs8;
const double clhs513 =             clhs23*clhs512;
const double clhs514 =             clhs513*clhs7;
const double clhs515 =             clhs31*clhs502;
const double clhs516 =             clhs10*clhs515;
const double clhs517 =             clhs516*clhs7;
const double clhs518 =             2*N[2]*clhs14;
const double clhs519 =             clhs39*clhs518;
const double clhs520 =             clhs519*clhs7;
const double clhs521 =             clhs47*clhs502;
const double clhs522 =             clhs10*clhs521;
const double clhs523 =             clhs522*clhs7;
const double clhs524 =             clhs502*clhs55;
const double clhs525 =             clhs524*clhs58;
const double clhs526 =             clhs525*clhs7;
const double clhs527 =             -clhs526;
const double clhs528 =             2*N[2]*clhs67*clhs8;
const double clhs529 =             clhs528*clhs66;
const double clhs530 =             clhs529*clhs7;
const double clhs531 =             -clhs530;
const double clhs532 =             2*N[2]*clhs14*clhs67;
const double clhs533 =             clhs532*clhs76;
const double clhs534 =             clhs533*clhs7;
const double clhs535 =             -clhs534;
const double clhs536 =             N[2]*clhs83*clhs84;
const double clhs537 =             clhs536*clhs82;
const double clhs538 =             -clhs537;
const double clhs539 =             N[2]*clhs83*clhs90;
const double clhs540 =             clhs539*clhs89;
const double clhs541 =             -clhs540;
const double clhs542 =             DN(2,0)*clhs102;
const double clhs543 =             clhs542*clhs7;
const double clhs544 =             -clhs543;
const double clhs545 =             2*N[2]*clhs102;
const double clhs546 =             clhs107*clhs545;
const double clhs547 =             clhs507 + clhs511 + clhs514 + clhs517 + clhs520 + clhs523 + clhs527 + clhs531 + clhs535 + clhs538 + clhs541 + clhs544 + clhs546;
const double clhs548 =             clhs503 + clhs547;
const double clhs549 =             clhs113*clhs502;
const double clhs550 =             2*DN(2,0);
const double clhs551 =             clhs10*clhs550;
const double clhs552 =             clhs551*clhs8;
const double clhs553 =             clhs552*clhs7;
const double clhs554 =             clhs508*clhs8;
const double clhs555 =             clhs14*clhs554;
const double clhs556 =             clhs555*clhs7;
const double clhs557 =             clhs524*clhs8;
const double clhs558 =             clhs557*clhs7;
const double clhs559 =             2*N[2]*clhs10;
const double clhs560 =             clhs559*clhs66;
const double clhs561 =             clhs560*clhs7;
const double clhs562 =             clhs131*clhs518;
const double clhs563 =             clhs562*clhs7;
const double clhs564 =             clhs521*clhs8;
const double clhs565 =             clhs564*clhs7;
const double clhs566 =             clhs136*clhs515;
const double clhs567 =             -clhs566*clhs7;
const double clhs568 =             2*N[2]*clhs10*clhs67;
const double clhs569 =             clhs23*clhs568;
const double clhs570 =             clhs569*clhs7;
const double clhs571 =             -clhs570;
const double clhs572 =             clhs147*clhs532;
const double clhs573 =             clhs572*clhs7;
const double clhs574 =             -clhs573;
const double clhs575 =             N[2]*clhs375*clhs83;
const double clhs576 =             -clhs151*clhs575;
const double clhs577 =             -clhs155*clhs539;
const double clhs578 =             DN(2,1)*clhs158;
const double clhs579 =             -clhs578*clhs7;
const double clhs580 =             2*N[2]*clhs83;
const double clhs581 =             clhs162*clhs580 + clhs553 + clhs556 + clhs558 + clhs561 + clhs563 + clhs565 + clhs567 + clhs571 + clhs574 + clhs576 + clhs577 + clhs579;
const double clhs582 =             clhs549 + clhs581;
const double clhs583 =             clhs166*clhs502;
const double clhs584 =             clhs14*clhs551;
const double clhs585 =             clhs584*clhs7;
const double clhs586 =             clhs504*clhs8;
const double clhs587 =             clhs14*clhs586;
const double clhs588 =             clhs587*clhs7;
const double clhs589 =             clhs14*clhs524;
const double clhs590 =             clhs589*clhs7;
const double clhs591 =             clhs559*clhs76;
const double clhs592 =             clhs591*clhs7;
const double clhs593 =             clhs14*clhs515;
const double clhs594 =             clhs593*clhs7;
const double clhs595 =             clhs147*clhs512;
const double clhs596 =             clhs595*clhs7;
const double clhs597 =             clhs181*clhs521;
const double clhs598 =             -clhs597*clhs7;
const double clhs599 =             clhs39*clhs568;
const double clhs600 =             clhs599*clhs7;
const double clhs601 =             -clhs600;
const double clhs602 =             clhs131*clhs528;
const double clhs603 =             clhs602*clhs7;
const double clhs604 =             -clhs603;
const double clhs605 =             -clhs190*clhs575;
const double clhs606 =             -clhs193*clhs536;
const double clhs607 =             DN(2,2)*clhs196;
const double clhs608 =             -clhs607*clhs7;
const double clhs609 =             clhs199*clhs580 + clhs585 + clhs588 + clhs590 + clhs592 + clhs594 + clhs596 + clhs598 + clhs601 + clhs604 + clhs605 + clhs606 + clhs608;
const double clhs610 =             clhs583 + clhs609;
const double clhs611 =             DN(2,0)*N[0];
const double clhs612 =             -clhs611;
const double clhs613 =             DN(2,0)*clhs8;
const double clhs614 =             N[2]*clhs66;
const double clhs615 =             DN(2,1)*clhs10;
const double clhs616 =             clhs615*clhs67;
const double clhs617 =             N[2]*clhs23;
const double clhs618 =             clhs617*clhs67;
const double clhs619 =             -clhs618;
const double clhs620 =             N[2]*clhs203;
const double clhs621 =             clhs151*clhs620;
const double clhs622 =             clhs620*clhs82;
const double clhs623 =             clhs613 + clhs614 - clhs616 + clhs619 - clhs621 + clhs622*clhs67;
const double clhs624 =             DN(2,0)*clhs14;
const double clhs625 =             N[2]*clhs76;
const double clhs626 =             DN(2,2)*clhs10;
const double clhs627 =             clhs626*clhs67;
const double clhs628 =             N[2]*clhs39;
const double clhs629 =             clhs628*clhs67;
const double clhs630 =             -clhs629;
const double clhs631 =             clhs190*clhs620;
const double clhs632 =             clhs620*clhs89;
const double clhs633 =             clhs624 + clhs625 - clhs627 + clhs630 - clhs631 + clhs632*clhs67;
const double clhs634 =             DN(2,2)*clhs14;
const double clhs635 =             clhs203*clhs634;
const double clhs636 =             N[2]*clhs47;
const double clhs637 =             clhs203*clhs636;
const double clhs638 =             N[2]*clhs7;
const double clhs639 =             clhs233*clhs638;
const double clhs640 =             -clhs635 - clhs637 + clhs639;
const double clhs641 =             N[2]*bdf0;
const double clhs642 =             -clhs641;
const double clhs643 =             DN(2,1)*clhs8;
const double clhs644 =             clhs203*clhs643;
const double clhs645 =             -clhs644;
const double clhs646 =             N[2]*clhs31;
const double clhs647 =             clhs203*clhs646;
const double clhs648 =             -clhs647;
const double clhs649 =             DN(2,0)*clhs10;
const double clhs650 =             clhs57*clhs649;
const double clhs651 =             clhs203*clhs650;
const double clhs652 =             N[2]*clhs55;
const double clhs653 =             clhs57*clhs652;
const double clhs654 =             clhs203*clhs653;
const double clhs655 =             clhs250*clhs638;
const double clhs656 =             clhs252*clhs638;
const double clhs657 =             clhs640 + clhs642 + clhs645 + clhs648 + clhs651 + clhs654 + clhs655 - clhs656;
const double clhs658 =             DN(2,1)*N[0];
const double clhs659 =             -clhs658;
const double clhs660 =             clhs614*clhs67;
const double clhs661 =             -clhs617 + clhs660;
const double clhs662 =             clhs613*clhs67 - clhs615 - clhs621*clhs67 + clhs622 + clhs661;
const double clhs663 =             DN(2,1)*clhs14;
const double clhs664 =             N[2]*clhs147;
const double clhs665 =             DN(2,2)*clhs8;
const double clhs666 =             clhs665*clhs67;
const double clhs667 =             N[2]*clhs131;
const double clhs668 =             clhs667*clhs67;
const double clhs669 =             -clhs668;
const double clhs670 =             clhs193*clhs620;
const double clhs671 =             clhs155*clhs620;
const double clhs672 =             clhs663 + clhs664 - clhs666 + clhs669 + clhs67*clhs671 - clhs670;
const double clhs673 =             clhs635 + clhs637 - clhs639;
const double clhs674 =             clhs203*clhs649;
const double clhs675 =             clhs203*clhs652;
const double clhs676 =             clhs57*clhs643;
const double clhs677 =             clhs203*clhs676;
const double clhs678 =             clhs57*clhs646;
const double clhs679 =             clhs203*clhs678;
const double clhs680 =             clhs278*clhs638;
const double clhs681 =             -clhs680;
const double clhs682 =             clhs281*clhs638;
const double clhs683 =             clhs641 + clhs673 + clhs674 + clhs675 - clhs677 - clhs679 + clhs681 + clhs682;
const double clhs684 =             DN(2,2)*N[0];
const double clhs685 =             -clhs684;
const double clhs686 =             clhs625*clhs67;
const double clhs687 =             -clhs628 + clhs686;
const double clhs688 =             clhs624*clhs67 - clhs626 - clhs631*clhs67 + clhs632 + clhs687;
const double clhs689 =             clhs664*clhs67;
const double clhs690 =             -clhs667 + clhs689;
const double clhs691 =             clhs663*clhs67 - clhs665 - clhs67*clhs670 + clhs671 + clhs690;
const double clhs692 =             -clhs655;
const double clhs693 =             clhs57*clhs634;
const double clhs694 =             clhs203*clhs693;
const double clhs695 =             clhs57*clhs636;
const double clhs696 =             clhs203*clhs695;
const double clhs697 =             clhs297*clhs638;
const double clhs698 =             clhs641 + clhs644 + clhs647 + clhs674 + clhs675 + clhs681 + clhs692 - clhs694 - clhs696 + clhs697;
const double clhs699 =             -clhs300*(DN(0,0)*DN(2,0) + DN(0,1)*DN(2,1) + DN(0,2)*DN(2,2));
const double clhs700 =             -N[3]*clhs236;
const double clhs701 =             2*N[3];
const double clhs702 =             clhs4*clhs701;
const double clhs703 =             2*DN(3,1);
const double clhs704 =             clhs10*clhs703;
const double clhs705 =             clhs704*clhs8;
const double clhs706 =             clhs7*clhs705;
const double clhs707 =             2*DN(3,2);
const double clhs708 =             clhs10*clhs707;
const double clhs709 =             clhs14*clhs708;
const double clhs710 =             clhs7*clhs709;
const double clhs711 =             2*N[3]*clhs8;
const double clhs712 =             clhs23*clhs711;
const double clhs713 =             clhs7*clhs712;
const double clhs714 =             clhs31*clhs701;
const double clhs715 =             clhs10*clhs714;
const double clhs716 =             clhs7*clhs715;
const double clhs717 =             2*N[3]*clhs14;
const double clhs718 =             clhs39*clhs717;
const double clhs719 =             clhs7*clhs718;
const double clhs720 =             clhs47*clhs701;
const double clhs721 =             clhs10*clhs720;
const double clhs722 =             clhs7*clhs721;
const double clhs723 =             clhs55*clhs701;
const double clhs724 =             clhs58*clhs723;
const double clhs725 =             clhs7*clhs724;
const double clhs726 =             -clhs725;
const double clhs727 =             2*N[3]*clhs67*clhs8;
const double clhs728 =             clhs66*clhs727;
const double clhs729 =             clhs7*clhs728;
const double clhs730 =             -clhs729;
const double clhs731 =             2*N[3]*clhs14*clhs67;
const double clhs732 =             clhs731*clhs76;
const double clhs733 =             clhs7*clhs732;
const double clhs734 =             -clhs733;
const double clhs735 =             N[3]*clhs83*clhs84;
const double clhs736 =             clhs735*clhs82;
const double clhs737 =             -clhs736;
const double clhs738 =             N[3]*clhs83*clhs90;
const double clhs739 =             clhs738*clhs89;
const double clhs740 =             -clhs739;
const double clhs741 =             DN(3,0)*clhs102;
const double clhs742 =             clhs7*clhs741;
const double clhs743 =             -clhs742;
const double clhs744 =             2*N[3]*clhs102;
const double clhs745 =             clhs107*clhs744;
const double clhs746 =             clhs706 + clhs710 + clhs713 + clhs716 + clhs719 + clhs722 + clhs726 + clhs730 + clhs734 + clhs737 + clhs740 + clhs743 + clhs745;
const double clhs747 =             clhs702 + clhs746;
const double clhs748 =             clhs113*clhs701;
const double clhs749 =             2*DN(3,0);
const double clhs750 =             clhs10*clhs749;
const double clhs751 =             clhs750*clhs8;
const double clhs752 =             clhs7*clhs751;
const double clhs753 =             clhs707*clhs8;
const double clhs754 =             clhs14*clhs753;
const double clhs755 =             clhs7*clhs754;
const double clhs756 =             clhs723*clhs8;
const double clhs757 =             clhs7*clhs756;
const double clhs758 =             2*N[3]*clhs10;
const double clhs759 =             clhs66*clhs758;
const double clhs760 =             clhs7*clhs759;
const double clhs761 =             clhs131*clhs717;
const double clhs762 =             clhs7*clhs761;
const double clhs763 =             clhs720*clhs8;
const double clhs764 =             clhs7*clhs763;
const double clhs765 =             clhs136*clhs714;
const double clhs766 =             -clhs7*clhs765;
const double clhs767 =             2*N[3]*clhs10*clhs67;
const double clhs768 =             clhs23*clhs767;
const double clhs769 =             clhs7*clhs768;
const double clhs770 =             -clhs769;
const double clhs771 =             clhs147*clhs731;
const double clhs772 =             clhs7*clhs771;
const double clhs773 =             -clhs772;
const double clhs774 =             N[3]*clhs375*clhs83;
const double clhs775 =             -clhs151*clhs774;
const double clhs776 =             -clhs155*clhs738;
const double clhs777 =             DN(3,1)*clhs158;
const double clhs778 =             -clhs7*clhs777;
const double clhs779 =             2*N[3]*clhs83;
const double clhs780 =             clhs162*clhs779 + clhs752 + clhs755 + clhs757 + clhs760 + clhs762 + clhs764 + clhs766 + clhs770 + clhs773 + clhs775 + clhs776 + clhs778;
const double clhs781 =             clhs748 + clhs780;
const double clhs782 =             clhs166*clhs701;
const double clhs783 =             clhs14*clhs750;
const double clhs784 =             clhs7*clhs783;
const double clhs785 =             clhs703*clhs8;
const double clhs786 =             clhs14*clhs785;
const double clhs787 =             clhs7*clhs786;
const double clhs788 =             clhs14*clhs723;
const double clhs789 =             clhs7*clhs788;
const double clhs790 =             clhs758*clhs76;
const double clhs791 =             clhs7*clhs790;
const double clhs792 =             clhs14*clhs714;
const double clhs793 =             clhs7*clhs792;
const double clhs794 =             clhs147*clhs711;
const double clhs795 =             clhs7*clhs794;
const double clhs796 =             clhs181*clhs720;
const double clhs797 =             -clhs7*clhs796;
const double clhs798 =             clhs39*clhs767;
const double clhs799 =             clhs7*clhs798;
const double clhs800 =             -clhs799;
const double clhs801 =             clhs131*clhs727;
const double clhs802 =             clhs7*clhs801;
const double clhs803 =             -clhs802;
const double clhs804 =             -clhs190*clhs774;
const double clhs805 =             -clhs193*clhs735;
const double clhs806 =             DN(3,2)*clhs196;
const double clhs807 =             -clhs7*clhs806;
const double clhs808 =             clhs199*clhs779 + clhs784 + clhs787 + clhs789 + clhs791 + clhs793 + clhs795 + clhs797 + clhs800 + clhs803 + clhs804 + clhs805 + clhs807;
const double clhs809 =             clhs782 + clhs808;
const double clhs810 =             DN(3,0)*N[0];
const double clhs811 =             -clhs810;
const double clhs812 =             DN(3,0)*clhs8;
const double clhs813 =             N[3]*clhs66;
const double clhs814 =             DN(3,1)*clhs10;
const double clhs815 =             clhs67*clhs814;
const double clhs816 =             N[3]*clhs23;
const double clhs817 =             clhs67*clhs816;
const double clhs818 =             -clhs817;
const double clhs819 =             N[3]*clhs203;
const double clhs820 =             clhs151*clhs819;
const double clhs821 =             clhs819*clhs82;
const double clhs822 =             clhs67*clhs821 + clhs812 + clhs813 - clhs815 + clhs818 - clhs820;
const double clhs823 =             DN(3,0)*clhs14;
const double clhs824 =             N[3]*clhs76;
const double clhs825 =             DN(3,2)*clhs10;
const double clhs826 =             clhs67*clhs825;
const double clhs827 =             N[3]*clhs39;
const double clhs828 =             clhs67*clhs827;
const double clhs829 =             -clhs828;
const double clhs830 =             clhs190*clhs819;
const double clhs831 =             clhs819*clhs89;
const double clhs832 =             clhs67*clhs831 + clhs823 + clhs824 - clhs826 + clhs829 - clhs830;
const double clhs833 =             DN(3,2)*clhs14;
const double clhs834 =             clhs203*clhs833;
const double clhs835 =             N[3]*clhs47;
const double clhs836 =             clhs203*clhs835;
const double clhs837 =             N[3]*clhs7;
const double clhs838 =             clhs233*clhs837;
const double clhs839 =             -clhs834 - clhs836 + clhs838;
const double clhs840 =             N[3]*bdf0;
const double clhs841 =             -clhs840;
const double clhs842 =             DN(3,1)*clhs8;
const double clhs843 =             clhs203*clhs842;
const double clhs844 =             -clhs843;
const double clhs845 =             N[3]*clhs31;
const double clhs846 =             clhs203*clhs845;
const double clhs847 =             -clhs846;
const double clhs848 =             DN(3,0)*clhs10;
const double clhs849 =             clhs57*clhs848;
const double clhs850 =             clhs203*clhs849;
const double clhs851 =             N[3]*clhs55;
const double clhs852 =             clhs57*clhs851;
const double clhs853 =             clhs203*clhs852;
const double clhs854 =             clhs250*clhs837;
const double clhs855 =             clhs252*clhs837;
const double clhs856 =             clhs839 + clhs841 + clhs844 + clhs847 + clhs850 + clhs853 + clhs854 - clhs855;
const double clhs857 =             DN(3,1)*N[0];
const double clhs858 =             -clhs857;
const double clhs859 =             clhs67*clhs813;
const double clhs860 =             -clhs816 + clhs859;
const double clhs861 =             clhs67*clhs812 - clhs67*clhs820 - clhs814 + clhs821 + clhs860;
const double clhs862 =             DN(3,1)*clhs14;
const double clhs863 =             N[3]*clhs147;
const double clhs864 =             DN(3,2)*clhs8;
const double clhs865 =             clhs67*clhs864;
const double clhs866 =             N[3]*clhs131;
const double clhs867 =             clhs67*clhs866;
const double clhs868 =             -clhs867;
const double clhs869 =             clhs193*clhs819;
const double clhs870 =             clhs155*clhs819;
const double clhs871 =             clhs67*clhs870 + clhs862 + clhs863 - clhs865 + clhs868 - clhs869;
const double clhs872 =             clhs834 + clhs836 - clhs838;
const double clhs873 =             clhs203*clhs848;
const double clhs874 =             clhs203*clhs851;
const double clhs875 =             clhs57*clhs842;
const double clhs876 =             clhs203*clhs875;
const double clhs877 =             clhs57*clhs845;
const double clhs878 =             clhs203*clhs877;
const double clhs879 =             clhs278*clhs837;
const double clhs880 =             -clhs879;
const double clhs881 =             clhs281*clhs837;
const double clhs882 =             clhs840 + clhs872 + clhs873 + clhs874 - clhs876 - clhs878 + clhs880 + clhs881;
const double clhs883 =             DN(3,2)*N[0];
const double clhs884 =             -clhs883;
const double clhs885 =             clhs67*clhs824;
const double clhs886 =             -clhs827 + clhs885;
const double clhs887 =             clhs67*clhs823 - clhs67*clhs830 - clhs825 + clhs831 + clhs886;
const double clhs888 =             clhs67*clhs863;
const double clhs889 =             -clhs866 + clhs888;
const double clhs890 =             clhs67*clhs862 - clhs67*clhs869 - clhs864 + clhs870 + clhs889;
const double clhs891 =             -clhs854;
const double clhs892 =             clhs57*clhs833;
const double clhs893 =             clhs203*clhs892;
const double clhs894 =             clhs57*clhs835;
const double clhs895 =             clhs203*clhs894;
const double clhs896 =             clhs297*clhs837;
const double clhs897 =             clhs840 + clhs843 + clhs846 + clhs873 + clhs874 + clhs880 + clhs891 - clhs893 - clhs895 + clhs896;
const double clhs898 =             -clhs300*(DN(0,0)*DN(3,0) + DN(0,1)*DN(3,1) + DN(0,2)*DN(3,2));
const double clhs899 =             clhs0*clhs4;
const double clhs900 =             DN(0,1)*N[0]*clhs203*v_sc;
const double clhs901 =             clhs151 + clhs82;
const double clhs902 =             -clhs19 - clhs20 + clhs203*clhs901 - clhs21 - clhs22 - clhs62 - clhs63 - clhs64 - clhs65;
const double clhs903 =             DN(0,2)*N[0]*clhs203*v_sc;
const double clhs904 =             clhs190 + clhs89;
const double clhs905 =             clhs203*clhs904 - clhs35 - clhs36 - clhs37 - clhs38 - clhs72 - clhs73 - clhs74 - clhs75;
const double clhs906 =             clhs6*v_sc/mu + 1;
const double clhs907 =             DN(0,1)*clhs7*clhs906*mu;
const double clhs908 =             clhs23 + clhs66;
const double clhs909 =             2*N[0]*clhs203;
const double clhs910 =             N[0]*clhs908 + clhs205 + clhs207 - clhs901*clhs909;
const double clhs911 =             DN(0,2)*clhs7*clhs906*mu;
const double clhs912 =             clhs39 + clhs76;
const double clhs913 =             N[0]*clhs912 + clhs217 + clhs219 - clhs904*clhs909;
const double clhs914 =             DN(0,0)*N[0]*clhs203*v_sc;
const double clhs915 =             -4.0L/3.0L*DN(0,0)*U(0,1) + (2.0L/3.0L)*DN(0,1)*U(0,2) + (2.0L/3.0L)*DN(0,2)*U(0,3) - 4.0L/3.0L*DN(1,0)*U(1,1) + (2.0L/3.0L)*DN(1,1)*U(1,2) + (2.0L/3.0L)*DN(1,2)*U(1,3) - 4.0L/3.0L*DN(2,0)*U(2,1) + (2.0L/3.0L)*DN(2,1)*U(2,2) + (2.0L/3.0L)*DN(2,2)*U(2,3) - 4.0L/3.0L*DN(3,0)*U(3,1) + (2.0L/3.0L)*DN(3,1)*U(3,2) + (2.0L/3.0L)*DN(3,2)*U(3,3) + (4.0L/3.0L)*clhs10*clhs106*clhs203 - 2.0L/3.0L*clhs14*clhs203*clhs88 - 2.0L/3.0L*clhs203*clhs8*clhs81;
const double clhs916 =             (2.0L/3.0L)*DN(0,0)*clhs7*clhs906*mu;
const double clhs917 =             2*N[0]*clhs14*clhs203;
const double clhs918 =             clhs88*clhs917;
const double clhs919 =             -clhs918;
const double clhs920 =             clhs228 + clhs230 + clhs919;
const double clhs921 =             clhs106*clhs212;
const double clhs922 =             clhs10*clhs921;
const double clhs923 =             2*N[0]*clhs203*clhs8;
const double clhs924 =             clhs81*clhs923;
const double clhs925 =             -clhs924;
const double clhs926 =             -clhs116 + clhs238 + clhs241 - clhs56 + clhs920 + 4*clhs922 + clhs925;
const double clhs927 =             (1.0L/2.0L)*N[0]*clhs7;
const double clhs928 =             clhs106*clhs203;
const double clhs929 =             clhs103 - clhs108*clhs928 - clhs12 - clhs17 + clhs214*clhs84 + clhs225*clhs90 - clhs25 - clhs33 - clhs41 - clhs49 + clhs59 + clhs69 + clhs78;
const double clhs930 =             (1.0L/2.0L)*N[0]*bdf0*tau1;
const double clhs931 =             clhs104 - clhs13 - clhs18 - clhs26 - clhs34 - clhs42 - clhs50 + clhs60 + clhs70 + clhs79 + clhs86 + clhs92;
const double clhs932 =             -clhs109 + clhs5 + clhs931;
const double clhs933 =             clhs27 + clhs28 + clhs29 + clhs30 + clhs43 + clhs44 + clhs45 + clhs46;
const double clhs934 =             N[0]*(U(0,0)*bdf0 + Un(0,0)*bdf1 + Unn(0,0)*bdf2) + N[1]*(U(1,0)*bdf0 + Un(1,0)*bdf1 + Unn(1,0)*bdf2) + N[2]*(U(2,0)*bdf0 + Un(2,0)*bdf1 + Unn(2,0)*bdf2) + N[3]*(U(3,0)*bdf0 + Un(3,0)*bdf1 + Unn(3,0)*bdf2) + clhs51 + clhs52 + clhs53 + clhs54 + clhs933;
const double clhs935 =             N[0]*clhs83*clhs934*tau1;
const double clhs936 =             6*N[0]*U(0,2) + 6*N[1]*U(1,2) + 6*N[2]*U(2,2) + 6*N[3]*U(3,2);
const double clhs937 =             6*N[0]*U(0,3) + 6*N[1]*U(1,3) + 6*N[2]*U(2,3) + 6*N[3]*U(3,3);
const double clhs938 =             -clhs100 - clhs96 - clhs98;
const double clhs939 =             clhs938 + clhs95;
const double clhs940 =             3*clhs67*clhs94;
const double clhs941 =             3*clhs67*clhs97;
const double clhs942 =             3*clhs67*clhs99;
const double clhs943 =             -6*clhs94 + clhs940 + clhs941 + clhs942;
const double clhs944 =             clhs115*clhs8;
const double clhs945 =             clhs82*clhs909;
const double clhs946 =             clhs151*clhs909;
const double clhs947 =             -clhs11 + clhs258 + clhs67*clhs944 - clhs67*clhs946 + clhs945;
const double clhs948 =             clhs3*(U(0,2)*bdf0 + Un(0,2)*bdf1 + Unn(0,2)*bdf2);
const double clhs949 =             clhs302*(U(1,2)*bdf0 + Un(1,2)*bdf1 + Unn(1,2)*bdf2);
const double clhs950 =             clhs502*(U(2,2)*bdf0 + Un(2,2)*bdf1 + Unn(2,2)*bdf2);
const double clhs951 =             clhs701*(U(3,2)*bdf0 + Un(3,2)*bdf1 + Unn(3,2)*bdf2);
const double clhs952 =             2*gamma - 2;
const double clhs953 =             DN(0,1)*U(0,4);
const double clhs954 =             DN(1,1)*U(1,4);
const double clhs955 =             DN(2,1)*U(2,4);
const double clhs956 =             DN(3,1)*U(3,4);
const double clhs957 =             clhs953 + clhs954 + clhs955 + clhs956;
const double clhs958 =             clhs952*clhs957;
const double clhs959 =             2*N[0]*U(0,0) + 2*N[1]*U(1,0) + 2*N[2]*U(2,0) + 2*N[3]*U(3,0);
const double clhs960 =             clhs113*clhs959;
const double clhs961 =             2*N[0]*U(0,2) + 2*N[1]*U(1,2) + 2*N[2]*U(2,2) + 2*N[3]*U(3,2);
const double clhs962 =             clhs203*clhs55;
const double clhs963 =             clhs961*clhs962;
const double clhs964 =             2*N[0]*U(0,1) + 2*N[1]*U(1,1) + 2*N[2]*U(2,1) + 2*N[3]*U(3,1);
const double clhs965 =             clhs203*clhs964;
const double clhs966 =             clhs66*clhs965;
const double clhs967 =             2*N[0]*U(0,3) + 2*N[1]*U(1,3) + 2*N[2]*U(2,3) + 2*N[3]*U(3,3);
const double clhs968 =             clhs203*clhs967;
const double clhs969 =             clhs131*clhs968;
const double clhs970 =             clhs203*clhs961;
const double clhs971 =             clhs47*clhs970;
const double clhs972 =             clhs31*clhs57*clhs970;
const double clhs973 =             clhs10*clhs203;
const double clhs974 =             clhs23*clhs973;
const double clhs975 =             clhs952*clhs974;
const double clhs976 =             clhs14*clhs203;
const double clhs977 =             clhs147*clhs976;
const double clhs978 =             clhs952*clhs977;
const double clhs979 =             clhs7*clhs964;
const double clhs980 =             clhs151*clhs979;
const double clhs981 =             clhs7*clhs967;
const double clhs982 =             clhs155*clhs981;
const double clhs983 =             clhs7*clhs81;
const double clhs984 =             clhs158*clhs983;
const double clhs985 =             clhs948 + clhs949 + clhs950 + clhs951 + clhs958 - clhs960 + clhs963 + clhs966 + clhs969 + clhs971 - clhs972 - clhs975 - clhs978 - clhs980 - clhs982 + clhs984;
const double clhs986 =             (1.0L/2.0L)*N[0]*clhs7*clhs985*tau2;
const double clhs987 =             clhs115*clhs14;
const double clhs988 =             clhs89*clhs909;
const double clhs989 =             clhs190*clhs909;
const double clhs990 =             -clhs16 + clhs286 + clhs67*clhs987 - clhs67*clhs989 + clhs988;
const double clhs991 =             clhs3*(U(0,3)*bdf0 + Un(0,3)*bdf1 + Unn(0,3)*bdf2);
const double clhs992 =             clhs302*(U(1,3)*bdf0 + Un(1,3)*bdf1 + Unn(1,3)*bdf2);
const double clhs993 =             clhs502*(U(2,3)*bdf0 + Un(2,3)*bdf1 + Unn(2,3)*bdf2);
const double clhs994 =             clhs701*(U(3,3)*bdf0 + Un(3,3)*bdf1 + Unn(3,3)*bdf2);
const double clhs995 =             DN(0,2)*U(0,4);
const double clhs996 =             DN(1,2)*U(1,4);
const double clhs997 =             DN(2,2)*U(2,4);
const double clhs998 =             DN(3,2)*U(3,4);
const double clhs999 =             clhs995 + clhs996 + clhs997 + clhs998;
const double clhs1000 =             clhs952*clhs999;
const double clhs1001 =             clhs166*clhs959;
const double clhs1002 =             clhs962*clhs967;
const double clhs1003 =             clhs76*clhs965;
const double clhs1004 =             clhs31*clhs968;
const double clhs1005 =             clhs147*clhs970;
const double clhs1006 =             clhs47*clhs57*clhs968;
const double clhs1007 =             clhs39*clhs973;
const double clhs1008 =             clhs1007*clhs952;
const double clhs1009 =             clhs203*clhs8;
const double clhs1010 =             clhs1009*clhs131;
const double clhs1011 =             clhs1010*clhs952;
const double clhs1012 =             clhs190*clhs979;
const double clhs1013 =             clhs7*clhs961;
const double clhs1014 =             clhs1013*clhs193;
const double clhs1015 =             clhs7*clhs88;
const double clhs1016 =             clhs1015*clhs196;
const double clhs1017 =             clhs1000 - clhs1001 + clhs1002 + clhs1003 + clhs1004 + clhs1005 - clhs1006 - clhs1008 - clhs1011 - clhs1012 - clhs1014 + clhs1016 + clhs991 + clhs992 + clhs993 + clhs994;
const double clhs1018 =             (1.0L/2.0L)*N[0]*clhs1017*clhs7*tau2;
const double clhs1019 =             -clhs170;
const double clhs1020 =             clhs14*clhs15;
const double clhs1021 =             -clhs1020;
const double clhs1022 =             -clhs241;
const double clhs1023 =             -clhs230;
const double clhs1024 =             clhs124*clhs928;
const double clhs1025 =             clhs1019 + clhs1021 + clhs1022 + clhs1023 - clhs1024*clhs57 + clhs116*clhs57 + clhs248 + clhs918 + clhs924;
const double clhs1026 =             clhs3*(U(0,1)*bdf0 + Un(0,1)*bdf1 + Unn(0,1)*bdf2);
const double clhs1027 =             clhs302*(U(1,1)*bdf0 + Un(1,1)*bdf1 + Unn(1,1)*bdf2);
const double clhs1028 =             clhs502*(U(2,1)*bdf0 + Un(2,1)*bdf1 + Unn(2,1)*bdf2);
const double clhs1029 =             clhs701*(U(3,1)*bdf0 + Un(3,1)*bdf1 + Unn(3,1)*bdf2);
const double clhs1030 =             DN(0,0)*U(0,4);
const double clhs1031 =             DN(1,0)*U(1,4);
const double clhs1032 =             DN(2,0)*U(2,4);
const double clhs1033 =             DN(3,0)*U(3,4);
const double clhs1034 =             clhs1030 + clhs1031 + clhs1032 + clhs1033;
const double clhs1035 =             clhs1034*clhs952;
const double clhs1036 =             clhs4*clhs959;
const double clhs1037 =             clhs23*clhs970;
const double clhs1038 =             clhs31*clhs965;
const double clhs1039 =             clhs39*clhs968;
const double clhs1040 =             clhs47*clhs965;
const double clhs1041 =             clhs57*clhs962*clhs964;
const double clhs1042 =             clhs1009*clhs66;
const double clhs1043 =             clhs1042*clhs952;
const double clhs1044 =             clhs76*clhs976;
const double clhs1045 =             clhs1044*clhs952;
const double clhs1046 =             clhs1013*clhs82;
const double clhs1047 =             clhs89*clhs981;
const double clhs1048 =             clhs106*clhs7;
const double clhs1049 =             clhs102*clhs1048;
const double clhs1050 =             clhs1026 + clhs1027 + clhs1028 + clhs1029 + clhs1035 - clhs1036 + clhs1037 + clhs1038 + clhs1039 + clhs1040 - clhs1041 - clhs1043 - clhs1045 - clhs1046 - clhs1047 + clhs1049;
const double clhs1051 =             (1.0L/2.0L)*N[0]*clhs1050*clhs7*tau2;
const double clhs1052 =             (1.0L/2.0L)*clhs203*clhs259*tau2;
const double clhs1053 =             (1.0L/2.0L)*clhs203*clhs287*tau2;
const double clhs1054 =             clhs212*clhs88;
const double clhs1055 =             clhs1054*clhs14;
const double clhs1056 =             -clhs1055 + clhs228 + clhs230;
const double clhs1057 =             clhs212*clhs81;
const double clhs1058 =             clhs1057*clhs8;
const double clhs1059 =             -clhs1058;
const double clhs1060 =             clhs58*clhs921;
const double clhs1061 =             clhs1056 + clhs1059 + clhs1060 + clhs238 + clhs241 - clhs245 - clhs248;
const double clhs1062 =             (1.0L/2.0L)*clhs1061*clhs203*tau2;
const double clhs1063 =             -clhs5;
const double clhs1064 =             clhs106*clhs83*clhs939;
const double clhs1065 =             clhs1064*clhs3;
const double clhs1066 =             clhs1063 + clhs1065 + clhs931;
const double clhs1067 =             (1.0L/2.0L)*DN(0,0)*clhs67*tau3;
const double clhs1068 =             N[0]*r[0] + N[1]*r[1] + N[2]*r[2] + N[3]*r[3];
const double clhs1069 =             clhs1068*clhs3;
const double clhs1070 =             -clhs1069;
const double clhs1071 =             2*N[0]*gamma;
const double clhs1072 =             clhs1034*clhs1071;
const double clhs1073 =             clhs10*clhs1072;
const double clhs1074 =             clhs1073*clhs7;
const double clhs1075 =             clhs1071*clhs957;
const double clhs1076 =             clhs1075*clhs8;
const double clhs1077 =             clhs1076*clhs7;
const double clhs1078 =             clhs1071*clhs999;
const double clhs1079 =             clhs1078*clhs14;
const double clhs1080 =             clhs1079*clhs7;
const double clhs1081 =             clhs10*clhs83;
const double clhs1082 =             N[0]*clhs66*clhs67*clhs84;
const double clhs1083 =             clhs1081*clhs1082;
const double clhs1084 =             N[0]*clhs67*clhs76*clhs90;
const double clhs1085 =             clhs1081*clhs1084;
const double clhs1086 =             N[0]*clhs23*clhs67*clhs84;
const double clhs1087 =             clhs1081*clhs1086;
const double clhs1088 =             clhs14*clhs83;
const double clhs1089 =             N[0]*clhs147*clhs67*clhs84;
const double clhs1090 =             clhs1088*clhs1089;
const double clhs1091 =             N[0]*clhs39*clhs67*clhs90;
const double clhs1092 =             clhs1081*clhs1091;
const double clhs1093 =             N[0]*clhs131*clhs67*clhs84;
const double clhs1094 =             clhs1088*clhs1093;
const double clhs1095 =             clhs100*clhs203;
const double clhs1096 =             N[0]*U(0,4);
const double clhs1097 =             2*clhs1096;
const double clhs1098 =             -clhs1097;
const double clhs1099 =             N[1]*U(1,4);
const double clhs1100 =             2*clhs1099;
const double clhs1101 =             -clhs1100;
const double clhs1102 =             N[2]*U(2,4);
const double clhs1103 =             2*clhs1102;
const double clhs1104 =             -clhs1103;
const double clhs1105 =             N[3]*U(3,4);
const double clhs1106 =             2*clhs1105;
const double clhs1107 =             -clhs1106;
const double clhs1108 =             clhs1096 + clhs1099 + clhs1102 + clhs1105;
const double clhs1109 =             clhs1108*clhs952;
const double clhs1110 =             -clhs1109;
const double clhs1111 =             clhs203*clhs96;
const double clhs1112 =             clhs203*clhs98;
const double clhs1113 =             clhs203*clhs67;
const double clhs1114 =             clhs94 + clhs97 + clhs99;
const double clhs1115 =             clhs1113*clhs1114;
const double clhs1116 =             clhs1098 + clhs1101 + clhs1104 + clhs1107 + clhs1110 + clhs1111 + clhs1112 + clhs1115;
const double clhs1117 =             clhs1095 + clhs1116;
const double clhs1118 =             clhs1117*clhs244;
const double clhs1119 =             clhs1118*clhs7;
const double clhs1120 =             clhs1117*clhs238;
const double clhs1121 =             clhs1120*clhs7;
const double clhs1122 =             clhs1117*clhs228;
const double clhs1123 =             clhs1122*clhs7;
const double clhs1124 =             clhs1095 + clhs1098 + clhs1101 + clhs1104 + clhs1107 + clhs1110 + clhs1112;
const double clhs1125 =             5*clhs1111 + clhs1115 + clhs1124;
const double clhs1126 =             clhs1125*clhs247;
const double clhs1127 =             clhs1126*clhs7;
const double clhs1128 =             clhs1095 + clhs1098 + clhs1101 + clhs1104 + clhs1107 + clhs1110 + clhs1111;
const double clhs1129 =             5*clhs1112 + clhs1115 + clhs1128;
const double clhs1130 =             clhs1129*clhs241;
const double clhs1131 =             clhs1130*clhs7;
const double clhs1132 =             5*clhs1095 + clhs1116;
const double clhs1133 =             clhs1132*clhs230;
const double clhs1134 =             clhs1133*clhs7;
const double clhs1135 =             -clhs1111;
const double clhs1136 =             -clhs1112;
const double clhs1137 =             -clhs1095;
const double clhs1138 =             -clhs1114*clhs203*clhs952;
const double clhs1139 =             clhs1097 + clhs1100 + clhs1103 + clhs1106 + clhs1109 + clhs1136 + clhs1137 + clhs1138;
const double clhs1140 =             clhs1135 + clhs1139;
const double clhs1141 =             clhs106*clhs1140*clhs83;
const double clhs1142 =             clhs1141*clhs124;
const double clhs1143 =             clhs1140*clhs81*clhs83;
const double clhs1144 =             clhs1143*clhs24;
const double clhs1145 =             clhs1140*clhs83*clhs88;
const double clhs1146 =             clhs1145*clhs40;
const double clhs1147 =             clhs1070 - clhs1074 - clhs1077 - clhs1080 + clhs1083 + clhs1085 + clhs1087 + clhs1090 + clhs1092 + clhs1094 + clhs1119 + clhs1121 + clhs1123 + clhs1127 + clhs1131 + clhs1134 + clhs1142 + clhs1144 + clhs1146;
const double clhs1148 =             DN(0,1)*clhs203;
const double clhs1149 =             -clhs1057;
const double clhs1150 =             DN(0,1) + clhs1149;
const double clhs1151 =             clhs1150*clhs906*mu;
const double clhs1152 =             clhs1148*clhs1151;
const double clhs1153 =             -clhs1152;
const double clhs1154 =             clhs226*clhs7*tau2;
const double clhs1155 =             clhs1154*clhs287;
const double clhs1156 =             DN(0,0)*clhs203;
const double clhs1157 =             -clhs921;
const double clhs1158 =             DN(0,0) + clhs1157;
const double clhs1159 =             clhs1158*clhs906*mu;
const double clhs1160 =             clhs1156*clhs1159;
const double clhs1161 =             DN(0,2)*clhs203;
const double clhs1162 =             -clhs1054;
const double clhs1163 =             DN(0,2) + clhs1162;
const double clhs1164 =             clhs1163*clhs906*mu;
const double clhs1165 =             clhs1161*clhs1164;
const double clhs1166 =             -clhs1165;
const double clhs1167 =             clhs215*clhs7*tau2;
const double clhs1168 =             clhs1167*clhs259;
const double clhs1169 =             N[0]*clhs7*clhs934*tau1;
const double clhs1170 =             clhs1061*clhs203*tau2;
const double clhs1171 =             clhs236 + clhs239 + clhs242 - clhs246 - clhs249 + clhs253 + clhs271 + clhs292;
const double clhs1172 =             clhs1149 + clhs9;
const double clhs1173 =             (1.0L/2.0L)*N[0]*clhs1172*clhs203*tau2;
const double clhs1174 =             clhs1162 + clhs15;
const double clhs1175 =             (1.0L/2.0L)*N[0]*clhs1174*clhs203*tau2;
const double clhs1176 =             clhs115 + clhs1157;
const double clhs1177 =             (1.0L/2.0L)*N[0]*clhs1176*clhs203*tau2;
const double clhs1178 =             clhs1050*clhs57;
const double clhs1179 =             (1.0L/2.0L)*DN(0,0)*tau1;
const double clhs1180 =             clhs11*clhs67;
const double clhs1181 =             clhs7*clhs8;
const double clhs1182 =             clhs16*clhs67;
const double clhs1183 =             clhs14*clhs7;
const double clhs1184 =             6*N[0]*U(0,1) + 6*N[1]*U(1,1) + 6*N[2]*U(2,1) + 6*N[3]*U(3,1);
const double clhs1185 =             clhs1184*clhs67*clhs7;
const double clhs1186 =             3*clhs1111;
const double clhs1187 =             clhs1124 + clhs1186;
const double clhs1188 =             clhs106*clhs1125;
const double clhs1189 =             -N[0]*clhs10*clhs67*clhs81*clhs83*clhs84 - N[0]*clhs10*clhs67*clhs83*clhs88*clhs90 - clhs1072*clhs203 + clhs1156*clhs1187 + clhs1180*clhs1181 + clhs1182*clhs1183 + clhs1185*clhs247 - clhs1188*clhs232 + clhs26*clhs67 + clhs34*clhs67 + clhs42*clhs67 + clhs50*clhs67 + clhs70 + clhs79;
const double clhs1190 =             clhs1189 + clhs5;
const double clhs1191 =             clhs1151*clhs1156;
const double clhs1192 =             clhs1148*clhs1159;
const double clhs1193 =             clhs269*clhs7*tau2;
const double clhs1194 =             clhs259*clhs7*tau2;
const double clhs1195 =             clhs203*clhs259*tau2;
const double clhs1196 =             (1.0L/2.0L)*N[0]*clhs203*clhs67*clhs985*tau2;
const double clhs1197 =             (1.0L/2.0L)*DN(0,1)*tau1;
const double clhs1198 =             clhs119*clhs67;
const double clhs1199 =             clhs67*clhs7*clhs936;
const double clhs1200 =             3*clhs1112;
const double clhs1201 =             clhs1128 + clhs1200;
const double clhs1202 =             clhs1129*clhs81;
const double clhs1203 =             -N[0]*clhs67*clhs8*clhs83*clhs88*clhs90 - clhs1075*clhs203 + clhs1148*clhs1201 + clhs118*clhs67 + clhs1183*clhs1198 + clhs1199*clhs241 - clhs1202*clhs232 + clhs123*clhs67 + clhs126*clhs67 + clhs133*clhs67 + clhs135*clhs67 + clhs141 + clhs149 - clhs153*clhs67;
const double clhs1204 =             clhs114 + clhs1203;
const double clhs1205 =             clhs1156*clhs1164;
const double clhs1206 =             clhs1159*clhs1161;
const double clhs1207 =             clhs287*clhs7*tau2;
const double clhs1208 =             clhs1023 + clhs1055 - clhs228;
const double clhs1209 =             -clhs238;
const double clhs1210 =             clhs1022 + clhs1058 - clhs1060 + clhs1208 + clhs1209 + clhs245 + clhs248;
const double clhs1211 =             clhs203*clhs287*tau2;
const double clhs1212 =             -clhs272;
const double clhs1213 =             -clhs273;
const double clhs1214 =             clhs1212 + clhs1213 + clhs237 + clhs240 + clhs243 + clhs251 + clhs279 + clhs294 + clhs296 - clhs298;
const double clhs1215 =             -clhs1026 - clhs1027 - clhs1028 - clhs1029 - clhs1035 + clhs1036 - clhs1037 - clhs1038 - clhs1039 - clhs1040 + clhs1041 + clhs1043 + clhs1045 + clhs1046 + clhs1047 - clhs1049;
const double clhs1216 =             (1.0L/2.0L)*N[0]*clhs203*clhs67*tau2;
const double clhs1217 =             -clhs1000 + clhs1001 - clhs1002 - clhs1003 - clhs1004 - clhs1005 + clhs1006 + clhs1008 + clhs1011 + clhs1012 + clhs1014 - clhs1016 - clhs991 - clhs992 - clhs993 - clhs994;
const double clhs1218 =             (1.0L/2.0L)*DN(0,2)*tau1;
const double clhs1219 =             clhs1063 + clhs110;
const double clhs1220 =             clhs67*clhs7*clhs937;
const double clhs1221 =             3*clhs1095;
const double clhs1222 =             clhs1098 + clhs1101 + clhs1104 + clhs1107 + clhs1110 + clhs1111 + clhs1112 + clhs1221;
const double clhs1223 =             clhs1132*clhs88;
const double clhs1224 =             -N[0]*clhs14*clhs67*clhs81*clhs83*clhs84 - clhs1078*clhs203 + clhs1161*clhs1222 + clhs1220*clhs230 - clhs1223*clhs232 + clhs169*clhs67 + clhs172*clhs67 + clhs174*clhs67 + clhs176*clhs67 + clhs178*clhs67 + clhs180*clhs67 + clhs185 + clhs188 - clhs191*clhs67;
const double clhs1225 =             clhs1224 + clhs167;
const double clhs1226 =             DN(0,0)*tau3;
const double clhs1227 =             clhs229*gamma + clhs231*gamma - clhs234*gamma + clhs236 + clhs239*gamma + clhs242*gamma - clhs251*gamma + clhs272*gamma + clhs273*gamma - clhs279*gamma;
const double clhs1228 =             N[0]*N[1];
const double clhs1229 =             clhs1228*clhs4;
const double clhs1230 =             DN(0,1)*N[1];
const double clhs1231 =             clhs10*clhs1230;
const double clhs1232 =             clhs10*clhs459;
const double clhs1233 =             N[1]*clhs209;
const double clhs1234 =             DN(0,0)*N[1];
const double clhs1235 =             clhs1234*clhs8;
const double clhs1236 =             clhs412*clhs8;
const double clhs1237 =             N[1]*clhs206;
const double clhs1238 =             clhs1237*clhs67;
const double clhs1239 =             clhs3*clhs423;
const double clhs1240 =             clhs3*clhs422;
const double clhs1241 =             -clhs1231 - clhs1232 - clhs1233 + clhs1235*clhs67 + clhs1236*clhs67 + clhs1238 + clhs1239 - clhs1240*clhs67;
const double clhs1242 =             (1.0L/2.0L)*clhs7*clhs985*tau2;
const double clhs1243 =             DN(0,2)*N[1];
const double clhs1244 =             clhs10*clhs1243;
const double clhs1245 =             clhs10*clhs485;
const double clhs1246 =             N[1]*clhs221;
const double clhs1247 =             clhs1234*clhs14;
const double clhs1248 =             clhs14*clhs412;
const double clhs1249 =             N[1]*clhs218;
const double clhs1250 =             clhs1249*clhs67;
const double clhs1251 =             clhs3*clhs433;
const double clhs1252 =             clhs3*clhs432;
const double clhs1253 =             -clhs1244 - clhs1245 - clhs1246 + clhs1247*clhs67 + clhs1248*clhs67 + clhs1250 + clhs1251 - clhs1252*clhs67;
const double clhs1254 =             (1.0L/2.0L)*clhs1017*clhs7*tau2;
const double clhs1255 =             clhs10*clhs1234;
const double clhs1256 =             clhs10*clhs412;
const double clhs1257 =             N[1]*clhs247;
const double clhs1258 =             clhs106*clhs421;
const double clhs1259 =             clhs124*clhs1258;
const double clhs1260 =             clhs1230*clhs8;
const double clhs1261 =             clhs1243*clhs14;
const double clhs1262 =             clhs459*clhs8;
const double clhs1263 =             clhs14*clhs485;
const double clhs1264 =             N[1]*clhs241;
const double clhs1265 =             N[1]*clhs230;
const double clhs1266 =             clhs421*clhs81;
const double clhs1267 =             clhs1266*clhs24;
const double clhs1268 =             -clhs1267;
const double clhs1269 =             clhs421*clhs88;
const double clhs1270 =             clhs1269*clhs40;
const double clhs1271 =             -clhs1270;
const double clhs1272 =             clhs1260 + clhs1261 + clhs1262 + clhs1263 + clhs1264 + clhs1265 + clhs1268 + clhs1271;
const double clhs1273 =             -clhs1255*clhs57 - clhs1256*clhs57 - clhs1257*clhs57 + clhs1259*clhs57 + clhs1272;
const double clhs1274 =             (1.0L/2.0L)*clhs1050*clhs7*tau2;
const double clhs1275 =             clhs1229 - clhs1241*clhs1242 - clhs1253*clhs1254 + clhs1273*clhs1274;
const double clhs1276 =             DN(0,1)*N[1]*clhs203*v_sc;
const double clhs1277 =             DN(0,2)*N[1]*clhs203*v_sc;
const double clhs1278 =             2*N[1]*clhs203;
const double clhs1279 =             N[1]*clhs908 - clhs1278*clhs901 + clhs414 + clhs416;
const double clhs1280 =             N[1]*clhs912 - clhs1278*clhs904 + clhs425 + clhs427;
const double clhs1281 =             DN(0,0)*N[1]*clhs203*v_sc;
const double clhs1282 =             2*N[1]*clhs14*clhs203;
const double clhs1283 =             clhs1282*clhs88;
const double clhs1284 =             -clhs1283;
const double clhs1285 =             clhs1284 + clhs435 + clhs437;
const double clhs1286 =             clhs10*clhs1258;
const double clhs1287 =             2*N[1]*clhs203*clhs8;
const double clhs1288 =             clhs1287*clhs81;
const double clhs1289 =             -clhs1288;
const double clhs1290 =             clhs1285 + 4*clhs1286 + clhs1289 - clhs324 - clhs351 + clhs444 + clhs447;
const double clhs1291 =             -clhs306 - clhs310 - clhs313 - clhs316 - clhs319 - clhs322 + clhs325 + clhs329 + clhs333 + clhs342 - clhs345*clhs928 + clhs423*clhs84 + clhs433*clhs90;
const double clhs1292 =             (1.0L/2.0L)*N[1]*bdf0*tau1;
const double clhs1293 =             clhs83*clhs934*tau1;
const double clhs1294 =             N[0]*N[1]*clhs203*clhs936;
const double clhs1295 =             N[0]*N[1]*clhs203*clhs937;
const double clhs1296 =             N[0]*N[1]*clhs203;
const double clhs1297 =             clhs106*clhs1296;
const double clhs1298 =             N[0]*clhs306 + N[0]*clhs310 + N[1]*clhs12 + N[1]*clhs17 + clhs124*clhs437 + clhs124*clhs447 - clhs124*clhs454 - clhs1294*clhs82 - clhs1295*clhs89 + clhs1297*clhs943 + clhs24*clhs418 - clhs24*clhs461 + clhs40*clhs429 - clhs40*clhs487;
const double clhs1299 =             -clhs303;
const double clhs1300 =             clhs1064*clhs302;
const double clhs1301 =             -clhs307 - clhs311 - clhs314 - clhs317 - clhs320 - clhs323 + clhs326 + clhs330 + clhs334 + clhs337 + clhs340 + clhs343;
const double clhs1302 =             clhs1299 + clhs1300 + clhs1301;
const double clhs1303 =             clhs1068*clhs302;
const double clhs1304 =             -clhs1303;
const double clhs1305 =             2*N[1]*gamma;
const double clhs1306 =             clhs1034*clhs1305;
const double clhs1307 =             clhs10*clhs1306;
const double clhs1308 =             clhs1307*clhs7;
const double clhs1309 =             clhs1305*clhs957;
const double clhs1310 =             clhs1309*clhs8;
const double clhs1311 =             clhs1310*clhs7;
const double clhs1312 =             clhs1305*clhs999;
const double clhs1313 =             clhs1312*clhs14;
const double clhs1314 =             clhs1313*clhs7;
const double clhs1315 =             N[1]*clhs66*clhs67*clhs84;
const double clhs1316 =             clhs1081*clhs1315;
const double clhs1317 =             N[1]*clhs67*clhs76*clhs90;
const double clhs1318 =             clhs1081*clhs1317;
const double clhs1319 =             N[1]*clhs23*clhs67*clhs84;
const double clhs1320 =             clhs1081*clhs1319;
const double clhs1321 =             N[1]*clhs147*clhs67*clhs84;
const double clhs1322 =             clhs1088*clhs1321;
const double clhs1323 =             N[1]*clhs39*clhs67*clhs90;
const double clhs1324 =             clhs1081*clhs1323;
const double clhs1325 =             N[1]*clhs131*clhs67*clhs84;
const double clhs1326 =             clhs1088*clhs1325;
const double clhs1327 =             clhs1117*clhs450;
const double clhs1328 =             clhs1327*clhs7;
const double clhs1329 =             clhs1117*clhs444;
const double clhs1330 =             clhs1329*clhs7;
const double clhs1331 =             clhs1117*clhs435;
const double clhs1332 =             clhs1331*clhs7;
const double clhs1333 =             clhs1125*clhs453;
const double clhs1334 =             clhs1333*clhs7;
const double clhs1335 =             clhs1129*clhs447;
const double clhs1336 =             clhs1335*clhs7;
const double clhs1337 =             clhs1132*clhs437;
const double clhs1338 =             clhs1337*clhs7;
const double clhs1339 =             clhs1141*clhs359;
const double clhs1340 =             clhs1143*clhs312;
const double clhs1341 =             clhs1145*clhs318;
const double clhs1342 =             clhs1304 - clhs1308 - clhs1311 - clhs1314 + clhs1316 + clhs1318 + clhs1320 + clhs1322 + clhs1324 + clhs1326 + clhs1328 + clhs1330 + clhs1332 + clhs1334 + clhs1336 + clhs1338 + clhs1339 + clhs1340 + clhs1341;
const double clhs1343 =             clhs7*clhs934*tau1;
const double clhs1344 =             clhs1296*clhs81;
const double clhs1345 =             clhs1230 - clhs1344 + clhs459;
const double clhs1346 =             (1.0L/2.0L)*clhs1345*clhs203*tau2;
const double clhs1347 =             clhs1346*clhs985;
const double clhs1348 =             -clhs1347;
const double clhs1349 =             clhs1296*clhs88;
const double clhs1350 =             clhs1243 - clhs1349 + clhs485;
const double clhs1351 =             (1.0L/2.0L)*clhs1350*clhs203*tau2;
const double clhs1352 =             clhs1017*clhs1351;
const double clhs1353 =             -clhs1352;
const double clhs1354 =             clhs1234 - clhs1297 + clhs412;
const double clhs1355 =             (1.0L/2.0L)*clhs1354*clhs203*tau2;
const double clhs1356 =             clhs1050*clhs1355;
const double clhs1357 =             clhs1273*clhs1343 + clhs1348 + clhs1353 + clhs1356*clhs57 + clhs301;
const double clhs1358 =             -clhs1258;
const double clhs1359 =             DN(1,0) + clhs1358;
const double clhs1360 =             clhs1359*clhs906*mu;
const double clhs1361 =             clhs1156*clhs1360;
const double clhs1362 =             -clhs1266;
const double clhs1363 =             DN(1,1) + clhs1362;
const double clhs1364 =             clhs1363*clhs906*mu;
const double clhs1365 =             clhs1148*clhs1364;
const double clhs1366 =             -clhs1365;
const double clhs1367 =             -clhs1269;
const double clhs1368 =             DN(1,2) + clhs1367;
const double clhs1369 =             clhs1368*clhs906*mu;
const double clhs1370 =             clhs1161*clhs1369;
const double clhs1371 =             -clhs1370;
const double clhs1372 =             clhs1269*clhs14;
const double clhs1373 =             -clhs1372 + clhs435 + clhs437;
const double clhs1374 =             clhs1266*clhs8;
const double clhs1375 =             -clhs1374;
const double clhs1376 =             clhs1258*clhs58;
const double clhs1377 =             clhs1373 + clhs1375 + clhs1376 + clhs444 + clhs447 - clhs451 - clhs454;
const double clhs1378 =             clhs424*clhs7*tau2;
const double clhs1379 =             clhs1378*clhs259;
const double clhs1380 =             clhs434*clhs7*tau2;
const double clhs1381 =             clhs1380*clhs287;
const double clhs1382 =             clhs442 + clhs445 + clhs448 - clhs452 - clhs455 + clhs457 + clhs474 + clhs493;
const double clhs1383 =             (1.0L/2.0L)*DN(1,0)*tau1;
const double clhs1384 =             clhs305*clhs67;
const double clhs1385 =             clhs309*clhs67;
const double clhs1386 =             DN(1,0)*clhs203;
const double clhs1387 =             -N[1]*clhs10*clhs67*clhs81*clhs83*clhs84 - N[1]*clhs10*clhs67*clhs83*clhs88*clhs90 + clhs1181*clhs1384 + clhs1183*clhs1385 + clhs1185*clhs453 + clhs1187*clhs1386 - clhs1188*clhs439 - clhs1306*clhs203 + clhs314*clhs67 + clhs317*clhs67 + clhs320*clhs67 + clhs323*clhs67 + clhs330 + clhs334;
const double clhs1388 =             clhs1387 + clhs303;
const double clhs1389 =             (1.0L/2.0L)*clhs203*clhs67*clhs985*tau2;
const double clhs1390 =             -clhs1050*clhs1346 - clhs1241*clhs1343 + clhs1354*clhs1389;
const double clhs1391 =             clhs1156*clhs1364;
const double clhs1392 =             clhs1148*clhs1360;
const double clhs1393 =             clhs473*clhs7*tau2;
const double clhs1394 =             clhs463*clhs7*tau2;
const double clhs1395 =             (1.0L/2.0L)*DN(1,1)*tau1;
const double clhs1396 =             clhs354*clhs67;
const double clhs1397 =             DN(1,1)*clhs203;
const double clhs1398 =             -N[1]*clhs106*clhs375*clhs67*clhs8*clhs83 - N[1]*clhs67*clhs8*clhs83*clhs88*clhs90 + clhs1183*clhs1396 + clhs1199*clhs447 + clhs1201*clhs1397 - clhs1202*clhs439 - clhs1309*clhs203 + clhs353*clhs67 + clhs358*clhs67 + clhs361*clhs67 + clhs363*clhs67 + clhs365*clhs67 + clhs370 + clhs373;
const double clhs1399 =             clhs1398 + clhs349;
const double clhs1400 =             (1.0L/2.0L)*clhs1217*clhs203*clhs67*tau2;
const double clhs1401 =             clhs1215*clhs1351 - clhs1253*clhs1343 - clhs1354*clhs1400;
const double clhs1402 =             clhs1156*clhs1369;
const double clhs1403 =             clhs1161*clhs1360;
const double clhs1404 =             clhs489*clhs7*tau2;
const double clhs1405 =             -clhs475;
const double clhs1406 =             -clhs476;
const double clhs1407 =             clhs1405 + clhs1406 + clhs443 + clhs446 + clhs449 + clhs456 + clhs481 + clhs495 + clhs497 - clhs498;
const double clhs1408 =             (1.0L/2.0L)*DN(1,2)*tau1;
const double clhs1409 =             DN(1,2)*clhs203;
const double clhs1410 =             -N[1]*clhs106*clhs14*clhs375*clhs67*clhs83 - N[1]*clhs14*clhs67*clhs81*clhs83*clhs84 + clhs1220*clhs437 + clhs1222*clhs1409 - clhs1223*clhs439 - clhs1312*clhs203 + clhs386*clhs67 + clhs389*clhs67 + clhs391*clhs67 + clhs393*clhs67 + clhs395*clhs67 + clhs397*clhs67 + clhs401 + clhs404;
const double clhs1411 =             clhs1410 + clhs384;
const double clhs1412 =             DN(1,1)*clhs203*tau2;
const double clhs1413 =             DN(1,2)*clhs203*tau2;
const double clhs1414 =             DN(1,0)*clhs203*tau2;
const double clhs1415 =             clhs436*gamma + clhs438*gamma - clhs440*gamma + clhs442 + clhs445*gamma + clhs448*gamma - clhs456*gamma + clhs475*gamma + clhs476*gamma - clhs481*gamma;
const double clhs1416 =             N[0]*N[2];
const double clhs1417 =             clhs1416*clhs4;
const double clhs1418 =             DN(0,1)*N[2];
const double clhs1419 =             clhs10*clhs1418;
const double clhs1420 =             clhs10*clhs658;
const double clhs1421 =             N[2]*clhs209;
const double clhs1422 =             DN(0,0)*N[2];
const double clhs1423 =             clhs1422*clhs8;
const double clhs1424 =             clhs611*clhs8;
const double clhs1425 =             N[2]*clhs206;
const double clhs1426 =             clhs1425*clhs67;
const double clhs1427 =             clhs3*clhs622;
const double clhs1428 =             clhs3*clhs621;
const double clhs1429 =             -clhs1419 - clhs1420 - clhs1421 + clhs1423*clhs67 + clhs1424*clhs67 + clhs1426 + clhs1427 - clhs1428*clhs67;
const double clhs1430 =             DN(0,2)*N[2];
const double clhs1431 =             clhs10*clhs1430;
const double clhs1432 =             clhs10*clhs684;
const double clhs1433 =             N[2]*clhs221;
const double clhs1434 =             clhs14*clhs1422;
const double clhs1435 =             clhs14*clhs611;
const double clhs1436 =             N[2]*clhs218;
const double clhs1437 =             clhs1436*clhs67;
const double clhs1438 =             clhs3*clhs632;
const double clhs1439 =             clhs3*clhs631;
const double clhs1440 =             -clhs1431 - clhs1432 - clhs1433 + clhs1434*clhs67 + clhs1435*clhs67 + clhs1437 + clhs1438 - clhs1439*clhs67;
const double clhs1441 =             clhs10*clhs1422;
const double clhs1442 =             clhs10*clhs611;
const double clhs1443 =             N[2]*clhs247;
const double clhs1444 =             clhs106*clhs620;
const double clhs1445 =             clhs124*clhs1444;
const double clhs1446 =             clhs1418*clhs8;
const double clhs1447 =             clhs14*clhs1430;
const double clhs1448 =             clhs658*clhs8;
const double clhs1449 =             clhs14*clhs684;
const double clhs1450 =             N[2]*clhs241;
const double clhs1451 =             N[2]*clhs230;
const double clhs1452 =             clhs620*clhs81;
const double clhs1453 =             clhs1452*clhs24;
const double clhs1454 =             -clhs1453;
const double clhs1455 =             clhs620*clhs88;
const double clhs1456 =             clhs1455*clhs40;
const double clhs1457 =             -clhs1456;
const double clhs1458 =             clhs1446 + clhs1447 + clhs1448 + clhs1449 + clhs1450 + clhs1451 + clhs1454 + clhs1457;
const double clhs1459 =             -clhs1441*clhs57 - clhs1442*clhs57 - clhs1443*clhs57 + clhs1445*clhs57 + clhs1458;
const double clhs1460 =             -clhs1242*clhs1429 - clhs1254*clhs1440 + clhs1274*clhs1459 + clhs1417;
const double clhs1461 =             DN(0,1)*N[2]*clhs203*v_sc;
const double clhs1462 =             DN(0,2)*N[2]*clhs203*v_sc;
const double clhs1463 =             2*N[2]*clhs203;
const double clhs1464 =             N[2]*clhs908 - clhs1463*clhs901 + clhs613 + clhs615;
const double clhs1465 =             N[2]*clhs912 - clhs1463*clhs904 + clhs624 + clhs626;
const double clhs1466 =             DN(0,0)*N[2]*clhs203*v_sc;
const double clhs1467 =             2*N[2]*clhs14*clhs203;
const double clhs1468 =             clhs1467*clhs88;
const double clhs1469 =             -clhs1468;
const double clhs1470 =             clhs1469 + clhs634 + clhs636;
const double clhs1471 =             clhs10*clhs1444;
const double clhs1472 =             2*N[2]*clhs203*clhs8;
const double clhs1473 =             clhs1472*clhs81;
const double clhs1474 =             -clhs1473;
const double clhs1475 =             clhs1470 + 4*clhs1471 + clhs1474 - clhs524 - clhs551 + clhs643 + clhs646;
const double clhs1476 =             -clhs506 - clhs510 - clhs513 - clhs516 - clhs519 - clhs522 + clhs525 + clhs529 + clhs533 + clhs542 - clhs545*clhs928 + clhs622*clhs84 + clhs632*clhs90;
const double clhs1477 =             (1.0L/2.0L)*N[2]*bdf0*tau1;
const double clhs1478 =             N[0]*N[2]*clhs203*clhs936;
const double clhs1479 =             N[0]*N[2]*clhs203*clhs937;
const double clhs1480 =             N[0]*N[2]*clhs203;
const double clhs1481 =             clhs106*clhs1480;
const double clhs1482 =             N[0]*clhs506 + N[0]*clhs510 + N[2]*clhs12 + N[2]*clhs17 + clhs124*clhs636 + clhs124*clhs646 - clhs124*clhs653 - clhs1478*clhs82 - clhs1479*clhs89 + clhs1481*clhs943 + clhs24*clhs617 - clhs24*clhs660 + clhs40*clhs628 - clhs40*clhs686;
const double clhs1483 =             -clhs503;
const double clhs1484 =             clhs1064*clhs502;
const double clhs1485 =             -clhs507 - clhs511 - clhs514 - clhs517 - clhs520 - clhs523 + clhs526 + clhs530 + clhs534 + clhs537 + clhs540 + clhs543;
const double clhs1486 =             clhs1483 + clhs1484 + clhs1485;
const double clhs1487 =             clhs1068*clhs502;
const double clhs1488 =             -clhs1487;
const double clhs1489 =             2*N[2]*gamma;
const double clhs1490 =             clhs1034*clhs1489;
const double clhs1491 =             clhs10*clhs1490;
const double clhs1492 =             clhs1491*clhs7;
const double clhs1493 =             clhs1489*clhs957;
const double clhs1494 =             clhs1493*clhs8;
const double clhs1495 =             clhs1494*clhs7;
const double clhs1496 =             clhs1489*clhs999;
const double clhs1497 =             clhs14*clhs1496;
const double clhs1498 =             clhs1497*clhs7;
const double clhs1499 =             N[2]*clhs66*clhs67*clhs84;
const double clhs1500 =             clhs1081*clhs1499;
const double clhs1501 =             N[2]*clhs67*clhs76*clhs90;
const double clhs1502 =             clhs1081*clhs1501;
const double clhs1503 =             N[2]*clhs23*clhs67*clhs84;
const double clhs1504 =             clhs1081*clhs1503;
const double clhs1505 =             N[2]*clhs147*clhs67*clhs84;
const double clhs1506 =             clhs1088*clhs1505;
const double clhs1507 =             N[2]*clhs39*clhs67*clhs90;
const double clhs1508 =             clhs1081*clhs1507;
const double clhs1509 =             N[2]*clhs131*clhs67*clhs84;
const double clhs1510 =             clhs1088*clhs1509;
const double clhs1511 =             clhs1117*clhs649;
const double clhs1512 =             clhs1511*clhs7;
const double clhs1513 =             clhs1117*clhs643;
const double clhs1514 =             clhs1513*clhs7;
const double clhs1515 =             clhs1117*clhs634;
const double clhs1516 =             clhs1515*clhs7;
const double clhs1517 =             clhs1125*clhs652;
const double clhs1518 =             clhs1517*clhs7;
const double clhs1519 =             clhs1129*clhs646;
const double clhs1520 =             clhs1519*clhs7;
const double clhs1521 =             clhs1132*clhs636;
const double clhs1522 =             clhs1521*clhs7;
const double clhs1523 =             clhs1141*clhs559;
const double clhs1524 =             clhs1143*clhs512;
const double clhs1525 =             clhs1145*clhs518;
const double clhs1526 =             clhs1488 - clhs1492 - clhs1495 - clhs1498 + clhs1500 + clhs1502 + clhs1504 + clhs1506 + clhs1508 + clhs1510 + clhs1512 + clhs1514 + clhs1516 + clhs1518 + clhs1520 + clhs1522 + clhs1523 + clhs1524 + clhs1525;
const double clhs1527 =             clhs1480*clhs81;
const double clhs1528 =             clhs1418 - clhs1527 + clhs658;
const double clhs1529 =             (1.0L/2.0L)*clhs1528*clhs203*tau2;
const double clhs1530 =             clhs1529*clhs985;
const double clhs1531 =             -clhs1530;
const double clhs1532 =             clhs1480*clhs88;
const double clhs1533 =             clhs1430 - clhs1532 + clhs684;
const double clhs1534 =             (1.0L/2.0L)*clhs1533*clhs203*tau2;
const double clhs1535 =             clhs1017*clhs1534;
const double clhs1536 =             -clhs1535;
const double clhs1537 =             clhs1422 - clhs1481 + clhs611;
const double clhs1538 =             (1.0L/2.0L)*clhs1537*clhs203*tau2;
const double clhs1539 =             clhs1050*clhs1538;
const double clhs1540 =             clhs1343*clhs1459 + clhs1531 + clhs1536 + clhs1539*clhs57 + clhs501;
const double clhs1541 =             -clhs1444;
const double clhs1542 =             DN(2,0) + clhs1541;
const double clhs1543 =             clhs1542*clhs906*mu;
const double clhs1544 =             clhs1156*clhs1543;
const double clhs1545 =             -clhs1452;
const double clhs1546 =             DN(2,1) + clhs1545;
const double clhs1547 =             clhs1546*clhs906*mu;
const double clhs1548 =             clhs1148*clhs1547;
const double clhs1549 =             -clhs1548;
const double clhs1550 =             -clhs1455;
const double clhs1551 =             DN(2,2) + clhs1550;
const double clhs1552 =             clhs1551*clhs906*mu;
const double clhs1553 =             clhs1161*clhs1552;
const double clhs1554 =             -clhs1553;
const double clhs1555 =             clhs14*clhs1455;
const double clhs1556 =             -clhs1555 + clhs634 + clhs636;
const double clhs1557 =             clhs1452*clhs8;
const double clhs1558 =             -clhs1557;
const double clhs1559 =             clhs1444*clhs58;
const double clhs1560 =             clhs1556 + clhs1558 + clhs1559 + clhs643 + clhs646 - clhs650 - clhs653;
const double clhs1561 =             clhs623*clhs7*tau2;
const double clhs1562 =             clhs1561*clhs259;
const double clhs1563 =             clhs633*clhs7*tau2;
const double clhs1564 =             clhs1563*clhs287;
const double clhs1565 =             clhs641 + clhs644 + clhs647 - clhs651 - clhs654 + clhs656 + clhs673 + clhs692;
const double clhs1566 =             (1.0L/2.0L)*DN(2,0)*tau1;
const double clhs1567 =             clhs505*clhs67;
const double clhs1568 =             clhs509*clhs67;
const double clhs1569 =             DN(2,0)*clhs203;
const double clhs1570 =             -N[2]*clhs10*clhs67*clhs81*clhs83*clhs84 - N[2]*clhs10*clhs67*clhs83*clhs88*clhs90 + clhs1181*clhs1567 + clhs1183*clhs1568 + clhs1185*clhs652 + clhs1187*clhs1569 - clhs1188*clhs638 - clhs1490*clhs203 + clhs514*clhs67 + clhs517*clhs67 + clhs520*clhs67 + clhs523*clhs67 + clhs530 + clhs534;
const double clhs1571 =             clhs1570 + clhs503;
const double clhs1572 =             -clhs1050*clhs1529 - clhs1343*clhs1429 + clhs1389*clhs1537;
const double clhs1573 =             clhs1156*clhs1547;
const double clhs1574 =             clhs1148*clhs1543;
const double clhs1575 =             clhs672*clhs7*tau2;
const double clhs1576 =             clhs662*clhs7*tau2;
const double clhs1577 =             (1.0L/2.0L)*DN(2,1)*tau1;
const double clhs1578 =             clhs554*clhs67;
const double clhs1579 =             DN(2,1)*clhs203;
const double clhs1580 =             -N[2]*clhs106*clhs375*clhs67*clhs8*clhs83 - N[2]*clhs67*clhs8*clhs83*clhs88*clhs90 + clhs1183*clhs1578 + clhs1199*clhs646 + clhs1201*clhs1579 - clhs1202*clhs638 - clhs1493*clhs203 + clhs553*clhs67 + clhs558*clhs67 + clhs561*clhs67 + clhs563*clhs67 + clhs565*clhs67 + clhs570 + clhs573;
const double clhs1581 =             clhs1580 + clhs549;
const double clhs1582 =             clhs1215*clhs1534 - clhs1343*clhs1440 - clhs1400*clhs1537;
const double clhs1583 =             clhs1156*clhs1552;
const double clhs1584 =             clhs1161*clhs1543;
const double clhs1585 =             clhs688*clhs7*tau2;
const double clhs1586 =             -clhs674;
const double clhs1587 =             -clhs675;
const double clhs1588 =             clhs1586 + clhs1587 + clhs642 + clhs645 + clhs648 + clhs655 + clhs680 + clhs694 + clhs696 - clhs697;
const double clhs1589 =             (1.0L/2.0L)*DN(2,2)*tau1;
const double clhs1590 =             DN(2,2)*clhs203;
const double clhs1591 =             -N[2]*clhs106*clhs14*clhs375*clhs67*clhs83 - N[2]*clhs14*clhs67*clhs81*clhs83*clhs84 + clhs1220*clhs636 + clhs1222*clhs1590 - clhs1223*clhs638 - clhs1496*clhs203 + clhs585*clhs67 + clhs588*clhs67 + clhs590*clhs67 + clhs592*clhs67 + clhs594*clhs67 + clhs596*clhs67 + clhs600 + clhs603;
const double clhs1592 =             clhs1591 + clhs583;
const double clhs1593 =             DN(2,1)*clhs203*tau2;
const double clhs1594 =             DN(2,2)*clhs203*tau2;
const double clhs1595 =             DN(2,0)*clhs203*tau2;
const double clhs1596 =             clhs635*gamma + clhs637*gamma - clhs639*gamma + clhs641 + clhs644*gamma + clhs647*gamma - clhs655*gamma + clhs674*gamma + clhs675*gamma - clhs680*gamma;
const double clhs1597 =             N[0]*N[3];
const double clhs1598 =             clhs1597*clhs4;
const double clhs1599 =             DN(0,1)*N[3];
const double clhs1600 =             clhs10*clhs1599;
const double clhs1601 =             clhs10*clhs857;
const double clhs1602 =             N[3]*clhs209;
const double clhs1603 =             DN(0,0)*N[3];
const double clhs1604 =             clhs1603*clhs8;
const double clhs1605 =             clhs8*clhs810;
const double clhs1606 =             N[3]*clhs206;
const double clhs1607 =             clhs1606*clhs67;
const double clhs1608 =             clhs3*clhs821;
const double clhs1609 =             clhs3*clhs820;
const double clhs1610 =             -clhs1600 - clhs1601 - clhs1602 + clhs1604*clhs67 + clhs1605*clhs67 + clhs1607 + clhs1608 - clhs1609*clhs67;
const double clhs1611 =             DN(0,2)*N[3];
const double clhs1612 =             clhs10*clhs1611;
const double clhs1613 =             clhs10*clhs883;
const double clhs1614 =             N[3]*clhs221;
const double clhs1615 =             clhs14*clhs1603;
const double clhs1616 =             clhs14*clhs810;
const double clhs1617 =             N[3]*clhs218;
const double clhs1618 =             clhs1617*clhs67;
const double clhs1619 =             clhs3*clhs831;
const double clhs1620 =             clhs3*clhs830;
const double clhs1621 =             -clhs1612 - clhs1613 - clhs1614 + clhs1615*clhs67 + clhs1616*clhs67 + clhs1618 + clhs1619 - clhs1620*clhs67;
const double clhs1622 =             clhs10*clhs1603;
const double clhs1623 =             clhs10*clhs810;
const double clhs1624 =             N[3]*clhs247;
const double clhs1625 =             clhs106*clhs819;
const double clhs1626 =             clhs124*clhs1625;
const double clhs1627 =             clhs1599*clhs8;
const double clhs1628 =             clhs14*clhs1611;
const double clhs1629 =             clhs8*clhs857;
const double clhs1630 =             clhs14*clhs883;
const double clhs1631 =             N[3]*clhs241;
const double clhs1632 =             N[3]*clhs230;
const double clhs1633 =             clhs81*clhs819;
const double clhs1634 =             clhs1633*clhs24;
const double clhs1635 =             -clhs1634;
const double clhs1636 =             clhs819*clhs88;
const double clhs1637 =             clhs1636*clhs40;
const double clhs1638 =             -clhs1637;
const double clhs1639 =             clhs1627 + clhs1628 + clhs1629 + clhs1630 + clhs1631 + clhs1632 + clhs1635 + clhs1638;
const double clhs1640 =             -clhs1622*clhs57 - clhs1623*clhs57 - clhs1624*clhs57 + clhs1626*clhs57 + clhs1639;
const double clhs1641 =             -clhs1242*clhs1610 - clhs1254*clhs1621 + clhs1274*clhs1640 + clhs1598;
const double clhs1642 =             DN(0,1)*N[3]*clhs203*v_sc;
const double clhs1643 =             DN(0,2)*N[3]*clhs203*v_sc;
const double clhs1644 =             2*N[3]*clhs203;
const double clhs1645 =             N[3]*clhs908 - clhs1644*clhs901 + clhs812 + clhs814;
const double clhs1646 =             N[3]*clhs912 - clhs1644*clhs904 + clhs823 + clhs825;
const double clhs1647 =             DN(0,0)*N[3]*clhs203*v_sc;
const double clhs1648 =             2*N[3]*clhs14*clhs203;
const double clhs1649 =             clhs1648*clhs88;
const double clhs1650 =             -clhs1649;
const double clhs1651 =             clhs1650 + clhs833 + clhs835;
const double clhs1652 =             clhs10*clhs1625;
const double clhs1653 =             2*N[3]*clhs203*clhs8;
const double clhs1654 =             clhs1653*clhs81;
const double clhs1655 =             -clhs1654;
const double clhs1656 =             clhs1651 + 4*clhs1652 + clhs1655 - clhs723 - clhs750 + clhs842 + clhs845;
const double clhs1657 =             -clhs705 - clhs709 - clhs712 - clhs715 - clhs718 - clhs721 + clhs724 + clhs728 + clhs732 + clhs741 - clhs744*clhs928 + clhs821*clhs84 + clhs831*clhs90;
const double clhs1658 =             (1.0L/2.0L)*N[3]*bdf0*tau1;
const double clhs1659 =             N[0]*N[3]*clhs203*clhs936;
const double clhs1660 =             N[0]*N[3]*clhs203*clhs937;
const double clhs1661 =             N[0]*N[3]*clhs203;
const double clhs1662 =             clhs106*clhs1661;
const double clhs1663 =             N[0]*clhs705 + N[0]*clhs709 + N[3]*clhs12 + N[3]*clhs17 + clhs124*clhs835 + clhs124*clhs845 - clhs124*clhs852 - clhs1659*clhs82 - clhs1660*clhs89 + clhs1662*clhs943 + clhs24*clhs816 - clhs24*clhs859 + clhs40*clhs827 - clhs40*clhs885;
const double clhs1664 =             -clhs702;
const double clhs1665 =             clhs1064*clhs701;
const double clhs1666 =             -clhs706 - clhs710 - clhs713 - clhs716 - clhs719 - clhs722 + clhs725 + clhs729 + clhs733 + clhs736 + clhs739 + clhs742;
const double clhs1667 =             clhs1664 + clhs1665 + clhs1666;
const double clhs1668 =             clhs1068*clhs701;
const double clhs1669 =             -clhs1668;
const double clhs1670 =             2*N[3]*gamma;
const double clhs1671 =             clhs1034*clhs1670;
const double clhs1672 =             clhs10*clhs1671;
const double clhs1673 =             clhs1672*clhs7;
const double clhs1674 =             clhs1670*clhs957;
const double clhs1675 =             clhs1674*clhs8;
const double clhs1676 =             clhs1675*clhs7;
const double clhs1677 =             clhs1670*clhs999;
const double clhs1678 =             clhs14*clhs1677;
const double clhs1679 =             clhs1678*clhs7;
const double clhs1680 =             N[3]*clhs66*clhs67*clhs84;
const double clhs1681 =             clhs1081*clhs1680;
const double clhs1682 =             N[3]*clhs67*clhs76*clhs90;
const double clhs1683 =             clhs1081*clhs1682;
const double clhs1684 =             N[3]*clhs23*clhs67*clhs84;
const double clhs1685 =             clhs1081*clhs1684;
const double clhs1686 =             N[3]*clhs147*clhs67*clhs84;
const double clhs1687 =             clhs1088*clhs1686;
const double clhs1688 =             N[3]*clhs39*clhs67*clhs90;
const double clhs1689 =             clhs1081*clhs1688;
const double clhs1690 =             N[3]*clhs131*clhs67*clhs84;
const double clhs1691 =             clhs1088*clhs1690;
const double clhs1692 =             clhs1117*clhs848;
const double clhs1693 =             clhs1692*clhs7;
const double clhs1694 =             clhs1117*clhs842;
const double clhs1695 =             clhs1694*clhs7;
const double clhs1696 =             clhs1117*clhs833;
const double clhs1697 =             clhs1696*clhs7;
const double clhs1698 =             clhs1125*clhs851;
const double clhs1699 =             clhs1698*clhs7;
const double clhs1700 =             clhs1129*clhs845;
const double clhs1701 =             clhs1700*clhs7;
const double clhs1702 =             clhs1132*clhs835;
const double clhs1703 =             clhs1702*clhs7;
const double clhs1704 =             clhs1141*clhs758;
const double clhs1705 =             clhs1143*clhs711;
const double clhs1706 =             clhs1145*clhs717;
const double clhs1707 =             clhs1669 - clhs1673 - clhs1676 - clhs1679 + clhs1681 + clhs1683 + clhs1685 + clhs1687 + clhs1689 + clhs1691 + clhs1693 + clhs1695 + clhs1697 + clhs1699 + clhs1701 + clhs1703 + clhs1704 + clhs1705 + clhs1706;
const double clhs1708 =             clhs1661*clhs81;
const double clhs1709 =             clhs1599 - clhs1708 + clhs857;
const double clhs1710 =             (1.0L/2.0L)*clhs1709*clhs203*tau2;
const double clhs1711 =             clhs1710*clhs985;
const double clhs1712 =             -clhs1711;
const double clhs1713 =             clhs1661*clhs88;
const double clhs1714 =             clhs1611 - clhs1713 + clhs883;
const double clhs1715 =             (1.0L/2.0L)*clhs1714*clhs203*tau2;
const double clhs1716 =             clhs1017*clhs1715;
const double clhs1717 =             -clhs1716;
const double clhs1718 =             clhs1603 - clhs1662 + clhs810;
const double clhs1719 =             (1.0L/2.0L)*clhs1718*clhs203*tau2;
const double clhs1720 =             clhs1050*clhs1719;
const double clhs1721 =             clhs1343*clhs1640 + clhs1712 + clhs1717 + clhs1720*clhs57 + clhs700;
const double clhs1722 =             -clhs1625;
const double clhs1723 =             DN(3,0) + clhs1722;
const double clhs1724 =             clhs1723*clhs906*mu;
const double clhs1725 =             clhs1156*clhs1724;
const double clhs1726 =             -clhs1633;
const double clhs1727 =             DN(3,1) + clhs1726;
const double clhs1728 =             clhs1727*clhs906*mu;
const double clhs1729 =             clhs1148*clhs1728;
const double clhs1730 =             -clhs1729;
const double clhs1731 =             -clhs1636;
const double clhs1732 =             DN(3,2) + clhs1731;
const double clhs1733 =             clhs1732*clhs906*mu;
const double clhs1734 =             clhs1161*clhs1733;
const double clhs1735 =             -clhs1734;
const double clhs1736 =             clhs14*clhs1636;
const double clhs1737 =             -clhs1736 + clhs833 + clhs835;
const double clhs1738 =             clhs1633*clhs8;
const double clhs1739 =             -clhs1738;
const double clhs1740 =             clhs1625*clhs58;
const double clhs1741 =             clhs1737 + clhs1739 + clhs1740 + clhs842 + clhs845 - clhs849 - clhs852;
const double clhs1742 =             clhs7*clhs822*tau2;
const double clhs1743 =             clhs1742*clhs259;
const double clhs1744 =             clhs7*clhs832*tau2;
const double clhs1745 =             clhs1744*clhs287;
const double clhs1746 =             clhs840 + clhs843 + clhs846 - clhs850 - clhs853 + clhs855 + clhs872 + clhs891;
const double clhs1747 =             (1.0L/2.0L)*DN(3,0)*tau1;
const double clhs1748 =             clhs67*clhs704;
const double clhs1749 =             clhs67*clhs708;
const double clhs1750 =             DN(3,0)*clhs203;
const double clhs1751 =             -N[3]*clhs10*clhs67*clhs81*clhs83*clhs84 - N[3]*clhs10*clhs67*clhs83*clhs88*clhs90 + clhs1181*clhs1748 + clhs1183*clhs1749 + clhs1185*clhs851 + clhs1187*clhs1750 - clhs1188*clhs837 - clhs1671*clhs203 + clhs67*clhs713 + clhs67*clhs716 + clhs67*clhs719 + clhs67*clhs722 + clhs729 + clhs733;
const double clhs1752 =             clhs1751 + clhs702;
const double clhs1753 =             -clhs1050*clhs1710 - clhs1343*clhs1610 + clhs1389*clhs1718;
const double clhs1754 =             clhs1156*clhs1728;
const double clhs1755 =             clhs1148*clhs1724;
const double clhs1756 =             clhs7*clhs871*tau2;
const double clhs1757 =             clhs7*clhs861*tau2;
const double clhs1758 =             (1.0L/2.0L)*DN(3,1)*tau1;
const double clhs1759 =             clhs67*clhs753;
const double clhs1760 =             DN(3,1)*clhs203;
const double clhs1761 =             -N[3]*clhs106*clhs375*clhs67*clhs8*clhs83 - N[3]*clhs67*clhs8*clhs83*clhs88*clhs90 + clhs1183*clhs1759 + clhs1199*clhs845 + clhs1201*clhs1760 - clhs1202*clhs837 - clhs1674*clhs203 + clhs67*clhs752 + clhs67*clhs757 + clhs67*clhs760 + clhs67*clhs762 + clhs67*clhs764 + clhs769 + clhs772;
const double clhs1762 =             clhs1761 + clhs748;
const double clhs1763 =             clhs1215*clhs1715 - clhs1343*clhs1621 - clhs1400*clhs1718;
const double clhs1764 =             clhs1156*clhs1733;
const double clhs1765 =             clhs1161*clhs1724;
const double clhs1766 =             clhs7*clhs887*tau2;
const double clhs1767 =             -clhs873;
const double clhs1768 =             -clhs874;
const double clhs1769 =             clhs1767 + clhs1768 + clhs841 + clhs844 + clhs847 + clhs854 + clhs879 + clhs893 + clhs895 - clhs896;
const double clhs1770 =             (1.0L/2.0L)*DN(3,2)*tau1;
const double clhs1771 =             DN(3,2)*clhs203;
const double clhs1772 =             -N[3]*clhs106*clhs14*clhs375*clhs67*clhs83 - N[3]*clhs14*clhs67*clhs81*clhs83*clhs84 + clhs1220*clhs835 + clhs1222*clhs1771 - clhs1223*clhs837 - clhs1677*clhs203 + clhs67*clhs784 + clhs67*clhs787 + clhs67*clhs789 + clhs67*clhs791 + clhs67*clhs793 + clhs67*clhs795 + clhs799 + clhs802;
const double clhs1773 =             clhs1772 + clhs782;
const double clhs1774 =             DN(3,1)*clhs203*tau2;
const double clhs1775 =             DN(3,2)*clhs203*tau2;
const double clhs1776 =             DN(3,0)*clhs203*tau2;
const double clhs1777 =             clhs834*gamma + clhs836*gamma - clhs838*gamma + clhs840 + clhs843*gamma + clhs846*gamma - clhs854*gamma + clhs873*gamma + clhs874*gamma - clhs879*gamma;
const double clhs1778 =             clhs0*clhs113;
const double clhs1779 =             clhs155 + clhs193;
const double clhs1780 =             -clhs127 - clhs128 - clhs129 - clhs130 - clhs143 - clhs144 - clhs145 - clhs146 + clhs1779*clhs203;
const double clhs1781 =             DN(0,0)*clhs7*clhs906*mu;
const double clhs1782 =             clhs131 + clhs147;
const double clhs1783 =             N[0]*clhs1782 - clhs1779*clhs909 + clhs260 + clhs262;
const double clhs1784 =             -2.0L/3.0L*DN(0,0)*U(0,1);
const double clhs1785 =             -2.0L/3.0L*DN(1,0)*U(1,1);
const double clhs1786 =             -2.0L/3.0L*DN(2,0)*U(2,1);
const double clhs1787 =             -2.0L/3.0L*DN(3,0)*U(3,1);
const double clhs1788 =             (2.0L/3.0L)*clhs10*clhs106*clhs203;
const double clhs1789 =             (4.0L/3.0L)*DN(0,1)*U(0,2) - 2.0L/3.0L*DN(0,2)*U(0,3) + (4.0L/3.0L)*DN(1,1)*U(1,2) - 2.0L/3.0L*DN(1,2)*U(1,3) + (4.0L/3.0L)*DN(2,1)*U(2,2) - 2.0L/3.0L*DN(2,2)*U(2,3) + (4.0L/3.0L)*DN(3,1)*U(3,2) - 2.0L/3.0L*DN(3,2)*U(3,3) + (2.0L/3.0L)*clhs14*clhs203*clhs88 + clhs1784 + clhs1785 + clhs1786 + clhs1787 + clhs1788 - 4.0L/3.0L*clhs203*clhs8*clhs81;
const double clhs1790 =             (2.0L/3.0L)*DN(0,1)*clhs7*clhs906*mu;
const double clhs1791 =             -clhs1024;
const double clhs1792 =             clhs1019 + clhs1057*clhs84 + clhs1791 + clhs244 + clhs247 - clhs32 + clhs920;
const double clhs1793 =             -clhs117 - clhs120 - clhs122 - clhs125 - clhs132 - clhs134 + clhs137 + clhs140 + clhs148 + clhs159 - clhs162*clhs909 + clhs213*clhs375 + clhs268*clhs90;
const double clhs1794 =             -clhs114;
const double clhs1795 =             clhs163 + clhs1794;
const double clhs1796 =             4*DN(0,0)*clhs10;
const double clhs1797 =             clhs1796*clhs8;
const double clhs1798 =             clhs157 + clhs938;
const double clhs1799 =             clhs940 + clhs941 + clhs942 - 6*clhs97;
const double clhs1800 =             -clhs946;
const double clhs1801 =             clhs1800 + clhs206 + clhs944;
const double clhs1802 =             -clhs1180 + clhs1801 + clhs211 + clhs67*clhs945;
const double clhs1803 =             clhs14*clhs9;
const double clhs1804 =             clhs155*clhs909;
const double clhs1805 =             clhs193*clhs909;
const double clhs1806 =             -clhs119 + clhs1803*clhs67 + clhs1804 - clhs1805*clhs67 + clhs289;
const double clhs1807 =             -clhs276;
const double clhs1808 =             clhs1020 + clhs116 + clhs1791 + clhs230 + clhs247 + clhs919;
const double clhs1809 =             -clhs170*clhs57 + clhs1807 + clhs1808 + clhs57*clhs924;
const double clhs1810 =             (1.0L/2.0L)*clhs203*clhs290*tau2;
const double clhs1811 =             (1.0L/2.0L)*clhs203*clhs215*tau2;
const double clhs1812 =             -clhs922;
const double clhs1813 =             clhs1057*clhs136;
const double clhs1814 =             clhs1056 + clhs1807 + clhs1812 + clhs1813 + clhs244 + clhs247 - clhs274;
const double clhs1815 =             (1.0L/2.0L)*clhs1814*clhs203*tau2;
const double clhs1816 =             (1.0L/2.0L)*DN(0,1)*clhs67*tau3;
const double clhs1817 =             clhs203*clhs215*tau2;
const double clhs1818 =             (1.0L/2.0L)*N[0]*clhs1172*clhs203*clhs67*tau2;
const double clhs1819 =             -clhs1160 + clhs1177*clhs1215 + clhs1193*clhs290;
const double clhs1820 =             -clhs244;
const double clhs1821 =             -clhs247;
const double clhs1822 =             clhs1208 - clhs1813 + clhs1820 + clhs1821 + clhs274 + clhs276 + clhs922;
const double clhs1823 =             clhs1822*clhs203*tau2;
const double clhs1824 =             clhs1212 + clhs1213 + clhs235 + clhs237 + clhs275 + clhs277 + clhs279 - clhs282;
const double clhs1825 =             clhs1175*clhs1217;
const double clhs1826 =             -clhs948 - clhs949 - clhs950 - clhs951 - clhs958 + clhs960 - clhs963 - clhs966 - clhs969 - clhs971 + clhs972 + clhs975 + clhs978 + clhs980 + clhs982 - clhs984;
const double clhs1827 =             clhs1173*clhs1826;
const double clhs1828 =             clhs1148*clhs1164;
const double clhs1829 =             clhs1151*clhs1161;
const double clhs1830 =             clhs290*clhs7*tau2;
const double clhs1831 =             clhs203*clhs290*tau2;
const double clhs1832 =             DN(0,1)*tau3;
const double clhs1833 =             clhs113*clhs1228;
const double clhs1834 =             clhs1235 + clhs1236 + clhs1237 - clhs1240;
const double clhs1835 =             clhs1231*clhs67;
const double clhs1836 =             clhs1232*clhs67;
const double clhs1837 =             clhs1233*clhs67;
const double clhs1838 =             clhs1239*clhs67 + clhs1834 - clhs1835 - clhs1836 - clhs1837;
const double clhs1839 =             clhs1243*clhs8;
const double clhs1840 =             clhs485*clhs8;
const double clhs1841 =             N[1]*clhs264;
const double clhs1842 =             clhs1230*clhs14;
const double clhs1843 =             clhs14*clhs459;
const double clhs1844 =             N[1]*clhs261;
const double clhs1845 =             clhs1844*clhs67;
const double clhs1846 =             clhs3*clhs472;
const double clhs1847 =             clhs3*clhs471;
const double clhs1848 =             -clhs1839 - clhs1840 - clhs1841 + clhs1842*clhs67 + clhs1843*clhs67 + clhs1845 + clhs1846 - clhs1847*clhs67;
const double clhs1849 =             -clhs1259;
const double clhs1850 =             clhs1255 + clhs1256 + clhs1257 + clhs1261 + clhs1263 + clhs1265 + clhs1271 + clhs1849;
const double clhs1851 =             -clhs1260*clhs57 - clhs1262*clhs57 - clhs1264*clhs57 + clhs1267*clhs57 + clhs1850;
const double clhs1852 =             clhs1242*clhs1851 - clhs1254*clhs1848 + clhs1274*clhs1838 + clhs1833;
const double clhs1853 =             N[1]*clhs1782 - clhs1278*clhs1779 + clhs464 + clhs466;
const double clhs1854 =             -clhs387;
const double clhs1855 =             clhs359*clhs928;
const double clhs1856 =             -clhs1855;
const double clhs1857 =             clhs1266*clhs84 + clhs1285 + clhs1854 + clhs1856 - clhs315 + clhs450 + clhs453;
const double clhs1858 =             -clhs1278*clhs162 - clhs352 - clhs355 - clhs357 - clhs360 - clhs362 - clhs364 + clhs366 + clhs369 + clhs372 + clhs375*clhs422 + clhs379 + clhs472*clhs90;
const double clhs1859 =             N[0]*N[1]*clhs1184*clhs203;
const double clhs1860 =             N[0]*clhs352 + N[0]*clhs355 + N[1]*clhs117 + N[1]*clhs120 + clhs124*clhs415 - clhs124*clhs419 - clhs1295*clhs155 + clhs1344*clhs1799 - clhs151*clhs1859 + clhs24*clhs437 + clhs24*clhs453 - clhs24*clhs479 + clhs40*clhs468 - clhs40*clhs490;
const double clhs1861 =             (1.0L/2.0L)*clhs1345*clhs203*clhs67*tau2;
const double clhs1862 =             clhs1050*clhs1861 + clhs1343*clhs1838 - clhs1355*clhs985;
const double clhs1863 =             clhs1215*clhs1355 - clhs1361 + clhs301;
const double clhs1864 =             -clhs479;
const double clhs1865 =             -clhs1286;
const double clhs1866 =             clhs1266*clhs136;
const double clhs1867 =             clhs1373 + clhs1864 + clhs1865 + clhs1866 + clhs450 + clhs453 - clhs477;
const double clhs1868 =             clhs1167*clhs463;
const double clhs1869 =             clhs1393*clhs290;
const double clhs1870 =             clhs1343*clhs1851;
const double clhs1871 =             clhs1405 + clhs1406 + clhs441 + clhs443 + clhs478 + clhs480 + clhs481 - clhs483;
const double clhs1872 =             clhs1217*clhs1351;
const double clhs1873 =             clhs1346*clhs1826;
const double clhs1874 =             clhs1017*clhs1861 - clhs1343*clhs1848 - clhs1351*clhs985;
const double clhs1875 =             clhs1148*clhs1369;
const double clhs1876 =             clhs1161*clhs1364;
const double clhs1877 =             clhs492*clhs7*tau2;
const double clhs1878 =             clhs113*clhs1416;
const double clhs1879 =             clhs1423 + clhs1424 + clhs1425 - clhs1428;
const double clhs1880 =             clhs1419*clhs67;
const double clhs1881 =             clhs1420*clhs67;
const double clhs1882 =             clhs1421*clhs67;
const double clhs1883 =             clhs1427*clhs67 + clhs1879 - clhs1880 - clhs1881 - clhs1882;
const double clhs1884 =             clhs1430*clhs8;
const double clhs1885 =             clhs684*clhs8;
const double clhs1886 =             N[2]*clhs264;
const double clhs1887 =             clhs14*clhs1418;
const double clhs1888 =             clhs14*clhs658;
const double clhs1889 =             N[2]*clhs261;
const double clhs1890 =             clhs1889*clhs67;
const double clhs1891 =             clhs3*clhs671;
const double clhs1892 =             clhs3*clhs670;
const double clhs1893 =             -clhs1884 - clhs1885 - clhs1886 + clhs1887*clhs67 + clhs1888*clhs67 + clhs1890 + clhs1891 - clhs1892*clhs67;
const double clhs1894 =             -clhs1445;
const double clhs1895 =             clhs1441 + clhs1442 + clhs1443 + clhs1447 + clhs1449 + clhs1451 + clhs1457 + clhs1894;
const double clhs1896 =             -clhs1446*clhs57 - clhs1448*clhs57 - clhs1450*clhs57 + clhs1453*clhs57 + clhs1895;
const double clhs1897 =             clhs1242*clhs1896 - clhs1254*clhs1893 + clhs1274*clhs1883 + clhs1878;
const double clhs1898 =             N[2]*clhs1782 - clhs1463*clhs1779 + clhs663 + clhs665;
const double clhs1899 =             -clhs586;
const double clhs1900 =             clhs559*clhs928;
const double clhs1901 =             -clhs1900;
const double clhs1902 =             clhs1452*clhs84 + clhs1470 + clhs1899 + clhs1901 - clhs515 + clhs649 + clhs652;
const double clhs1903 =             -clhs1463*clhs162 + clhs375*clhs621 - clhs552 - clhs555 - clhs557 - clhs560 - clhs562 - clhs564 + clhs566 + clhs569 + clhs572 + clhs578 + clhs671*clhs90;
const double clhs1904 =             N[0]*N[2]*clhs1184*clhs203;
const double clhs1905 =             N[0]*clhs552 + N[0]*clhs555 + N[2]*clhs117 + N[2]*clhs120 + clhs124*clhs614 - clhs124*clhs618 - clhs1479*clhs155 - clhs151*clhs1904 + clhs1527*clhs1799 + clhs24*clhs636 + clhs24*clhs652 - clhs24*clhs678 + clhs40*clhs667 - clhs40*clhs689;
const double clhs1906 =             (1.0L/2.0L)*clhs1528*clhs203*clhs67*tau2;
const double clhs1907 =             clhs1050*clhs1906 + clhs1343*clhs1883 - clhs1538*clhs985;
const double clhs1908 =             clhs1215*clhs1538 - clhs1544 + clhs501;
const double clhs1909 =             -clhs678;
const double clhs1910 =             -clhs1471;
const double clhs1911 =             clhs136*clhs1452;
const double clhs1912 =             clhs1556 + clhs1909 + clhs1910 + clhs1911 + clhs649 + clhs652 - clhs676;
const double clhs1913 =             clhs1167*clhs662;
const double clhs1914 =             clhs1575*clhs290;
const double clhs1915 =             clhs1343*clhs1896;
const double clhs1916 =             clhs1586 + clhs1587 + clhs640 + clhs642 + clhs677 + clhs679 + clhs680 - clhs682;
const double clhs1917 =             clhs1217*clhs1534;
const double clhs1918 =             clhs1529*clhs1826;
const double clhs1919 =             clhs1017*clhs1906 - clhs1343*clhs1893 - clhs1534*clhs985;
const double clhs1920 =             clhs1148*clhs1552;
const double clhs1921 =             clhs1161*clhs1547;
const double clhs1922 =             clhs691*clhs7*tau2;
const double clhs1923 =             clhs113*clhs1597;
const double clhs1924 =             clhs1604 + clhs1605 + clhs1606 - clhs1609;
const double clhs1925 =             clhs1600*clhs67;
const double clhs1926 =             clhs1601*clhs67;
const double clhs1927 =             clhs1602*clhs67;
const double clhs1928 =             clhs1608*clhs67 + clhs1924 - clhs1925 - clhs1926 - clhs1927;
const double clhs1929 =             clhs1611*clhs8;
const double clhs1930 =             clhs8*clhs883;
const double clhs1931 =             N[3]*clhs264;
const double clhs1932 =             clhs14*clhs1599;
const double clhs1933 =             clhs14*clhs857;
const double clhs1934 =             N[3]*clhs261;
const double clhs1935 =             clhs1934*clhs67;
const double clhs1936 =             clhs3*clhs870;
const double clhs1937 =             clhs3*clhs869;
const double clhs1938 =             -clhs1929 - clhs1930 - clhs1931 + clhs1932*clhs67 + clhs1933*clhs67 + clhs1935 + clhs1936 - clhs1937*clhs67;
const double clhs1939 =             -clhs1626;
const double clhs1940 =             clhs1622 + clhs1623 + clhs1624 + clhs1628 + clhs1630 + clhs1632 + clhs1638 + clhs1939;
const double clhs1941 =             -clhs1627*clhs57 - clhs1629*clhs57 - clhs1631*clhs57 + clhs1634*clhs57 + clhs1940;
const double clhs1942 =             clhs1242*clhs1941 - clhs1254*clhs1938 + clhs1274*clhs1928 + clhs1923;
const double clhs1943 =             N[3]*clhs1782 - clhs1644*clhs1779 + clhs862 + clhs864;
const double clhs1944 =             -clhs785;
const double clhs1945 =             clhs758*clhs928;
const double clhs1946 =             -clhs1945;
const double clhs1947 =             clhs1633*clhs84 + clhs1651 + clhs1944 + clhs1946 - clhs714 + clhs848 + clhs851;
const double clhs1948 =             -clhs162*clhs1644 + clhs375*clhs820 - clhs751 - clhs754 - clhs756 - clhs759 - clhs761 - clhs763 + clhs765 + clhs768 + clhs771 + clhs777 + clhs870*clhs90;
const double clhs1949 =             N[0]*N[3]*clhs1184*clhs203;
const double clhs1950 =             N[0]*clhs751 + N[0]*clhs754 + N[3]*clhs117 + N[3]*clhs120 + clhs124*clhs813 - clhs124*clhs817 - clhs151*clhs1949 - clhs155*clhs1660 + clhs1708*clhs1799 + clhs24*clhs835 + clhs24*clhs851 - clhs24*clhs877 + clhs40*clhs866 - clhs40*clhs888;
const double clhs1951 =             (1.0L/2.0L)*clhs1709*clhs203*clhs67*tau2;
const double clhs1952 =             clhs1050*clhs1951 + clhs1343*clhs1928 - clhs1719*clhs985;
const double clhs1953 =             clhs1215*clhs1719 - clhs1725 + clhs700;
const double clhs1954 =             -clhs877;
const double clhs1955 =             -clhs1652;
const double clhs1956 =             clhs136*clhs1633;
const double clhs1957 =             clhs1737 + clhs1954 + clhs1955 + clhs1956 + clhs848 + clhs851 - clhs875;
const double clhs1958 =             clhs1167*clhs861;
const double clhs1959 =             clhs1756*clhs290;
const double clhs1960 =             clhs1343*clhs1941;
const double clhs1961 =             clhs1767 + clhs1768 + clhs839 + clhs841 + clhs876 + clhs878 + clhs879 - clhs881;
const double clhs1962 =             clhs1217*clhs1715;
const double clhs1963 =             clhs1710*clhs1826;
const double clhs1964 =             clhs1017*clhs1951 - clhs1343*clhs1938 - clhs1715*clhs985;
const double clhs1965 =             clhs1148*clhs1733;
const double clhs1966 =             clhs1161*clhs1728;
const double clhs1967 =             clhs7*clhs890*tau2;
const double clhs1968 =             clhs0*clhs166;
const double clhs1969 =             -2.0L/3.0L*DN(0,1)*U(0,2) + (4.0L/3.0L)*DN(0,2)*U(0,3) - 2.0L/3.0L*DN(1,1)*U(1,2) + (4.0L/3.0L)*DN(1,2)*U(1,3) - 2.0L/3.0L*DN(2,1)*U(2,2) + (4.0L/3.0L)*DN(2,2)*U(2,3) - 2.0L/3.0L*DN(3,1)*U(3,2) + (4.0L/3.0L)*DN(3,2)*U(3,3) - 4.0L/3.0L*clhs14*clhs203*clhs88 + clhs1784 + clhs1785 + clhs1786 + clhs1787 + clhs1788 + (2.0L/3.0L)*clhs203*clhs8*clhs81;
const double clhs1970 =             (2.0L/3.0L)*DN(0,2)*clhs7*clhs906*mu;
const double clhs1971 =             clhs238 + clhs241 + clhs925;
const double clhs1972 =             clhs1021 + clhs1054*clhs90 + clhs1791 + clhs1971 + clhs244 + clhs247 - clhs48;
const double clhs1973 =             -clhs168 - clhs171 - clhs173 - clhs175 - clhs177 - clhs179 + clhs182 + clhs184 + clhs187 + clhs197 - clhs199*clhs909 + clhs224*clhs375 + clhs267*clhs84;
const double clhs1974 =             -clhs167;
const double clhs1975 =             clhs1974 + clhs200;
const double clhs1976 =             clhs14*clhs1796;
const double clhs1977 =             clhs195 + clhs938;
const double clhs1978 =             clhs940 + clhs941 + clhs942 - 6*clhs99;
const double clhs1979 =             -clhs989;
const double clhs1980 =             clhs1979 + clhs218 + clhs987;
const double clhs1981 =             -clhs1182 + clhs1980 + clhs223 + clhs67*clhs988;
const double clhs1982 =             -clhs1805;
const double clhs1983 =             clhs1803 + clhs1982 + clhs261;
const double clhs1984 =             -clhs1198 + clhs1804*clhs67 + clhs1983 + clhs266;
const double clhs1985 =             -clhs295;
const double clhs1986 =             clhs116 + clhs170 + clhs1791 + clhs241 + clhs247 + clhs925;
const double clhs1987 =             -clhs1020*clhs57 + clhs1985 + clhs1986 + clhs57*clhs918;
const double clhs1988 =             (1.0L/2.0L)*clhs203*clhs269*tau2;
const double clhs1989 =             (1.0L/2.0L)*clhs203*clhs226*tau2;
const double clhs1990 =             clhs1054*clhs181;
const double clhs1991 =             clhs1059 + clhs1812 + clhs238 + clhs241 + clhs244 + clhs247;
const double clhs1992 =             clhs1985 + clhs1990 + clhs1991 - clhs293;
const double clhs1993 =             (1.0L/2.0L)*clhs1992*clhs203*tau2;
const double clhs1994 =             (1.0L/2.0L)*DN(0,2)*clhs67*tau3;
const double clhs1995 =             clhs203*clhs226*tau2;
const double clhs1996 =             clhs203*clhs269*tau2;
const double clhs1997 =             clhs1022 + clhs1058 + clhs1209 + clhs1820 + clhs1821 - clhs1990 + clhs293 + clhs295 + clhs922;
const double clhs1998 =             clhs1997*clhs203*tau2;
const double clhs1999 =             DN(0,2)*tau3;
const double clhs2000 =             clhs1228*clhs166;
const double clhs2001 =             clhs1247 + clhs1248 + clhs1249 - clhs1252;
const double clhs2002 =             clhs1244*clhs67;
const double clhs2003 =             clhs1245*clhs67;
const double clhs2004 =             clhs1246*clhs67;
const double clhs2005 =             clhs1251*clhs67 + clhs2001 - clhs2002 - clhs2003 - clhs2004;
const double clhs2006 =             clhs1842 + clhs1843 + clhs1844 - clhs1847;
const double clhs2007 =             clhs1839*clhs67;
const double clhs2008 =             clhs1840*clhs67;
const double clhs2009 =             clhs1841*clhs67;
const double clhs2010 =             clhs1846*clhs67 + clhs2006 - clhs2007 - clhs2008 - clhs2009;
const double clhs2011 =             clhs1255 + clhs1256 + clhs1257 + clhs1260 + clhs1262 + clhs1264 + clhs1268 + clhs1849;
const double clhs2012 =             -clhs1261*clhs57 - clhs1263*clhs57 - clhs1265*clhs57 + clhs1270*clhs57 + clhs2011;
const double clhs2013 =             clhs1242*clhs2010 + clhs1254*clhs2012 + clhs1274*clhs2005 + clhs2000;
const double clhs2014 =             clhs1289 + clhs444 + clhs447;
const double clhs2015 =             clhs14*clhs308;
const double clhs2016 =             -clhs2015;
const double clhs2017 =             clhs1269*clhs90 + clhs1856 + clhs2014 + clhs2016 - clhs321 + clhs450 + clhs453;
const double clhs2018 =             -clhs1278*clhs199 + clhs375*clhs432 - clhs385 - clhs388 - clhs390 - clhs392 - clhs394 - clhs396 + clhs398 + clhs400 + clhs403 + clhs408 + clhs471*clhs84;
const double clhs2019 =             N[0]*clhs385 + N[0]*clhs388 + N[1]*clhs168 + N[1]*clhs171 + clhs124*clhs426 - clhs124*clhs430 - clhs1294*clhs193 + clhs1349*clhs1978 - clhs1859*clhs190 + clhs24*clhs465 - clhs24*clhs469 + clhs40*clhs447 + clhs40*clhs453 - clhs40*clhs496;
const double clhs2020 =             (1.0L/2.0L)*clhs1050*clhs203*clhs67*tau2;
const double clhs2021 =             -clhs1017*clhs1355 + clhs1343*clhs2005 + clhs1350*clhs2020;
const double clhs2022 =             -clhs1017*clhs1346 + clhs1343*clhs2010 + clhs1350*clhs1389;
const double clhs2023 =             -clhs496;
const double clhs2024 =             clhs1269*clhs181;
const double clhs2025 =             clhs1375 + clhs1865 + clhs444 + clhs447 + clhs450 + clhs453;
const double clhs2026 =             clhs2023 + clhs2024 + clhs2025 - clhs494;
const double clhs2027 =             clhs1154*clhs489;
const double clhs2028 =             clhs1193*clhs492;
const double clhs2029 =             clhs1343*clhs2012;
const double clhs2030 =             clhs1416*clhs166;
const double clhs2031 =             clhs1434 + clhs1435 + clhs1436 - clhs1439;
const double clhs2032 =             clhs1431*clhs67;
const double clhs2033 =             clhs1432*clhs67;
const double clhs2034 =             clhs1433*clhs67;
const double clhs2035 =             clhs1438*clhs67 + clhs2031 - clhs2032 - clhs2033 - clhs2034;
const double clhs2036 =             clhs1887 + clhs1888 + clhs1889 - clhs1892;
const double clhs2037 =             clhs1884*clhs67;
const double clhs2038 =             clhs1885*clhs67;
const double clhs2039 =             clhs1886*clhs67;
const double clhs2040 =             clhs1891*clhs67 + clhs2036 - clhs2037 - clhs2038 - clhs2039;
const double clhs2041 =             clhs1441 + clhs1442 + clhs1443 + clhs1446 + clhs1448 + clhs1450 + clhs1454 + clhs1894;
const double clhs2042 =             -clhs1447*clhs57 - clhs1449*clhs57 - clhs1451*clhs57 + clhs1456*clhs57 + clhs2041;
const double clhs2043 =             clhs1242*clhs2040 + clhs1254*clhs2042 + clhs1274*clhs2035 + clhs2030;
const double clhs2044 =             clhs1474 + clhs643 + clhs646;
const double clhs2045 =             clhs14*clhs508;
const double clhs2046 =             -clhs2045;
const double clhs2047 =             clhs1455*clhs90 + clhs1901 + clhs2044 + clhs2046 - clhs521 + clhs649 + clhs652;
const double clhs2048 =             -clhs1463*clhs199 + clhs375*clhs631 - clhs584 - clhs587 - clhs589 - clhs591 - clhs593 - clhs595 + clhs597 + clhs599 + clhs602 + clhs607 + clhs670*clhs84;
const double clhs2049 =             N[0]*clhs584 + N[0]*clhs587 + N[2]*clhs168 + N[2]*clhs171 + clhs124*clhs625 - clhs124*clhs629 - clhs1478*clhs193 + clhs1532*clhs1978 - clhs190*clhs1904 + clhs24*clhs664 - clhs24*clhs668 + clhs40*clhs646 + clhs40*clhs652 - clhs40*clhs695;
const double clhs2050 =             -clhs1017*clhs1538 + clhs1343*clhs2035 + clhs1533*clhs2020;
const double clhs2051 =             -clhs1017*clhs1529 + clhs1343*clhs2040 + clhs1389*clhs1533;
const double clhs2052 =             -clhs695;
const double clhs2053 =             clhs1455*clhs181;
const double clhs2054 =             clhs1558 + clhs1910 + clhs643 + clhs646 + clhs649 + clhs652;
const double clhs2055 =             clhs2052 + clhs2053 + clhs2054 - clhs693;
const double clhs2056 =             clhs1154*clhs688;
const double clhs2057 =             clhs1193*clhs691;
const double clhs2058 =             clhs1343*clhs2042;
const double clhs2059 =             clhs1597*clhs166;
const double clhs2060 =             clhs1615 + clhs1616 + clhs1617 - clhs1620;
const double clhs2061 =             clhs1612*clhs67;
const double clhs2062 =             clhs1613*clhs67;
const double clhs2063 =             clhs1614*clhs67;
const double clhs2064 =             clhs1619*clhs67 + clhs2060 - clhs2061 - clhs2062 - clhs2063;
const double clhs2065 =             clhs1932 + clhs1933 + clhs1934 - clhs1937;
const double clhs2066 =             clhs1929*clhs67;
const double clhs2067 =             clhs1930*clhs67;
const double clhs2068 =             clhs1931*clhs67;
const double clhs2069 =             clhs1936*clhs67 + clhs2065 - clhs2066 - clhs2067 - clhs2068;
const double clhs2070 =             clhs1622 + clhs1623 + clhs1624 + clhs1627 + clhs1629 + clhs1631 + clhs1635 + clhs1939;
const double clhs2071 =             -clhs1628*clhs57 - clhs1630*clhs57 - clhs1632*clhs57 + clhs1637*clhs57 + clhs2070;
const double clhs2072 =             clhs1242*clhs2069 + clhs1254*clhs2071 + clhs1274*clhs2064 + clhs2059;
const double clhs2073 =             clhs1655 + clhs842 + clhs845;
const double clhs2074 =             clhs14*clhs707;
const double clhs2075 =             -clhs2074;
const double clhs2076 =             clhs1636*clhs90 + clhs1946 + clhs2073 + clhs2075 - clhs720 + clhs848 + clhs851;
const double clhs2077 =             -clhs1644*clhs199 + clhs375*clhs830 - clhs783 - clhs786 - clhs788 - clhs790 - clhs792 - clhs794 + clhs796 + clhs798 + clhs801 + clhs806 + clhs84*clhs869;
const double clhs2078 =             N[0]*clhs783 + N[0]*clhs786 + N[3]*clhs168 + N[3]*clhs171 + clhs124*clhs824 - clhs124*clhs828 - clhs1659*clhs193 + clhs1713*clhs1978 - clhs190*clhs1949 + clhs24*clhs863 - clhs24*clhs867 + clhs40*clhs845 + clhs40*clhs851 - clhs40*clhs894;
const double clhs2079 =             -clhs1017*clhs1719 + clhs1343*clhs2064 + clhs1714*clhs2020;
const double clhs2080 =             -clhs1017*clhs1710 + clhs1343*clhs2069 + clhs1389*clhs1714;
const double clhs2081 =             -clhs894;
const double clhs2082 =             clhs1636*clhs181;
const double clhs2083 =             clhs1739 + clhs1955 + clhs842 + clhs845 + clhs848 + clhs851;
const double clhs2084 =             clhs2081 + clhs2082 + clhs2083 - clhs892;
const double clhs2085 =             clhs1154*clhs887;
const double clhs2086 =             clhs1193*clhs890;
const double clhs2087 =             clhs1343*clhs2071;
const double clhs2088 =             (1.0L/2.0L)*N[0]*clhs7*tau2;
const double clhs2089 =             clhs1184*clhs67;
const double clhs2090 =             -clhs1113*clhs25 - clhs1113*clhs33 - clhs1113*clhs41 - clhs1113*clhs49 - clhs203*clhs69 - clhs203*clhs78 - clhs2089*clhs273;
const double clhs2091 =             N[0]*clhs1034;
const double clhs2092 =             clhs203*clhs84;
const double clhs2093 =             clhs203*clhs90;
const double clhs2094 =             clhs67*clhs936;
const double clhs2095 =             N[0]*clhs10*clhs7*clhs81;
const double clhs2096 =             clhs67*clhs937;
const double clhs2097 =             N[0]*clhs10*clhs7*clhs88;
const double clhs2098 =             -7*clhs1111 + clhs1139;
const double clhs2099 =             -DN(0,0)*clhs1125 - clhs208*clhs2092 + clhs2090 + clhs2091*gamma - clhs2093*clhs220 + clhs2094*clhs2095 + clhs2096*clhs2097 - clhs2098*clhs921;
const double clhs2100 =             -clhs1113*clhs122 - clhs1113*clhs125 - clhs1113*clhs132 - clhs1113*clhs134 - clhs140*clhs203 - clhs148*clhs203 - clhs2094*clhs242;
const double clhs2101 =             N[0]*clhs957;
const double clhs2102 =             N[0]*clhs106*clhs7*clhs8;
const double clhs2103 =             N[0]*clhs7*clhs8*clhs88;
const double clhs2104 =             clhs1097 + clhs1100 + clhs1103 + clhs1106 + clhs1109 - 7*clhs1112 + clhs1135 + clhs1137 + clhs1138;
const double clhs2105 =             -DN(0,1)*clhs1129 - clhs1057*clhs2104 - clhs1113*clhs1797 + clhs2089*clhs2102 - clhs2093*clhs263 + clhs2096*clhs2103 + clhs2100 + clhs2101*gamma;
const double clhs2106 =             -clhs1113*clhs173 - clhs1113*clhs175 - clhs1113*clhs177 - clhs1113*clhs179 - clhs184*clhs203 - clhs187*clhs203 - clhs2096*clhs231;
const double clhs2107 =             N[0]*clhs999;
const double clhs2108 =             N[0]*clhs106*clhs14*clhs7;
const double clhs2109 =             N[0]*clhs14*clhs7*clhs81;
const double clhs2110 =             -7*clhs1095 + clhs1097 + clhs1100 + clhs1103 + clhs1106 + clhs1109 + clhs1135 + clhs1136 + clhs1138;
const double clhs2111 =             -DN(0,1)*clhs1113*clhs8*clhs90 - DN(0,2)*clhs1132 - clhs1054*clhs2110 - clhs1113*clhs1976 + clhs2089*clhs2108 + clhs2094*clhs2109 + clhs2106 + clhs2107*gamma;
const double clhs2112 =             (1.0L/4.0L)*tau2*(clhs105 - clhs1065 + clhs13 + clhs18 + clhs26 + clhs34 + clhs42 + clhs5 + clhs50 + clhs61 + clhs71 + clhs80 + clhs87 + clhs93);
const double clhs2113 =             clhs1063 + clhs1189;
const double clhs2114 =             clhs1798*clhs81;
const double clhs2115 =             (1.0L/4.0L)*tau2*(clhs114 + clhs118 + clhs121 + clhs123 + clhs126 + clhs133 + clhs135 + clhs138 + clhs142 + clhs150 + clhs154 + clhs156 + clhs160 - clhs161*clhs2114);
const double clhs2116 =             clhs1203 + clhs1794;
const double clhs2117 =             clhs1977*clhs88;
const double clhs2118 =             (1.0L/4.0L)*tau2*(-clhs161*clhs2117 + clhs167 + clhs169 + clhs172 + clhs174 + clhs176 + clhs178 + clhs180 + clhs183 + clhs186 + clhs189 + clhs192 + clhs194 + clhs198);
const double clhs2119 =             clhs1224 + clhs1974;
const double clhs2120 =             clhs1020 + clhs230 + clhs919;
const double clhs2121 =             clhs1986 + clhs2120;
const double clhs2122 =             clhs10*clhs7*clhs8*clhs952;
const double clhs2123 =             clhs10*clhs14*clhs7*clhs952;
const double clhs2124 =             clhs14*clhs7*clhs8*clhs952;
const double clhs2125 =             clhs1117*clhs7;
const double clhs2126 =             -clhs1034*clhs965*gamma + clhs1068*clhs959 + clhs113*clhs961 + clhs1187*clhs962 + clhs1201*clhs203*clhs31 + clhs1222*clhs203*clhs47 + clhs131*clhs2124 + clhs147*clhs2124 + clhs166*clhs967 + clhs2122*clhs23 + clhs2122*clhs66 + clhs2123*clhs39 + clhs2123*clhs76 - clhs2125*clhs233 - clhs2125*clhs250 - clhs2125*clhs278 - clhs3*(U(0,4)*bdf0 + Un(0,4)*bdf1 + Unn(0,4)*bdf2) - clhs302*(U(1,4)*bdf0 + Un(1,4)*bdf1 + Unn(1,4)*bdf2) + clhs4*clhs964 - clhs502*(U(2,4)*bdf0 + Un(2,4)*bdf1 + Unn(2,4)*bdf2) - clhs701*(U(3,4)*bdf0 + Un(3,4)*bdf1 + Unn(3,4)*bdf2) - clhs957*clhs970*gamma - clhs968*clhs999*gamma;
const double clhs2127 =             clhs2126*tau3;
const double clhs2128 =             (1.0L/3.0L)*DN(0,0)*clhs203;
const double clhs2129 =             3*N[0]*clhs203*clhs8*v_sc;
const double clhs2130 =             3*N[0]*clhs14*clhs203*v_sc;
const double clhs2131 =             3*N[0]*clhs7*clhs8*clhs906*mu;
const double clhs2132 =             3*N[0]*clhs14*clhs7*clhs906*mu;
const double clhs2133 =             3*clhs906*clhs910*mu;
const double clhs2134 =             3*clhs906*clhs913*mu;
const double clhs2135 =             2*N[0]*clhs10*clhs203;
const double clhs2136 =             clhs1009*clhs81;
const double clhs2137 =             clhs88*clhs976;
const double clhs2138 =             -U(0,1)*clhs115 - U(1,1)*clhs350 - U(2,1)*clhs550 - U(3,1)*clhs749 - clhs2136 - clhs2137 + clhs928*clhs964 + clhs933;
const double clhs2139 =             clhs2138*v_sc;
const double clhs2140 =             clhs2138*clhs7*clhs906*mu;
const double clhs2141 =             clhs7*clhs906*clhs964*mu;
const double clhs2142 =             3*N[0]*k_sc;
const double clhs2143 =             clhs1108*clhs928;
const double clhs2144 =             -clhs1030 - clhs1031 - clhs1032 - clhs1033 + clhs1042 + clhs1044 - clhs1048*clhs94 - clhs1048*clhs97 - clhs1048*clhs99 + clhs2143 + clhs55*clhs973;
const double clhs2145 =             1.0/c_v;
const double clhs2146 =             c_v*clhs6*k_sc/lambda + 1;
const double clhs2147 =             3*clhs203*clhs2145*clhs2146*lambda;
const double clhs2148 =             3*N[0]*clhs106*clhs7;
const double clhs2149 =             -clhs1181*clhs2133 - clhs1183*clhs2134 + clhs124*clhs2140 - clhs2129*clhs902 - clhs2130*clhs905 + clhs2131*clhs902 + clhs2132*clhs905 - clhs2135*clhs2139 + clhs2141*clhs926 - clhs2142*clhs2144 + clhs2147*(-DN(0,0)*clhs1108 + clhs1156*clhs94 + clhs1156*clhs97 + clhs1156*clhs99 - clhs2091 + clhs2143*clhs3 - clhs2148*clhs94 - clhs2148*clhs97 - clhs2148*clhs99 + clhs56*clhs973 + clhs66*clhs923 + clhs76*clhs917);
const double clhs2150 =             (1.0L/3.0L)*DN(0,1)*clhs203;
const double clhs2151 =             3*N[0]*clhs10*clhs203*v_sc;
const double clhs2152 =             3*N[0]*clhs10*clhs7*clhs906*mu;
const double clhs2153 =             clhs10*clhs7;
const double clhs2154 =             3*clhs1783*clhs906*mu;
const double clhs2155 =             clhs10*clhs928 - clhs51 - clhs52 - clhs53 - clhs54;
const double clhs2156 =             U(0,2)*clhs9 + U(1,2)*clhs304 + U(2,2)*clhs504 + U(3,2)*clhs703 + clhs2137 + clhs2155 - clhs43 - clhs44 - clhs45 - clhs46 - clhs81*clhs970;
const double clhs2157 =             clhs2156*v_sc;
const double clhs2158 =             clhs2156*clhs7*clhs906*mu;
const double clhs2159 =             clhs7*clhs906*clhs961*mu;
const double clhs2160 =             clhs1108*clhs203;
const double clhs2161 =             clhs2160*clhs81;
const double clhs2162 =             clhs1009*clhs31 + clhs2161 - clhs94*clhs983 - clhs953 - clhs954 - clhs955 - clhs956 - clhs97*clhs983 + clhs974 + clhs977 - clhs983*clhs99;
const double clhs2163 =             3*N[0]*clhs7*clhs81;
const double clhs2164 =             -clhs1183*clhs2154 - clhs1780*clhs2130 + clhs1780*clhs2132 + clhs1792*clhs2159 - clhs2133*clhs2153 - clhs2142*clhs2162 + clhs2147*(-DN(0,1)*clhs1108 + clhs1009*clhs32 + clhs1148*clhs94 + clhs1148*clhs97 + clhs1148*clhs99 + clhs147*clhs917 - clhs2101 + clhs2135*clhs23 + clhs2161*clhs3 - clhs2163*clhs94 - clhs2163*clhs97 - clhs2163*clhs99) - clhs2151*clhs902 + clhs2152*clhs902 + clhs2157*clhs923 - clhs2158*clhs24;
const double clhs2165 =             (1.0L/3.0L)*DN(0,2)*clhs203;
const double clhs2166 =             U(0,3)*clhs15 + U(1,3)*clhs308 + U(2,3)*clhs508 + U(3,3)*clhs707 + clhs2136 + clhs2155 - clhs27 - clhs28 - clhs29 - clhs30 - clhs88*clhs968;
const double clhs2167 =             clhs2166*v_sc;
const double clhs2168 =             clhs2166*clhs7*clhs906*mu;
const double clhs2169 =             clhs7*clhs906*clhs967*mu;
const double clhs2170 =             clhs2160*clhs88;
const double clhs2171 =             clhs1007 + clhs1010 - clhs1015*clhs94 - clhs1015*clhs97 - clhs1015*clhs99 + clhs2170 + clhs47*clhs976 - clhs995 - clhs996 - clhs997 - clhs998;
const double clhs2172 =             3*N[0]*clhs7*clhs88;
const double clhs2173 =             -clhs1181*clhs2154 - clhs1780*clhs2129 + clhs1780*clhs2131 + clhs1972*clhs2169 - clhs2134*clhs2153 - clhs2142*clhs2171 + clhs2147*(-DN(0,2)*clhs1108 + clhs1161*clhs94 + clhs1161*clhs97 + clhs1161*clhs99 + clhs131*clhs923 - clhs2107 + clhs2135*clhs39 + clhs2170*clhs3 - clhs2172*clhs94 - clhs2172*clhs97 - clhs2172*clhs99 + clhs48*clhs976) - clhs2151*clhs905 + clhs2152*clhs905 + clhs2167*clhs917 - clhs2168*clhs40;
const double clhs2174 =             -clhs1073 - clhs1076 - clhs1079;
const double clhs2175 =             clhs1024*clhs1140 + clhs1082*clhs973 + clhs1084*clhs973 + clhs1086*clhs973 + clhs1089*clhs976 + clhs1091*clhs973 + clhs1093*clhs976 + clhs1118 + clhs1120 + clhs1122 + clhs1126 + clhs1130 + clhs1133 + clhs1140*clhs918 + clhs1140*clhs924 + clhs2174;
const double clhs2176 =             clhs10*clhs203*clhs936;
const double clhs2177 =             clhs10*clhs203*clhs937;
const double clhs2178 =             clhs203*clhs8*clhs937;
const double clhs2179 =             -6*N[0]*U(0,4) - 6*N[1]*U(1,4) - 6*N[2]*U(2,4) - 6*N[3]*U(3,4) - 6*clhs1108*clhs67 + 9*clhs1114*clhs203*clhs67 + clhs1186 + clhs1200 + clhs1221;
const double clhs2180 =             clhs1074 + clhs1077 + clhs1080 - clhs1083 - clhs1085 - clhs1087 - clhs1090 - clhs1092 - clhs1094 - clhs1119 - clhs1121 - clhs1123 - clhs1127 - clhs1131 - clhs1134 - clhs1142 - clhs1144 - clhs1146;
const double clhs2181 =             clhs1070 + clhs2180;
const double clhs2182 =             clhs1056 + clhs1991;
const double clhs2183 =             (1.0L/2.0L)*clhs203*clhs2182*gamma*tau3;
const double clhs2184 =             clhs1069 + clhs2180;
const double clhs2185 =             (1.0L/3.0L)*DN(0,1)*clhs7;
const double clhs2186 =             3*clhs1150*clhs906*mu;
const double clhs2187 =             3*clhs2145*clhs2146*lambda;
const double clhs2188 =             -clhs945;
const double clhs2189 =             3*N[0]*clhs906*mu;
const double clhs2190 =             -clhs2189*clhs902;
const double clhs2191 =             clhs10*clhs2186 - clhs1159*clhs961 - clhs2187*(clhs207 + clhs209 + clhs2188) + clhs2190;
const double clhs2192 =             (1.0L/3.0L)*DN(0,2)*clhs7;
const double clhs2193 =             3*clhs1163*clhs906*mu;
const double clhs2194 =             -clhs988;
const double clhs2195 =             -clhs2189*clhs905;
const double clhs2196 =             clhs10*clhs2193 - clhs1159*clhs967 - clhs2187*(clhs219 + clhs2194 + clhs221) + clhs2195;
const double clhs2197 =             (1.0L/3.0L)*DN(0,0)*clhs7;
const double clhs2198 =             clhs2186*clhs8;
const double clhs2199 =             clhs14*clhs2193;
const double clhs2200 =             clhs2138*clhs906*mu;
const double clhs2201 =             clhs1159*clhs375 - clhs2187*(clhs1791 + clhs244 + clhs247) + clhs2198 + clhs2199 - clhs2200*clhs3;
const double clhs2202 =             (1.0L/2.0L)*N[0]*clhs67*clhs7*tau2*(clhs11 + clhs1801 + clhs209 + clhs2188);
const double clhs2203 =             (1.0L/2.0L)*N[0]*clhs67*clhs7*tau2*(clhs16 + clhs1980 + clhs2194 + clhs221);
const double clhs2204 =             (1.0L/2.0L)*N[0]*clhs67*clhs7*tau2;
const double clhs2205 =             3*clhs247;
const double clhs2206 =             (1.0L/2.0L)*N[0]*clhs203;
const double clhs2207 =             clhs67*clhs84;
const double clhs2208 =             clhs67*clhs90;
const double clhs2209 =             -DN(0,0)*clhs1187 - clhs1009*clhs1180 + clhs1072 + clhs1125*clhs921 - clhs1182*clhs976 + clhs2090 + clhs2095*clhs2207 + clhs2097*clhs2208;
const double clhs2210 =             (1.0L/2.0L)*clhs254*tau2;
const double clhs2211 =             (1.0L/2.0L)*N[0]*clhs203*clhs2126*gamma*tau3;
const double clhs2212 =             3*clhs1158*clhs906*mu;
const double clhs2213 =             -clhs1151*clhs964 - clhs2187*(clhs1800 + clhs205 + clhs206) + clhs2190 + clhs2212*clhs8;
const double clhs2214 =             -clhs1804;
const double clhs2215 =             -clhs1780*clhs2189;
const double clhs2216 =             -clhs1151*clhs967 - clhs2187*(clhs2214 + clhs262 + clhs264) + clhs2193*clhs8 + clhs2215;
const double clhs2217 =             clhs10*clhs2212;
const double clhs2218 =             2*N[0]*clhs906*mu;
const double clhs2219 =             clhs1151*clhs84 - clhs1971*clhs2187 + clhs2156*clhs2218 + clhs2199 + clhs2217;
const double clhs2220 =             (1.0L/2.0L)*N[0]*clhs67*clhs7*tau2*(clhs119 + clhs1983 + clhs2214 + clhs264);
const double clhs2221 =             3*clhs241;
const double clhs2222 =             clhs375*clhs67;
const double clhs2223 =             -DN(0,1)*clhs1201 + clhs1057*clhs1129 + clhs1075 - clhs1113*clhs117 - clhs1198*clhs976 + clhs2100 + clhs2102*clhs2222 + clhs2103*clhs2208;
const double clhs2224 =             (1.0L/2.0L)*clhs1824*tau2;
const double clhs2225 =             -clhs1164*clhs964 + clhs14*clhs2212 - clhs2187*(clhs1979 + clhs217 + clhs218) + clhs2195;
const double clhs2226 =             -clhs1164*clhs961 + clhs14*clhs2186 - clhs2187*(clhs1982 + clhs260 + clhs261) + clhs2215;
const double clhs2227 =             clhs1164*clhs90 + clhs2166*clhs2218 - clhs2187*clhs920 + clhs2198 + clhs2217;
const double clhs2228 =             3*clhs230;
const double clhs2229 =             -DN(0,2)*clhs1222 + clhs1054*clhs1132 + clhs1078 - clhs1113*clhs168 - clhs1113*clhs171 + clhs2106 + clhs2108*clhs2222 + clhs2109*clhs2207;
const double clhs2230 =             (1.0L/2.0L)*clhs1214*tau2;
const double clhs2231 =             DN(0,0)*clhs203*clhs2145*clhs2146*lambda;
const double clhs2232 =             DN(0,1)*clhs203*clhs2145*clhs2146*lambda;
const double clhs2233 =             DN(0,2)*clhs203*clhs2145*clhs2146*lambda;
const double clhs2234 =             N[0]*clhs203*gamma;
const double clhs2235 =             clhs203*clhs2182*gamma*tau3;
const double clhs2236 =             (1.0L/2.0L)*N[0]*clhs203*gamma*tau2;
const double clhs2237 =             (1.0L/2.0L)*DN(0,0)*clhs67*tau2;
const double clhs2238 =             (1.0L/2.0L)*DN(0,1)*clhs67*tau2;
const double clhs2239 =             (1.0L/2.0L)*DN(0,2)*clhs67*tau2;
const double clhs2240 =             (1.0L/4.0L)*clhs1215*clhs7*tau2;
const double clhs2241 =             N[1]*clhs1034;
const double clhs2242 =             clhs2241*gamma;
const double clhs2243 =             clhs2242*clhs3;
const double clhs2244 =             12*clhs10*clhs203*clhs67;
const double clhs2245 =             4*clhs10*clhs203*clhs67;
const double clhs2246 =             12*gamma - 12;
const double clhs2247 =             N[0]*N[1]*clhs2246*clhs7*clhs8;
const double clhs2248 =             N[0]*N[1]*clhs14*clhs2246*clhs7;
const double clhs2249 =             clhs1258*clhs2098;
const double clhs2250 =             -clhs1125*clhs1234 - clhs1125*clhs412 - clhs1238*clhs2092 - clhs1250*clhs2093 - clhs1257*clhs2244 - clhs1264*clhs2245 - clhs1265*clhs2245 - clhs1835*clhs2092 - clhs1836*clhs2092 - clhs1837*clhs2092 - clhs2002*clhs2093 - clhs2003*clhs2093 - clhs2004*clhs2093 + clhs2243 + clhs2247*clhs82 + clhs2248*clhs89 - clhs2249*clhs3;
const double clhs2251 =             (1.0L/4.0L)*clhs1826*clhs7*tau2;
const double clhs2252 =             N[1]*clhs957;
const double clhs2253 =             clhs2252*gamma;
const double clhs2254 =             clhs2253*clhs3;
const double clhs2255 =             DN(0,0)*N[1]*clhs10*clhs203*clhs67;
const double clhs2256 =             clhs203*clhs67*clhs84;
const double clhs2257 =             4*clhs10*clhs203;
const double clhs2258 =             clhs203*clhs2246*clhs8;
const double clhs2259 =             N[0]*N[1]*clhs10*clhs2246*clhs7;
const double clhs2260 =             clhs1266*clhs2104;
const double clhs2261 =             -clhs1129*clhs1230 - clhs1129*clhs459 - clhs1238*clhs2257 - clhs1256*clhs2256 - clhs1257*clhs2256 - clhs1264*clhs2258 - clhs1265*clhs2256 + clhs151*clhs2259 + clhs155*clhs2248 - clhs1837*clhs2257 - clhs1845*clhs2093 - clhs2007*clhs2093 - clhs2008*clhs2093 - clhs2009*clhs2093 + clhs2254 - clhs2255*clhs84 - clhs2260*clhs3;
const double clhs2262 =             (1.0L/4.0L)*clhs1217*clhs7*tau2;
const double clhs2263 =             N[1]*clhs999;
const double clhs2264 =             clhs2263*gamma;
const double clhs2265 =             clhs2264*clhs3;
const double clhs2266 =             clhs203*clhs67*clhs90;
const double clhs2267 =             clhs14*clhs203*clhs2246;
const double clhs2268 =             clhs1269*clhs2110;
const double clhs2269 =             -clhs1132*clhs1243 - clhs1132*clhs485 - clhs1250*clhs2257 - clhs1256*clhs2266 - clhs1257*clhs2266 - clhs1260*clhs2266 - clhs1262*clhs2266 - clhs1264*clhs2266 - clhs1265*clhs2267 - clhs1845*clhs2092 + clhs190*clhs2259 + clhs193*clhs2247 - clhs2004*clhs2257 - clhs2009*clhs2092 - clhs2255*clhs90 + clhs2265 - clhs2268*clhs3;
const double clhs2270 =             clhs1261 + clhs1263 + clhs1265 + clhs1271 + clhs2011;
const double clhs2271 =             (1.0L/2.0L)*clhs2126*clhs7*gamma*tau3;
const double clhs2272 =             clhs10*clhs2179;
const double clhs2273 =             clhs2179*clhs8;
const double clhs2274 =             clhs14*clhs2179;
const double clhs2275 =             clhs1068*clhs1228 + clhs1293*(-clhs10*clhs2243 - clhs1140*clhs1255 - clhs1140*clhs1256 - clhs1140*clhs1260 - clhs1140*clhs1261 - clhs1140*clhs1262 - clhs1140*clhs1263 + clhs1238*clhs2176 + clhs1250*clhs2177 - clhs1257*clhs2098 - clhs1264*clhs2104 - clhs1265*clhs2110 - clhs1297*clhs2272 - clhs1344*clhs2273 - clhs1349*clhs2274 - clhs14*clhs2265 + clhs1837*clhs2176 + clhs1845*clhs2178 + clhs2004*clhs2177 + clhs2009*clhs2178 - clhs2254*clhs8) - clhs2240*clhs2250 - clhs2251*clhs2261 - clhs2262*clhs2269 - clhs2270*clhs2271;
const double clhs2276 =             (1.0L/4.0L)*tau2*(-clhs1300 + clhs303 + clhs307 + clhs311 + clhs314 + clhs317 + clhs320 + clhs323 + clhs327 + clhs331 + clhs335 + clhs338 + clhs341 + clhs344);
const double clhs2277 =             (1.0L/4.0L)*tau2*(-clhs2114*clhs381 + clhs349 + clhs353 + clhs356 + clhs358 + clhs361 + clhs363 + clhs365 + clhs367 + clhs371 + clhs374 + clhs377 + clhs378 + clhs380);
const double clhs2278 =             (1.0L/4.0L)*tau2*(-clhs2117*clhs381 + clhs384 + clhs386 + clhs389 + clhs391 + clhs393 + clhs395 + clhs397 + clhs399 + clhs402 + clhs405 + clhs406 + clhs407 + clhs409);
const double clhs2279 =             3*N[1]*clhs203*clhs8*v_sc;
const double clhs2280 =             3*N[1]*clhs14*clhs203*v_sc;
const double clhs2281 =             3*N[1]*clhs7*clhs8*clhs906*mu;
const double clhs2282 =             3*N[1]*clhs14*clhs7*clhs906*mu;
const double clhs2283 =             3*clhs1279*clhs906*mu;
const double clhs2284 =             3*clhs1280*clhs906*mu;
const double clhs2285 =             2*N[1]*clhs10*clhs203;
const double clhs2286 =             3*N[1]*k_sc;
const double clhs2287 =             3*N[1]*clhs106*clhs7;
const double clhs2288 =             -clhs1181*clhs2283 - clhs1183*clhs2284 + clhs1290*clhs2141 - clhs2139*clhs2285 + clhs2140*clhs359 - clhs2144*clhs2286 + clhs2147*(-DN(1,0)*clhs1108 + clhs1282*clhs76 + clhs1287*clhs66 + clhs1386*clhs94 + clhs1386*clhs97 + clhs1386*clhs99 + clhs2143*clhs302 - clhs2241 - clhs2287*clhs94 - clhs2287*clhs97 - clhs2287*clhs99 + clhs324*clhs973) - clhs2279*clhs902 - clhs2280*clhs905 + clhs2281*clhs902 + clhs2282*clhs905;
const double clhs2289 =             3*N[1]*clhs10*clhs203*v_sc;
const double clhs2290 =             3*N[1]*clhs10*clhs7*clhs906*mu;
const double clhs2291 =             3*clhs1853*clhs906*mu;
const double clhs2292 =             3*N[1]*clhs7*clhs81;
const double clhs2293 =             -clhs1183*clhs2291 + clhs1287*clhs2157 - clhs1780*clhs2280 + clhs1780*clhs2282 + clhs1857*clhs2159 + clhs2147*(-DN(1,1)*clhs1108 + clhs1009*clhs315 + clhs1282*clhs147 + clhs1397*clhs94 + clhs1397*clhs97 + clhs1397*clhs99 + clhs2161*clhs302 - clhs2252 + clhs2285*clhs23 - clhs2292*clhs94 - clhs2292*clhs97 - clhs2292*clhs99) - clhs2153*clhs2283 - clhs2158*clhs312 - clhs2162*clhs2286 - clhs2289*clhs902 + clhs2290*clhs902;
const double clhs2294 =             3*N[1]*clhs7*clhs88;
const double clhs2295 =             -clhs1181*clhs2291 + clhs1282*clhs2167 - clhs1780*clhs2279 + clhs1780*clhs2281 + clhs2017*clhs2169 + clhs2147*(-DN(1,2)*clhs1108 + clhs1287*clhs131 + clhs1409*clhs94 + clhs1409*clhs97 + clhs1409*clhs99 + clhs2170*clhs302 - clhs2263 + clhs2285*clhs39 - clhs2294*clhs94 - clhs2294*clhs97 - clhs2294*clhs99 + clhs321*clhs976) - clhs2153*clhs2284 - clhs2168*clhs318 - clhs2171*clhs2286 - clhs2289*clhs905 + clhs2290*clhs905;
const double clhs2296 =             -clhs1307 - clhs1310 - clhs1313;
const double clhs2297 =             clhs1140*clhs1283 + clhs1140*clhs1288 + clhs1140*clhs1855 + clhs1315*clhs973 + clhs1317*clhs973 + clhs1319*clhs973 + clhs1321*clhs976 + clhs1323*clhs973 + clhs1325*clhs976 + clhs1327 + clhs1329 + clhs1331 + clhs1333 + clhs1335 + clhs1337 + clhs2296;
const double clhs2298 =             clhs1308 + clhs1311 + clhs1314 - clhs1316 - clhs1318 - clhs1320 - clhs1322 - clhs1324 - clhs1326 - clhs1328 - clhs1330 - clhs1332 - clhs1334 - clhs1336 - clhs1338 - clhs1339 - clhs1340 - clhs1341;
const double clhs2299 =             clhs1303 + clhs2298;
const double clhs2300 =             (1.0L/2.0L)*clhs67*clhs7*tau2*(clhs1231 + clhs1232 + clhs1233 - clhs1239 + clhs1834);
const double clhs2301 =             (1.0L/2.0L)*clhs67*clhs7*tau2*(clhs1244 + clhs1245 + clhs1246 - clhs1251 + clhs2001);
const double clhs2302 =             (1.0L/2.0L)*clhs1215*clhs67*clhs7*tau2;
const double clhs2303 =             (1.0L/2.0L)*clhs7*clhs934*tau1;
const double clhs2304 =             (1.0L/2.0L)*clhs203*clhs2126*gamma*tau3;
const double clhs2305 =             -clhs1217*clhs2301 + clhs1229 + clhs1354*clhs2304 - clhs1826*clhs2300 + clhs2250*clhs2303 - clhs2302*(N[1]*clhs2205 - clhs1184*clhs1297 + 3*clhs1255 + 3*clhs1256 + clhs1272);
const double clhs2306 =             3*clhs1363*clhs906*mu;
const double clhs2307 =             clhs1278*clhs82;
const double clhs2308 =             -clhs2307;
const double clhs2309 =             3*N[1]*clhs906*mu;
const double clhs2310 =             -clhs2309*clhs902;
const double clhs2311 =             clhs10*clhs2306 - clhs1360*clhs961 - clhs2187*(clhs2308 + clhs416 + clhs418) + clhs2310;
const double clhs2312 =             3*clhs1368*clhs906*mu;
const double clhs2313 =             clhs1278*clhs89;
const double clhs2314 =             -clhs2313;
const double clhs2315 =             -clhs2309*clhs905;
const double clhs2316 =             clhs10*clhs2312 - clhs1360*clhs967 - clhs2187*(clhs2314 + clhs427 + clhs429) + clhs2315;
const double clhs2317 =             clhs2306*clhs8;
const double clhs2318 =             clhs14*clhs2312;
const double clhs2319 =             clhs1360*clhs375 - clhs2187*(clhs1856 + clhs450 + clhs453) - clhs2200*clhs302 + clhs2317 + clhs2318;
const double clhs2320 =             -clhs1113*clhs313 - clhs1113*clhs316 - clhs1113*clhs319 - clhs1113*clhs322 - clhs203*clhs329 - clhs203*clhs333 - clhs2089*clhs476;
const double clhs2321 =             N[1]*clhs10*clhs7*clhs81;
const double clhs2322 =             N[1]*clhs10*clhs7*clhs88;
const double clhs2323 =             -DN(1,0)*clhs1187 - clhs1009*clhs1384 + clhs1125*clhs1258 + clhs1306 - clhs1385*clhs976 + clhs2207*clhs2321 + clhs2208*clhs2322 + clhs2320;
const double clhs2324 =             (1.0L/2.0L)*clhs203*clhs424*tau2;
const double clhs2325 =             (1.0L/2.0L)*clhs203*clhs434*tau2;
const double clhs2326 =             (1.0L/2.0L)*clhs458*tau2;
const double clhs2327 =             (1.0L/2.0L)*clhs67*clhs7*tau2*(clhs1839 + clhs1840 + clhs1841 - clhs1846 + clhs2006);
const double clhs2328 =             (1.0L/2.0L)*clhs1826*clhs67*clhs7*tau2;
const double clhs2329 =             -clhs1215*clhs2300 - clhs1217*clhs2327 + clhs1345*clhs2304 + clhs1833 + clhs2261*clhs2303 - clhs2328*(N[1]*clhs2221 + 3*clhs1260 + 3*clhs1262 - clhs1344*clhs936 + clhs1850);
const double clhs2330 =             3*clhs1359*clhs906*mu;
const double clhs2331 =             clhs1278*clhs151;
const double clhs2332 =             -clhs2331;
const double clhs2333 =             -clhs1364*clhs964 - clhs2187*(clhs2332 + clhs414 + clhs415) + clhs2310 + clhs2330*clhs8;
const double clhs2334 =             clhs1278*clhs155;
const double clhs2335 =             -clhs2334;
const double clhs2336 =             -clhs1780*clhs2309;
const double clhs2337 =             -clhs1364*clhs967 - clhs2187*(clhs2335 + clhs466 + clhs468) + clhs2312*clhs8 + clhs2336;
const double clhs2338 =             clhs10*clhs2330;
const double clhs2339 =             2*N[1]*clhs906*mu;
const double clhs2340 =             clhs1364*clhs84 - clhs2014*clhs2187 + clhs2156*clhs2339 + clhs2318 + clhs2338;
const double clhs2341 =             -clhs1113*clhs357 - clhs1113*clhs360 - clhs1113*clhs362 - clhs1113*clhs364 - clhs203*clhs369 - clhs203*clhs372 - clhs2094*clhs448;
const double clhs2342 =             N[1]*clhs106*clhs7*clhs8;
const double clhs2343 =             N[1]*clhs7*clhs8*clhs88;
const double clhs2344 =             -DN(1,1)*clhs1201 - clhs1113*clhs352 + clhs1129*clhs1266 + clhs1309 - clhs1396*clhs976 + clhs2208*clhs2343 + clhs2222*clhs2342 + clhs2341;
const double clhs2345 =             (1.0L/2.0L)*clhs203*clhs473*tau2;
const double clhs2346 =             (1.0L/2.0L)*clhs203*clhs463*tau2;
const double clhs2347 =             (1.0L/2.0L)*clhs1871*tau2;
const double clhs2348 =             (1.0L/2.0L)*clhs1217*clhs67*clhs7*tau2;
const double clhs2349 =             -clhs1215*clhs2301 + clhs1350*clhs2304 - clhs1826*clhs2327 + clhs2000 + clhs2269*clhs2303 - clhs2348*(N[1]*clhs2228 + 3*clhs1261 + 3*clhs1263 - clhs1349*clhs937 + clhs2011);
const double clhs2350 =             clhs1278*clhs190;
const double clhs2351 =             -clhs2350;
const double clhs2352 =             -clhs1369*clhs964 + clhs14*clhs2330 - clhs2187*(clhs2351 + clhs425 + clhs426) + clhs2315;
const double clhs2353 =             clhs1278*clhs193;
const double clhs2354 =             -clhs2353;
const double clhs2355 =             -clhs1369*clhs961 + clhs14*clhs2306 - clhs2187*(clhs2354 + clhs464 + clhs465) + clhs2336;
const double clhs2356 =             -clhs1285*clhs2187 + clhs1369*clhs90 + clhs2166*clhs2339 + clhs2317 + clhs2338;
const double clhs2357 =             -clhs1113*clhs390 - clhs1113*clhs392 - clhs1113*clhs394 - clhs1113*clhs396 - clhs203*clhs400 - clhs203*clhs403 - clhs2096*clhs438;
const double clhs2358 =             N[1]*clhs106*clhs14*clhs7;
const double clhs2359 =             N[1]*clhs14*clhs7*clhs81;
const double clhs2360 =             -DN(1,2)*clhs1222 - clhs1113*clhs385 - clhs1113*clhs388 + clhs1132*clhs1269 + clhs1312 + clhs2207*clhs2359 + clhs2222*clhs2358 + clhs2357;
const double clhs2361 =             (1.0L/2.0L)*clhs203*clhs489*tau2;
const double clhs2362 =             (1.0L/2.0L)*clhs203*clhs492*tau2;
const double clhs2363 =             (1.0L/2.0L)*clhs1407*tau2;
const double clhs2364 =             clhs7*clhs934*gamma*tau1;
const double clhs2365 =             -clhs1347*gamma - clhs1352*gamma - clhs1356*gamma + clhs2270*clhs2364 + clhs301;
const double clhs2366 =             clhs1373 + clhs2025;
const double clhs2367 =             (1.0L/2.0L)*DN(1,0)*clhs67*tau2;
const double clhs2368 =             (1.0L/2.0L)*DN(1,1)*clhs67*tau2;
const double clhs2369 =             (1.0L/2.0L)*DN(1,2)*clhs67*tau2;
const double clhs2370 =             N[2]*clhs1034;
const double clhs2371 =             clhs2370*gamma;
const double clhs2372 =             clhs2371*clhs3;
const double clhs2373 =             N[0]*N[2]*clhs2246*clhs7*clhs8;
const double clhs2374 =             N[0]*N[2]*clhs14*clhs2246*clhs7;
const double clhs2375 =             clhs1444*clhs2098;
const double clhs2376 =             -clhs1125*clhs1422 - clhs1125*clhs611 - clhs1426*clhs2092 - clhs1437*clhs2093 - clhs1443*clhs2244 - clhs1450*clhs2245 - clhs1451*clhs2245 - clhs1880*clhs2092 - clhs1881*clhs2092 - clhs1882*clhs2092 - clhs2032*clhs2093 - clhs2033*clhs2093 - clhs2034*clhs2093 + clhs2372 + clhs2373*clhs82 + clhs2374*clhs89 - clhs2375*clhs3;
const double clhs2377 =             N[2]*clhs957;
const double clhs2378 =             clhs2377*gamma;
const double clhs2379 =             clhs2378*clhs3;
const double clhs2380 =             DN(0,0)*N[2]*clhs10*clhs203*clhs67;
const double clhs2381 =             N[0]*N[2]*clhs10*clhs2246*clhs7;
const double clhs2382 =             clhs1452*clhs2104;
const double clhs2383 =             -clhs1129*clhs1418 - clhs1129*clhs658 - clhs1426*clhs2257 - clhs1442*clhs2256 - clhs1443*clhs2256 - clhs1450*clhs2258 - clhs1451*clhs2256 + clhs151*clhs2381 + clhs155*clhs2374 - clhs1882*clhs2257 - clhs1890*clhs2093 - clhs2037*clhs2093 - clhs2038*clhs2093 - clhs2039*clhs2093 + clhs2379 - clhs2380*clhs84 - clhs2382*clhs3;
const double clhs2384 =             N[2]*clhs999;
const double clhs2385 =             clhs2384*gamma;
const double clhs2386 =             clhs2385*clhs3;
const double clhs2387 =             clhs1455*clhs2110;
const double clhs2388 =             -clhs1132*clhs1430 - clhs1132*clhs684 - clhs1437*clhs2257 - clhs1442*clhs2266 - clhs1443*clhs2266 - clhs1446*clhs2266 - clhs1448*clhs2266 - clhs1450*clhs2266 - clhs1451*clhs2267 - clhs1890*clhs2092 + clhs190*clhs2381 + clhs193*clhs2373 - clhs2034*clhs2257 - clhs2039*clhs2092 - clhs2380*clhs90 + clhs2386 - clhs2387*clhs3;
const double clhs2389 =             clhs1447 + clhs1449 + clhs1451 + clhs1457 + clhs2041;
const double clhs2390 =             clhs1068*clhs1416 + clhs1293*(-clhs10*clhs2372 - clhs1140*clhs1441 - clhs1140*clhs1442 - clhs1140*clhs1446 - clhs1140*clhs1447 - clhs1140*clhs1448 - clhs1140*clhs1449 - clhs14*clhs2386 + clhs1426*clhs2176 + clhs1437*clhs2177 - clhs1443*clhs2098 - clhs1450*clhs2104 - clhs1451*clhs2110 - clhs1481*clhs2272 - clhs1527*clhs2273 - clhs1532*clhs2274 + clhs1882*clhs2176 + clhs1890*clhs2178 + clhs2034*clhs2177 + clhs2039*clhs2178 - clhs2379*clhs8) - clhs2240*clhs2376 - clhs2251*clhs2383 - clhs2262*clhs2388 - clhs2271*clhs2389;
const double clhs2391 =             (1.0L/4.0L)*tau2*(-clhs1484 + clhs503 + clhs507 + clhs511 + clhs514 + clhs517 + clhs520 + clhs523 + clhs527 + clhs531 + clhs535 + clhs538 + clhs541 + clhs544);
const double clhs2392 =             (1.0L/4.0L)*tau2*(-clhs2114*clhs580 + clhs549 + clhs553 + clhs556 + clhs558 + clhs561 + clhs563 + clhs565 + clhs567 + clhs571 + clhs574 + clhs576 + clhs577 + clhs579);
const double clhs2393 =             (1.0L/4.0L)*tau2*(-clhs2117*clhs580 + clhs583 + clhs585 + clhs588 + clhs590 + clhs592 + clhs594 + clhs596 + clhs598 + clhs601 + clhs604 + clhs605 + clhs606 + clhs608);
const double clhs2394 =             3*N[2]*clhs203*clhs8*v_sc;
const double clhs2395 =             3*N[2]*clhs14*clhs203*v_sc;
const double clhs2396 =             3*N[2]*clhs7*clhs8*clhs906*mu;
const double clhs2397 =             3*N[2]*clhs14*clhs7*clhs906*mu;
const double clhs2398 =             3*clhs1464*clhs906*mu;
const double clhs2399 =             3*clhs1465*clhs906*mu;
const double clhs2400 =             2*N[2]*clhs10*clhs203;
const double clhs2401 =             3*N[2]*k_sc;
const double clhs2402 =             3*N[2]*clhs106*clhs7;
const double clhs2403 =             -clhs1181*clhs2398 - clhs1183*clhs2399 + clhs1475*clhs2141 - clhs2139*clhs2400 + clhs2140*clhs559 - clhs2144*clhs2401 + clhs2147*(-DN(2,0)*clhs1108 + clhs1467*clhs76 + clhs1472*clhs66 + clhs1569*clhs94 + clhs1569*clhs97 + clhs1569*clhs99 + clhs2143*clhs502 - clhs2370 - clhs2402*clhs94 - clhs2402*clhs97 - clhs2402*clhs99 + clhs524*clhs973) - clhs2394*clhs902 - clhs2395*clhs905 + clhs2396*clhs902 + clhs2397*clhs905;
const double clhs2404 =             3*N[2]*clhs10*clhs203*v_sc;
const double clhs2405 =             3*N[2]*clhs10*clhs7*clhs906*mu;
const double clhs2406 =             3*clhs1898*clhs906*mu;
const double clhs2407 =             3*N[2]*clhs7*clhs81;
const double clhs2408 =             -clhs1183*clhs2406 + clhs1472*clhs2157 - clhs1780*clhs2395 + clhs1780*clhs2397 + clhs1902*clhs2159 + clhs2147*(-DN(2,1)*clhs1108 + clhs1009*clhs515 + clhs1467*clhs147 + clhs1579*clhs94 + clhs1579*clhs97 + clhs1579*clhs99 + clhs2161*clhs502 + clhs23*clhs2400 - clhs2377 - clhs2407*clhs94 - clhs2407*clhs97 - clhs2407*clhs99) - clhs2153*clhs2398 - clhs2158*clhs512 - clhs2162*clhs2401 - clhs2404*clhs902 + clhs2405*clhs902;
const double clhs2409 =             3*N[2]*clhs7*clhs88;
const double clhs2410 =             -clhs1181*clhs2406 + clhs1467*clhs2167 - clhs1780*clhs2394 + clhs1780*clhs2396 + clhs2047*clhs2169 + clhs2147*(-DN(2,2)*clhs1108 + clhs131*clhs1472 + clhs1590*clhs94 + clhs1590*clhs97 + clhs1590*clhs99 + clhs2170*clhs502 - clhs2384 + clhs2400*clhs39 - clhs2409*clhs94 - clhs2409*clhs97 - clhs2409*clhs99 + clhs521*clhs976) - clhs2153*clhs2399 - clhs2168*clhs518 - clhs2171*clhs2401 - clhs2404*clhs905 + clhs2405*clhs905;
const double clhs2411 =             -clhs1491 - clhs1494 - clhs1497;
const double clhs2412 =             clhs1140*clhs1468 + clhs1140*clhs1473 + clhs1140*clhs1900 + clhs1499*clhs973 + clhs1501*clhs973 + clhs1503*clhs973 + clhs1505*clhs976 + clhs1507*clhs973 + clhs1509*clhs976 + clhs1511 + clhs1513 + clhs1515 + clhs1517 + clhs1519 + clhs1521 + clhs2411;
const double clhs2413 =             clhs1492 + clhs1495 + clhs1498 - clhs1500 - clhs1502 - clhs1504 - clhs1506 - clhs1508 - clhs1510 - clhs1512 - clhs1514 - clhs1516 - clhs1518 - clhs1520 - clhs1522 - clhs1523 - clhs1524 - clhs1525;
const double clhs2414 =             clhs1487 + clhs2413;
const double clhs2415 =             (1.0L/2.0L)*clhs67*clhs7*tau2*(clhs1419 + clhs1420 + clhs1421 - clhs1427 + clhs1879);
const double clhs2416 =             (1.0L/2.0L)*clhs67*clhs7*tau2*(clhs1431 + clhs1432 + clhs1433 - clhs1438 + clhs2031);
const double clhs2417 =             -clhs1217*clhs2416 + clhs1417 + clhs1537*clhs2304 - clhs1826*clhs2415 - clhs2302*(N[2]*clhs2205 - clhs1184*clhs1481 + 3*clhs1441 + 3*clhs1442 + clhs1458) + clhs2303*clhs2376;
const double clhs2418 =             3*clhs1546*clhs906*mu;
const double clhs2419 =             clhs1463*clhs82;
const double clhs2420 =             -clhs2419;
const double clhs2421 =             3*N[2]*clhs906*mu;
const double clhs2422 =             -clhs2421*clhs902;
const double clhs2423 =             clhs10*clhs2418 - clhs1543*clhs961 - clhs2187*(clhs2420 + clhs615 + clhs617) + clhs2422;
const double clhs2424 =             3*clhs1551*clhs906*mu;
const double clhs2425 =             clhs1463*clhs89;
const double clhs2426 =             -clhs2425;
const double clhs2427 =             -clhs2421*clhs905;
const double clhs2428 =             clhs10*clhs2424 - clhs1543*clhs967 - clhs2187*(clhs2426 + clhs626 + clhs628) + clhs2427;
const double clhs2429 =             clhs2418*clhs8;
const double clhs2430 =             clhs14*clhs2424;
const double clhs2431 =             clhs1543*clhs375 - clhs2187*(clhs1901 + clhs649 + clhs652) - clhs2200*clhs502 + clhs2429 + clhs2430;
const double clhs2432 =             -clhs1113*clhs513 - clhs1113*clhs516 - clhs1113*clhs519 - clhs1113*clhs522 - clhs203*clhs529 - clhs203*clhs533 - clhs2089*clhs675;
const double clhs2433 =             N[2]*clhs10*clhs7*clhs81;
const double clhs2434 =             N[2]*clhs10*clhs7*clhs88;
const double clhs2435 =             -DN(2,0)*clhs1187 - clhs1009*clhs1567 + clhs1125*clhs1444 + clhs1490 - clhs1568*clhs976 + clhs2207*clhs2433 + clhs2208*clhs2434 + clhs2432;
const double clhs2436 =             (1.0L/2.0L)*clhs203*clhs623*tau2;
const double clhs2437 =             (1.0L/2.0L)*clhs203*clhs633*tau2;
const double clhs2438 =             (1.0L/2.0L)*clhs657*tau2;
const double clhs2439 =             (1.0L/2.0L)*clhs67*clhs7*tau2*(clhs1884 + clhs1885 + clhs1886 - clhs1891 + clhs2036);
const double clhs2440 =             -clhs1215*clhs2415 - clhs1217*clhs2439 + clhs1528*clhs2304 + clhs1878 + clhs2303*clhs2383 - clhs2328*(N[2]*clhs2221 + 3*clhs1446 + 3*clhs1448 - clhs1527*clhs936 + clhs1895);
const double clhs2441 =             3*clhs1542*clhs906*mu;
const double clhs2442 =             clhs1463*clhs151;
const double clhs2443 =             -clhs2442;
const double clhs2444 =             -clhs1547*clhs964 - clhs2187*(clhs2443 + clhs613 + clhs614) + clhs2422 + clhs2441*clhs8;
const double clhs2445 =             clhs1463*clhs155;
const double clhs2446 =             -clhs2445;
const double clhs2447 =             -clhs1780*clhs2421;
const double clhs2448 =             -clhs1547*clhs967 - clhs2187*(clhs2446 + clhs665 + clhs667) + clhs2424*clhs8 + clhs2447;
const double clhs2449 =             clhs10*clhs2441;
const double clhs2450 =             2*N[2]*clhs906*mu;
const double clhs2451 =             clhs1547*clhs84 - clhs2044*clhs2187 + clhs2156*clhs2450 + clhs2430 + clhs2449;
const double clhs2452 =             -clhs1113*clhs557 - clhs1113*clhs560 - clhs1113*clhs562 - clhs1113*clhs564 - clhs203*clhs569 - clhs203*clhs572 - clhs2094*clhs647;
const double clhs2453 =             N[2]*clhs106*clhs7*clhs8;
const double clhs2454 =             N[2]*clhs7*clhs8*clhs88;
const double clhs2455 =             -DN(2,1)*clhs1201 - clhs1113*clhs552 + clhs1129*clhs1452 + clhs1493 - clhs1578*clhs976 + clhs2208*clhs2454 + clhs2222*clhs2453 + clhs2452;
const double clhs2456 =             (1.0L/2.0L)*clhs203*clhs672*tau2;
const double clhs2457 =             (1.0L/2.0L)*clhs203*clhs662*tau2;
const double clhs2458 =             (1.0L/2.0L)*clhs1916*tau2;
const double clhs2459 =             -clhs1215*clhs2416 + clhs1533*clhs2304 - clhs1826*clhs2439 + clhs2030 + clhs2303*clhs2388 - clhs2348*(N[2]*clhs2228 + 3*clhs1447 + 3*clhs1449 - clhs1532*clhs937 + clhs2041);
const double clhs2460 =             clhs1463*clhs190;
const double clhs2461 =             -clhs2460;
const double clhs2462 =             clhs14*clhs2441 - clhs1552*clhs964 - clhs2187*(clhs2461 + clhs624 + clhs625) + clhs2427;
const double clhs2463 =             clhs1463*clhs193;
const double clhs2464 =             -clhs2463;
const double clhs2465 =             clhs14*clhs2418 - clhs1552*clhs961 - clhs2187*(clhs2464 + clhs663 + clhs664) + clhs2447;
const double clhs2466 =             -clhs1470*clhs2187 + clhs1552*clhs90 + clhs2166*clhs2450 + clhs2429 + clhs2449;
const double clhs2467 =             -clhs1113*clhs589 - clhs1113*clhs591 - clhs1113*clhs593 - clhs1113*clhs595 - clhs203*clhs599 - clhs203*clhs602 - clhs2096*clhs637;
const double clhs2468 =             N[2]*clhs106*clhs14*clhs7;
const double clhs2469 =             N[2]*clhs14*clhs7*clhs81;
const double clhs2470 =             -DN(2,2)*clhs1222 - clhs1113*clhs584 - clhs1113*clhs587 + clhs1132*clhs1455 + clhs1496 + clhs2207*clhs2469 + clhs2222*clhs2468 + clhs2467;
const double clhs2471 =             (1.0L/2.0L)*clhs203*clhs688*tau2;
const double clhs2472 =             (1.0L/2.0L)*clhs203*clhs691*tau2;
const double clhs2473 =             (1.0L/2.0L)*clhs1588*tau2;
const double clhs2474 =             -clhs1530*gamma - clhs1535*gamma - clhs1539*gamma + clhs2364*clhs2389 + clhs501;
const double clhs2475 =             clhs1556 + clhs2054;
const double clhs2476 =             (1.0L/2.0L)*DN(2,0)*clhs67*tau2;
const double clhs2477 =             (1.0L/2.0L)*DN(2,1)*clhs67*tau2;
const double clhs2478 =             (1.0L/2.0L)*DN(2,2)*clhs67*tau2;
const double clhs2479 =             N[3]*clhs1034;
const double clhs2480 =             clhs2479*gamma;
const double clhs2481 =             clhs2480*clhs3;
const double clhs2482 =             N[0]*N[3]*clhs2246*clhs7*clhs8;
const double clhs2483 =             N[0]*N[3]*clhs14*clhs2246*clhs7;
const double clhs2484 =             clhs1625*clhs2098;
const double clhs2485 =             -clhs1125*clhs1603 - clhs1125*clhs810 - clhs1607*clhs2092 - clhs1618*clhs2093 - clhs1624*clhs2244 - clhs1631*clhs2245 - clhs1632*clhs2245 - clhs1925*clhs2092 - clhs1926*clhs2092 - clhs1927*clhs2092 - clhs2061*clhs2093 - clhs2062*clhs2093 - clhs2063*clhs2093 + clhs2481 + clhs2482*clhs82 + clhs2483*clhs89 - clhs2484*clhs3;
const double clhs2486 =             N[3]*clhs957;
const double clhs2487 =             clhs2486*gamma;
const double clhs2488 =             clhs2487*clhs3;
const double clhs2489 =             DN(0,0)*N[3]*clhs10*clhs203*clhs67;
const double clhs2490 =             N[0]*N[3]*clhs10*clhs2246*clhs7;
const double clhs2491 =             clhs1633*clhs2104;
const double clhs2492 =             -clhs1129*clhs1599 - clhs1129*clhs857 + clhs151*clhs2490 + clhs155*clhs2483 - clhs1607*clhs2257 - clhs1623*clhs2256 - clhs1624*clhs2256 - clhs1631*clhs2258 - clhs1632*clhs2256 - clhs1927*clhs2257 - clhs1935*clhs2093 - clhs2066*clhs2093 - clhs2067*clhs2093 - clhs2068*clhs2093 + clhs2488 - clhs2489*clhs84 - clhs2491*clhs3;
const double clhs2493 =             N[3]*clhs999;
const double clhs2494 =             clhs2493*gamma;
const double clhs2495 =             clhs2494*clhs3;
const double clhs2496 =             clhs1636*clhs2110;
const double clhs2497 =             -clhs1132*clhs1611 - clhs1132*clhs883 - clhs1618*clhs2257 - clhs1623*clhs2266 - clhs1624*clhs2266 - clhs1627*clhs2266 - clhs1629*clhs2266 - clhs1631*clhs2266 - clhs1632*clhs2267 + clhs190*clhs2490 + clhs193*clhs2482 - clhs1935*clhs2092 - clhs2063*clhs2257 - clhs2068*clhs2092 - clhs2489*clhs90 + clhs2495 - clhs2496*clhs3;
const double clhs2498 =             clhs1628 + clhs1630 + clhs1632 + clhs1638 + clhs2070;
const double clhs2499 =             clhs1068*clhs1597 + clhs1293*(-clhs10*clhs2481 - clhs1140*clhs1622 - clhs1140*clhs1623 - clhs1140*clhs1627 - clhs1140*clhs1628 - clhs1140*clhs1629 - clhs1140*clhs1630 - clhs14*clhs2495 + clhs1607*clhs2176 + clhs1618*clhs2177 - clhs1624*clhs2098 - clhs1631*clhs2104 - clhs1632*clhs2110 - clhs1662*clhs2272 - clhs1708*clhs2273 - clhs1713*clhs2274 + clhs1927*clhs2176 + clhs1935*clhs2178 + clhs2063*clhs2177 + clhs2068*clhs2178 - clhs2488*clhs8) - clhs2240*clhs2485 - clhs2251*clhs2492 - clhs2262*clhs2497 - clhs2271*clhs2498;
const double clhs2500 =             (1.0L/4.0L)*tau2*(-clhs1665 + clhs702 + clhs706 + clhs710 + clhs713 + clhs716 + clhs719 + clhs722 + clhs726 + clhs730 + clhs734 + clhs737 + clhs740 + clhs743);
const double clhs2501 =             (1.0L/4.0L)*tau2*(-clhs2114*clhs779 + clhs748 + clhs752 + clhs755 + clhs757 + clhs760 + clhs762 + clhs764 + clhs766 + clhs770 + clhs773 + clhs775 + clhs776 + clhs778);
const double clhs2502 =             (1.0L/4.0L)*tau2*(-clhs2117*clhs779 + clhs782 + clhs784 + clhs787 + clhs789 + clhs791 + clhs793 + clhs795 + clhs797 + clhs800 + clhs803 + clhs804 + clhs805 + clhs807);
const double clhs2503 =             3*N[3]*clhs203*clhs8*v_sc;
const double clhs2504 =             3*N[3]*clhs14*clhs203*v_sc;
const double clhs2505 =             3*N[3]*clhs7*clhs8*clhs906*mu;
const double clhs2506 =             3*N[3]*clhs14*clhs7*clhs906*mu;
const double clhs2507 =             3*clhs1645*clhs906*mu;
const double clhs2508 =             3*clhs1646*clhs906*mu;
const double clhs2509 =             2*N[3]*clhs10*clhs203;
const double clhs2510 =             3*N[3]*k_sc;
const double clhs2511 =             3*N[3]*clhs106*clhs7;
const double clhs2512 =             -clhs1181*clhs2507 - clhs1183*clhs2508 + clhs1656*clhs2141 - clhs2139*clhs2509 + clhs2140*clhs758 - clhs2144*clhs2510 + clhs2147*(-DN(3,0)*clhs1108 + clhs1648*clhs76 + clhs1653*clhs66 + clhs1750*clhs94 + clhs1750*clhs97 + clhs1750*clhs99 + clhs2143*clhs701 - clhs2479 - clhs2511*clhs94 - clhs2511*clhs97 - clhs2511*clhs99 + clhs723*clhs973) - clhs2503*clhs902 - clhs2504*clhs905 + clhs2505*clhs902 + clhs2506*clhs905;
const double clhs2513 =             3*N[3]*clhs10*clhs203*v_sc;
const double clhs2514 =             3*N[3]*clhs10*clhs7*clhs906*mu;
const double clhs2515 =             3*clhs1943*clhs906*mu;
const double clhs2516 =             3*N[3]*clhs7*clhs81;
const double clhs2517 =             -clhs1183*clhs2515 + clhs1653*clhs2157 - clhs1780*clhs2504 + clhs1780*clhs2506 + clhs1947*clhs2159 + clhs2147*(-DN(3,1)*clhs1108 + clhs1009*clhs714 + clhs147*clhs1648 + clhs1760*clhs94 + clhs1760*clhs97 + clhs1760*clhs99 + clhs2161*clhs701 + clhs23*clhs2509 - clhs2486 - clhs2516*clhs94 - clhs2516*clhs97 - clhs2516*clhs99) - clhs2153*clhs2507 - clhs2158*clhs711 - clhs2162*clhs2510 - clhs2513*clhs902 + clhs2514*clhs902;
const double clhs2518 =             3*N[3]*clhs7*clhs88;
const double clhs2519 =             -clhs1181*clhs2515 + clhs1648*clhs2167 - clhs1780*clhs2503 + clhs1780*clhs2505 + clhs2076*clhs2169 + clhs2147*(-DN(3,2)*clhs1108 + clhs131*clhs1653 + clhs1771*clhs94 + clhs1771*clhs97 + clhs1771*clhs99 + clhs2170*clhs701 - clhs2493 + clhs2509*clhs39 - clhs2518*clhs94 - clhs2518*clhs97 - clhs2518*clhs99 + clhs720*clhs976) - clhs2153*clhs2508 - clhs2168*clhs717 - clhs2171*clhs2510 - clhs2513*clhs905 + clhs2514*clhs905;
const double clhs2520 =             -clhs1672 - clhs1675 - clhs1678;
const double clhs2521 =             clhs1140*clhs1649 + clhs1140*clhs1654 + clhs1140*clhs1945 + clhs1680*clhs973 + clhs1682*clhs973 + clhs1684*clhs973 + clhs1686*clhs976 + clhs1688*clhs973 + clhs1690*clhs976 + clhs1692 + clhs1694 + clhs1696 + clhs1698 + clhs1700 + clhs1702 + clhs2520;
const double clhs2522 =             clhs1673 + clhs1676 + clhs1679 - clhs1681 - clhs1683 - clhs1685 - clhs1687 - clhs1689 - clhs1691 - clhs1693 - clhs1695 - clhs1697 - clhs1699 - clhs1701 - clhs1703 - clhs1704 - clhs1705 - clhs1706;
const double clhs2523 =             clhs1668 + clhs2522;
const double clhs2524 =             (1.0L/2.0L)*clhs67*clhs7*tau2*(clhs1600 + clhs1601 + clhs1602 - clhs1608 + clhs1924);
const double clhs2525 =             (1.0L/2.0L)*clhs67*clhs7*tau2*(clhs1612 + clhs1613 + clhs1614 - clhs1619 + clhs2060);
const double clhs2526 =             -clhs1217*clhs2525 + clhs1598 + clhs1718*clhs2304 - clhs1826*clhs2524 - clhs2302*(N[3]*clhs2205 - clhs1184*clhs1662 + 3*clhs1622 + 3*clhs1623 + clhs1639) + clhs2303*clhs2485;
const double clhs2527 =             3*clhs1727*clhs906*mu;
const double clhs2528 =             clhs1644*clhs82;
const double clhs2529 =             -clhs2528;
const double clhs2530 =             3*N[3]*clhs906*mu;
const double clhs2531 =             -clhs2530*clhs902;
const double clhs2532 =             clhs10*clhs2527 - clhs1724*clhs961 - clhs2187*(clhs2529 + clhs814 + clhs816) + clhs2531;
const double clhs2533 =             3*clhs1732*clhs906*mu;
const double clhs2534 =             clhs1644*clhs89;
const double clhs2535 =             -clhs2534;
const double clhs2536 =             -clhs2530*clhs905;
const double clhs2537 =             clhs10*clhs2533 - clhs1724*clhs967 - clhs2187*(clhs2535 + clhs825 + clhs827) + clhs2536;
const double clhs2538 =             clhs2527*clhs8;
const double clhs2539 =             clhs14*clhs2533;
const double clhs2540 =             clhs1724*clhs375 - clhs2187*(clhs1946 + clhs848 + clhs851) - clhs2200*clhs701 + clhs2538 + clhs2539;
const double clhs2541 =             -clhs1113*clhs712 - clhs1113*clhs715 - clhs1113*clhs718 - clhs1113*clhs721 - clhs203*clhs728 - clhs203*clhs732 - clhs2089*clhs874;
const double clhs2542 =             N[3]*clhs10*clhs7*clhs81;
const double clhs2543 =             N[3]*clhs10*clhs7*clhs88;
const double clhs2544 =             -DN(3,0)*clhs1187 - clhs1009*clhs1748 + clhs1125*clhs1625 + clhs1671 - clhs1749*clhs976 + clhs2207*clhs2542 + clhs2208*clhs2543 + clhs2541;
const double clhs2545 =             (1.0L/2.0L)*clhs203*clhs822*tau2;
const double clhs2546 =             (1.0L/2.0L)*clhs203*clhs832*tau2;
const double clhs2547 =             (1.0L/2.0L)*clhs856*tau2;
const double clhs2548 =             (1.0L/2.0L)*clhs67*clhs7*tau2*(clhs1929 + clhs1930 + clhs1931 - clhs1936 + clhs2065);
const double clhs2549 =             -clhs1215*clhs2524 - clhs1217*clhs2548 + clhs1709*clhs2304 + clhs1923 + clhs2303*clhs2492 - clhs2328*(N[3]*clhs2221 + 3*clhs1627 + 3*clhs1629 - clhs1708*clhs936 + clhs1940);
const double clhs2550 =             3*clhs1723*clhs906*mu;
const double clhs2551 =             clhs151*clhs1644;
const double clhs2552 =             -clhs2551;
const double clhs2553 =             -clhs1728*clhs964 - clhs2187*(clhs2552 + clhs812 + clhs813) + clhs2531 + clhs2550*clhs8;
const double clhs2554 =             clhs155*clhs1644;
const double clhs2555 =             -clhs2554;
const double clhs2556 =             -clhs1780*clhs2530;
const double clhs2557 =             -clhs1728*clhs967 - clhs2187*(clhs2555 + clhs864 + clhs866) + clhs2533*clhs8 + clhs2556;
const double clhs2558 =             clhs10*clhs2550;
const double clhs2559 =             2*N[3]*clhs906*mu;
const double clhs2560 =             clhs1728*clhs84 - clhs2073*clhs2187 + clhs2156*clhs2559 + clhs2539 + clhs2558;
const double clhs2561 =             -clhs1113*clhs756 - clhs1113*clhs759 - clhs1113*clhs761 - clhs1113*clhs763 - clhs203*clhs768 - clhs203*clhs771 - clhs2094*clhs846;
const double clhs2562 =             N[3]*clhs106*clhs7*clhs8;
const double clhs2563 =             N[3]*clhs7*clhs8*clhs88;
const double clhs2564 =             -DN(3,1)*clhs1201 - clhs1113*clhs751 + clhs1129*clhs1633 + clhs1674 - clhs1759*clhs976 + clhs2208*clhs2563 + clhs2222*clhs2562 + clhs2561;
const double clhs2565 =             (1.0L/2.0L)*clhs203*clhs871*tau2;
const double clhs2566 =             (1.0L/2.0L)*clhs203*clhs861*tau2;
const double clhs2567 =             (1.0L/2.0L)*clhs1961*tau2;
const double clhs2568 =             -clhs1215*clhs2525 + clhs1714*clhs2304 - clhs1826*clhs2548 + clhs2059 + clhs2303*clhs2497 - clhs2348*(N[3]*clhs2228 + 3*clhs1628 + 3*clhs1630 - clhs1713*clhs937 + clhs2070);
const double clhs2569 =             clhs1644*clhs190;
const double clhs2570 =             -clhs2569;
const double clhs2571 =             clhs14*clhs2550 - clhs1733*clhs964 - clhs2187*(clhs2570 + clhs823 + clhs824) + clhs2536;
const double clhs2572 =             clhs1644*clhs193;
const double clhs2573 =             -clhs2572;
const double clhs2574 =             clhs14*clhs2527 - clhs1733*clhs961 - clhs2187*(clhs2573 + clhs862 + clhs863) + clhs2556;
const double clhs2575 =             -clhs1651*clhs2187 + clhs1733*clhs90 + clhs2166*clhs2559 + clhs2538 + clhs2558;
const double clhs2576 =             -clhs1113*clhs788 - clhs1113*clhs790 - clhs1113*clhs792 - clhs1113*clhs794 - clhs203*clhs798 - clhs203*clhs801 - clhs2096*clhs836;
const double clhs2577 =             N[3]*clhs106*clhs14*clhs7;
const double clhs2578 =             N[3]*clhs14*clhs7*clhs81;
const double clhs2579 =             -DN(3,2)*clhs1222 - clhs1113*clhs783 - clhs1113*clhs786 + clhs1132*clhs1636 + clhs1677 + clhs2207*clhs2578 + clhs2222*clhs2577 + clhs2576;
const double clhs2580 =             (1.0L/2.0L)*clhs203*clhs887*tau2;
const double clhs2581 =             (1.0L/2.0L)*clhs203*clhs890*tau2;
const double clhs2582 =             (1.0L/2.0L)*clhs1769*tau2;
const double clhs2583 =             -clhs1711*gamma - clhs1716*gamma - clhs1720*gamma + clhs2364*clhs2498 + clhs700;
const double clhs2584 =             clhs1737 + clhs2083;
const double clhs2585 =             (1.0L/2.0L)*DN(3,0)*clhs67*tau2;
const double clhs2586 =             (1.0L/2.0L)*DN(3,1)*clhs67*tau2;
const double clhs2587 =             (1.0L/2.0L)*DN(3,2)*clhs67*tau2;
const double clhs2588 =             (1.0L/2.0L)*DN(1,0)*tau2;
const double clhs2589 =             (1.0L/2.0L)*DN(1,1)*tau2;
const double clhs2590 =             (1.0L/2.0L)*DN(1,2)*tau2;
const double clhs2591 =             -clhs1234;
const double clhs2592 =             DN(1,0)*tau2;
const double clhs2593 =             -clhs1230;
const double clhs2594 =             DN(1,1)*tau2;
const double clhs2595 =             -clhs1243;
const double clhs2596 =             DN(1,2)*tau2;
const double clhs2597 =             pow(N[1], 2);
const double clhs2598 =             -bdf0*clhs2597;
const double clhs2599 =             -DN(1,0)*N[1];
const double clhs2600 =             -DN(1,1)*N[1];
const double clhs2601 =             -DN(1,2)*N[1];
const double clhs2602 =             -N[2]*clhs442;
const double clhs2603 =             DN(2,0)*N[1];
const double clhs2604 =             -clhs2603;
const double clhs2605 =             DN(2,1)*N[1];
const double clhs2606 =             -clhs2605;
const double clhs2607 =             DN(2,2)*N[1];
const double clhs2608 =             -clhs2607;
const double clhs2609 =             -clhs300*(DN(1,0)*DN(2,0) + DN(1,1)*DN(2,1) + DN(1,2)*DN(2,2));
const double clhs2610 =             -N[3]*clhs442;
const double clhs2611 =             DN(3,0)*N[1];
const double clhs2612 =             -clhs2611;
const double clhs2613 =             DN(3,1)*N[1];
const double clhs2614 =             -clhs2613;
const double clhs2615 =             DN(3,2)*N[1];
const double clhs2616 =             -clhs2615;
const double clhs2617 =             -clhs300*(DN(1,0)*DN(3,0) + DN(1,1)*DN(3,1) + DN(1,2)*DN(3,2));
const double clhs2618 =             DN(1,1)*N[0]*clhs203*v_sc;
const double clhs2619 =             DN(1,2)*N[0]*clhs203*v_sc;
const double clhs2620 =             DN(1,1)*clhs7*clhs906*mu;
const double clhs2621 =             DN(1,2)*clhs7*clhs906*mu;
const double clhs2622 =             DN(1,0)*N[0]*clhs203*v_sc;
const double clhs2623 =             (2.0L/3.0L)*DN(1,0)*clhs7*clhs906*mu;
const double clhs2624 =             (1.0L/2.0L)*N[1]*clhs7;
const double clhs2625 =             clhs1301 + clhs303 - clhs346;
const double clhs2626 =             (1.0L/2.0L)*clhs1377*clhs203*tau2;
const double clhs2627 =             (1.0L/2.0L)*DN(1,0)*clhs67*tau3;
const double clhs2628 =             clhs1159*clhs1386;
const double clhs2629 =             clhs1151*clhs1397;
const double clhs2630 =             -clhs2629;
const double clhs2631 =             clhs1164*clhs1409;
const double clhs2632 =             -clhs2631;
const double clhs2633 =             clhs1377*clhs203*tau2;
const double clhs2634 =             clhs1151*clhs1386;
const double clhs2635 =             clhs1159*clhs1397;
const double clhs2636 =             clhs203*clhs463*tau2;
const double clhs2637 =             clhs1164*clhs1386;
const double clhs2638 =             clhs1159*clhs1409;
const double clhs2639 =             -clhs437;
const double clhs2640 =             clhs1372 + clhs2639 - clhs435;
const double clhs2641 =             -clhs444;
const double clhs2642 =             -clhs447;
const double clhs2643 =             clhs1374 - clhs1376 + clhs2640 + clhs2641 + clhs2642 + clhs451 + clhs454;
const double clhs2644 =             clhs203*clhs489*tau2;
const double clhs2645 =             clhs1299 + clhs347;
const double clhs2646 =             DN(1,0)*tau3;
const double clhs2647 =             clhs2597*clhs4;
const double clhs2648 =             DN(1,1)*N[1]*clhs203*v_sc;
const double clhs2649 =             DN(1,2)*N[1]*clhs203*v_sc;
const double clhs2650 =             DN(1,0)*N[1]*clhs203*v_sc;
const double clhs2651 =             N[1]*clhs83*clhs934*tau1;
const double clhs2652 =             clhs350*clhs8;
const double clhs2653 =             clhs2307 - clhs2331*clhs67 + clhs2652*clhs67 - clhs305 + clhs462;
const double clhs2654 =             (1.0L/2.0L)*N[1]*clhs7*clhs985*tau2;
const double clhs2655 =             clhs14*clhs350;
const double clhs2656 =             clhs2313 - clhs2350*clhs67 + clhs2655*clhs67 - clhs309 + clhs488;
const double clhs2657 =             (1.0L/2.0L)*N[1]*clhs1017*clhs7*tau2;
const double clhs2658 =             clhs1283 + clhs1288 + clhs1854 - clhs1855*clhs57 + clhs2016 + clhs2639 + clhs2642 + clhs351*clhs57 + clhs454;
const double clhs2659 =             (1.0L/2.0L)*N[1]*clhs1050*clhs7*tau2;
const double clhs2660 =             clhs1364*clhs1397;
const double clhs2661 =             -clhs2660;
const double clhs2662 =             clhs1380*clhs489;
const double clhs2663 =             clhs1360*clhs1386;
const double clhs2664 =             clhs1369*clhs1409;
const double clhs2665 =             -clhs2664;
const double clhs2666 =             clhs1378*clhs463;
const double clhs2667 =             N[1]*clhs7*clhs934*tau1;
const double clhs2668 =             clhs1362 + clhs304;
const double clhs2669 =             (1.0L/2.0L)*N[1]*clhs203*clhs2668*tau2;
const double clhs2670 =             clhs1367 + clhs308;
const double clhs2671 =             (1.0L/2.0L)*N[1]*clhs203*clhs2670*tau2;
const double clhs2672 =             clhs1358 + clhs350;
const double clhs2673 =             (1.0L/2.0L)*N[1]*clhs203*clhs2672*tau2;
const double clhs2674 =             clhs1364*clhs1386;
const double clhs2675 =             clhs1360*clhs1397;
const double clhs2676 =             (1.0L/2.0L)*N[1]*clhs203*clhs67*clhs985*tau2;
const double clhs2677 =             clhs1369*clhs1386;
const double clhs2678 =             clhs1360*clhs1409;
const double clhs2679 =             (1.0L/2.0L)*N[1]*clhs203*clhs67*tau2;
const double clhs2680 =             N[1]*N[2];
const double clhs2681 =             clhs2680*clhs4;
const double clhs2682 =             DN(1,1)*N[2];
const double clhs2683 =             clhs10*clhs2682;
const double clhs2684 =             clhs10*clhs2605;
const double clhs2685 =             N[2]*clhs418;
const double clhs2686 =             DN(1,0)*N[2];
const double clhs2687 =             clhs2686*clhs8;
const double clhs2688 =             clhs2603*clhs8;
const double clhs2689 =             N[2]*clhs415;
const double clhs2690 =             clhs2689*clhs67;
const double clhs2691 =             clhs302*clhs622;
const double clhs2692 =             clhs302*clhs621;
const double clhs2693 =             -clhs2683 - clhs2684 - clhs2685 + clhs2687*clhs67 + clhs2688*clhs67 + clhs2690 + clhs2691 - clhs2692*clhs67;
const double clhs2694 =             DN(1,2)*N[2];
const double clhs2695 =             clhs10*clhs2694;
const double clhs2696 =             clhs10*clhs2607;
const double clhs2697 =             N[2]*clhs429;
const double clhs2698 =             clhs14*clhs2686;
const double clhs2699 =             clhs14*clhs2603;
const double clhs2700 =             N[2]*clhs426;
const double clhs2701 =             clhs2700*clhs67;
const double clhs2702 =             clhs302*clhs632;
const double clhs2703 =             clhs302*clhs631;
const double clhs2704 =             -clhs2695 - clhs2696 - clhs2697 + clhs2698*clhs67 + clhs2699*clhs67 + clhs2701 + clhs2702 - clhs2703*clhs67;
const double clhs2705 =             clhs10*clhs2686;
const double clhs2706 =             clhs10*clhs2603;
const double clhs2707 =             N[2]*clhs453;
const double clhs2708 =             clhs1444*clhs359;
const double clhs2709 =             clhs2682*clhs8;
const double clhs2710 =             clhs14*clhs2694;
const double clhs2711 =             clhs2605*clhs8;
const double clhs2712 =             clhs14*clhs2607;
const double clhs2713 =             N[2]*clhs447;
const double clhs2714 =             N[2]*clhs437;
const double clhs2715 =             clhs1452*clhs312;
const double clhs2716 =             -clhs2715;
const double clhs2717 =             clhs1455*clhs318;
const double clhs2718 =             -clhs2717;
const double clhs2719 =             clhs2709 + clhs2710 + clhs2711 + clhs2712 + clhs2713 + clhs2714 + clhs2716 + clhs2718;
const double clhs2720 =             -clhs2705*clhs57 - clhs2706*clhs57 - clhs2707*clhs57 + clhs2708*clhs57 + clhs2719;
const double clhs2721 =             -clhs1242*clhs2693 - clhs1254*clhs2704 + clhs1274*clhs2720 + clhs2681;
const double clhs2722 =             DN(1,1)*N[2]*clhs203*v_sc;
const double clhs2723 =             DN(1,2)*N[2]*clhs203*v_sc;
const double clhs2724 =             DN(1,0)*N[2]*clhs203*v_sc;
const double clhs2725 =             N[1]*N[2]*clhs203*clhs936;
const double clhs2726 =             N[1]*N[2]*clhs203*clhs937;
const double clhs2727 =             N[1]*N[2]*clhs203;
const double clhs2728 =             clhs106*clhs2727;
const double clhs2729 =             N[1]*clhs506 + N[1]*clhs510 + N[2]*clhs306 + N[2]*clhs310 - clhs2725*clhs82 - clhs2726*clhs89 + clhs2728*clhs943 + clhs312*clhs617 - clhs312*clhs660 + clhs318*clhs628 - clhs318*clhs686 + clhs359*clhs636 + clhs359*clhs646 - clhs359*clhs653;
const double clhs2730 =             clhs2727*clhs81;
const double clhs2731 =             clhs2605 + clhs2682 - clhs2730;
const double clhs2732 =             (1.0L/2.0L)*clhs203*clhs2731*tau2;
const double clhs2733 =             clhs2732*clhs985;
const double clhs2734 =             -clhs2733;
const double clhs2735 =             clhs2727*clhs88;
const double clhs2736 =             clhs2607 + clhs2694 - clhs2735;
const double clhs2737 =             (1.0L/2.0L)*clhs203*clhs2736*tau2;
const double clhs2738 =             clhs1017*clhs2737;
const double clhs2739 =             -clhs2738;
const double clhs2740 =             clhs2603 + clhs2686 - clhs2728;
const double clhs2741 =             (1.0L/2.0L)*clhs203*clhs2740*tau2;
const double clhs2742 =             clhs1050*clhs2741;
const double clhs2743 =             clhs1343*clhs2720 + clhs2602 + clhs2734 + clhs2739 + clhs2742*clhs57;
const double clhs2744 =             clhs1386*clhs1543;
const double clhs2745 =             clhs1397*clhs1547;
const double clhs2746 =             -clhs2745;
const double clhs2747 =             clhs1409*clhs1552;
const double clhs2748 =             -clhs2747;
const double clhs2749 =             clhs1561*clhs463;
const double clhs2750 =             clhs1563*clhs489;
const double clhs2751 =             -clhs1050*clhs2732 - clhs1343*clhs2693 + clhs1389*clhs2740;
const double clhs2752 =             clhs1386*clhs1547;
const double clhs2753 =             clhs1397*clhs1543;
const double clhs2754 =             clhs1215*clhs2737 - clhs1343*clhs2704 - clhs1400*clhs2740;
const double clhs2755 =             clhs1386*clhs1552;
const double clhs2756 =             clhs1409*clhs1543;
const double clhs2757 =             N[1]*N[3];
const double clhs2758 =             clhs2757*clhs4;
const double clhs2759 =             DN(1,1)*N[3];
const double clhs2760 =             clhs10*clhs2759;
const double clhs2761 =             clhs10*clhs2613;
const double clhs2762 =             N[3]*clhs418;
const double clhs2763 =             DN(1,0)*N[3];
const double clhs2764 =             clhs2763*clhs8;
const double clhs2765 =             clhs2611*clhs8;
const double clhs2766 =             N[3]*clhs415;
const double clhs2767 =             clhs2766*clhs67;
const double clhs2768 =             clhs302*clhs821;
const double clhs2769 =             clhs302*clhs820;
const double clhs2770 =             -clhs2760 - clhs2761 - clhs2762 + clhs2764*clhs67 + clhs2765*clhs67 + clhs2767 + clhs2768 - clhs2769*clhs67;
const double clhs2771 =             DN(1,2)*N[3];
const double clhs2772 =             clhs10*clhs2771;
const double clhs2773 =             clhs10*clhs2615;
const double clhs2774 =             N[3]*clhs429;
const double clhs2775 =             clhs14*clhs2763;
const double clhs2776 =             clhs14*clhs2611;
const double clhs2777 =             N[3]*clhs426;
const double clhs2778 =             clhs2777*clhs67;
const double clhs2779 =             clhs302*clhs831;
const double clhs2780 =             clhs302*clhs830;
const double clhs2781 =             -clhs2772 - clhs2773 - clhs2774 + clhs2775*clhs67 + clhs2776*clhs67 + clhs2778 + clhs2779 - clhs2780*clhs67;
const double clhs2782 =             clhs10*clhs2763;
const double clhs2783 =             clhs10*clhs2611;
const double clhs2784 =             N[3]*clhs453;
const double clhs2785 =             clhs1625*clhs359;
const double clhs2786 =             clhs2759*clhs8;
const double clhs2787 =             clhs14*clhs2771;
const double clhs2788 =             clhs2613*clhs8;
const double clhs2789 =             clhs14*clhs2615;
const double clhs2790 =             N[3]*clhs447;
const double clhs2791 =             N[3]*clhs437;
const double clhs2792 =             clhs1633*clhs312;
const double clhs2793 =             -clhs2792;
const double clhs2794 =             clhs1636*clhs318;
const double clhs2795 =             -clhs2794;
const double clhs2796 =             clhs2786 + clhs2787 + clhs2788 + clhs2789 + clhs2790 + clhs2791 + clhs2793 + clhs2795;
const double clhs2797 =             -clhs2782*clhs57 - clhs2783*clhs57 - clhs2784*clhs57 + clhs2785*clhs57 + clhs2796;
const double clhs2798 =             -clhs1242*clhs2770 - clhs1254*clhs2781 + clhs1274*clhs2797 + clhs2758;
const double clhs2799 =             DN(1,1)*N[3]*clhs203*v_sc;
const double clhs2800 =             DN(1,2)*N[3]*clhs203*v_sc;
const double clhs2801 =             DN(1,0)*N[3]*clhs203*v_sc;
const double clhs2802 =             N[1]*N[3]*clhs203*clhs936;
const double clhs2803 =             N[1]*N[3]*clhs203*clhs937;
const double clhs2804 =             N[1]*N[3]*clhs203;
const double clhs2805 =             clhs106*clhs2804;
const double clhs2806 =             N[1]*clhs705 + N[1]*clhs709 + N[3]*clhs306 + N[3]*clhs310 - clhs2802*clhs82 - clhs2803*clhs89 + clhs2805*clhs943 + clhs312*clhs816 - clhs312*clhs859 + clhs318*clhs827 - clhs318*clhs885 + clhs359*clhs835 + clhs359*clhs845 - clhs359*clhs852;
const double clhs2807 =             clhs2804*clhs81;
const double clhs2808 =             clhs2613 + clhs2759 - clhs2807;
const double clhs2809 =             (1.0L/2.0L)*clhs203*clhs2808*tau2;
const double clhs2810 =             clhs2809*clhs985;
const double clhs2811 =             -clhs2810;
const double clhs2812 =             clhs2804*clhs88;
const double clhs2813 =             clhs2615 + clhs2771 - clhs2812;
const double clhs2814 =             (1.0L/2.0L)*clhs203*clhs2813*tau2;
const double clhs2815 =             clhs1017*clhs2814;
const double clhs2816 =             -clhs2815;
const double clhs2817 =             clhs2611 + clhs2763 - clhs2805;
const double clhs2818 =             (1.0L/2.0L)*clhs203*clhs2817*tau2;
const double clhs2819 =             clhs1050*clhs2818;
const double clhs2820 =             clhs1343*clhs2797 + clhs2610 + clhs2811 + clhs2816 + clhs2819*clhs57;
const double clhs2821 =             clhs1386*clhs1724;
const double clhs2822 =             clhs1397*clhs1728;
const double clhs2823 =             -clhs2822;
const double clhs2824 =             clhs1409*clhs1733;
const double clhs2825 =             -clhs2824;
const double clhs2826 =             clhs1742*clhs463;
const double clhs2827 =             clhs1744*clhs489;
const double clhs2828 =             -clhs1050*clhs2809 - clhs1343*clhs2770 + clhs1389*clhs2817;
const double clhs2829 =             clhs1386*clhs1728;
const double clhs2830 =             clhs1397*clhs1724;
const double clhs2831 =             clhs1215*clhs2814 - clhs1343*clhs2781 - clhs1400*clhs2817;
const double clhs2832 =             clhs1386*clhs1733;
const double clhs2833 =             clhs1409*clhs1724;
const double clhs2834 =             DN(1,0)*clhs7*clhs906*mu;
const double clhs2835 =             (2.0L/3.0L)*DN(1,1)*clhs7*clhs906*mu;
const double clhs2836 =             -clhs349;
const double clhs2837 =             clhs2836 + clhs382;
const double clhs2838 =             (1.0L/2.0L)*clhs1867*clhs203*tau2;
const double clhs2839 =             (1.0L/2.0L)*DN(1,1)*clhs67*tau3;
const double clhs2840 =             clhs203*clhs424*tau2;
const double clhs2841 =             -clhs1356 - clhs2628 + clhs301;
const double clhs2842 =             clhs203*clhs283*tau2;
const double clhs2843 =             clhs1164*clhs1397;
const double clhs2844 =             clhs1151*clhs1409;
const double clhs2845 =             clhs203*clhs492*tau2;
const double clhs2846 =             -clhs450;
const double clhs2847 =             -clhs453;
const double clhs2848 =             clhs1286 - clhs1866 + clhs2640 + clhs2846 + clhs2847 + clhs477 + clhs479;
const double clhs2849 =             DN(1,1)*tau3;
const double clhs2850 =             clhs113*clhs2597;
const double clhs2851 =             clhs2332 + clhs2652 + clhs415;
const double clhs2852 =             -clhs1384 + clhs2307*clhs67 + clhs2851 + clhs420;
const double clhs2853 =             clhs14*clhs304;
const double clhs2854 =             clhs2334 - clhs2353*clhs67 + clhs2853*clhs67 - clhs354 + clhs491;
const double clhs2855 =             clhs1284 + clhs1856 + clhs2015 + clhs351 + clhs437 + clhs453;
const double clhs2856 =             clhs1288*clhs57 + clhs1864 + clhs2855 - clhs387*clhs57;
const double clhs2857 =             (1.0L/2.0L)*N[1]*clhs203*clhs2668*clhs67*tau2;
const double clhs2858 =             clhs1215*clhs2673 + clhs1393*clhs492 - clhs2663;
const double clhs2859 =             clhs203*clhs2848*tau2;
const double clhs2860 =             clhs1217*clhs2671;
const double clhs2861 =             clhs1826*clhs2669;
const double clhs2862 =             clhs1369*clhs1397;
const double clhs2863 =             clhs1364*clhs1409;
const double clhs2864 =             clhs113*clhs2680;
const double clhs2865 =             clhs2687 + clhs2688 + clhs2689 - clhs2692;
const double clhs2866 =             clhs2683*clhs67;
const double clhs2867 =             clhs2684*clhs67;
const double clhs2868 =             clhs2685*clhs67;
const double clhs2869 =             clhs2691*clhs67 + clhs2865 - clhs2866 - clhs2867 - clhs2868;
const double clhs2870 =             clhs2694*clhs8;
const double clhs2871 =             clhs2607*clhs8;
const double clhs2872 =             N[2]*clhs468;
const double clhs2873 =             clhs14*clhs2682;
const double clhs2874 =             clhs14*clhs2605;
const double clhs2875 =             N[2]*clhs465;
const double clhs2876 =             clhs2875*clhs67;
const double clhs2877 =             clhs302*clhs671;
const double clhs2878 =             clhs302*clhs670;
const double clhs2879 =             -clhs2870 - clhs2871 - clhs2872 + clhs2873*clhs67 + clhs2874*clhs67 + clhs2876 + clhs2877 - clhs2878*clhs67;
const double clhs2880 =             -clhs2708;
const double clhs2881 =             clhs2705 + clhs2706 + clhs2707 + clhs2710 + clhs2712 + clhs2714 + clhs2718 + clhs2880;
const double clhs2882 =             -clhs2709*clhs57 - clhs2711*clhs57 - clhs2713*clhs57 + clhs2715*clhs57 + clhs2881;
const double clhs2883 =             clhs1242*clhs2882 - clhs1254*clhs2879 + clhs1274*clhs2869 + clhs2864;
const double clhs2884 =             N[1]*N[2]*clhs1184*clhs203;
const double clhs2885 =             N[1]*clhs552 + N[1]*clhs555 + N[2]*clhs352 + N[2]*clhs355 - clhs151*clhs2884 - clhs155*clhs2726 + clhs1799*clhs2730 + clhs312*clhs636 + clhs312*clhs652 - clhs312*clhs678 + clhs318*clhs667 - clhs318*clhs689 + clhs359*clhs614 - clhs359*clhs618;
const double clhs2886 =             clhs1343*clhs2869 + clhs2020*clhs2731 - clhs2741*clhs985;
const double clhs2887 =             clhs1215*clhs2741 + clhs2602 - clhs2744;
const double clhs2888 =             clhs1378*clhs662;
const double clhs2889 =             clhs1575*clhs492;
const double clhs2890 =             clhs1343*clhs2882;
const double clhs2891 =             clhs1217*clhs2737;
const double clhs2892 =             clhs1826*clhs2732;
const double clhs2893 =             (1.0L/2.0L)*clhs1017*clhs203*clhs67*tau2;
const double clhs2894 =             -clhs1343*clhs2879 + clhs2731*clhs2893 - clhs2737*clhs985;
const double clhs2895 =             clhs1397*clhs1552;
const double clhs2896 =             clhs1409*clhs1547;
const double clhs2897 =             clhs113*clhs2757;
const double clhs2898 =             clhs2764 + clhs2765 + clhs2766 - clhs2769;
const double clhs2899 =             clhs2760*clhs67;
const double clhs2900 =             clhs2761*clhs67;
const double clhs2901 =             clhs2762*clhs67;
const double clhs2902 =             clhs2768*clhs67 + clhs2898 - clhs2899 - clhs2900 - clhs2901;
const double clhs2903 =             clhs2771*clhs8;
const double clhs2904 =             clhs2615*clhs8;
const double clhs2905 =             N[3]*clhs468;
const double clhs2906 =             clhs14*clhs2759;
const double clhs2907 =             clhs14*clhs2613;
const double clhs2908 =             N[3]*clhs465;
const double clhs2909 =             clhs2908*clhs67;
const double clhs2910 =             clhs302*clhs870;
const double clhs2911 =             clhs302*clhs869;
const double clhs2912 =             -clhs2903 - clhs2904 - clhs2905 + clhs2906*clhs67 + clhs2907*clhs67 + clhs2909 + clhs2910 - clhs2911*clhs67;
const double clhs2913 =             -clhs2785;
const double clhs2914 =             clhs2782 + clhs2783 + clhs2784 + clhs2787 + clhs2789 + clhs2791 + clhs2795 + clhs2913;
const double clhs2915 =             -clhs2786*clhs57 - clhs2788*clhs57 - clhs2790*clhs57 + clhs2792*clhs57 + clhs2914;
const double clhs2916 =             clhs1242*clhs2915 - clhs1254*clhs2912 + clhs1274*clhs2902 + clhs2897;
const double clhs2917 =             N[1]*N[3]*clhs1184*clhs203;
const double clhs2918 =             N[1]*clhs751 + N[1]*clhs754 + N[3]*clhs352 + N[3]*clhs355 - clhs151*clhs2917 - clhs155*clhs2803 + clhs1799*clhs2807 + clhs312*clhs835 + clhs312*clhs851 - clhs312*clhs877 + clhs318*clhs866 - clhs318*clhs888 + clhs359*clhs813 - clhs359*clhs817;
const double clhs2919 =             clhs1343*clhs2902 + clhs2020*clhs2808 - clhs2818*clhs985;
const double clhs2920 =             clhs1215*clhs2818 + clhs2610 - clhs2821;
const double clhs2921 =             clhs1378*clhs861;
const double clhs2922 =             clhs1756*clhs492;
const double clhs2923 =             clhs1343*clhs2915;
const double clhs2924 =             clhs1217*clhs2814;
const double clhs2925 =             clhs1826*clhs2809;
const double clhs2926 =             -clhs1343*clhs2912 + clhs2808*clhs2893 - clhs2814*clhs985;
const double clhs2927 =             clhs1397*clhs1733;
const double clhs2928 =             clhs1409*clhs1728;
const double clhs2929 =             (2.0L/3.0L)*DN(1,2)*clhs7*clhs906*mu;
const double clhs2930 =             -clhs384;
const double clhs2931 =             clhs2930 + clhs410;
const double clhs2932 =             (1.0L/2.0L)*clhs2026*clhs203*tau2;
const double clhs2933 =             (1.0L/2.0L)*DN(1,2)*clhs67*tau3;
const double clhs2934 =             clhs203*clhs434*tau2;
const double clhs2935 =             clhs203*clhs473*tau2;
const double clhs2936 =             clhs203*clhs299*tau2;
const double clhs2937 =             clhs1286 + clhs1374 - clhs2024 + clhs2641 + clhs2642 + clhs2846 + clhs2847 + clhs494 + clhs496;
const double clhs2938 =             DN(1,2)*tau3;
const double clhs2939 =             clhs166*clhs2597;
const double clhs2940 =             clhs2351 + clhs2655 + clhs426;
const double clhs2941 =             -clhs1385 + clhs2313*clhs67 + clhs2940 + clhs431;
const double clhs2942 =             clhs2354 + clhs2853 + clhs465;
const double clhs2943 =             -clhs1396 + clhs2334*clhs67 + clhs2942 + clhs470;
const double clhs2944 =             clhs1289 + clhs1856 + clhs351 + clhs387 + clhs447 + clhs453;
const double clhs2945 =             clhs1283*clhs57 - clhs2015*clhs57 + clhs2023 + clhs2944;
const double clhs2946 =             clhs203*clhs2937*tau2;
const double clhs2947 =             clhs166*clhs2680;
const double clhs2948 =             clhs2698 + clhs2699 + clhs2700 - clhs2703;
const double clhs2949 =             clhs2695*clhs67;
const double clhs2950 =             clhs2696*clhs67;
const double clhs2951 =             clhs2697*clhs67;
const double clhs2952 =             clhs2702*clhs67 + clhs2948 - clhs2949 - clhs2950 - clhs2951;
const double clhs2953 =             clhs2873 + clhs2874 + clhs2875 - clhs2878;
const double clhs2954 =             clhs2870*clhs67;
const double clhs2955 =             clhs2871*clhs67;
const double clhs2956 =             clhs2872*clhs67;
const double clhs2957 =             clhs2877*clhs67 + clhs2953 - clhs2954 - clhs2955 - clhs2956;
const double clhs2958 =             clhs2705 + clhs2706 + clhs2707 + clhs2709 + clhs2711 + clhs2713 + clhs2716 + clhs2880;
const double clhs2959 =             -clhs2710*clhs57 - clhs2712*clhs57 - clhs2714*clhs57 + clhs2717*clhs57 + clhs2958;
const double clhs2960 =             clhs1242*clhs2957 + clhs1254*clhs2959 + clhs1274*clhs2952 + clhs2947;
const double clhs2961 =             N[1]*clhs584 + N[1]*clhs587 + N[2]*clhs385 + N[2]*clhs388 - clhs190*clhs2884 - clhs193*clhs2725 + clhs1978*clhs2735 + clhs312*clhs664 - clhs312*clhs668 + clhs318*clhs646 + clhs318*clhs652 - clhs318*clhs695 + clhs359*clhs625 - clhs359*clhs629;
const double clhs2962 =             -clhs1017*clhs2741 + clhs1343*clhs2952 + clhs2020*clhs2736;
const double clhs2963 =             -clhs1017*clhs2732 + clhs1343*clhs2957 + clhs1389*clhs2736;
const double clhs2964 =             clhs1380*clhs688;
const double clhs2965 =             clhs1393*clhs691;
const double clhs2966 =             clhs1343*clhs2959;
const double clhs2967 =             clhs166*clhs2757;
const double clhs2968 =             clhs2775 + clhs2776 + clhs2777 - clhs2780;
const double clhs2969 =             clhs2772*clhs67;
const double clhs2970 =             clhs2773*clhs67;
const double clhs2971 =             clhs2774*clhs67;
const double clhs2972 =             clhs2779*clhs67 + clhs2968 - clhs2969 - clhs2970 - clhs2971;
const double clhs2973 =             clhs2906 + clhs2907 + clhs2908 - clhs2911;
const double clhs2974 =             clhs2903*clhs67;
const double clhs2975 =             clhs2904*clhs67;
const double clhs2976 =             clhs2905*clhs67;
const double clhs2977 =             clhs2910*clhs67 + clhs2973 - clhs2974 - clhs2975 - clhs2976;
const double clhs2978 =             clhs2782 + clhs2783 + clhs2784 + clhs2786 + clhs2788 + clhs2790 + clhs2793 + clhs2913;
const double clhs2979 =             -clhs2787*clhs57 - clhs2789*clhs57 - clhs2791*clhs57 + clhs2794*clhs57 + clhs2978;
const double clhs2980 =             clhs1242*clhs2977 + clhs1254*clhs2979 + clhs1274*clhs2972 + clhs2967;
const double clhs2981 =             N[1]*clhs783 + N[1]*clhs786 + N[3]*clhs385 + N[3]*clhs388 - clhs190*clhs2917 - clhs193*clhs2802 + clhs1978*clhs2812 + clhs312*clhs863 - clhs312*clhs867 + clhs318*clhs845 + clhs318*clhs851 - clhs318*clhs894 + clhs359*clhs824 - clhs359*clhs828;
const double clhs2982 =             -clhs1017*clhs2818 + clhs1343*clhs2972 + clhs2020*clhs2813;
const double clhs2983 =             -clhs1017*clhs2809 + clhs1343*clhs2977 + clhs1389*clhs2813;
const double clhs2984 =             clhs1380*clhs887;
const double clhs2985 =             clhs1393*clhs890;
const double clhs2986 =             clhs1343*clhs2979;
const double clhs2987 =             clhs1299 + clhs1387;
const double clhs2988 =             clhs1398 + clhs2836;
const double clhs2989 =             clhs1410 + clhs2930;
const double clhs2990 =             (1.0L/3.0L)*DN(1,0)*clhs203;
const double clhs2991 =             (1.0L/3.0L)*DN(1,1)*clhs203;
const double clhs2992 =             (1.0L/3.0L)*DN(1,2)*clhs203;
const double clhs2993 =             clhs1304 + clhs2298;
const double clhs2994 =             (1.0L/2.0L)*clhs203*clhs2366*gamma*tau3;
const double clhs2995 =             (1.0L/3.0L)*DN(1,1)*clhs7;
const double clhs2996 =             (1.0L/3.0L)*DN(1,2)*clhs7;
const double clhs2997 =             (1.0L/3.0L)*DN(1,0)*clhs7;
const double clhs2998 =             (1.0L/2.0L)*N[1]*clhs203;
const double clhs2999 =             DN(1,0)*clhs203*clhs2145*clhs2146*lambda;
const double clhs3000 =             DN(1,1)*clhs203*clhs2145*clhs2146*lambda;
const double clhs3001 =             DN(1,2)*clhs203*clhs2145*clhs2146*lambda;
const double clhs3002 =             N[1]*clhs203*gamma;
const double clhs3003 =             clhs203*clhs2366*gamma*tau3;
const double clhs3004 =             (1.0L/2.0L)*N[1]*clhs7*tau2;
const double clhs3005 =             -DN(1,0)*clhs1125 - clhs2092*clhs417 - clhs2093*clhs428 + clhs2094*clhs2321 + clhs2096*clhs2322 + clhs2242 - clhs2249 + clhs2320;
const double clhs3006 =             -DN(1,0)*clhs10*clhs1113*clhs84 - DN(1,1)*clhs1129 + clhs2089*clhs2342 - clhs2093*clhs467 + clhs2096*clhs2343 + clhs2253 - clhs2260 + clhs2341;
const double clhs3007 =             -DN(1,0)*clhs10*clhs1113*clhs90 - DN(1,1)*clhs1113*clhs8*clhs90 - DN(1,2)*clhs1132 + clhs2089*clhs2358 + clhs2094*clhs2359 + clhs2264 - clhs2268 + clhs2357;
const double clhs3008 =             clhs1284 + clhs2015 + clhs437;
const double clhs3009 =             clhs2944 + clhs3008;
const double clhs3010 =             (1.0L/2.0L)*N[1]*clhs67*clhs7*tau2*(clhs2308 + clhs2851 + clhs305 + clhs418);
const double clhs3011 =             (1.0L/2.0L)*N[1]*clhs67*clhs7*tau2*(clhs2314 + clhs2940 + clhs309 + clhs429);
const double clhs3012 =             (1.0L/2.0L)*N[1]*clhs67*clhs7*tau2;
const double clhs3013 =             3*clhs453;
const double clhs3014 =             (1.0L/2.0L)*N[1]*clhs203*clhs2126*gamma*tau3;
const double clhs3015 =             (1.0L/2.0L)*N[1]*clhs67*clhs7*tau2*(clhs2335 + clhs2942 + clhs354 + clhs468);
const double clhs3016 =             3*clhs447;
const double clhs3017 =             3*clhs437;
const double clhs3018 =             (1.0L/2.0L)*N[1]*clhs203*gamma*tau2;
const double clhs3019 =             clhs2371*clhs302;
const double clhs3020 =             N[1]*N[2]*clhs2246*clhs7*clhs8;
const double clhs3021 =             N[1]*N[2]*clhs14*clhs2246*clhs7;
const double clhs3022 =             -clhs1125*clhs2603 - clhs1125*clhs2686 - clhs2092*clhs2690 - clhs2092*clhs2866 - clhs2092*clhs2867 - clhs2092*clhs2868 - clhs2093*clhs2701 - clhs2093*clhs2949 - clhs2093*clhs2950 - clhs2093*clhs2951 - clhs2244*clhs2707 - clhs2245*clhs2713 - clhs2245*clhs2714 - clhs2375*clhs302 + clhs3019 + clhs3020*clhs82 + clhs3021*clhs89;
const double clhs3023 =             clhs2378*clhs302;
const double clhs3024 =             N[1]*N[2]*clhs10*clhs2246*clhs7;
const double clhs3025 =             -clhs1129*clhs2605 - clhs1129*clhs2682 + clhs151*clhs3024 + clhs155*clhs3021 - clhs2093*clhs2876 - clhs2093*clhs2954 - clhs2093*clhs2955 - clhs2093*clhs2956 - clhs2256*clhs2705 - clhs2256*clhs2706 - clhs2256*clhs2707 - clhs2256*clhs2714 - clhs2257*clhs2690 - clhs2257*clhs2868 - clhs2258*clhs2713 - clhs2382*clhs302 + clhs3023;
const double clhs3026 =             clhs2385*clhs302;
const double clhs3027 =             -clhs1132*clhs2607 - clhs1132*clhs2694 + clhs190*clhs3024 + clhs193*clhs3020 - clhs2092*clhs2876 - clhs2092*clhs2956 - clhs2257*clhs2701 - clhs2257*clhs2951 - clhs2266*clhs2705 - clhs2266*clhs2706 - clhs2266*clhs2707 - clhs2266*clhs2709 - clhs2266*clhs2711 - clhs2266*clhs2713 - clhs2267*clhs2714 - clhs2387*clhs302 + clhs3026;
const double clhs3028 =             clhs2710 + clhs2712 + clhs2714 + clhs2718 + clhs2958;
const double clhs3029 =             clhs1068*clhs2680 + clhs1293*(-clhs10*clhs3019 - clhs1140*clhs2705 - clhs1140*clhs2706 - clhs1140*clhs2709 - clhs1140*clhs2710 - clhs1140*clhs2711 - clhs1140*clhs2712 - clhs14*clhs3026 - clhs2098*clhs2707 - clhs2104*clhs2713 - clhs2110*clhs2714 + clhs2176*clhs2690 + clhs2176*clhs2868 + clhs2177*clhs2701 + clhs2177*clhs2951 + clhs2178*clhs2876 + clhs2178*clhs2956 - clhs2272*clhs2728 - clhs2273*clhs2730 - clhs2274*clhs2735 - clhs3023*clhs8) - clhs2240*clhs3022 - clhs2251*clhs3025 - clhs2262*clhs3027 - clhs2271*clhs3028;
const double clhs3030 =             (1.0L/2.0L)*clhs67*clhs7*tau2*(clhs2683 + clhs2684 + clhs2685 - clhs2691 + clhs2865);
const double clhs3031 =             (1.0L/2.0L)*clhs67*clhs7*tau2*(clhs2695 + clhs2696 + clhs2697 - clhs2702 + clhs2948);
const double clhs3032 =             -clhs1217*clhs3031 - clhs1826*clhs3030 - clhs2302*(N[2]*clhs3013 - clhs1184*clhs2728 + 3*clhs2705 + 3*clhs2706 + clhs2719) + clhs2303*clhs3022 + clhs2304*clhs2740 + clhs2681;
const double clhs3033 =             (1.0L/2.0L)*clhs67*clhs7*tau2*(clhs2870 + clhs2871 + clhs2872 - clhs2877 + clhs2953);
const double clhs3034 =             -clhs1215*clhs3030 - clhs1217*clhs3033 + clhs2303*clhs3025 + clhs2304*clhs2731 - clhs2328*(N[2]*clhs3016 + 3*clhs2709 + 3*clhs2711 - clhs2730*clhs936 + clhs2881) + clhs2864;
const double clhs3035 =             -clhs1215*clhs3031 - clhs1826*clhs3033 + clhs2303*clhs3027 + clhs2304*clhs2736 - clhs2348*(N[2]*clhs3017 + 3*clhs2710 + 3*clhs2712 - clhs2735*clhs937 + clhs2958) + clhs2947;
const double clhs3036 =             clhs2364*clhs3028 + clhs2602 - clhs2733*gamma - clhs2738*gamma - clhs2742*gamma;
const double clhs3037 =             clhs2480*clhs302;
const double clhs3038 =             N[1]*N[3]*clhs2246*clhs7*clhs8;
const double clhs3039 =             N[1]*N[3]*clhs14*clhs2246*clhs7;
const double clhs3040 =             -clhs1125*clhs2611 - clhs1125*clhs2763 - clhs2092*clhs2767 - clhs2092*clhs2899 - clhs2092*clhs2900 - clhs2092*clhs2901 - clhs2093*clhs2778 - clhs2093*clhs2969 - clhs2093*clhs2970 - clhs2093*clhs2971 - clhs2244*clhs2784 - clhs2245*clhs2790 - clhs2245*clhs2791 - clhs2484*clhs302 + clhs3037 + clhs3038*clhs82 + clhs3039*clhs89;
const double clhs3041 =             clhs2487*clhs302;
const double clhs3042 =             N[1]*N[3]*clhs10*clhs2246*clhs7;
const double clhs3043 =             -clhs1129*clhs2613 - clhs1129*clhs2759 + clhs151*clhs3042 + clhs155*clhs3039 - clhs2093*clhs2909 - clhs2093*clhs2974 - clhs2093*clhs2975 - clhs2093*clhs2976 - clhs2256*clhs2782 - clhs2256*clhs2783 - clhs2256*clhs2784 - clhs2256*clhs2791 - clhs2257*clhs2767 - clhs2257*clhs2901 - clhs2258*clhs2790 - clhs2491*clhs302 + clhs3041;
const double clhs3044 =             clhs2494*clhs302;
const double clhs3045 =             -clhs1132*clhs2615 - clhs1132*clhs2771 + clhs190*clhs3042 + clhs193*clhs3038 - clhs2092*clhs2909 - clhs2092*clhs2976 - clhs2257*clhs2778 - clhs2257*clhs2971 - clhs2266*clhs2782 - clhs2266*clhs2783 - clhs2266*clhs2784 - clhs2266*clhs2786 - clhs2266*clhs2788 - clhs2266*clhs2790 - clhs2267*clhs2791 - clhs2496*clhs302 + clhs3044;
const double clhs3046 =             clhs2787 + clhs2789 + clhs2791 + clhs2795 + clhs2978;
const double clhs3047 =             clhs1068*clhs2757 + clhs1293*(-clhs10*clhs3037 - clhs1140*clhs2782 - clhs1140*clhs2783 - clhs1140*clhs2786 - clhs1140*clhs2787 - clhs1140*clhs2788 - clhs1140*clhs2789 - clhs14*clhs3044 - clhs2098*clhs2784 - clhs2104*clhs2790 - clhs2110*clhs2791 + clhs2176*clhs2767 + clhs2176*clhs2901 + clhs2177*clhs2778 + clhs2177*clhs2971 + clhs2178*clhs2909 + clhs2178*clhs2976 - clhs2272*clhs2805 - clhs2273*clhs2807 - clhs2274*clhs2812 - clhs3041*clhs8) - clhs2240*clhs3040 - clhs2251*clhs3043 - clhs2262*clhs3045 - clhs2271*clhs3046;
const double clhs3048 =             (1.0L/2.0L)*clhs67*clhs7*tau2*(clhs2760 + clhs2761 + clhs2762 - clhs2768 + clhs2898);
const double clhs3049 =             (1.0L/2.0L)*clhs67*clhs7*tau2*(clhs2772 + clhs2773 + clhs2774 - clhs2779 + clhs2968);
const double clhs3050 =             -clhs1217*clhs3049 - clhs1826*clhs3048 - clhs2302*(N[3]*clhs3013 - clhs1184*clhs2805 + 3*clhs2782 + 3*clhs2783 + clhs2796) + clhs2303*clhs3040 + clhs2304*clhs2817 + clhs2758;
const double clhs3051 =             (1.0L/2.0L)*clhs67*clhs7*tau2*(clhs2903 + clhs2904 + clhs2905 - clhs2910 + clhs2973);
const double clhs3052 =             -clhs1215*clhs3048 - clhs1217*clhs3051 + clhs2303*clhs3043 + clhs2304*clhs2808 - clhs2328*(N[3]*clhs3016 + 3*clhs2786 + 3*clhs2788 - clhs2807*clhs936 + clhs2914) + clhs2897;
const double clhs3053 =             -clhs1215*clhs3049 - clhs1826*clhs3051 + clhs2303*clhs3045 + clhs2304*clhs2813 - clhs2348*(N[3]*clhs3017 + 3*clhs2787 + 3*clhs2789 - clhs2812*clhs937 + clhs2978) + clhs2967;
const double clhs3054 =             clhs2364*clhs3046 + clhs2610 - clhs2810*gamma - clhs2815*gamma - clhs2819*gamma;
const double clhs3055 =             (1.0L/2.0L)*DN(2,0)*tau2;
const double clhs3056 =             (1.0L/2.0L)*DN(2,1)*tau2;
const double clhs3057 =             (1.0L/2.0L)*DN(2,2)*tau2;
const double clhs3058 =             -clhs1422;
const double clhs3059 =             DN(2,0)*tau2;
const double clhs3060 =             -clhs1418;
const double clhs3061 =             DN(2,1)*tau2;
const double clhs3062 =             -clhs1430;
const double clhs3063 =             DN(2,2)*tau2;
const double clhs3064 =             -clhs2686;
const double clhs3065 =             -clhs2682;
const double clhs3066 =             -clhs2694;
const double clhs3067 =             pow(N[2], 2);
const double clhs3068 =             -bdf0*clhs3067;
const double clhs3069 =             -DN(2,0)*N[2];
const double clhs3070 =             -DN(2,1)*N[2];
const double clhs3071 =             -DN(2,2)*N[2];
const double clhs3072 =             -N[3]*clhs641;
const double clhs3073 =             DN(3,0)*N[2];
const double clhs3074 =             -clhs3073;
const double clhs3075 =             DN(3,1)*N[2];
const double clhs3076 =             -clhs3075;
const double clhs3077 =             DN(3,2)*N[2];
const double clhs3078 =             -clhs3077;
const double clhs3079 =             -clhs300*(DN(2,0)*DN(3,0) + DN(2,1)*DN(3,1) + DN(2,2)*DN(3,2));
const double clhs3080 =             DN(2,1)*N[0]*clhs203*v_sc;
const double clhs3081 =             DN(2,2)*N[0]*clhs203*v_sc;
const double clhs3082 =             DN(2,1)*clhs7*clhs906*mu;
const double clhs3083 =             DN(2,2)*clhs7*clhs906*mu;
const double clhs3084 =             DN(2,0)*N[0]*clhs203*v_sc;
const double clhs3085 =             (2.0L/3.0L)*DN(2,0)*clhs7*clhs906*mu;
const double clhs3086 =             (1.0L/2.0L)*N[2]*clhs7;
const double clhs3087 =             clhs1485 + clhs503 - clhs546;
const double clhs3088 =             (1.0L/2.0L)*clhs1560*clhs203*tau2;
const double clhs3089 =             (1.0L/2.0L)*DN(2,0)*clhs67*tau3;
const double clhs3090 =             clhs1159*clhs1569;
const double clhs3091 =             clhs1151*clhs1579;
const double clhs3092 =             -clhs3091;
const double clhs3093 =             clhs1164*clhs1590;
const double clhs3094 =             -clhs3093;
const double clhs3095 =             clhs1560*clhs203*tau2;
const double clhs3096 =             clhs1151*clhs1569;
const double clhs3097 =             clhs1159*clhs1579;
const double clhs3098 =             clhs203*clhs662*tau2;
const double clhs3099 =             clhs1164*clhs1569;
const double clhs3100 =             clhs1159*clhs1590;
const double clhs3101 =             -clhs636;
const double clhs3102 =             clhs1555 + clhs3101 - clhs634;
const double clhs3103 =             -clhs643;
const double clhs3104 =             -clhs646;
const double clhs3105 =             clhs1557 - clhs1559 + clhs3102 + clhs3103 + clhs3104 + clhs650 + clhs653;
const double clhs3106 =             clhs203*clhs688*tau2;
const double clhs3107 =             clhs1483 + clhs547;
const double clhs3108 =             DN(2,0)*tau3;
const double clhs3109 =             DN(2,1)*N[1]*clhs203*v_sc;
const double clhs3110 =             DN(2,2)*N[1]*clhs203*v_sc;
const double clhs3111 =             DN(2,0)*N[1]*clhs203*v_sc;
const double clhs3112 =             clhs1360*clhs1569;
const double clhs3113 =             clhs1364*clhs1579;
const double clhs3114 =             -clhs3113;
const double clhs3115 =             clhs1369*clhs1590;
const double clhs3116 =             -clhs3115;
const double clhs3117 =             clhs1364*clhs1569;
const double clhs3118 =             clhs1360*clhs1579;
const double clhs3119 =             clhs1369*clhs1569;
const double clhs3120 =             clhs1360*clhs1590;
const double clhs3121 =             clhs3067*clhs4;
const double clhs3122 =             DN(2,1)*N[2]*clhs203*v_sc;
const double clhs3123 =             DN(2,2)*N[2]*clhs203*v_sc;
const double clhs3124 =             DN(2,0)*N[2]*clhs203*v_sc;
const double clhs3125 =             N[2]*clhs83*clhs934*tau1;
const double clhs3126 =             clhs550*clhs8;
const double clhs3127 =             clhs2419 - clhs2442*clhs67 + clhs3126*clhs67 - clhs505 + clhs661;
const double clhs3128 =             (1.0L/2.0L)*N[2]*clhs7*clhs985*tau2;
const double clhs3129 =             clhs14*clhs550;
const double clhs3130 =             clhs2425 - clhs2460*clhs67 + clhs3129*clhs67 - clhs509 + clhs687;
const double clhs3131 =             (1.0L/2.0L)*N[2]*clhs1017*clhs7*tau2;
const double clhs3132 =             clhs1468 + clhs1473 + clhs1899 - clhs1900*clhs57 + clhs2046 + clhs3101 + clhs3104 + clhs551*clhs57 + clhs653;
const double clhs3133 =             (1.0L/2.0L)*N[2]*clhs1050*clhs7*tau2;
const double clhs3134 =             clhs1547*clhs1579;
const double clhs3135 =             -clhs3134;
const double clhs3136 =             clhs1563*clhs688;
const double clhs3137 =             clhs1543*clhs1569;
const double clhs3138 =             clhs1552*clhs1590;
const double clhs3139 =             -clhs3138;
const double clhs3140 =             clhs1561*clhs662;
const double clhs3141 =             N[2]*clhs7*clhs934*tau1;
const double clhs3142 =             clhs1545 + clhs504;
const double clhs3143 =             (1.0L/2.0L)*N[2]*clhs203*clhs3142*tau2;
const double clhs3144 =             clhs1550 + clhs508;
const double clhs3145 =             (1.0L/2.0L)*N[2]*clhs203*clhs3144*tau2;
const double clhs3146 =             clhs1541 + clhs550;
const double clhs3147 =             (1.0L/2.0L)*N[2]*clhs203*clhs3146*tau2;
const double clhs3148 =             clhs1547*clhs1569;
const double clhs3149 =             clhs1543*clhs1579;
const double clhs3150 =             (1.0L/2.0L)*N[2]*clhs203*clhs67*clhs985*tau2;
const double clhs3151 =             clhs1552*clhs1569;
const double clhs3152 =             clhs1543*clhs1590;
const double clhs3153 =             (1.0L/2.0L)*N[2]*clhs203*clhs67*tau2;
const double clhs3154 =             N[2]*N[3];
const double clhs3155 =             clhs3154*clhs4;
const double clhs3156 =             DN(2,1)*N[3];
const double clhs3157 =             clhs10*clhs3156;
const double clhs3158 =             clhs10*clhs3075;
const double clhs3159 =             N[3]*clhs617;
const double clhs3160 =             DN(2,0)*N[3];
const double clhs3161 =             clhs3160*clhs8;
const double clhs3162 =             clhs3073*clhs8;
const double clhs3163 =             N[3]*clhs614;
const double clhs3164 =             clhs3163*clhs67;
const double clhs3165 =             clhs502*clhs821;
const double clhs3166 =             clhs502*clhs820;
const double clhs3167 =             -clhs3157 - clhs3158 - clhs3159 + clhs3161*clhs67 + clhs3162*clhs67 + clhs3164 + clhs3165 - clhs3166*clhs67;
const double clhs3168 =             DN(2,2)*N[3];
const double clhs3169 =             clhs10*clhs3168;
const double clhs3170 =             clhs10*clhs3077;
const double clhs3171 =             N[3]*clhs628;
const double clhs3172 =             clhs14*clhs3160;
const double clhs3173 =             clhs14*clhs3073;
const double clhs3174 =             N[3]*clhs625;
const double clhs3175 =             clhs3174*clhs67;
const double clhs3176 =             clhs502*clhs831;
const double clhs3177 =             clhs502*clhs830;
const double clhs3178 =             -clhs3169 - clhs3170 - clhs3171 + clhs3172*clhs67 + clhs3173*clhs67 + clhs3175 + clhs3176 - clhs3177*clhs67;
const double clhs3179 =             clhs10*clhs3160;
const double clhs3180 =             clhs10*clhs3073;
const double clhs3181 =             N[3]*clhs652;
const double clhs3182 =             clhs1625*clhs559;
const double clhs3183 =             clhs3156*clhs8;
const double clhs3184 =             clhs14*clhs3168;
const double clhs3185 =             clhs3075*clhs8;
const double clhs3186 =             clhs14*clhs3077;
const double clhs3187 =             N[3]*clhs646;
const double clhs3188 =             N[3]*clhs636;
const double clhs3189 =             clhs1633*clhs512;
const double clhs3190 =             -clhs3189;
const double clhs3191 =             clhs1636*clhs518;
const double clhs3192 =             -clhs3191;
const double clhs3193 =             clhs3183 + clhs3184 + clhs3185 + clhs3186 + clhs3187 + clhs3188 + clhs3190 + clhs3192;
const double clhs3194 =             -clhs3179*clhs57 - clhs3180*clhs57 - clhs3181*clhs57 + clhs3182*clhs57 + clhs3193;
const double clhs3195 =             -clhs1242*clhs3167 - clhs1254*clhs3178 + clhs1274*clhs3194 + clhs3155;
const double clhs3196 =             DN(2,1)*N[3]*clhs203*v_sc;
const double clhs3197 =             DN(2,2)*N[3]*clhs203*v_sc;
const double clhs3198 =             DN(2,0)*N[3]*clhs203*v_sc;
const double clhs3199 =             N[2]*N[3]*clhs203*clhs936;
const double clhs3200 =             N[2]*N[3]*clhs203*clhs937;
const double clhs3201 =             N[2]*N[3]*clhs203;
const double clhs3202 =             clhs106*clhs3201;
const double clhs3203 =             N[2]*clhs705 + N[2]*clhs709 + N[3]*clhs506 + N[3]*clhs510 - clhs3199*clhs82 - clhs3200*clhs89 + clhs3202*clhs943 + clhs512*clhs816 - clhs512*clhs859 + clhs518*clhs827 - clhs518*clhs885 + clhs559*clhs835 + clhs559*clhs845 - clhs559*clhs852;
const double clhs3204 =             clhs3201*clhs81;
const double clhs3205 =             clhs3075 + clhs3156 - clhs3204;
const double clhs3206 =             (1.0L/2.0L)*clhs203*clhs3205*tau2;
const double clhs3207 =             clhs3206*clhs985;
const double clhs3208 =             -clhs3207;
const double clhs3209 =             clhs3201*clhs88;
const double clhs3210 =             clhs3077 + clhs3168 - clhs3209;
const double clhs3211 =             (1.0L/2.0L)*clhs203*clhs3210*tau2;
const double clhs3212 =             clhs1017*clhs3211;
const double clhs3213 =             -clhs3212;
const double clhs3214 =             clhs3073 + clhs3160 - clhs3202;
const double clhs3215 =             (1.0L/2.0L)*clhs203*clhs3214*tau2;
const double clhs3216 =             clhs1050*clhs3215;
const double clhs3217 =             clhs1343*clhs3194 + clhs3072 + clhs3208 + clhs3213 + clhs3216*clhs57;
const double clhs3218 =             clhs1569*clhs1724;
const double clhs3219 =             clhs1579*clhs1728;
const double clhs3220 =             -clhs3219;
const double clhs3221 =             clhs1590*clhs1733;
const double clhs3222 =             -clhs3221;
const double clhs3223 =             clhs1742*clhs662;
const double clhs3224 =             clhs1744*clhs688;
const double clhs3225 =             -clhs1050*clhs3206 - clhs1343*clhs3167 + clhs1389*clhs3214;
const double clhs3226 =             clhs1569*clhs1728;
const double clhs3227 =             clhs1579*clhs1724;
const double clhs3228 =             clhs1215*clhs3211 - clhs1343*clhs3178 - clhs1400*clhs3214;
const double clhs3229 =             clhs1569*clhs1733;
const double clhs3230 =             clhs1590*clhs1724;
const double clhs3231 =             DN(2,0)*clhs7*clhs906*mu;
const double clhs3232 =             (2.0L/3.0L)*DN(2,1)*clhs7*clhs906*mu;
const double clhs3233 =             -clhs549;
const double clhs3234 =             clhs3233 + clhs581;
const double clhs3235 =             (1.0L/2.0L)*clhs1912*clhs203*tau2;
const double clhs3236 =             (1.0L/2.0L)*DN(2,1)*clhs67*tau3;
const double clhs3237 =             clhs203*clhs623*tau2;
const double clhs3238 =             -clhs1539 - clhs3090 + clhs501;
const double clhs3239 =             clhs1164*clhs1579;
const double clhs3240 =             clhs1151*clhs1590;
const double clhs3241 =             clhs203*clhs691*tau2;
const double clhs3242 =             -clhs649;
const double clhs3243 =             -clhs652;
const double clhs3244 =             clhs1471 - clhs1911 + clhs3102 + clhs3242 + clhs3243 + clhs676 + clhs678;
const double clhs3245 =             DN(2,1)*tau3;
const double clhs3246 =             clhs2602 - clhs2742 - clhs3112;
const double clhs3247 =             clhs203*clhs484*tau2;
const double clhs3248 =             clhs1369*clhs1579;
const double clhs3249 =             clhs1364*clhs1590;
const double clhs3250 =             clhs113*clhs3067;
const double clhs3251 =             clhs2443 + clhs3126 + clhs614;
const double clhs3252 =             -clhs1567 + clhs2419*clhs67 + clhs3251 + clhs619;
const double clhs3253 =             clhs14*clhs504;
const double clhs3254 =             clhs2445 - clhs2463*clhs67 + clhs3253*clhs67 - clhs554 + clhs690;
const double clhs3255 =             clhs1469 + clhs1901 + clhs2045 + clhs551 + clhs636 + clhs652;
const double clhs3256 =             clhs1473*clhs57 + clhs1909 + clhs3255 - clhs57*clhs586;
const double clhs3257 =             (1.0L/2.0L)*N[2]*clhs203*clhs3142*clhs67*tau2;
const double clhs3258 =             clhs1215*clhs3147 + clhs1575*clhs691 - clhs3137;
const double clhs3259 =             clhs203*clhs3244*tau2;
const double clhs3260 =             clhs1217*clhs3145;
const double clhs3261 =             clhs1826*clhs3143;
const double clhs3262 =             clhs1552*clhs1579;
const double clhs3263 =             clhs1547*clhs1590;
const double clhs3264 =             clhs113*clhs3154;
const double clhs3265 =             clhs3161 + clhs3162 + clhs3163 - clhs3166;
const double clhs3266 =             clhs3157*clhs67;
const double clhs3267 =             clhs3158*clhs67;
const double clhs3268 =             clhs3159*clhs67;
const double clhs3269 =             clhs3165*clhs67 + clhs3265 - clhs3266 - clhs3267 - clhs3268;
const double clhs3270 =             clhs3168*clhs8;
const double clhs3271 =             clhs3077*clhs8;
const double clhs3272 =             N[3]*clhs667;
const double clhs3273 =             clhs14*clhs3156;
const double clhs3274 =             clhs14*clhs3075;
const double clhs3275 =             N[3]*clhs664;
const double clhs3276 =             clhs3275*clhs67;
const double clhs3277 =             clhs502*clhs870;
const double clhs3278 =             clhs502*clhs869;
const double clhs3279 =             -clhs3270 - clhs3271 - clhs3272 + clhs3273*clhs67 + clhs3274*clhs67 + clhs3276 + clhs3277 - clhs3278*clhs67;
const double clhs3280 =             -clhs3182;
const double clhs3281 =             clhs3179 + clhs3180 + clhs3181 + clhs3184 + clhs3186 + clhs3188 + clhs3192 + clhs3280;
const double clhs3282 =             -clhs3183*clhs57 - clhs3185*clhs57 - clhs3187*clhs57 + clhs3189*clhs57 + clhs3281;
const double clhs3283 =             clhs1242*clhs3282 - clhs1254*clhs3279 + clhs1274*clhs3269 + clhs3264;
const double clhs3284 =             N[2]*N[3]*clhs1184*clhs203;
const double clhs3285 =             N[2]*clhs751 + N[2]*clhs754 + N[3]*clhs552 + N[3]*clhs555 - clhs151*clhs3284 - clhs155*clhs3200 + clhs1799*clhs3204 + clhs512*clhs835 + clhs512*clhs851 - clhs512*clhs877 + clhs518*clhs866 - clhs518*clhs888 + clhs559*clhs813 - clhs559*clhs817;
const double clhs3286 =             clhs1343*clhs3269 + clhs2020*clhs3205 - clhs3215*clhs985;
const double clhs3287 =             clhs1215*clhs3215 + clhs3072 - clhs3218;
const double clhs3288 =             clhs1561*clhs861;
const double clhs3289 =             clhs1756*clhs691;
const double clhs3290 =             clhs1343*clhs3282;
const double clhs3291 =             clhs1217*clhs3211;
const double clhs3292 =             clhs1826*clhs3206;
const double clhs3293 =             -clhs1343*clhs3279 + clhs2893*clhs3205 - clhs3211*clhs985;
const double clhs3294 =             clhs1579*clhs1733;
const double clhs3295 =             clhs1590*clhs1728;
const double clhs3296 =             (2.0L/3.0L)*DN(2,2)*clhs7*clhs906*mu;
const double clhs3297 =             -clhs583;
const double clhs3298 =             clhs3297 + clhs609;
const double clhs3299 =             (1.0L/2.0L)*clhs203*clhs2055*tau2;
const double clhs3300 =             (1.0L/2.0L)*DN(2,2)*clhs67*tau3;
const double clhs3301 =             clhs203*clhs633*tau2;
const double clhs3302 =             clhs203*clhs672*tau2;
const double clhs3303 =             clhs1471 + clhs1557 - clhs2053 + clhs3103 + clhs3104 + clhs3242 + clhs3243 + clhs693 + clhs695;
const double clhs3304 =             DN(2,2)*tau3;
const double clhs3305 =             clhs203*clhs499*tau2;
const double clhs3306 =             clhs166*clhs3067;
const double clhs3307 =             clhs2461 + clhs3129 + clhs625;
const double clhs3308 =             -clhs1568 + clhs2425*clhs67 + clhs3307 + clhs630;
const double clhs3309 =             clhs2464 + clhs3253 + clhs664;
const double clhs3310 =             -clhs1578 + clhs2445*clhs67 + clhs3309 + clhs669;
const double clhs3311 =             clhs1474 + clhs1901 + clhs551 + clhs586 + clhs646 + clhs652;
const double clhs3312 =             clhs1468*clhs57 - clhs2045*clhs57 + clhs2052 + clhs3311;
const double clhs3313 =             clhs203*clhs3303*tau2;
const double clhs3314 =             clhs166*clhs3154;
const double clhs3315 =             clhs3172 + clhs3173 + clhs3174 - clhs3177;
const double clhs3316 =             clhs3169*clhs67;
const double clhs3317 =             clhs3170*clhs67;
const double clhs3318 =             clhs3171*clhs67;
const double clhs3319 =             clhs3176*clhs67 + clhs3315 - clhs3316 - clhs3317 - clhs3318;
const double clhs3320 =             clhs3273 + clhs3274 + clhs3275 - clhs3278;
const double clhs3321 =             clhs3270*clhs67;
const double clhs3322 =             clhs3271*clhs67;
const double clhs3323 =             clhs3272*clhs67;
const double clhs3324 =             clhs3277*clhs67 + clhs3320 - clhs3321 - clhs3322 - clhs3323;
const double clhs3325 =             clhs3179 + clhs3180 + clhs3181 + clhs3183 + clhs3185 + clhs3187 + clhs3190 + clhs3280;
const double clhs3326 =             -clhs3184*clhs57 - clhs3186*clhs57 - clhs3188*clhs57 + clhs3191*clhs57 + clhs3325;
const double clhs3327 =             clhs1242*clhs3324 + clhs1254*clhs3326 + clhs1274*clhs3319 + clhs3314;
const double clhs3328 =             N[2]*clhs783 + N[2]*clhs786 + N[3]*clhs584 + N[3]*clhs587 - clhs190*clhs3284 - clhs193*clhs3199 + clhs1978*clhs3209 + clhs512*clhs863 - clhs512*clhs867 + clhs518*clhs845 + clhs518*clhs851 - clhs518*clhs894 + clhs559*clhs824 - clhs559*clhs828;
const double clhs3329 =             -clhs1017*clhs3215 + clhs1343*clhs3319 + clhs2020*clhs3210;
const double clhs3330 =             -clhs1017*clhs3206 + clhs1343*clhs3324 + clhs1389*clhs3210;
const double clhs3331 =             clhs1563*clhs887;
const double clhs3332 =             clhs1575*clhs890;
const double clhs3333 =             clhs1343*clhs3326;
const double clhs3334 =             clhs1483 + clhs1570;
const double clhs3335 =             clhs1580 + clhs3233;
const double clhs3336 =             clhs1591 + clhs3297;
const double clhs3337 =             (1.0L/3.0L)*DN(2,0)*clhs203;
const double clhs3338 =             (1.0L/3.0L)*DN(2,1)*clhs203;
const double clhs3339 =             (1.0L/3.0L)*DN(2,2)*clhs203;
const double clhs3340 =             clhs1488 + clhs2413;
const double clhs3341 =             (1.0L/2.0L)*clhs203*clhs2475*gamma*tau3;
const double clhs3342 =             (1.0L/3.0L)*DN(2,1)*clhs7;
const double clhs3343 =             (1.0L/3.0L)*DN(2,2)*clhs7;
const double clhs3344 =             (1.0L/3.0L)*DN(2,0)*clhs7;
const double clhs3345 =             (1.0L/2.0L)*N[2]*clhs203;
const double clhs3346 =             DN(2,0)*clhs203*clhs2145*clhs2146*lambda;
const double clhs3347 =             DN(2,1)*clhs203*clhs2145*clhs2146*lambda;
const double clhs3348 =             DN(2,2)*clhs203*clhs2145*clhs2146*lambda;
const double clhs3349 =             N[2]*clhs203*gamma;
const double clhs3350 =             clhs203*clhs2475*gamma*tau3;
const double clhs3351 =             (1.0L/2.0L)*N[2]*clhs7*tau2;
const double clhs3352 =             -DN(2,0)*clhs1125 - clhs2092*clhs616 - clhs2093*clhs627 + clhs2094*clhs2433 + clhs2096*clhs2434 + clhs2371 - clhs2375 + clhs2432;
const double clhs3353 =             -DN(2,0)*clhs10*clhs1113*clhs84 - DN(2,1)*clhs1129 + clhs2089*clhs2453 - clhs2093*clhs666 + clhs2096*clhs2454 + clhs2378 - clhs2382 + clhs2452;
const double clhs3354 =             -DN(2,0)*clhs10*clhs1113*clhs90 - DN(2,1)*clhs1113*clhs8*clhs90 - DN(2,2)*clhs1132 + clhs2089*clhs2468 + clhs2094*clhs2469 + clhs2385 - clhs2387 + clhs2467;
const double clhs3355 =             clhs1469 + clhs2045 + clhs636;
const double clhs3356 =             clhs3311 + clhs3355;
const double clhs3357 =             (1.0L/2.0L)*N[2]*clhs67*clhs7*tau2*(clhs2420 + clhs3251 + clhs505 + clhs617);
const double clhs3358 =             (1.0L/2.0L)*N[2]*clhs67*clhs7*tau2*(clhs2426 + clhs3307 + clhs509 + clhs628);
const double clhs3359 =             (1.0L/2.0L)*N[2]*clhs67*clhs7*tau2;
const double clhs3360 =             3*clhs652;
const double clhs3361 =             (1.0L/2.0L)*N[2]*clhs203*clhs2126*gamma*tau3;
const double clhs3362 =             (1.0L/2.0L)*N[2]*clhs67*clhs7*tau2*(clhs2446 + clhs3309 + clhs554 + clhs667);
const double clhs3363 =             3*clhs646;
const double clhs3364 =             3*clhs636;
const double clhs3365 =             (1.0L/2.0L)*N[2]*clhs203*gamma*tau2;
const double clhs3366 =             clhs2480*clhs502;
const double clhs3367 =             N[2]*N[3]*clhs2246*clhs7*clhs8;
const double clhs3368 =             N[2]*N[3]*clhs14*clhs2246*clhs7;
const double clhs3369 =             -clhs1125*clhs3073 - clhs1125*clhs3160 - clhs2092*clhs3164 - clhs2092*clhs3266 - clhs2092*clhs3267 - clhs2092*clhs3268 - clhs2093*clhs3175 - clhs2093*clhs3316 - clhs2093*clhs3317 - clhs2093*clhs3318 - clhs2244*clhs3181 - clhs2245*clhs3187 - clhs2245*clhs3188 - clhs2484*clhs502 + clhs3366 + clhs3367*clhs82 + clhs3368*clhs89;
const double clhs3370 =             clhs2487*clhs502;
const double clhs3371 =             N[2]*N[3]*clhs10*clhs2246*clhs7;
const double clhs3372 =             -clhs1129*clhs3075 - clhs1129*clhs3156 + clhs151*clhs3371 + clhs155*clhs3368 - clhs2093*clhs3276 - clhs2093*clhs3321 - clhs2093*clhs3322 - clhs2093*clhs3323 - clhs2256*clhs3179 - clhs2256*clhs3180 - clhs2256*clhs3181 - clhs2256*clhs3188 - clhs2257*clhs3164 - clhs2257*clhs3268 - clhs2258*clhs3187 - clhs2491*clhs502 + clhs3370;
const double clhs3373 =             clhs2494*clhs502;
const double clhs3374 =             -clhs1132*clhs3077 - clhs1132*clhs3168 + clhs190*clhs3371 + clhs193*clhs3367 - clhs2092*clhs3276 - clhs2092*clhs3323 - clhs2257*clhs3175 - clhs2257*clhs3318 - clhs2266*clhs3179 - clhs2266*clhs3180 - clhs2266*clhs3181 - clhs2266*clhs3183 - clhs2266*clhs3185 - clhs2266*clhs3187 - clhs2267*clhs3188 - clhs2496*clhs502 + clhs3373;
const double clhs3375 =             clhs3184 + clhs3186 + clhs3188 + clhs3192 + clhs3325;
const double clhs3376 =             clhs1068*clhs3154 + clhs1293*(-clhs10*clhs3366 - clhs1140*clhs3179 - clhs1140*clhs3180 - clhs1140*clhs3183 - clhs1140*clhs3184 - clhs1140*clhs3185 - clhs1140*clhs3186 - clhs14*clhs3373 - clhs2098*clhs3181 - clhs2104*clhs3187 - clhs2110*clhs3188 + clhs2176*clhs3164 + clhs2176*clhs3268 + clhs2177*clhs3175 + clhs2177*clhs3318 + clhs2178*clhs3276 + clhs2178*clhs3323 - clhs2272*clhs3202 - clhs2273*clhs3204 - clhs2274*clhs3209 - clhs3370*clhs8) - clhs2240*clhs3369 - clhs2251*clhs3372 - clhs2262*clhs3374 - clhs2271*clhs3375;
const double clhs3377 =             (1.0L/2.0L)*clhs67*clhs7*tau2*(clhs3157 + clhs3158 + clhs3159 - clhs3165 + clhs3265);
const double clhs3378 =             (1.0L/2.0L)*clhs67*clhs7*tau2*(clhs3169 + clhs3170 + clhs3171 - clhs3176 + clhs3315);
const double clhs3379 =             -clhs1217*clhs3378 - clhs1826*clhs3377 - clhs2302*(N[3]*clhs3360 - clhs1184*clhs3202 + 3*clhs3179 + 3*clhs3180 + clhs3193) + clhs2303*clhs3369 + clhs2304*clhs3214 + clhs3155;
const double clhs3380 =             (1.0L/2.0L)*clhs67*clhs7*tau2*(clhs3270 + clhs3271 + clhs3272 - clhs3277 + clhs3320);
const double clhs3381 =             -clhs1215*clhs3377 - clhs1217*clhs3380 + clhs2303*clhs3372 + clhs2304*clhs3205 - clhs2328*(N[3]*clhs3363 + 3*clhs3183 + 3*clhs3185 - clhs3204*clhs936 + clhs3281) + clhs3264;
const double clhs3382 =             -clhs1215*clhs3378 - clhs1826*clhs3380 + clhs2303*clhs3374 + clhs2304*clhs3210 - clhs2348*(N[3]*clhs3364 + 3*clhs3184 + 3*clhs3186 - clhs3209*clhs937 + clhs3325) + clhs3314;
const double clhs3383 =             clhs2364*clhs3375 + clhs3072 - clhs3207*gamma - clhs3212*gamma - clhs3216*gamma;
const double clhs3384 =             (1.0L/2.0L)*DN(3,0)*tau2;
const double clhs3385 =             (1.0L/2.0L)*DN(3,1)*tau2;
const double clhs3386 =             (1.0L/2.0L)*DN(3,2)*tau2;
const double clhs3387 =             -clhs1603;
const double clhs3388 =             DN(3,0)*tau2;
const double clhs3389 =             -clhs1599;
const double clhs3390 =             DN(3,1)*tau2;
const double clhs3391 =             -clhs1611;
const double clhs3392 =             DN(3,2)*tau2;
const double clhs3393 =             -clhs2763;
const double clhs3394 =             -clhs2759;
const double clhs3395 =             -clhs2771;
const double clhs3396 =             -clhs3160;
const double clhs3397 =             -clhs3156;
const double clhs3398 =             -clhs3168;
const double clhs3399 =             pow(N[3], 2);
const double clhs3400 =             -bdf0*clhs3399;
const double clhs3401 =             -DN(3,0)*N[3];
const double clhs3402 =             -DN(3,1)*N[3];
const double clhs3403 =             -DN(3,2)*N[3];
const double clhs3404 =             DN(3,1)*N[0]*clhs203*v_sc;
const double clhs3405 =             DN(3,2)*N[0]*clhs203*v_sc;
const double clhs3406 =             DN(3,1)*clhs7*clhs906*mu;
const double clhs3407 =             DN(3,2)*clhs7*clhs906*mu;
const double clhs3408 =             DN(3,0)*N[0]*clhs203*v_sc;
const double clhs3409 =             (2.0L/3.0L)*DN(3,0)*clhs7*clhs906*mu;
const double clhs3410 =             (1.0L/2.0L)*N[3]*clhs7;
const double clhs3411 =             clhs1666 + clhs702 - clhs745;
const double clhs3412 =             (1.0L/2.0L)*clhs1741*clhs203*tau2;
const double clhs3413 =             (1.0L/2.0L)*DN(3,0)*clhs67*tau3;
const double clhs3414 =             clhs1159*clhs1750;
const double clhs3415 =             clhs1151*clhs1760;
const double clhs3416 =             -clhs3415;
const double clhs3417 =             clhs1164*clhs1771;
const double clhs3418 =             -clhs3417;
const double clhs3419 =             clhs1741*clhs203*tau2;
const double clhs3420 =             clhs1151*clhs1750;
const double clhs3421 =             clhs1159*clhs1760;
const double clhs3422 =             clhs203*clhs861*tau2;
const double clhs3423 =             clhs1164*clhs1750;
const double clhs3424 =             clhs1159*clhs1771;
const double clhs3425 =             -clhs835;
const double clhs3426 =             clhs1736 + clhs3425 - clhs833;
const double clhs3427 =             -clhs842;
const double clhs3428 =             -clhs845;
const double clhs3429 =             clhs1738 - clhs1740 + clhs3426 + clhs3427 + clhs3428 + clhs849 + clhs852;
const double clhs3430 =             clhs203*clhs887*tau2;
const double clhs3431 =             clhs1664 + clhs746;
const double clhs3432 =             DN(3,0)*tau3;
const double clhs3433 =             DN(3,1)*N[1]*clhs203*v_sc;
const double clhs3434 =             DN(3,2)*N[1]*clhs203*v_sc;
const double clhs3435 =             DN(3,0)*N[1]*clhs203*v_sc;
const double clhs3436 =             clhs1360*clhs1750;
const double clhs3437 =             clhs1364*clhs1760;
const double clhs3438 =             -clhs3437;
const double clhs3439 =             clhs1369*clhs1771;
const double clhs3440 =             -clhs3439;
const double clhs3441 =             clhs1364*clhs1750;
const double clhs3442 =             clhs1360*clhs1760;
const double clhs3443 =             clhs1369*clhs1750;
const double clhs3444 =             clhs1360*clhs1771;
const double clhs3445 =             DN(3,1)*N[2]*clhs203*v_sc;
const double clhs3446 =             DN(3,2)*N[2]*clhs203*v_sc;
const double clhs3447 =             DN(3,0)*N[2]*clhs203*v_sc;
const double clhs3448 =             clhs1543*clhs1750;
const double clhs3449 =             clhs1547*clhs1760;
const double clhs3450 =             -clhs3449;
const double clhs3451 =             clhs1552*clhs1771;
const double clhs3452 =             -clhs3451;
const double clhs3453 =             clhs1547*clhs1750;
const double clhs3454 =             clhs1543*clhs1760;
const double clhs3455 =             clhs1552*clhs1750;
const double clhs3456 =             clhs1543*clhs1771;
const double clhs3457 =             clhs3399*clhs4;
const double clhs3458 =             DN(3,1)*N[3]*clhs203*v_sc;
const double clhs3459 =             DN(3,2)*N[3]*clhs203*v_sc;
const double clhs3460 =             DN(3,0)*N[3]*clhs203*v_sc;
const double clhs3461 =             N[3]*clhs83*clhs934*tau1;
const double clhs3462 =             clhs749*clhs8;
const double clhs3463 =             clhs2528 - clhs2551*clhs67 + clhs3462*clhs67 - clhs704 + clhs860;
const double clhs3464 =             (1.0L/2.0L)*N[3]*clhs7*clhs985*tau2;
const double clhs3465 =             clhs14*clhs749;
const double clhs3466 =             clhs2534 - clhs2569*clhs67 + clhs3465*clhs67 - clhs708 + clhs886;
const double clhs3467 =             (1.0L/2.0L)*N[3]*clhs1017*clhs7*tau2;
const double clhs3468 =             clhs1649 + clhs1654 + clhs1944 - clhs1945*clhs57 + clhs2075 + clhs3425 + clhs3428 + clhs57*clhs750 + clhs852;
const double clhs3469 =             (1.0L/2.0L)*N[3]*clhs1050*clhs7*tau2;
const double clhs3470 =             clhs1728*clhs1760;
const double clhs3471 =             -clhs3470;
const double clhs3472 =             clhs1744*clhs887;
const double clhs3473 =             clhs1724*clhs1750;
const double clhs3474 =             clhs1733*clhs1771;
const double clhs3475 =             -clhs3474;
const double clhs3476 =             clhs1742*clhs861;
const double clhs3477 =             N[3]*clhs7*clhs934*tau1;
const double clhs3478 =             clhs1726 + clhs703;
const double clhs3479 =             (1.0L/2.0L)*N[3]*clhs203*clhs3478*tau2;
const double clhs3480 =             clhs1731 + clhs707;
const double clhs3481 =             (1.0L/2.0L)*N[3]*clhs203*clhs3480*tau2;
const double clhs3482 =             clhs1722 + clhs749;
const double clhs3483 =             (1.0L/2.0L)*N[3]*clhs203*clhs3482*tau2;
const double clhs3484 =             clhs1728*clhs1750;
const double clhs3485 =             clhs1724*clhs1760;
const double clhs3486 =             (1.0L/2.0L)*N[3]*clhs203*clhs67*clhs985*tau2;
const double clhs3487 =             clhs1733*clhs1750;
const double clhs3488 =             clhs1724*clhs1771;
const double clhs3489 =             (1.0L/2.0L)*N[3]*clhs203*clhs67*tau2;
const double clhs3490 =             DN(3,0)*clhs7*clhs906*mu;
const double clhs3491 =             (2.0L/3.0L)*DN(3,1)*clhs7*clhs906*mu;
const double clhs3492 =             -clhs748;
const double clhs3493 =             clhs3492 + clhs780;
const double clhs3494 =             (1.0L/2.0L)*clhs1957*clhs203*tau2;
const double clhs3495 =             (1.0L/2.0L)*DN(3,1)*clhs67*tau3;
const double clhs3496 =             clhs203*clhs822*tau2;
const double clhs3497 =             -clhs1720 - clhs3414 + clhs700;
const double clhs3498 =             clhs1164*clhs1760;
const double clhs3499 =             clhs1151*clhs1771;
const double clhs3500 =             clhs203*clhs890*tau2;
const double clhs3501 =             -clhs848;
const double clhs3502 =             -clhs851;
const double clhs3503 =             clhs1652 - clhs1956 + clhs3426 + clhs3501 + clhs3502 + clhs875 + clhs877;
const double clhs3504 =             DN(3,1)*tau3;
const double clhs3505 =             clhs2610 - clhs2819 - clhs3436;
const double clhs3506 =             clhs1369*clhs1760;
const double clhs3507 =             clhs1364*clhs1771;
const double clhs3508 =             clhs3072 - clhs3216 - clhs3448;
const double clhs3509 =             clhs1552*clhs1760;
const double clhs3510 =             clhs1547*clhs1771;
const double clhs3511 =             clhs113*clhs3399;
const double clhs3512 =             clhs2552 + clhs3462 + clhs813;
const double clhs3513 =             -clhs1748 + clhs2528*clhs67 + clhs3512 + clhs818;
const double clhs3514 =             clhs14*clhs703;
const double clhs3515 =             clhs2554 - clhs2572*clhs67 + clhs3514*clhs67 - clhs753 + clhs889;
const double clhs3516 =             clhs1650 + clhs1946 + clhs2074 + clhs750 + clhs835 + clhs851;
const double clhs3517 =             clhs1654*clhs57 + clhs1954 + clhs3516 - clhs57*clhs785;
const double clhs3518 =             (1.0L/2.0L)*N[3]*clhs203*clhs3478*clhs67*tau2;
const double clhs3519 =             clhs1215*clhs3483 + clhs1756*clhs890 - clhs3473;
const double clhs3520 =             clhs203*tau2;
const double clhs3521 =             clhs1217*clhs3481;
const double clhs3522 =             clhs1826*clhs3479;
const double clhs3523 =             clhs1733*clhs1760;
const double clhs3524 =             clhs1728*clhs1771;
const double clhs3525 =             (2.0L/3.0L)*DN(3,2)*clhs7*clhs906*mu;
const double clhs3526 =             -clhs782;
const double clhs3527 =             clhs3526 + clhs808;
const double clhs3528 =             (1.0L/2.0L)*clhs203*clhs2084*tau2;
const double clhs3529 =             (1.0L/2.0L)*DN(3,2)*clhs67*tau3;
const double clhs3530 =             clhs203*clhs832*tau2;
const double clhs3531 =             clhs203*clhs871*tau2;
const double clhs3532 =             clhs1652 + clhs1738 - clhs2082 + clhs3427 + clhs3428 + clhs3501 + clhs3502 + clhs892 + clhs894;
const double clhs3533 =             DN(3,2)*tau3;
const double clhs3534 =             clhs166*clhs3399;
const double clhs3535 =             clhs2570 + clhs3465 + clhs824;
const double clhs3536 =             -clhs1749 + clhs2534*clhs67 + clhs3535 + clhs829;
const double clhs3537 =             clhs2573 + clhs3514 + clhs863;
const double clhs3538 =             -clhs1759 + clhs2554*clhs67 + clhs3537 + clhs868;
const double clhs3539 =             clhs1655 + clhs1946 + clhs750 + clhs785 + clhs845 + clhs851;
const double clhs3540 =             clhs1649*clhs57 - clhs2074*clhs57 + clhs2081 + clhs3539;
const double clhs3541 =             clhs1664 + clhs1751;
const double clhs3542 =             clhs1761 + clhs3492;
const double clhs3543 =             clhs1772 + clhs3526;
const double clhs3544 =             (1.0L/3.0L)*DN(3,0)*clhs203;
const double clhs3545 =             (1.0L/3.0L)*DN(3,1)*clhs203;
const double clhs3546 =             (1.0L/3.0L)*DN(3,2)*clhs203;
const double clhs3547 =             clhs1669 + clhs2522;
const double clhs3548 =             (1.0L/2.0L)*clhs203*clhs2584*gamma*tau3;
const double clhs3549 =             (1.0L/3.0L)*DN(3,1)*clhs7;
const double clhs3550 =             (1.0L/3.0L)*DN(3,2)*clhs7;
const double clhs3551 =             (1.0L/3.0L)*DN(3,0)*clhs7;
const double clhs3552 =             (1.0L/2.0L)*N[3]*clhs203;
const double clhs3553 =             DN(3,0)*clhs203*clhs2145*clhs2146*lambda;
const double clhs3554 =             DN(3,1)*clhs203*clhs2145*clhs2146*lambda;
const double clhs3555 =             DN(3,2)*clhs203*clhs2145*clhs2146*lambda;
const double clhs3556 =             N[3]*clhs203*gamma;
const double clhs3557 =             clhs203*clhs2584*gamma*tau3;
const double clhs3558 =             (1.0L/2.0L)*N[3]*clhs7*tau2;
const double clhs3559 =             -DN(3,0)*clhs1125 - clhs2092*clhs815 - clhs2093*clhs826 + clhs2094*clhs2542 + clhs2096*clhs2543 + clhs2480 - clhs2484 + clhs2541;
const double clhs3560 =             -DN(3,0)*clhs10*clhs1113*clhs84 - DN(3,1)*clhs1129 + clhs2089*clhs2562 - clhs2093*clhs865 + clhs2096*clhs2563 + clhs2487 - clhs2491 + clhs2561;
const double clhs3561 =             -DN(3,0)*clhs10*clhs1113*clhs90 - DN(3,1)*clhs1113*clhs8*clhs90 - DN(3,2)*clhs1132 + clhs2089*clhs2577 + clhs2094*clhs2578 + clhs2494 - clhs2496 + clhs2576;
const double clhs3562 =             clhs1650 + clhs2074 + clhs835;
const double clhs3563 =             clhs3539 + clhs3562;
const double clhs3564 =             (1.0L/2.0L)*N[3]*clhs67*clhs7*tau2*(clhs2529 + clhs3512 + clhs704 + clhs816);
const double clhs3565 =             (1.0L/2.0L)*N[3]*clhs67*clhs7*tau2*(clhs2535 + clhs3535 + clhs708 + clhs827);
const double clhs3566 =             (1.0L/2.0L)*N[3]*clhs67*clhs7*tau2;
const double clhs3567 =             (1.0L/2.0L)*N[3]*clhs203*clhs2126*gamma*tau3;
const double clhs3568 =             (1.0L/2.0L)*N[3]*clhs67*clhs7*tau2*(clhs2555 + clhs3537 + clhs753 + clhs866);
const double clhs3569 =             (1.0L/2.0L)*N[3]*clhs203*gamma*tau2;
            lhs(0,0)=clhs1 + clhs111*clhs2 + clhs112*clhs164 + clhs165*clhs201;
            lhs(0,1)=clhs202 - clhs204*clhs215 - clhs216*clhs226 + clhs227*clhs254;
            lhs(0,2)=-clhs216*clhs269 + clhs255 + clhs256*clhs259 - clhs270*clhs283;
            lhs(0,3)=clhs204*clhs290 + clhs256*clhs287 + clhs284 - clhs291*clhs299;
            lhs(0,4)=-clhs300*(pow(DN(0,0), 2) + pow(DN(0,1), 2) + pow(DN(0,2), 2));
            lhs(0,5)=clhs112*clhs383 + clhs165*clhs411 + clhs2*clhs348 + clhs301;
            lhs(0,6)=-clhs204*clhs424 - clhs216*clhs434 + clhs227*clhs458 + clhs413;
            lhs(0,7)=-clhs216*clhs473 + clhs256*clhs463 - clhs270*clhs484 + clhs460;
            lhs(0,8)=clhs204*clhs492 + clhs256*clhs489 - clhs291*clhs499 + clhs486;
            lhs(0,9)=clhs500;
            lhs(0,10)=clhs112*clhs582 + clhs165*clhs610 + clhs2*clhs548 + clhs501;
            lhs(0,11)=-clhs204*clhs623 - clhs216*clhs633 + clhs227*clhs657 + clhs612;
            lhs(0,12)=-clhs216*clhs672 + clhs256*clhs662 - clhs270*clhs683 + clhs659;
            lhs(0,13)=clhs204*clhs691 + clhs256*clhs688 - clhs291*clhs698 + clhs685;
            lhs(0,14)=clhs699;
            lhs(0,15)=clhs112*clhs781 + clhs165*clhs809 + clhs2*clhs747 + clhs700;
            lhs(0,16)=-clhs204*clhs822 - clhs216*clhs832 + clhs227*clhs856 + clhs811;
            lhs(0,17)=-clhs216*clhs871 + clhs256*clhs861 - clhs270*clhs882 + clhs858;
            lhs(0,18)=clhs204*clhs890 + clhs256*clhs887 - clhs291*clhs897 + clhs884;
            lhs(0,19)=clhs898;
            lhs(1,0)=-clhs1018*clhs990 - clhs1025*clhs1051 - clhs1052*clhs164 - clhs1053*clhs201 - clhs1062*clhs1066 - clhs1067*clhs1147 + clhs899 + clhs900*clhs902 + clhs903*clhs905 + clhs907*clhs910 + clhs911*clhs913 + clhs914*clhs915 - clhs916*clhs926 - clhs927*clhs929 - clhs930*clhs932 - clhs935*(DN(0,0)*clhs939 - clhs103 + clhs207*clhs84 - clhs214*clhs936 + clhs219*clhs90 - clhs225*clhs937 + clhs25 + clhs33 + clhs41 + clhs49 - clhs59 - clhs69 - clhs78 + clhs921*clhs943) - clhs947*clhs986;
            lhs(1,1)=clhs1 - clhs1017*clhs1175 - clhs1025*clhs1169 - clhs1061*clhs212 + clhs1067*clhs1190 + clhs1153 + clhs1155 - 4.0L/3.0L*clhs1160 + clhs1166 + clhs1168 - clhs1170*clhs1171 - clhs1173*clhs985 + clhs1177*clhs1178 - clhs1179*clhs932;
            lhs(1,2)=-clhs1050*clhs1173 + clhs1061*clhs1194 + clhs1067*clhs1204 - clhs1169*clhs947 + clhs1176*clhs1196 + (2.0L/3.0L)*clhs1191 - clhs1192 + clhs1193*clhs287 + clhs1195*clhs283 - clhs1197*clhs932 + clhs212*clhs259;
            lhs(1,3)=clhs1067*clhs1225 - clhs1169*clhs990 + clhs1175*clhs1215 - clhs1176*clhs1216*clhs1217 - clhs1194*clhs290 + (2.0L/3.0L)*clhs1205 - clhs1206 - clhs1207*clhs1210 - clhs1211*clhs1214 + clhs1218*clhs1219 + clhs212*clhs287;
            lhs(1,4)=clhs67*(-clhs1061*clhs256 - clhs1226*clhs1227 + clhs202 + clhs204*clhs259 + clhs216*clhs287);
            lhs(1,5)=-clhs1052*clhs383 - clhs1053*clhs411 - clhs1062*clhs1302 - clhs1067*clhs1342 + clhs1275 + clhs1276*clhs902 + clhs1277*clhs905 + clhs1279*clhs907 + clhs1280*clhs911 + clhs1281*clhs915 - clhs1290*clhs916 - clhs1291*clhs927 - clhs1292*clhs932 - clhs1293*(-clhs102*clhs412 + clhs1234*clhs939 + clhs1298);
            lhs(1,6)=clhs1067*clhs1388 - clhs1170*clhs1382 + clhs1357 - 4.0L/3.0L*clhs1361 + clhs1366 + clhs1371 - clhs1377*clhs212 + clhs1379 + clhs1381 - clhs1383*clhs932;
            lhs(1,7)=clhs1061*clhs1394 + clhs1067*clhs1399 + clhs1195*clhs484 + clhs1390 + (2.0L/3.0L)*clhs1391 - clhs1392 + clhs1393*clhs287 - clhs1395*clhs932 + clhs212*clhs463;
            lhs(1,8)=clhs1067*clhs1411 - clhs1194*clhs492 - clhs1210*clhs1404 - clhs1211*clhs1407 + clhs1219*clhs1408 + clhs1401 + (2.0L/3.0L)*clhs1402 - clhs1403 + clhs212*clhs489;
            lhs(1,9)=clhs67*(-clhs1061*clhs1414 - clhs1226*clhs1415 + clhs1412*clhs259 + clhs1413*clhs287 + clhs413);
            lhs(1,10)=-clhs1052*clhs582 - clhs1053*clhs610 - clhs1062*clhs1486 - clhs1067*clhs1526 - clhs1293*(-clhs102*clhs611 + clhs1422*clhs939 + clhs1482) + clhs1460 + clhs1461*clhs902 + clhs1462*clhs905 + clhs1464*clhs907 + clhs1465*clhs911 + clhs1466*clhs915 - clhs1475*clhs916 - clhs1476*clhs927 - clhs1477*clhs932;
            lhs(1,11)=clhs1067*clhs1571 - clhs1170*clhs1565 + clhs1540 - 4.0L/3.0L*clhs1544 + clhs1549 + clhs1554 - clhs1560*clhs212 + clhs1562 + clhs1564 - clhs1566*clhs932;
            lhs(1,12)=clhs1061*clhs1576 + clhs1067*clhs1581 + clhs1195*clhs683 + clhs1572 + (2.0L/3.0L)*clhs1573 - clhs1574 + clhs1575*clhs287 - clhs1577*clhs932 + clhs212*clhs662;
            lhs(1,13)=clhs1067*clhs1592 - clhs1194*clhs691 - clhs1210*clhs1585 - clhs1211*clhs1588 + clhs1219*clhs1589 + clhs1582 + (2.0L/3.0L)*clhs1583 - clhs1584 + clhs212*clhs688;
            lhs(1,14)=clhs67*(-clhs1061*clhs1595 - clhs1226*clhs1596 + clhs1593*clhs259 + clhs1594*clhs287 + clhs612);
            lhs(1,15)=-clhs1052*clhs781 - clhs1053*clhs809 - clhs1062*clhs1667 - clhs1067*clhs1707 - clhs1293*(-clhs102*clhs810 + clhs1603*clhs939 + clhs1663) + clhs1641 + clhs1642*clhs902 + clhs1643*clhs905 + clhs1645*clhs907 + clhs1646*clhs911 + clhs1647*clhs915 - clhs1656*clhs916 - clhs1657*clhs927 - clhs1658*clhs932;
            lhs(1,16)=clhs1067*clhs1752 - clhs1170*clhs1746 + clhs1721 - 4.0L/3.0L*clhs1725 + clhs1730 + clhs1735 - clhs1741*clhs212 + clhs1743 + clhs1745 - clhs1747*clhs932;
            lhs(1,17)=clhs1061*clhs1757 + clhs1067*clhs1762 + clhs1195*clhs882 + clhs1753 + (2.0L/3.0L)*clhs1754 - clhs1755 + clhs1756*clhs287 - clhs1758*clhs932 + clhs212*clhs861;
            lhs(1,18)=clhs1067*clhs1773 - clhs1194*clhs890 - clhs1210*clhs1766 - clhs1211*clhs1769 + clhs1219*clhs1770 + clhs1763 + (2.0L/3.0L)*clhs1764 - clhs1765 + clhs212*clhs887;
            lhs(1,19)=clhs67*(-clhs1061*clhs1776 - clhs1226*clhs1777 + clhs1774*clhs259 + clhs1775*clhs287 + clhs811);
            lhs(2,0)=-clhs1018*clhs1806 + clhs1051*clhs1802 - clhs1066*clhs1811 - clhs1147*clhs1816 + clhs164*clhs1815 + clhs1778 + clhs1780*clhs903 + clhs1781*clhs910 + clhs1783*clhs911 - clhs1789*clhs900 - clhs1790*clhs1792 - clhs1793*clhs927 + clhs1795*clhs930 + clhs1809*clhs986 - clhs1810*clhs201 + clhs902*clhs914 - clhs935*(DN(0,1)*clhs1798 + clhs1057*clhs1799 - clhs1184*clhs213 + clhs122 + clhs125 + clhs132 + clhs134 - clhs137 - clhs140 - clhs148 - clhs159 + clhs1797 + clhs262*clhs90 - clhs268*clhs937);
            lhs(2,1)=clhs1050*clhs1818 + clhs1154*clhs290 - clhs1167*clhs1814 + clhs1169*clhs1802 - clhs1171*clhs1817 - clhs1177*clhs985 + clhs1179*clhs1795 + clhs1190*clhs1816 - clhs1191 + (2.0L/3.0L)*clhs1192 - clhs212*clhs215;
            lhs(2,2)=clhs1 - 4.0L/3.0L*clhs1152 + clhs1166 + clhs1168 + clhs1169*clhs1809 + clhs1197*clhs1795 + clhs1204*clhs1816 - clhs1814*clhs212 + clhs1819 - clhs1823*clhs1824 + clhs1825 - clhs1827*clhs57;
            lhs(2,3)=clhs1017*clhs1818 + clhs1167*clhs287 - clhs1169*clhs1806 - clhs1175*clhs985 + clhs1218*clhs1795 + clhs1225*clhs1816 + clhs1814*clhs1830 + (2.0L/3.0L)*clhs1828 - clhs1829 + clhs1831*clhs299 + clhs212*clhs290;
            lhs(2,4)=clhs67*(-clhs1227*clhs1832 + clhs1822*clhs204 - clhs215*clhs256 + clhs216*clhs290 + clhs255);
            lhs(2,5)=-clhs1276*clhs1789 + clhs1277*clhs1780 + clhs1279*clhs1781 + clhs1281*clhs902 + clhs1292*clhs1795 - clhs1293*(clhs1230*clhs1798 - clhs158*clhs459 + clhs1860) - clhs1302*clhs1811 - clhs1342*clhs1816 - clhs1790*clhs1857 - clhs1810*clhs411 + clhs1815*clhs383 + clhs1852 + clhs1853*clhs911 - clhs1858*clhs927;
            lhs(2,6)=-clhs1378*clhs1814 + clhs1380*clhs290 - clhs1382*clhs1817 + clhs1383*clhs1795 + clhs1388*clhs1816 - clhs1391 + (2.0L/3.0L)*clhs1392 + clhs1862 - clhs212*clhs424;
            lhs(2,7)=-4.0L/3.0L*clhs1365 + clhs1371 + clhs1395*clhs1795 + clhs1399*clhs1816 - clhs1823*clhs1871 + clhs1863 - clhs1867*clhs212 + clhs1868 + clhs1869 + clhs1870 + clhs1872 - clhs1873*clhs57;
            lhs(2,8)=clhs1167*clhs489 + clhs1408*clhs1795 + clhs1411*clhs1816 + clhs1814*clhs1877 + clhs1831*clhs499 + clhs1874 + (2.0L/3.0L)*clhs1875 - clhs1876 + clhs212*clhs492;
            lhs(2,9)=clhs67*(clhs1412*clhs1822 + clhs1413*clhs290 - clhs1414*clhs215 - clhs1415*clhs1832 + clhs460);
            lhs(2,10)=-clhs1293*(clhs1418*clhs1798 - clhs158*clhs658 + clhs1905) - clhs1461*clhs1789 + clhs1462*clhs1780 + clhs1464*clhs1781 + clhs1466*clhs902 + clhs1477*clhs1795 - clhs1486*clhs1811 - clhs1526*clhs1816 - clhs1790*clhs1902 - clhs1810*clhs610 + clhs1815*clhs582 + clhs1897 + clhs1898*clhs911 - clhs1903*clhs927;
            lhs(2,11)=-clhs1561*clhs1814 + clhs1563*clhs290 - clhs1565*clhs1817 + clhs1566*clhs1795 + clhs1571*clhs1816 - clhs1573 + (2.0L/3.0L)*clhs1574 + clhs1907 - clhs212*clhs623;
            lhs(2,12)=-4.0L/3.0L*clhs1548 + clhs1554 + clhs1577*clhs1795 + clhs1581*clhs1816 - clhs1823*clhs1916 + clhs1908 - clhs1912*clhs212 + clhs1913 + clhs1914 + clhs1915 + clhs1917 - clhs1918*clhs57;
            lhs(2,13)=clhs1167*clhs688 + clhs1589*clhs1795 + clhs1592*clhs1816 + clhs1814*clhs1922 + clhs1831*clhs698 + clhs1919 + (2.0L/3.0L)*clhs1920 - clhs1921 + clhs212*clhs691;
            lhs(2,14)=clhs67*(clhs1593*clhs1822 + clhs1594*clhs290 - clhs1595*clhs215 - clhs1596*clhs1832 + clhs659);
            lhs(2,15)=-clhs1293*(-clhs158*clhs857 + clhs1599*clhs1798 + clhs1950) - clhs1642*clhs1789 + clhs1643*clhs1780 + clhs1645*clhs1781 + clhs1647*clhs902 + clhs1658*clhs1795 - clhs1667*clhs1811 - clhs1707*clhs1816 - clhs1790*clhs1947 - clhs1810*clhs809 + clhs1815*clhs781 + clhs1942 + clhs1943*clhs911 - clhs1948*clhs927;
            lhs(2,16)=-clhs1742*clhs1814 + clhs1744*clhs290 - clhs1746*clhs1817 + clhs1747*clhs1795 + clhs1752*clhs1816 - clhs1754 + (2.0L/3.0L)*clhs1755 + clhs1952 - clhs212*clhs822;
            lhs(2,17)=-4.0L/3.0L*clhs1729 + clhs1735 + clhs1758*clhs1795 + clhs1762*clhs1816 - clhs1823*clhs1961 + clhs1953 - clhs1957*clhs212 + clhs1958 + clhs1959 + clhs1960 + clhs1962 - clhs1963*clhs57;
            lhs(2,18)=clhs1167*clhs887 + clhs1770*clhs1795 + clhs1773*clhs1816 + clhs1814*clhs1967 + clhs1831*clhs897 + clhs1964 + (2.0L/3.0L)*clhs1965 - clhs1966 + clhs212*clhs890;
            lhs(2,19)=clhs67*(clhs1774*clhs1822 + clhs1775*clhs290 - clhs1776*clhs215 - clhs1777*clhs1832 + clhs858);
            lhs(3,0)=clhs1018*clhs1987 + clhs1051*clhs1981 - clhs1066*clhs1989 - clhs1147*clhs1994 + clhs164*clhs1988 + clhs1780*clhs900 + clhs1781*clhs913 + clhs1783*clhs907 + clhs1968 - clhs1969*clhs903 - clhs1970*clhs1972 - clhs1973*clhs927 + clhs1975*clhs930 + clhs1984*clhs986 + clhs1993*clhs201 + clhs905*clhs914 - clhs935*(DN(0,2)*clhs1977 + clhs1054*clhs1978 - clhs1184*clhs224 + clhs173 + clhs175 + clhs177 + clhs179 - clhs182 - clhs184 - clhs187 - clhs197 + clhs1976 + clhs238*clhs90 - clhs267*clhs936);
            lhs(3,1)=-clhs1017*clhs1177 + clhs1050*clhs1174*clhs1216 - clhs1154*clhs1992 - clhs1167*clhs269 + clhs1169*clhs1981 - clhs1171*clhs1995 + clhs1179*clhs1975 + clhs1190*clhs1994 - clhs1205 + (2.0L/3.0L)*clhs1206 - clhs212*clhs226;
            lhs(3,2)=-clhs1017*clhs1173 + clhs1154*clhs259 + clhs1169*clhs1984 + clhs1174*clhs1196 - clhs1193*clhs1992 + clhs1197*clhs1975 + clhs1204*clhs1994 - clhs1828 + (2.0L/3.0L)*clhs1829 - clhs1996*clhs283 - clhs212*clhs269;
            lhs(3,3)=clhs1 + clhs1153 + clhs1155 - 4.0L/3.0L*clhs1165 + clhs1169*clhs1987 - clhs1214*clhs1998 + clhs1218*clhs1975 + clhs1225*clhs1994 + clhs1819 - clhs1825*clhs57 + clhs1827 - clhs1992*clhs212;
            lhs(3,4)=clhs67*(-clhs1227*clhs1999 + clhs1997*clhs216 - clhs204*clhs269 - clhs226*clhs256 + clhs284);
            lhs(3,5)=clhs1276*clhs1780 - clhs1277*clhs1969 + clhs1280*clhs1781 + clhs1281*clhs905 + clhs1292*clhs1975 - clhs1293*(clhs1243*clhs1977 - clhs196*clhs485 + clhs2019) - clhs1302*clhs1989 - clhs1342*clhs1994 + clhs1853*clhs907 - clhs1970*clhs2017 + clhs1988*clhs383 + clhs1993*clhs411 + clhs2013 - clhs2018*clhs927;
            lhs(3,6)=-clhs1193*clhs424 - clhs1380*clhs1992 - clhs1382*clhs1995 + clhs1383*clhs1975 + clhs1388*clhs1994 - clhs1402 + (2.0L/3.0L)*clhs1403 + clhs2021 - clhs212*clhs434;
            lhs(3,7)=clhs1154*clhs463 - clhs1393*clhs1992 + clhs1395*clhs1975 + clhs1399*clhs1994 - clhs1875 + (2.0L/3.0L)*clhs1876 - clhs1996*clhs484 + clhs2022 - clhs212*clhs473;
            lhs(3,8)=clhs1366 - 4.0L/3.0L*clhs1370 - clhs1407*clhs1998 + clhs1408*clhs1975 + clhs1411*clhs1994 + clhs1863 - clhs1872*clhs57 + clhs1873 - clhs2026*clhs212 + clhs2027 + clhs2028 + clhs2029;
            lhs(3,9)=clhs67*(-clhs1412*clhs269 + clhs1413*clhs1997 - clhs1414*clhs226 - clhs1415*clhs1999 + clhs486);
            lhs(3,10)=-clhs1293*(clhs1430*clhs1977 - clhs196*clhs684 + clhs2049) + clhs1461*clhs1780 - clhs1462*clhs1969 + clhs1465*clhs1781 + clhs1466*clhs905 + clhs1477*clhs1975 - clhs1486*clhs1989 - clhs1526*clhs1994 + clhs1898*clhs907 - clhs1970*clhs2047 + clhs1988*clhs582 + clhs1993*clhs610 + clhs2043 - clhs2048*clhs927;
            lhs(3,11)=-clhs1193*clhs623 - clhs1563*clhs1992 - clhs1565*clhs1995 + clhs1566*clhs1975 + clhs1571*clhs1994 - clhs1583 + (2.0L/3.0L)*clhs1584 + clhs2050 - clhs212*clhs633;
            lhs(3,12)=clhs1154*clhs662 - clhs1575*clhs1992 + clhs1577*clhs1975 + clhs1581*clhs1994 - clhs1920 + (2.0L/3.0L)*clhs1921 - clhs1996*clhs683 + clhs2051 - clhs212*clhs672;
            lhs(3,13)=clhs1549 - 4.0L/3.0L*clhs1553 - clhs1588*clhs1998 + clhs1589*clhs1975 + clhs1592*clhs1994 + clhs1908 - clhs1917*clhs57 + clhs1918 - clhs2055*clhs212 + clhs2056 + clhs2057 + clhs2058;
            lhs(3,14)=clhs67*(-clhs1593*clhs269 + clhs1594*clhs1997 - clhs1595*clhs226 - clhs1596*clhs1999 + clhs685);
            lhs(3,15)=-clhs1293*(clhs1611*clhs1977 - clhs196*clhs883 + clhs2078) + clhs1642*clhs1780 - clhs1643*clhs1969 + clhs1646*clhs1781 + clhs1647*clhs905 + clhs1658*clhs1975 - clhs1667*clhs1989 - clhs1707*clhs1994 + clhs1943*clhs907 - clhs1970*clhs2076 + clhs1988*clhs781 + clhs1993*clhs809 + clhs2072 - clhs2077*clhs927;
            lhs(3,16)=-clhs1193*clhs822 - clhs1744*clhs1992 - clhs1746*clhs1995 + clhs1747*clhs1975 + clhs1752*clhs1994 - clhs1764 + (2.0L/3.0L)*clhs1765 + clhs2079 - clhs212*clhs832;
            lhs(3,17)=clhs1154*clhs861 - clhs1756*clhs1992 + clhs1758*clhs1975 + clhs1762*clhs1994 - clhs1965 + (2.0L/3.0L)*clhs1966 - clhs1996*clhs882 + clhs2080 - clhs212*clhs871;
            lhs(3,18)=clhs1730 - 4.0L/3.0L*clhs1734 - clhs1769*clhs1998 + clhs1770*clhs1975 + clhs1773*clhs1994 + clhs1953 - clhs1962*clhs57 + clhs1963 - clhs2084*clhs212 + clhs2085 + clhs2086 + clhs2087;
            lhs(3,19)=clhs67*(-clhs1774*clhs269 + clhs1775*clhs1997 - clhs1776*clhs226 - clhs1777*clhs1999 + clhs884);
            lhs(4,0)=-1.0L/2.0L*N[0]*clhs2121*clhs2127*clhs7*gamma + clhs0*clhs1068 - clhs1215*clhs2088*clhs2099 - clhs1217*clhs2088*clhs2111 - clhs1826*clhs2088*clhs2105 - clhs2112*clhs2113 - clhs2115*clhs2116 - clhs2118*clhs2119 - clhs2128*clhs2149 - clhs2150*clhs2164 - clhs2165*clhs2173 - clhs2175*clhs927 + clhs2181*clhs930 + clhs2183*clhs2184 + clhs935*(-clhs1020*clhs1140 - clhs1055*clhs2179 - clhs1058*clhs2179 - clhs1140*clhs116 - clhs1140*clhs170 - clhs2098*clhs247 + clhs210*clhs2176 - clhs2104*clhs241 - clhs2110*clhs230 + clhs2174 + clhs2176*clhs257 + clhs2177*clhs222 + clhs2177*clhs285 + clhs2178*clhs265 + clhs2178*clhs288 - clhs2179*clhs922);
            lhs(4,1)=clhs1169*clhs2099 + clhs1176*clhs2211 + clhs1179*clhs2181 + clhs1190*clhs2183 - clhs1215*clhs2204*(clhs170 + clhs2120 + clhs2205 + clhs241 + 6*clhs244 - 6*clhs922 + clhs925) - clhs1217*clhs2203 + clhs1811*clhs2116 - clhs1826*clhs2202 + clhs1989*clhs2119 - clhs2113*clhs2210 - clhs2185*clhs2191 - clhs2192*clhs2196 - clhs2197*clhs2201 - clhs2206*clhs2209 + clhs899;
            lhs(4,2)=-clhs1052*clhs2113 + clhs1169*clhs2105 + clhs1172*clhs2211 + clhs1197*clhs2181 + clhs1204*clhs2183 - clhs1215*clhs2202 - clhs1217*clhs2220 + clhs1778 - clhs1826*clhs2204*(-6*clhs1058 + clhs1808 + clhs2221 + 6*clhs238) + clhs1988*clhs2119 - clhs2116*clhs2224 - clhs2185*clhs2219 - clhs2192*clhs2216 - clhs2197*clhs2213 - clhs2206*clhs2223;
            lhs(4,3)=-clhs1053*clhs2113 + clhs1169*clhs2111 + clhs1174*clhs2211 - clhs1215*clhs2203 - clhs1217*clhs2204*(-6*clhs1055 + clhs1986 + clhs2228 + 6*clhs228) + clhs1218*clhs2181 + clhs1225*clhs2183 - clhs1810*clhs2116 - clhs1826*clhs2220 + clhs1968 - clhs2119*clhs2230 - clhs2185*clhs2226 - clhs2192*clhs2227 - clhs2197*clhs2225 - clhs2206*clhs2229;
            lhs(4,4)=clhs1 - clhs1017*clhs1174*clhs2236 - clhs1050*clhs1176*clhs2236 - clhs1150*clhs2232 - clhs1158*clhs2231 - clhs1163*clhs2233 + clhs1169*clhs2121*gamma - clhs1172*clhs2236*clhs985 - clhs1227*clhs2235 + clhs2113*clhs2237 + clhs2116*clhs2238 + clhs2119*clhs2239 - clhs2182*clhs2234;
            lhs(4,5)=clhs1292*clhs2181 - clhs2113*clhs2276 - clhs2116*clhs2277 - clhs2119*clhs2278 - clhs2128*clhs2288 - clhs2150*clhs2293 - clhs2165*clhs2295 + clhs2183*clhs2299 + clhs2275 - clhs2297*clhs927;
            lhs(4,6)=clhs1383*clhs2181 + clhs1388*clhs2183 - clhs2113*clhs2326 + clhs2116*clhs2324 + clhs2119*clhs2325 - clhs2185*clhs2311 - clhs2192*clhs2316 - clhs2197*clhs2319 - clhs2206*clhs2323 + clhs2305;
            lhs(4,7)=clhs1395*clhs2181 + clhs1399*clhs2183 - clhs2113*clhs2346 - clhs2116*clhs2347 + clhs2119*clhs2345 - clhs2185*clhs2340 - clhs2192*clhs2337 - clhs2197*clhs2333 - clhs2206*clhs2344 + clhs2329;
            lhs(4,8)=clhs1408*clhs2181 + clhs1411*clhs2183 - clhs2113*clhs2361 - clhs2116*clhs2362 - clhs2119*clhs2363 - clhs2185*clhs2355 - clhs2192*clhs2356 - clhs2197*clhs2352 - clhs2206*clhs2360 + clhs2349;
            lhs(4,9)=-clhs1359*clhs2231 - clhs1363*clhs2232 - clhs1368*clhs2233 - clhs1415*clhs2235 + clhs2113*clhs2367 + clhs2116*clhs2368 + clhs2119*clhs2369 - clhs2234*clhs2366 + clhs2365;
            lhs(4,10)=clhs1477*clhs2181 - clhs2113*clhs2391 - clhs2116*clhs2392 - clhs2119*clhs2393 - clhs2128*clhs2403 - clhs2150*clhs2408 - clhs2165*clhs2410 + clhs2183*clhs2414 + clhs2390 - clhs2412*clhs927;
            lhs(4,11)=clhs1566*clhs2181 + clhs1571*clhs2183 - clhs2113*clhs2438 + clhs2116*clhs2436 + clhs2119*clhs2437 - clhs2185*clhs2423 - clhs2192*clhs2428 - clhs2197*clhs2431 - clhs2206*clhs2435 + clhs2417;
            lhs(4,12)=clhs1577*clhs2181 + clhs1581*clhs2183 - clhs2113*clhs2457 - clhs2116*clhs2458 + clhs2119*clhs2456 - clhs2185*clhs2451 - clhs2192*clhs2448 - clhs2197*clhs2444 - clhs2206*clhs2455 + clhs2440;
            lhs(4,13)=clhs1589*clhs2181 + clhs1592*clhs2183 - clhs2113*clhs2471 - clhs2116*clhs2472 - clhs2119*clhs2473 - clhs2185*clhs2465 - clhs2192*clhs2466 - clhs2197*clhs2462 - clhs2206*clhs2470 + clhs2459;
            lhs(4,14)=-clhs1542*clhs2231 - clhs1546*clhs2232 - clhs1551*clhs2233 - clhs1596*clhs2235 + clhs2113*clhs2476 + clhs2116*clhs2477 + clhs2119*clhs2478 - clhs2234*clhs2475 + clhs2474;
            lhs(4,15)=clhs1658*clhs2181 - clhs2113*clhs2500 - clhs2116*clhs2501 - clhs2119*clhs2502 - clhs2128*clhs2512 - clhs2150*clhs2517 - clhs2165*clhs2519 + clhs2183*clhs2523 + clhs2499 - clhs2521*clhs927;
            lhs(4,16)=clhs1747*clhs2181 + clhs1752*clhs2183 - clhs2113*clhs2547 + clhs2116*clhs2545 + clhs2119*clhs2546 - clhs2185*clhs2532 - clhs2192*clhs2537 - clhs2197*clhs2540 - clhs2206*clhs2544 + clhs2526;
            lhs(4,17)=clhs1758*clhs2181 + clhs1762*clhs2183 - clhs2113*clhs2566 - clhs2116*clhs2567 + clhs2119*clhs2565 - clhs2185*clhs2560 - clhs2192*clhs2557 - clhs2197*clhs2553 - clhs2206*clhs2564 + clhs2549;
            lhs(4,18)=clhs1770*clhs2181 + clhs1773*clhs2183 - clhs2113*clhs2580 - clhs2116*clhs2581 - clhs2119*clhs2582 - clhs2185*clhs2574 - clhs2192*clhs2575 - clhs2197*clhs2571 - clhs2206*clhs2579 + clhs2568;
            lhs(4,19)=-clhs1723*clhs2231 - clhs1727*clhs2232 - clhs1732*clhs2233 - clhs1777*clhs2235 + clhs2113*clhs2585 + clhs2116*clhs2586 + clhs2119*clhs2587 - clhs2234*clhs2584 + clhs2583;
            lhs(5,0)=clhs111*clhs2588 + clhs164*clhs2589 + clhs201*clhs2590 + clhs301;
            lhs(5,1)=-clhs1412*clhs215 - clhs1413*clhs226 + clhs254*clhs2592 + clhs2591;
            lhs(5,2)=-clhs1413*clhs269 + clhs1414*clhs259 + clhs1824*clhs2594 + clhs2593;
            lhs(5,3)=clhs1412*clhs290 + clhs1414*clhs287 + clhs2595 - clhs2596*clhs299;
            lhs(5,4)=clhs500;
            lhs(5,5)=clhs2588*clhs348 + clhs2589*clhs383 + clhs2590*clhs411 + clhs2598;
            lhs(5,6)=-clhs1412*clhs424 - clhs1413*clhs434 + clhs2592*clhs458 + clhs2599;
            lhs(5,7)=-clhs1413*clhs473 + clhs1414*clhs463 - clhs2594*clhs484 + clhs2600;
            lhs(5,8)=clhs1412*clhs492 + clhs1414*clhs489 - clhs2596*clhs499 + clhs2601;
            lhs(5,9)=-clhs300*(pow(DN(1,0), 2) + pow(DN(1,1), 2) + pow(DN(1,2), 2));
            lhs(5,10)=clhs2588*clhs548 + clhs2589*clhs582 + clhs2590*clhs610 + clhs2602;
            lhs(5,11)=-clhs1412*clhs623 - clhs1413*clhs633 + clhs2592*clhs657 + clhs2604;
            lhs(5,12)=-clhs1413*clhs672 + clhs1414*clhs662 - clhs2594*clhs683 + clhs2606;
            lhs(5,13)=clhs1412*clhs691 + clhs1414*clhs688 - clhs2596*clhs698 + clhs2608;
            lhs(5,14)=clhs2609;
            lhs(5,15)=clhs2588*clhs747 + clhs2589*clhs781 + clhs2590*clhs809 + clhs2610;
            lhs(5,16)=-clhs1412*clhs822 - clhs1413*clhs832 + clhs2592*clhs856 + clhs2612;
            lhs(5,17)=-clhs1413*clhs871 + clhs1414*clhs861 - clhs2594*clhs882 + clhs2614;
            lhs(5,18)=clhs1412*clhs890 + clhs1414*clhs887 - clhs2596*clhs897 + clhs2616;
            lhs(5,19)=clhs2617;
            lhs(6,0)=-clhs1066*clhs2626 - clhs1147*clhs2627 + clhs1275 - clhs1293*(-clhs102*clhs1234 + clhs1298 + clhs412*clhs939) - clhs164*clhs2346 - clhs201*clhs2361 + clhs2618*clhs902 + clhs2619*clhs905 + clhs2620*clhs910 + clhs2621*clhs913 + clhs2622*clhs915 - clhs2623*clhs926 - clhs2624*clhs929 - clhs2625*clhs930;
            lhs(6,1)=-clhs1061*clhs421 - clhs1171*clhs2633 - clhs1179*clhs2625 + clhs1190*clhs2627 + clhs1357 + clhs1868 + clhs2027 - 4.0L/3.0L*clhs2628 + clhs2630 + clhs2632;
            lhs(6,2)=clhs1193*clhs489 + clhs1194*clhs1377 - clhs1197*clhs2625 + clhs1204*clhs2627 + clhs1390 + clhs259*clhs421 + (2.0L/3.0L)*clhs2634 - clhs2635 + clhs2636*clhs283;
            lhs(6,3)=-clhs1207*clhs2643 - clhs1214*clhs2644 + clhs1218*clhs2645 + clhs1225*clhs2627 + clhs1401 - clhs1830*clhs463 + (2.0L/3.0L)*clhs2637 - clhs2638 + clhs287*clhs421;
            lhs(6,4)=clhs67*(-clhs1227*clhs2646 - clhs1377*clhs256 + clhs204*clhs463 + clhs216*clhs489 + clhs2591);
            lhs(6,5)=clhs1279*clhs2620 + clhs1280*clhs2621 - clhs1290*clhs2623 - clhs1291*clhs2624 - clhs1292*clhs2625 - clhs1302*clhs2626 - clhs1342*clhs2627 - clhs2346*clhs383 - clhs2361*clhs411 + clhs2647 + clhs2648*clhs902 + clhs2649*clhs905 + clhs2650*clhs915 - clhs2651*(DN(1,0)*clhs939 + clhs1258*clhs943 + clhs313 + clhs316 + clhs319 + clhs322 - clhs325 - clhs329 - clhs333 - clhs342 + clhs416*clhs84 - clhs423*clhs936 + clhs427*clhs90 - clhs433*clhs937) - clhs2653*clhs2654 - clhs2656*clhs2657 - clhs2658*clhs2659;
            lhs(6,6)=-clhs1017*clhs2671 + clhs1178*clhs2673 - clhs1377*clhs421 - clhs1382*clhs2633 - clhs1383*clhs2625 + clhs1388*clhs2627 + clhs2598 - clhs2658*clhs2667 + clhs2661 + clhs2662 - 4.0L/3.0L*clhs2663 + clhs2665 + clhs2666 - clhs2669*clhs985;
            lhs(6,7)=-clhs1050*clhs2669 + clhs1377*clhs1394 + clhs1393*clhs489 - clhs1395*clhs2625 + clhs1399*clhs2627 + clhs2636*clhs484 - clhs2653*clhs2667 + clhs2672*clhs2676 + (2.0L/3.0L)*clhs2674 - clhs2675 + clhs421*clhs463;
            lhs(6,8)=clhs1215*clhs2671 - clhs1217*clhs2672*clhs2679 - clhs1394*clhs492 - clhs1404*clhs2643 - clhs1407*clhs2644 + clhs1408*clhs2645 + clhs1411*clhs2627 - clhs2656*clhs2667 + (2.0L/3.0L)*clhs2677 - clhs2678 + clhs421*clhs489;
            lhs(6,9)=clhs67*(-clhs1377*clhs1414 + clhs1412*clhs463 + clhs1413*clhs489 - clhs1415*clhs2646 + clhs2599);
            lhs(6,10)=-clhs1293*(-clhs102*clhs2603 + clhs2686*clhs939 + clhs2729) + clhs1464*clhs2620 + clhs1465*clhs2621 - clhs1475*clhs2623 - clhs1476*clhs2624 - clhs1477*clhs2625 - clhs1486*clhs2626 - clhs1526*clhs2627 - clhs2346*clhs582 - clhs2361*clhs610 + clhs2721 + clhs2722*clhs902 + clhs2723*clhs905 + clhs2724*clhs915;
            lhs(6,11)=-clhs1560*clhs421 - clhs1565*clhs2633 - clhs1566*clhs2625 + clhs1571*clhs2627 + clhs2743 - 4.0L/3.0L*clhs2744 + clhs2746 + clhs2748 + clhs2749 + clhs2750;
            lhs(6,12)=clhs1377*clhs1576 + clhs1575*clhs489 - clhs1577*clhs2625 + clhs1581*clhs2627 + clhs2636*clhs683 + clhs2751 + (2.0L/3.0L)*clhs2752 - clhs2753 + clhs421*clhs662;
            lhs(6,13)=-clhs1394*clhs691 - clhs1585*clhs2643 - clhs1588*clhs2644 + clhs1589*clhs2645 + clhs1592*clhs2627 + clhs2754 + (2.0L/3.0L)*clhs2755 - clhs2756 + clhs421*clhs688;
            lhs(6,14)=clhs67*(-clhs1377*clhs1595 + clhs1593*clhs463 + clhs1594*clhs489 - clhs1596*clhs2646 + clhs2604);
            lhs(6,15)=-clhs1293*(-clhs102*clhs2611 + clhs2763*clhs939 + clhs2806) + clhs1645*clhs2620 + clhs1646*clhs2621 - clhs1656*clhs2623 - clhs1657*clhs2624 - clhs1658*clhs2625 - clhs1667*clhs2626 - clhs1707*clhs2627 - clhs2346*clhs781 - clhs2361*clhs809 + clhs2798 + clhs2799*clhs902 + clhs2800*clhs905 + clhs2801*clhs915;
            lhs(6,16)=-clhs1741*clhs421 - clhs1746*clhs2633 - clhs1747*clhs2625 + clhs1752*clhs2627 + clhs2820 - 4.0L/3.0L*clhs2821 + clhs2823 + clhs2825 + clhs2826 + clhs2827;
            lhs(6,17)=clhs1377*clhs1757 + clhs1756*clhs489 - clhs1758*clhs2625 + clhs1762*clhs2627 + clhs2636*clhs882 + clhs2828 + (2.0L/3.0L)*clhs2829 - clhs2830 + clhs421*clhs861;
            lhs(6,18)=-clhs1394*clhs890 - clhs1766*clhs2643 - clhs1769*clhs2644 + clhs1770*clhs2645 + clhs1773*clhs2627 + clhs2831 + (2.0L/3.0L)*clhs2832 - clhs2833 + clhs421*clhs887;
            lhs(6,19)=clhs67*(-clhs1377*clhs1776 + clhs1774*clhs463 + clhs1775*clhs489 - clhs1777*clhs2646 + clhs2612);
            lhs(7,0)=-clhs1066*clhs2324 - clhs1147*clhs2839 - clhs1293*(-clhs1230*clhs158 + clhs1798*clhs459 + clhs1860) + clhs164*clhs2838 + clhs1780*clhs2619 + clhs1783*clhs2621 - clhs1789*clhs2618 - clhs1792*clhs2835 - clhs1793*clhs2624 + clhs1852 - clhs201*clhs2362 + clhs2622*clhs902 + clhs2834*clhs910 + clhs2837*clhs930;
            lhs(7,1)=clhs1154*clhs492 - clhs1167*clhs1867 - clhs1171*clhs2840 + clhs1179*clhs2837 + clhs1190*clhs2839 + clhs1862 - clhs215*clhs421 - clhs2634 + (2.0L/3.0L)*clhs2635;
            lhs(7,2)=clhs1197*clhs2837 + clhs1204*clhs2839 + clhs1347*clhs57 + clhs1353 + clhs1379 - clhs1814*clhs421 - clhs1867*clhs2842 + clhs1870 + clhs2028 - 4.0L/3.0L*clhs2629 + clhs2632 + clhs2841;
            lhs(7,3)=clhs1218*clhs2837 + clhs1225*clhs2839 + clhs1378*clhs287 + clhs1830*clhs1867 + clhs1874 + (2.0L/3.0L)*clhs2843 - clhs2844 + clhs2845*clhs299 + clhs290*clhs421;
            lhs(7,4)=clhs67*(-clhs1227*clhs2849 + clhs204*clhs2848 + clhs216*clhs492 - clhs256*clhs424 + clhs2593);
            lhs(7,5)=clhs1279*clhs2834 + clhs1292*clhs2837 - clhs1302*clhs2324 - clhs1342*clhs2839 + clhs1780*clhs2649 - clhs1789*clhs2648 + clhs1853*clhs2621 - clhs1857*clhs2835 - clhs1858*clhs2624 - clhs2362*clhs411 + clhs2650*clhs902 - clhs2651*(DN(1,1)*clhs1798 - clhs1184*clhs422 + clhs1266*clhs1799 + clhs357 + clhs360 + clhs362 + clhs364 - clhs366 - clhs369 - clhs372 - clhs379 + clhs450*clhs84 + clhs466*clhs90 - clhs472*clhs937) + clhs2654*clhs2856 - clhs2657*clhs2854 + clhs2659*clhs2852 + clhs2838*clhs383 + clhs2850;
            lhs(7,6)=clhs1050*clhs2857 - clhs1378*clhs1867 + clhs1380*clhs492 - clhs1382*clhs2840 + clhs1383*clhs2837 + clhs1388*clhs2839 + clhs2667*clhs2852 - clhs2673*clhs985 - clhs2674 + (2.0L/3.0L)*clhs2675 - clhs421*clhs424;
            lhs(7,7)=clhs1395*clhs2837 + clhs1399*clhs2839 - clhs1867*clhs421 - clhs1871*clhs2859 + clhs2598 - 4.0L/3.0L*clhs2660 + clhs2665 + clhs2666 + clhs2667*clhs2856 + clhs2858 + clhs2860 - clhs2861*clhs57;
            lhs(7,8)=clhs1017*clhs2857 + clhs1378*clhs489 + clhs1408*clhs2837 + clhs1411*clhs2839 + clhs1867*clhs1877 - clhs2667*clhs2854 - clhs2671*clhs985 + clhs2845*clhs499 + (2.0L/3.0L)*clhs2862 - clhs2863 + clhs421*clhs492;
            lhs(7,9)=clhs67*(clhs1412*clhs2848 + clhs1413*clhs492 - clhs1414*clhs424 - clhs1415*clhs2849 + clhs2600);
            lhs(7,10)=-clhs1293*(-clhs158*clhs2605 + clhs1798*clhs2682 + clhs2885) + clhs1464*clhs2834 + clhs1477*clhs2837 - clhs1486*clhs2324 - clhs1526*clhs2839 + clhs1780*clhs2723 - clhs1789*clhs2722 + clhs1898*clhs2621 - clhs1902*clhs2835 - clhs1903*clhs2624 - clhs2362*clhs610 + clhs2724*clhs902 + clhs2838*clhs582 + clhs2883;
            lhs(7,11)=-clhs1561*clhs1867 + clhs1563*clhs492 - clhs1565*clhs2840 + clhs1566*clhs2837 + clhs1571*clhs2839 - clhs2752 + (2.0L/3.0L)*clhs2753 + clhs2886 - clhs421*clhs623;
            lhs(7,12)=clhs1577*clhs2837 + clhs1581*clhs2839 - clhs1912*clhs421 - clhs1916*clhs2859 - 4.0L/3.0L*clhs2745 + clhs2748 + clhs2887 + clhs2888 + clhs2889 + clhs2890 + clhs2891 - clhs2892*clhs57;
            lhs(7,13)=clhs1378*clhs688 + clhs1589*clhs2837 + clhs1592*clhs2839 + clhs1867*clhs1922 + clhs2845*clhs698 + clhs2894 + (2.0L/3.0L)*clhs2895 - clhs2896 + clhs421*clhs691;
            lhs(7,14)=clhs67*(clhs1593*clhs2848 + clhs1594*clhs492 - clhs1595*clhs424 - clhs1596*clhs2849 + clhs2606);
            lhs(7,15)=-clhs1293*(-clhs158*clhs2613 + clhs1798*clhs2759 + clhs2918) + clhs1645*clhs2834 + clhs1658*clhs2837 - clhs1667*clhs2324 - clhs1707*clhs2839 + clhs1780*clhs2800 - clhs1789*clhs2799 + clhs1943*clhs2621 - clhs1947*clhs2835 - clhs1948*clhs2624 - clhs2362*clhs809 + clhs2801*clhs902 + clhs2838*clhs781 + clhs2916;
            lhs(7,16)=-clhs1742*clhs1867 + clhs1744*clhs492 - clhs1746*clhs2840 + clhs1747*clhs2837 + clhs1752*clhs2839 - clhs2829 + (2.0L/3.0L)*clhs2830 + clhs2919 - clhs421*clhs822;
            lhs(7,17)=clhs1758*clhs2837 + clhs1762*clhs2839 - clhs1957*clhs421 - clhs1961*clhs2859 - 4.0L/3.0L*clhs2822 + clhs2825 + clhs2920 + clhs2921 + clhs2922 + clhs2923 + clhs2924 - clhs2925*clhs57;
            lhs(7,18)=clhs1378*clhs887 + clhs1770*clhs2837 + clhs1773*clhs2839 + clhs1867*clhs1967 + clhs2845*clhs897 + clhs2926 + (2.0L/3.0L)*clhs2927 - clhs2928 + clhs421*clhs890;
            lhs(7,19)=clhs67*(clhs1774*clhs2848 + clhs1775*clhs492 - clhs1776*clhs424 - clhs1777*clhs2849 + clhs2614);
            lhs(8,0)=-clhs1066*clhs2325 - clhs1147*clhs2933 - clhs1293*(-clhs1243*clhs196 + clhs1977*clhs485 + clhs2019) + clhs164*clhs2345 + clhs1780*clhs2618 + clhs1783*clhs2620 - clhs1969*clhs2619 - clhs1972*clhs2929 - clhs1973*clhs2624 + clhs201*clhs2932 + clhs2013 + clhs2622*clhs905 + clhs2834*clhs913 + clhs2931*clhs930;
            lhs(8,1)=-clhs1154*clhs2026 - clhs1167*clhs473 - clhs1171*clhs2934 + clhs1179*clhs2931 + clhs1190*clhs2933 + clhs2021 - clhs226*clhs421 - clhs2637 + (2.0L/3.0L)*clhs2638;
            lhs(8,2)=-clhs1193*clhs2026 + clhs1197*clhs2931 + clhs1204*clhs2933 + clhs1380*clhs259 + clhs2022 - clhs269*clhs421 - clhs283*clhs2935 - clhs2843 + (2.0L/3.0L)*clhs2844;
            lhs(8,3)=clhs1218*clhs2931 + clhs1225*clhs2933 + clhs1348 + clhs1352*clhs57 + clhs1381 + clhs1869 - clhs1992*clhs421 - clhs2026*clhs2936 + clhs2029 + clhs2630 - 4.0L/3.0L*clhs2631 + clhs2841;
            lhs(8,4)=clhs67*(-clhs1227*clhs2938 - clhs204*clhs473 + clhs216*clhs2937 - clhs256*clhs434 + clhs2595);
            lhs(8,5)=clhs1280*clhs2834 + clhs1292*clhs2931 - clhs1302*clhs2325 - clhs1342*clhs2933 + clhs1780*clhs2648 + clhs1853*clhs2620 - clhs1969*clhs2649 - clhs2017*clhs2929 - clhs2018*clhs2624 + clhs2345*clhs383 + clhs2650*clhs905 - clhs2651*(DN(1,2)*clhs1977 - clhs1184*clhs432 + clhs1269*clhs1978 + clhs390 + clhs392 + clhs394 + clhs396 - clhs398 - clhs400 - clhs403 - clhs408 + clhs444*clhs90 + clhs450*clhs90 - clhs471*clhs936) + clhs2654*clhs2943 + clhs2657*clhs2945 + clhs2659*clhs2941 + clhs2932*clhs411 + clhs2939;
            lhs(8,6)=-clhs1017*clhs2673 + clhs1050*clhs2670*clhs2679 - clhs1378*clhs473 - clhs1380*clhs2026 - clhs1382*clhs2934 + clhs1383*clhs2931 + clhs1388*clhs2933 + clhs2667*clhs2941 - clhs2677 + (2.0L/3.0L)*clhs2678 - clhs421*clhs434;
            lhs(8,7)=-clhs1017*clhs2669 + clhs1380*clhs463 - clhs1393*clhs2026 + clhs1395*clhs2931 + clhs1399*clhs2933 + clhs2667*clhs2943 + clhs2670*clhs2676 - clhs2862 + (2.0L/3.0L)*clhs2863 - clhs2935*clhs484 - clhs421*clhs473;
            lhs(8,8)=-clhs1407*clhs2946 + clhs1408*clhs2931 + clhs1411*clhs2933 - clhs2026*clhs421 + clhs2598 + clhs2661 + clhs2662 - 4.0L/3.0L*clhs2664 + clhs2667*clhs2945 + clhs2858 - clhs2860*clhs57 + clhs2861;
            lhs(8,9)=clhs67*(-clhs1412*clhs473 + clhs1413*clhs2937 - clhs1414*clhs434 - clhs1415*clhs2938 + clhs2601);
            lhs(8,10)=-clhs1293*(-clhs196*clhs2607 + clhs1977*clhs2694 + clhs2961) + clhs1465*clhs2834 + clhs1477*clhs2931 - clhs1486*clhs2325 - clhs1526*clhs2933 + clhs1780*clhs2722 + clhs1898*clhs2620 - clhs1969*clhs2723 - clhs2047*clhs2929 - clhs2048*clhs2624 + clhs2345*clhs582 + clhs2724*clhs905 + clhs2932*clhs610 + clhs2960;
            lhs(8,11)=-clhs1393*clhs623 - clhs1563*clhs2026 - clhs1565*clhs2934 + clhs1566*clhs2931 + clhs1571*clhs2933 - clhs2755 + (2.0L/3.0L)*clhs2756 + clhs2962 - clhs421*clhs633;
            lhs(8,12)=clhs1380*clhs662 - clhs1575*clhs2026 + clhs1577*clhs2931 + clhs1581*clhs2933 - clhs2895 + (2.0L/3.0L)*clhs2896 - clhs2935*clhs683 + clhs2963 - clhs421*clhs672;
            lhs(8,13)=-clhs1588*clhs2946 + clhs1589*clhs2931 + clhs1592*clhs2933 - clhs2055*clhs421 + clhs2746 - 4.0L/3.0L*clhs2747 + clhs2887 - clhs2891*clhs57 + clhs2892 + clhs2964 + clhs2965 + clhs2966;
            lhs(8,14)=clhs67*(-clhs1593*clhs473 + clhs1594*clhs2937 - clhs1595*clhs434 - clhs1596*clhs2938 + clhs2608);
            lhs(8,15)=-clhs1293*(-clhs196*clhs2615 + clhs1977*clhs2771 + clhs2981) + clhs1646*clhs2834 + clhs1658*clhs2931 - clhs1667*clhs2325 - clhs1707*clhs2933 + clhs1780*clhs2799 + clhs1943*clhs2620 - clhs1969*clhs2800 - clhs2076*clhs2929 - clhs2077*clhs2624 + clhs2345*clhs781 + clhs2801*clhs905 + clhs2932*clhs809 + clhs2980;
            lhs(8,16)=-clhs1393*clhs822 - clhs1744*clhs2026 - clhs1746*clhs2934 + clhs1747*clhs2931 + clhs1752*clhs2933 - clhs2832 + (2.0L/3.0L)*clhs2833 + clhs2982 - clhs421*clhs832;
            lhs(8,17)=clhs1380*clhs861 - clhs1756*clhs2026 + clhs1758*clhs2931 + clhs1762*clhs2933 - clhs2927 + (2.0L/3.0L)*clhs2928 - clhs2935*clhs882 + clhs2983 - clhs421*clhs871;
            lhs(8,18)=-clhs1769*clhs2946 + clhs1770*clhs2931 + clhs1773*clhs2933 - clhs2084*clhs421 + clhs2823 - 4.0L/3.0L*clhs2824 + clhs2920 - clhs2924*clhs57 + clhs2925 + clhs2984 + clhs2985 + clhs2986;
            lhs(8,19)=clhs67*(-clhs1774*clhs473 + clhs1775*clhs2937 - clhs1776*clhs434 - clhs1777*clhs2938 + clhs2616);
            lhs(9,0)=-clhs2112*clhs2987 - clhs2115*clhs2988 - clhs2118*clhs2989 - clhs2149*clhs2990 - clhs2164*clhs2991 - clhs2173*clhs2992 - clhs2175*clhs2624 + clhs2184*clhs2994 + clhs2275 + clhs2993*clhs930;
            lhs(9,1)=clhs1179*clhs2993 + clhs1190*clhs2994 + clhs1811*clhs2988 + clhs1989*clhs2989 - clhs2191*clhs2995 - clhs2196*clhs2996 - clhs2201*clhs2997 - clhs2209*clhs2998 - clhs2210*clhs2987 + clhs2305;
            lhs(9,2)=-clhs1052*clhs2987 + clhs1197*clhs2993 + clhs1204*clhs2994 + clhs1988*clhs2989 - clhs2213*clhs2997 - clhs2216*clhs2996 - clhs2219*clhs2995 - clhs2223*clhs2998 - clhs2224*clhs2988 + clhs2329;
            lhs(9,3)=-clhs1053*clhs2987 + clhs1218*clhs2993 + clhs1225*clhs2994 - clhs1810*clhs2988 - clhs2225*clhs2997 - clhs2226*clhs2995 - clhs2227*clhs2996 - clhs2229*clhs2998 - clhs2230*clhs2989 + clhs2349;
            lhs(9,4)=-clhs1150*clhs3000 - clhs1158*clhs2999 - clhs1163*clhs3001 - clhs1227*clhs3003 - clhs2182*clhs3002 + clhs2237*clhs2987 + clhs2238*clhs2988 + clhs2239*clhs2989 + clhs2365;
            lhs(9,5)=-1.0L/2.0L*N[1]*clhs2127*clhs3009*clhs7*gamma + clhs1068*clhs2597 - clhs1215*clhs3004*clhs3005 - clhs1217*clhs3004*clhs3007 + clhs1292*clhs2993 - clhs1826*clhs3004*clhs3006 - clhs2276*clhs2987 - clhs2277*clhs2988 - clhs2278*clhs2989 - clhs2288*clhs2990 - clhs2293*clhs2991 - clhs2295*clhs2992 - clhs2297*clhs2624 + clhs2299*clhs2994 + clhs2651*(-clhs1140*clhs2015 - clhs1140*clhs351 - clhs1140*clhs387 - clhs1286*clhs2179 - clhs1372*clhs2179 - clhs1374*clhs2179 - clhs2098*clhs453 - clhs2104*clhs447 - clhs2110*clhs437 + clhs2176*clhs419 + clhs2176*clhs461 + clhs2177*clhs430 + clhs2177*clhs487 + clhs2178*clhs469 + clhs2178*clhs490 + clhs2296);
            lhs(9,6)=-clhs1215*clhs3012*(-6*clhs1286 + clhs1289 + clhs3008 + clhs3013 + clhs387 + clhs447 + 6*clhs450) - clhs1217*clhs3011 + clhs1383*clhs2993 + clhs1388*clhs2994 - clhs1826*clhs3010 - clhs2311*clhs2995 - clhs2316*clhs2996 - clhs2319*clhs2997 - clhs2323*clhs2998 + clhs2324*clhs2988 + clhs2325*clhs2989 - clhs2326*clhs2987 + clhs2647 + clhs2667*clhs3005 + clhs2672*clhs3014;
            lhs(9,7)=-clhs1215*clhs3010 - clhs1217*clhs3015 + clhs1395*clhs2993 + clhs1399*clhs2994 - clhs1826*clhs3012*(-6*clhs1374 + clhs2855 + clhs3016 + 6*clhs444) - clhs2333*clhs2997 - clhs2337*clhs2996 - clhs2340*clhs2995 - clhs2344*clhs2998 + clhs2345*clhs2989 - clhs2346*clhs2987 - clhs2347*clhs2988 + clhs2667*clhs3006 + clhs2668*clhs3014 + clhs2850;
            lhs(9,8)=-clhs1215*clhs3011 - clhs1217*clhs3012*(-6*clhs1372 + clhs2944 + clhs3017 + 6*clhs435) + clhs1408*clhs2993 + clhs1411*clhs2994 - clhs1826*clhs3015 - clhs2352*clhs2997 - clhs2355*clhs2995 - clhs2356*clhs2996 - clhs2360*clhs2998 - clhs2361*clhs2987 - clhs2362*clhs2988 - clhs2363*clhs2989 + clhs2667*clhs3007 + clhs2670*clhs3014 + clhs2939;
            lhs(9,9)=-clhs1017*clhs2670*clhs3018 - clhs1050*clhs2672*clhs3018 - clhs1359*clhs2999 - clhs1363*clhs3000 - clhs1368*clhs3001 - clhs1415*clhs3003 - clhs2366*clhs3002 + clhs2367*clhs2987 + clhs2368*clhs2988 + clhs2369*clhs2989 + clhs2598 + clhs2667*clhs3009*gamma - clhs2668*clhs3018*clhs985;
            lhs(9,10)=clhs1477*clhs2993 - clhs2391*clhs2987 - clhs2392*clhs2988 - clhs2393*clhs2989 - clhs2403*clhs2990 - clhs2408*clhs2991 - clhs2410*clhs2992 - clhs2412*clhs2624 + clhs2414*clhs2994 + clhs3029;
            lhs(9,11)=clhs1566*clhs2993 + clhs1571*clhs2994 - clhs2423*clhs2995 - clhs2428*clhs2996 - clhs2431*clhs2997 - clhs2435*clhs2998 + clhs2436*clhs2988 + clhs2437*clhs2989 - clhs2438*clhs2987 + clhs3032;
            lhs(9,12)=clhs1577*clhs2993 + clhs1581*clhs2994 - clhs2444*clhs2997 - clhs2448*clhs2996 - clhs2451*clhs2995 - clhs2455*clhs2998 + clhs2456*clhs2989 - clhs2457*clhs2987 - clhs2458*clhs2988 + clhs3034;
            lhs(9,13)=clhs1589*clhs2993 + clhs1592*clhs2994 - clhs2462*clhs2997 - clhs2465*clhs2995 - clhs2466*clhs2996 - clhs2470*clhs2998 - clhs2471*clhs2987 - clhs2472*clhs2988 - clhs2473*clhs2989 + clhs3035;
            lhs(9,14)=-clhs1542*clhs2999 - clhs1546*clhs3000 - clhs1551*clhs3001 - clhs1596*clhs3003 - clhs2475*clhs3002 + clhs2476*clhs2987 + clhs2477*clhs2988 + clhs2478*clhs2989 + clhs3036;
            lhs(9,15)=clhs1658*clhs2993 - clhs2500*clhs2987 - clhs2501*clhs2988 - clhs2502*clhs2989 - clhs2512*clhs2990 - clhs2517*clhs2991 - clhs2519*clhs2992 - clhs2521*clhs2624 + clhs2523*clhs2994 + clhs3047;
            lhs(9,16)=clhs1747*clhs2993 + clhs1752*clhs2994 - clhs2532*clhs2995 - clhs2537*clhs2996 - clhs2540*clhs2997 - clhs2544*clhs2998 + clhs2545*clhs2988 + clhs2546*clhs2989 - clhs2547*clhs2987 + clhs3050;
            lhs(9,17)=clhs1758*clhs2993 + clhs1762*clhs2994 - clhs2553*clhs2997 - clhs2557*clhs2996 - clhs2560*clhs2995 - clhs2564*clhs2998 + clhs2565*clhs2989 - clhs2566*clhs2987 - clhs2567*clhs2988 + clhs3052;
            lhs(9,18)=clhs1770*clhs2993 + clhs1773*clhs2994 - clhs2571*clhs2997 - clhs2574*clhs2995 - clhs2575*clhs2996 - clhs2579*clhs2998 - clhs2580*clhs2987 - clhs2581*clhs2988 - clhs2582*clhs2989 + clhs3053;
            lhs(9,19)=-clhs1723*clhs2999 - clhs1727*clhs3000 - clhs1732*clhs3001 - clhs1777*clhs3003 - clhs2584*clhs3002 + clhs2585*clhs2987 + clhs2586*clhs2988 + clhs2587*clhs2989 + clhs3054;
            lhs(10,0)=clhs111*clhs3055 + clhs164*clhs3056 + clhs201*clhs3057 + clhs501;
            lhs(10,1)=-clhs1593*clhs215 - clhs1594*clhs226 + clhs254*clhs3059 + clhs3058;
            lhs(10,2)=-clhs1594*clhs269 + clhs1595*clhs259 + clhs1824*clhs3061 + clhs3060;
            lhs(10,3)=clhs1593*clhs290 + clhs1595*clhs287 - clhs299*clhs3063 + clhs3062;
            lhs(10,4)=clhs699;
            lhs(10,5)=clhs2602 + clhs3055*clhs348 + clhs3056*clhs383 + clhs3057*clhs411;
            lhs(10,6)=-clhs1593*clhs424 - clhs1594*clhs434 + clhs3059*clhs458 + clhs3064;
            lhs(10,7)=-clhs1594*clhs473 + clhs1595*clhs463 + clhs1871*clhs3061 + clhs3065;
            lhs(10,8)=clhs1593*clhs492 + clhs1595*clhs489 - clhs3063*clhs499 + clhs3066;
            lhs(10,9)=clhs2609;
            lhs(10,10)=clhs3055*clhs548 + clhs3056*clhs582 + clhs3057*clhs610 + clhs3068;
            lhs(10,11)=-clhs1593*clhs623 - clhs1594*clhs633 + clhs3059*clhs657 + clhs3069;
            lhs(10,12)=-clhs1594*clhs672 + clhs1595*clhs662 - clhs3061*clhs683 + clhs3070;
            lhs(10,13)=clhs1593*clhs691 + clhs1595*clhs688 - clhs3063*clhs698 + clhs3071;
            lhs(10,14)=-clhs300*(pow(DN(2,0), 2) + pow(DN(2,1), 2) + pow(DN(2,2), 2));
            lhs(10,15)=clhs3055*clhs747 + clhs3056*clhs781 + clhs3057*clhs809 + clhs3072;
            lhs(10,16)=-clhs1593*clhs822 - clhs1594*clhs832 + clhs3059*clhs856 + clhs3074;
            lhs(10,17)=-clhs1594*clhs871 + clhs1595*clhs861 - clhs3061*clhs882 + clhs3076;
            lhs(10,18)=clhs1593*clhs890 + clhs1595*clhs887 - clhs3063*clhs897 + clhs3078;
            lhs(10,19)=clhs3079;
            lhs(11,0)=-clhs1066*clhs3088 - clhs1147*clhs3089 - clhs1293*(-clhs102*clhs1422 + clhs1482 + clhs611*clhs939) + clhs1460 - clhs164*clhs2457 - clhs201*clhs2471 + clhs3080*clhs902 + clhs3081*clhs905 + clhs3082*clhs910 + clhs3083*clhs913 + clhs3084*clhs915 - clhs3085*clhs926 - clhs3086*clhs929 - clhs3087*clhs930;
            lhs(11,1)=-clhs1061*clhs620 - clhs1171*clhs3095 - clhs1179*clhs3087 + clhs1190*clhs3089 + clhs1540 + clhs1913 + clhs2056 - 4.0L/3.0L*clhs3090 + clhs3092 + clhs3094;
            lhs(11,2)=clhs1193*clhs688 + clhs1194*clhs1560 - clhs1197*clhs3087 + clhs1204*clhs3089 + clhs1572 + clhs259*clhs620 + clhs283*clhs3098 + (2.0L/3.0L)*clhs3096 - clhs3097;
            lhs(11,3)=-clhs1207*clhs3105 - clhs1214*clhs3106 + clhs1218*clhs3107 + clhs1225*clhs3089 + clhs1582 - clhs1830*clhs662 + clhs287*clhs620 + (2.0L/3.0L)*clhs3099 - clhs3100;
            lhs(11,4)=clhs67*(-clhs1227*clhs3108 - clhs1560*clhs256 + clhs204*clhs662 + clhs216*clhs688 + clhs3058);
            lhs(11,5)=clhs1279*clhs3082 + clhs1280*clhs3083 - clhs1290*clhs3085 - clhs1291*clhs3086 - clhs1292*clhs3087 - clhs1293*(-clhs102*clhs2686 + clhs2603*clhs939 + clhs2729) - clhs1302*clhs3088 - clhs1342*clhs3089 - clhs2457*clhs383 - clhs2471*clhs411 + clhs2721 + clhs3109*clhs902 + clhs3110*clhs905 + clhs3111*clhs915;
            lhs(11,6)=-clhs1377*clhs620 - clhs1382*clhs3095 - clhs1383*clhs3087 + clhs1388*clhs3089 + clhs2743 + clhs2888 + clhs2964 - 4.0L/3.0L*clhs3112 + clhs3114 + clhs3116;
            lhs(11,7)=clhs1393*clhs688 + clhs1394*clhs1560 - clhs1395*clhs3087 + clhs1399*clhs3089 + clhs2751 + clhs3098*clhs484 + (2.0L/3.0L)*clhs3117 - clhs3118 + clhs463*clhs620;
            lhs(11,8)=-clhs1404*clhs3105 - clhs1407*clhs3106 + clhs1408*clhs3107 + clhs1411*clhs3089 - clhs1877*clhs662 + clhs2754 + (2.0L/3.0L)*clhs3119 - clhs3120 + clhs489*clhs620;
            lhs(11,9)=clhs67*(clhs1412*clhs662 + clhs1413*clhs688 - clhs1414*clhs1560 - clhs1415*clhs3108 + clhs3064);
            lhs(11,10)=clhs1464*clhs3082 + clhs1465*clhs3083 - clhs1475*clhs3085 - clhs1476*clhs3086 - clhs1477*clhs3087 - clhs1486*clhs3088 - clhs1526*clhs3089 - clhs2457*clhs582 - clhs2471*clhs610 + clhs3121 + clhs3122*clhs902 + clhs3123*clhs905 + clhs3124*clhs915 - clhs3125*(DN(2,0)*clhs939 + clhs1444*clhs943 + clhs513 + clhs516 + clhs519 + clhs522 - clhs525 - clhs529 - clhs533 - clhs542 + clhs615*clhs84 - clhs622*clhs936 + clhs626*clhs90 - clhs632*clhs937) - clhs3127*clhs3128 - clhs3130*clhs3131 - clhs3132*clhs3133;
            lhs(11,11)=-clhs1017*clhs3145 + clhs1178*clhs3147 - clhs1560*clhs620 - clhs1565*clhs3095 - clhs1566*clhs3087 + clhs1571*clhs3089 + clhs3068 - clhs3132*clhs3141 + clhs3135 + clhs3136 - 4.0L/3.0L*clhs3137 + clhs3139 + clhs3140 - clhs3143*clhs985;
            lhs(11,12)=-clhs1050*clhs3143 + clhs1560*clhs1576 + clhs1575*clhs688 - clhs1577*clhs3087 + clhs1581*clhs3089 + clhs3098*clhs683 - clhs3127*clhs3141 + clhs3146*clhs3150 + (2.0L/3.0L)*clhs3148 - clhs3149 + clhs620*clhs662;
            lhs(11,13)=clhs1215*clhs3145 - clhs1217*clhs3146*clhs3153 - clhs1576*clhs691 - clhs1585*clhs3105 - clhs1588*clhs3106 + clhs1589*clhs3107 + clhs1592*clhs3089 - clhs3130*clhs3141 + (2.0L/3.0L)*clhs3151 - clhs3152 + clhs620*clhs688;
            lhs(11,14)=clhs67*(-clhs1560*clhs1595 + clhs1593*clhs662 + clhs1594*clhs688 - clhs1596*clhs3108 + clhs3069);
            lhs(11,15)=-clhs1293*(-clhs102*clhs3073 + clhs3160*clhs939 + clhs3203) + clhs1645*clhs3082 + clhs1646*clhs3083 - clhs1656*clhs3085 - clhs1657*clhs3086 - clhs1658*clhs3087 - clhs1667*clhs3088 - clhs1707*clhs3089 - clhs2457*clhs781 - clhs2471*clhs809 + clhs3195 + clhs3196*clhs902 + clhs3197*clhs905 + clhs3198*clhs915;
            lhs(11,16)=-clhs1741*clhs620 - clhs1746*clhs3095 - clhs1747*clhs3087 + clhs1752*clhs3089 + clhs3217 - 4.0L/3.0L*clhs3218 + clhs3220 + clhs3222 + clhs3223 + clhs3224;
            lhs(11,17)=clhs1560*clhs1757 + clhs1756*clhs688 - clhs1758*clhs3087 + clhs1762*clhs3089 + clhs3098*clhs882 + clhs3225 + (2.0L/3.0L)*clhs3226 - clhs3227 + clhs620*clhs861;
            lhs(11,18)=-clhs1576*clhs890 - clhs1766*clhs3105 - clhs1769*clhs3106 + clhs1770*clhs3107 + clhs1773*clhs3089 + clhs3228 + (2.0L/3.0L)*clhs3229 - clhs3230 + clhs620*clhs887;
            lhs(11,19)=clhs67*(-clhs1560*clhs1776 + clhs1774*clhs662 + clhs1775*clhs688 - clhs1777*clhs3108 + clhs3074);
            lhs(12,0)=-clhs1066*clhs2436 - clhs1147*clhs3236 - clhs1293*(-clhs1418*clhs158 + clhs1798*clhs658 + clhs1905) + clhs164*clhs3235 + clhs1780*clhs3081 + clhs1783*clhs3083 - clhs1789*clhs3080 - clhs1792*clhs3232 - clhs1793*clhs3086 + clhs1897 - clhs201*clhs2472 + clhs3084*clhs902 + clhs3231*clhs910 + clhs3234*clhs930;
            lhs(12,1)=clhs1154*clhs691 - clhs1167*clhs1912 - clhs1171*clhs3237 + clhs1179*clhs3234 + clhs1190*clhs3236 + clhs1907 - clhs215*clhs620 - clhs3096 + (2.0L/3.0L)*clhs3097;
            lhs(12,2)=clhs1197*clhs3234 + clhs1204*clhs3236 + clhs1530*clhs57 + clhs1536 + clhs1562 - clhs1814*clhs620 - clhs1912*clhs2842 + clhs1915 + clhs2057 - 4.0L/3.0L*clhs3091 + clhs3094 + clhs3238;
            lhs(12,3)=clhs1218*clhs3234 + clhs1225*clhs3236 + clhs1561*clhs287 + clhs1830*clhs1912 + clhs1919 + clhs290*clhs620 + clhs299*clhs3241 + (2.0L/3.0L)*clhs3239 - clhs3240;
            lhs(12,4)=clhs67*(-clhs1227*clhs3245 + clhs204*clhs3244 + clhs216*clhs691 - clhs256*clhs623 + clhs3060);
            lhs(12,5)=clhs1279*clhs3231 + clhs1292*clhs3234 - clhs1293*(-clhs158*clhs2682 + clhs1798*clhs2605 + clhs2885) - clhs1302*clhs2436 - clhs1342*clhs3236 + clhs1780*clhs3110 - clhs1789*clhs3109 + clhs1853*clhs3083 - clhs1857*clhs3232 - clhs1858*clhs3086 - clhs2472*clhs411 + clhs2883 + clhs3111*clhs902 + clhs3235*clhs383;
            lhs(12,6)=-clhs1378*clhs1912 + clhs1380*clhs691 - clhs1382*clhs3237 + clhs1383*clhs3234 + clhs1388*clhs3236 + clhs2886 - clhs3117 + (2.0L/3.0L)*clhs3118 - clhs424*clhs620;
            lhs(12,7)=clhs1395*clhs3234 + clhs1399*clhs3236 - clhs1867*clhs620 - clhs1912*clhs3247 + clhs2733*clhs57 + clhs2739 + clhs2749 + clhs2890 + clhs2965 - 4.0L/3.0L*clhs3113 + clhs3116 + clhs3246;
            lhs(12,8)=clhs1408*clhs3234 + clhs1411*clhs3236 + clhs1561*clhs489 + clhs1877*clhs1912 + clhs2894 + clhs3241*clhs499 + (2.0L/3.0L)*clhs3248 - clhs3249 + clhs492*clhs620;
            lhs(12,9)=clhs67*(clhs1412*clhs3244 + clhs1413*clhs691 - clhs1414*clhs623 - clhs1415*clhs3245 + clhs3065);
            lhs(12,10)=clhs1464*clhs3231 + clhs1477*clhs3234 - clhs1486*clhs2436 - clhs1526*clhs3236 + clhs1780*clhs3123 - clhs1789*clhs3122 + clhs1898*clhs3083 - clhs1902*clhs3232 - clhs1903*clhs3086 - clhs2472*clhs610 + clhs3124*clhs902 - clhs3125*(DN(2,1)*clhs1798 - clhs1184*clhs621 + clhs1452*clhs1799 + clhs557 + clhs560 + clhs562 + clhs564 - clhs566 - clhs569 - clhs572 - clhs578 + clhs649*clhs84 + clhs665*clhs90 - clhs671*clhs937) + clhs3128*clhs3256 - clhs3131*clhs3254 + clhs3133*clhs3252 + clhs3235*clhs582 + clhs3250;
            lhs(12,11)=clhs1050*clhs3257 - clhs1561*clhs1912 + clhs1563*clhs691 - clhs1565*clhs3237 + clhs1566*clhs3234 + clhs1571*clhs3236 + clhs3141*clhs3252 - clhs3147*clhs985 - clhs3148 + (2.0L/3.0L)*clhs3149 - clhs620*clhs623;
            lhs(12,12)=clhs1577*clhs3234 + clhs1581*clhs3236 - clhs1912*clhs620 - clhs1916*clhs3259 + clhs3068 - 4.0L/3.0L*clhs3134 + clhs3139 + clhs3140 + clhs3141*clhs3256 + clhs3258 + clhs3260 - clhs3261*clhs57;
            lhs(12,13)=clhs1017*clhs3257 + clhs1561*clhs688 + clhs1589*clhs3234 + clhs1592*clhs3236 + clhs1912*clhs1922 - clhs3141*clhs3254 - clhs3145*clhs985 + clhs3241*clhs698 + (2.0L/3.0L)*clhs3262 - clhs3263 + clhs620*clhs691;
            lhs(12,14)=clhs67*(clhs1593*clhs3244 + clhs1594*clhs691 - clhs1595*clhs623 - clhs1596*clhs3245 + clhs3070);
            lhs(12,15)=-clhs1293*(-clhs158*clhs3075 + clhs1798*clhs3156 + clhs3285) + clhs1645*clhs3231 + clhs1658*clhs3234 - clhs1667*clhs2436 - clhs1707*clhs3236 + clhs1780*clhs3197 - clhs1789*clhs3196 + clhs1943*clhs3083 - clhs1947*clhs3232 - clhs1948*clhs3086 - clhs2472*clhs809 + clhs3198*clhs902 + clhs3235*clhs781 + clhs3283;
            lhs(12,16)=-clhs1742*clhs1912 + clhs1744*clhs691 - clhs1746*clhs3237 + clhs1747*clhs3234 + clhs1752*clhs3236 - clhs3226 + (2.0L/3.0L)*clhs3227 + clhs3286 - clhs620*clhs822;
            lhs(12,17)=clhs1758*clhs3234 + clhs1762*clhs3236 - clhs1957*clhs620 - clhs1961*clhs3259 - 4.0L/3.0L*clhs3219 + clhs3222 + clhs3287 + clhs3288 + clhs3289 + clhs3290 + clhs3291 - clhs3292*clhs57;
            lhs(12,18)=clhs1561*clhs887 + clhs1770*clhs3234 + clhs1773*clhs3236 + clhs1912*clhs1967 + clhs3241*clhs897 + clhs3293 + (2.0L/3.0L)*clhs3294 - clhs3295 + clhs620*clhs890;
            lhs(12,19)=clhs67*(clhs1774*clhs3244 + clhs1775*clhs691 - clhs1776*clhs623 - clhs1777*clhs3245 + clhs3076);
            lhs(13,0)=-clhs1066*clhs2437 - clhs1147*clhs3300 - clhs1293*(-clhs1430*clhs196 + clhs1977*clhs684 + clhs2049) + clhs164*clhs2456 + clhs1780*clhs3080 + clhs1783*clhs3082 - clhs1969*clhs3081 - clhs1972*clhs3296 - clhs1973*clhs3086 + clhs201*clhs3299 + clhs2043 + clhs3084*clhs905 + clhs3231*clhs913 + clhs3298*clhs930;
            lhs(13,1)=-clhs1154*clhs2055 - clhs1167*clhs672 - clhs1171*clhs3301 + clhs1179*clhs3298 + clhs1190*clhs3300 + clhs2050 - clhs226*clhs620 - clhs3099 + (2.0L/3.0L)*clhs3100;
            lhs(13,2)=-clhs1193*clhs2055 + clhs1197*clhs3298 + clhs1204*clhs3300 + clhs1563*clhs259 + clhs2051 - clhs269*clhs620 - clhs283*clhs3302 - clhs3239 + (2.0L/3.0L)*clhs3240;
            lhs(13,3)=clhs1218*clhs3298 + clhs1225*clhs3300 + clhs1531 + clhs1535*clhs57 + clhs1564 + clhs1914 - clhs1992*clhs620 - clhs2055*clhs2936 + clhs2058 + clhs3092 - 4.0L/3.0L*clhs3093 + clhs3238;
            lhs(13,4)=clhs67*(-clhs1227*clhs3304 - clhs204*clhs672 + clhs216*clhs3303 - clhs256*clhs633 + clhs3062);
            lhs(13,5)=clhs1280*clhs3231 + clhs1292*clhs3298 - clhs1293*(-clhs196*clhs2694 + clhs1977*clhs2607 + clhs2961) - clhs1302*clhs2437 - clhs1342*clhs3300 + clhs1780*clhs3109 + clhs1853*clhs3082 - clhs1969*clhs3110 - clhs2017*clhs3296 - clhs2018*clhs3086 + clhs2456*clhs383 + clhs2960 + clhs3111*clhs905 + clhs3299*clhs411;
            lhs(13,6)=-clhs1378*clhs672 - clhs1380*clhs2055 - clhs1382*clhs3301 + clhs1383*clhs3298 + clhs1388*clhs3300 + clhs2962 - clhs3119 + (2.0L/3.0L)*clhs3120 - clhs434*clhs620;
            lhs(13,7)=-clhs1393*clhs2055 + clhs1395*clhs3298 + clhs1399*clhs3300 + clhs1563*clhs463 + clhs2963 - clhs3248 + (2.0L/3.0L)*clhs3249 - clhs3302*clhs484 - clhs473*clhs620;
            lhs(13,8)=clhs1408*clhs3298 + clhs1411*clhs3300 - clhs2026*clhs620 - clhs2055*clhs3305 + clhs2734 + clhs2738*clhs57 + clhs2750 + clhs2889 + clhs2966 + clhs3114 - 4.0L/3.0L*clhs3115 + clhs3246;
            lhs(13,9)=clhs67*(-clhs1412*clhs672 + clhs1413*clhs3303 - clhs1414*clhs633 - clhs1415*clhs3304 + clhs3066);
            lhs(13,10)=clhs1465*clhs3231 + clhs1477*clhs3298 - clhs1486*clhs2437 - clhs1526*clhs3300 + clhs1780*clhs3122 + clhs1898*clhs3082 - clhs1969*clhs3123 - clhs2047*clhs3296 - clhs2048*clhs3086 + clhs2456*clhs582 + clhs3124*clhs905 - clhs3125*(DN(2,2)*clhs1977 - clhs1184*clhs631 + clhs1455*clhs1978 + clhs589 + clhs591 + clhs593 + clhs595 - clhs597 - clhs599 - clhs602 - clhs607 + clhs643*clhs90 + clhs649*clhs90 - clhs670*clhs936) + clhs3128*clhs3310 + clhs3131*clhs3312 + clhs3133*clhs3308 + clhs3299*clhs610 + clhs3306;
            lhs(13,11)=-clhs1017*clhs3147 + clhs1050*clhs3144*clhs3153 - clhs1561*clhs672 - clhs1563*clhs2055 - clhs1565*clhs3301 + clhs1566*clhs3298 + clhs1571*clhs3300 + clhs3141*clhs3308 - clhs3151 + (2.0L/3.0L)*clhs3152 - clhs620*clhs633;
            lhs(13,12)=-clhs1017*clhs3143 + clhs1563*clhs662 - clhs1575*clhs2055 + clhs1577*clhs3298 + clhs1581*clhs3300 + clhs3141*clhs3310 + clhs3144*clhs3150 - clhs3262 + (2.0L/3.0L)*clhs3263 - clhs3302*clhs683 - clhs620*clhs672;
            lhs(13,13)=-clhs1588*clhs3313 + clhs1589*clhs3298 + clhs1592*clhs3300 - clhs2055*clhs620 + clhs3068 + clhs3135 + clhs3136 - 4.0L/3.0L*clhs3138 + clhs3141*clhs3312 + clhs3258 - clhs3260*clhs57 + clhs3261;
            lhs(13,14)=clhs67*(-clhs1593*clhs672 + clhs1594*clhs3303 - clhs1595*clhs633 - clhs1596*clhs3304 + clhs3071);
            lhs(13,15)=-clhs1293*(-clhs196*clhs3077 + clhs1977*clhs3168 + clhs3328) + clhs1646*clhs3231 + clhs1658*clhs3298 - clhs1667*clhs2437 - clhs1707*clhs3300 + clhs1780*clhs3196 + clhs1943*clhs3082 - clhs1969*clhs3197 - clhs2076*clhs3296 - clhs2077*clhs3086 + clhs2456*clhs781 + clhs3198*clhs905 + clhs3299*clhs809 + clhs3327;
            lhs(13,16)=-clhs1575*clhs822 - clhs1744*clhs2055 - clhs1746*clhs3301 + clhs1747*clhs3298 + clhs1752*clhs3300 - clhs3229 + (2.0L/3.0L)*clhs3230 + clhs3329 - clhs620*clhs832;
            lhs(13,17)=clhs1563*clhs861 - clhs1756*clhs2055 + clhs1758*clhs3298 + clhs1762*clhs3300 - clhs3294 + (2.0L/3.0L)*clhs3295 - clhs3302*clhs882 + clhs3330 - clhs620*clhs871;
            lhs(13,18)=-clhs1769*clhs3313 + clhs1770*clhs3298 + clhs1773*clhs3300 - clhs2084*clhs620 + clhs3220 - 4.0L/3.0L*clhs3221 + clhs3287 - clhs3291*clhs57 + clhs3292 + clhs3331 + clhs3332 + clhs3333;
            lhs(13,19)=clhs67*(-clhs1774*clhs672 + clhs1775*clhs3303 - clhs1776*clhs633 - clhs1777*clhs3304 + clhs3078);
            lhs(14,0)=-clhs2112*clhs3334 - clhs2115*clhs3335 - clhs2118*clhs3336 - clhs2149*clhs3337 - clhs2164*clhs3338 - clhs2173*clhs3339 - clhs2175*clhs3086 + clhs2184*clhs3341 + clhs2390 + clhs3340*clhs930;
            lhs(14,1)=clhs1179*clhs3340 + clhs1190*clhs3341 + clhs1811*clhs3335 + clhs1989*clhs3336 - clhs2191*clhs3342 - clhs2196*clhs3343 - clhs2201*clhs3344 - clhs2209*clhs3345 - clhs2210*clhs3334 + clhs2417;
            lhs(14,2)=-clhs1052*clhs3334 + clhs1197*clhs3340 + clhs1204*clhs3341 + clhs1988*clhs3336 - clhs2213*clhs3344 - clhs2216*clhs3343 - clhs2219*clhs3342 - clhs2223*clhs3345 - clhs2224*clhs3335 + clhs2440;
            lhs(14,3)=-clhs1053*clhs3334 + clhs1218*clhs3340 + clhs1225*clhs3341 - clhs1810*clhs3335 - clhs2225*clhs3344 - clhs2226*clhs3342 - clhs2227*clhs3343 - clhs2229*clhs3345 - clhs2230*clhs3336 + clhs2459;
            lhs(14,4)=-clhs1150*clhs3347 - clhs1158*clhs3346 - clhs1163*clhs3348 - clhs1227*clhs3350 - clhs2182*clhs3349 + clhs2237*clhs3334 + clhs2238*clhs3335 + clhs2239*clhs3336 + clhs2474;
            lhs(14,5)=clhs1292*clhs3340 - clhs2276*clhs3334 - clhs2277*clhs3335 - clhs2278*clhs3336 - clhs2288*clhs3337 - clhs2293*clhs3338 - clhs2295*clhs3339 - clhs2297*clhs3086 + clhs2299*clhs3341 + clhs3029;
            lhs(14,6)=clhs1383*clhs3340 + clhs1388*clhs3341 - clhs2311*clhs3342 - clhs2316*clhs3343 - clhs2319*clhs3344 - clhs2323*clhs3345 + clhs2324*clhs3335 + clhs2325*clhs3336 - clhs2326*clhs3334 + clhs3032;
            lhs(14,7)=clhs1395*clhs3340 + clhs1399*clhs3341 - clhs2333*clhs3344 - clhs2337*clhs3343 - clhs2340*clhs3342 - clhs2344*clhs3345 + clhs2345*clhs3336 - clhs2346*clhs3334 - clhs2347*clhs3335 + clhs3034;
            lhs(14,8)=clhs1408*clhs3340 + clhs1411*clhs3341 - clhs2352*clhs3344 - clhs2355*clhs3342 - clhs2356*clhs3343 - clhs2360*clhs3345 - clhs2361*clhs3334 - clhs2362*clhs3335 - clhs2363*clhs3336 + clhs3035;
            lhs(14,9)=-clhs1359*clhs3346 - clhs1363*clhs3347 - clhs1368*clhs3348 - clhs1415*clhs3350 - clhs2366*clhs3349 + clhs2367*clhs3334 + clhs2368*clhs3335 + clhs2369*clhs3336 + clhs3036;
            lhs(14,10)=-1.0L/2.0L*N[2]*clhs2127*clhs3356*clhs7*gamma + clhs1068*clhs3067 - clhs1215*clhs3351*clhs3352 - clhs1217*clhs3351*clhs3354 + clhs1477*clhs3340 - clhs1826*clhs3351*clhs3353 - clhs2391*clhs3334 - clhs2392*clhs3335 - clhs2393*clhs3336 - clhs2403*clhs3337 - clhs2408*clhs3338 - clhs2410*clhs3339 - clhs2412*clhs3086 + clhs2414*clhs3341 + clhs3125*(-clhs1140*clhs2045 - clhs1140*clhs551 - clhs1140*clhs586 - clhs1471*clhs2179 - clhs1555*clhs2179 - clhs1557*clhs2179 - clhs2098*clhs652 - clhs2104*clhs646 - clhs2110*clhs636 + clhs2176*clhs618 + clhs2176*clhs660 + clhs2177*clhs629 + clhs2177*clhs686 + clhs2178*clhs668 + clhs2178*clhs689 + clhs2411);
            lhs(14,11)=-clhs1215*clhs3359*(-6*clhs1471 + clhs1474 + clhs3355 + clhs3360 + clhs586 + clhs646 + 6*clhs649) - clhs1217*clhs3358 + clhs1566*clhs3340 + clhs1571*clhs3341 - clhs1826*clhs3357 - clhs2423*clhs3342 - clhs2428*clhs3343 - clhs2431*clhs3344 - clhs2435*clhs3345 + clhs2436*clhs3335 + clhs2437*clhs3336 - clhs2438*clhs3334 + clhs3121 + clhs3141*clhs3352 + clhs3146*clhs3361;
            lhs(14,12)=-clhs1215*clhs3357 - clhs1217*clhs3362 + clhs1577*clhs3340 + clhs1581*clhs3341 - clhs1826*clhs3359*(-6*clhs1557 + clhs3255 + clhs3363 + 6*clhs643) - clhs2444*clhs3344 - clhs2448*clhs3343 - clhs2451*clhs3342 - clhs2455*clhs3345 + clhs2456*clhs3336 - clhs2457*clhs3334 - clhs2458*clhs3335 + clhs3141*clhs3353 + clhs3142*clhs3361 + clhs3250;
            lhs(14,13)=-clhs1215*clhs3358 - clhs1217*clhs3359*(-6*clhs1555 + clhs3311 + clhs3364 + 6*clhs634) + clhs1589*clhs3340 + clhs1592*clhs3341 - clhs1826*clhs3362 - clhs2462*clhs3344 - clhs2465*clhs3342 - clhs2466*clhs3343 - clhs2470*clhs3345 - clhs2471*clhs3334 - clhs2472*clhs3335 - clhs2473*clhs3336 + clhs3141*clhs3354 + clhs3144*clhs3361 + clhs3306;
            lhs(14,14)=-clhs1017*clhs3144*clhs3365 - clhs1050*clhs3146*clhs3365 - clhs1542*clhs3346 - clhs1546*clhs3347 - clhs1551*clhs3348 - clhs1596*clhs3350 - clhs2475*clhs3349 + clhs2476*clhs3334 + clhs2477*clhs3335 + clhs2478*clhs3336 + clhs3068 + clhs3141*clhs3356*gamma - clhs3142*clhs3365*clhs985;
            lhs(14,15)=clhs1658*clhs3340 - clhs2500*clhs3334 - clhs2501*clhs3335 - clhs2502*clhs3336 - clhs2512*clhs3337 - clhs2517*clhs3338 - clhs2519*clhs3339 - clhs2521*clhs3086 + clhs2523*clhs3341 + clhs3376;
            lhs(14,16)=clhs1747*clhs3340 + clhs1752*clhs3341 - clhs2532*clhs3342 - clhs2537*clhs3343 - clhs2540*clhs3344 - clhs2544*clhs3345 + clhs2545*clhs3335 + clhs2546*clhs3336 - clhs2547*clhs3334 + clhs3379;
            lhs(14,17)=clhs1758*clhs3340 + clhs1762*clhs3341 - clhs2553*clhs3344 - clhs2557*clhs3343 - clhs2560*clhs3342 - clhs2564*clhs3345 + clhs2565*clhs3336 - clhs2566*clhs3334 - clhs2567*clhs3335 + clhs3381;
            lhs(14,18)=clhs1770*clhs3340 + clhs1773*clhs3341 - clhs2571*clhs3344 - clhs2574*clhs3342 - clhs2575*clhs3343 - clhs2579*clhs3345 - clhs2580*clhs3334 - clhs2581*clhs3335 - clhs2582*clhs3336 + clhs3382;
            lhs(14,19)=-clhs1723*clhs3346 - clhs1727*clhs3347 - clhs1732*clhs3348 - clhs1777*clhs3350 - clhs2584*clhs3349 + clhs2585*clhs3334 + clhs2586*clhs3335 + clhs2587*clhs3336 + clhs3383;
            lhs(15,0)=clhs111*clhs3384 + clhs164*clhs3385 + clhs201*clhs3386 + clhs700;
            lhs(15,1)=-clhs1774*clhs215 - clhs1775*clhs226 + clhs254*clhs3388 + clhs3387;
            lhs(15,2)=-clhs1775*clhs269 + clhs1776*clhs259 + clhs1824*clhs3390 + clhs3389;
            lhs(15,3)=clhs1774*clhs290 + clhs1776*clhs287 - clhs299*clhs3392 + clhs3391;
            lhs(15,4)=clhs898;
            lhs(15,5)=clhs2610 + clhs3384*clhs348 + clhs3385*clhs383 + clhs3386*clhs411;
            lhs(15,6)=-clhs1774*clhs424 - clhs1775*clhs434 + clhs3388*clhs458 + clhs3393;
            lhs(15,7)=-clhs1775*clhs473 + clhs1776*clhs463 + clhs1871*clhs3390 + clhs3394;
            lhs(15,8)=clhs1774*clhs492 + clhs1776*clhs489 - clhs3392*clhs499 + clhs3395;
            lhs(15,9)=clhs2617;
            lhs(15,10)=clhs3072 + clhs3384*clhs548 + clhs3385*clhs582 + clhs3386*clhs610;
            lhs(15,11)=-clhs1774*clhs623 - clhs1775*clhs633 + clhs3388*clhs657 + clhs3396;
            lhs(15,12)=-clhs1775*clhs672 + clhs1776*clhs662 + clhs1916*clhs3390 + clhs3397;
            lhs(15,13)=clhs1774*clhs691 + clhs1776*clhs688 - clhs3392*clhs698 + clhs3398;
            lhs(15,14)=clhs3079;
            lhs(15,15)=clhs3384*clhs747 + clhs3385*clhs781 + clhs3386*clhs809 + clhs3400;
            lhs(15,16)=-clhs1774*clhs822 - clhs1775*clhs832 + clhs3388*clhs856 + clhs3401;
            lhs(15,17)=-clhs1775*clhs871 + clhs1776*clhs861 - clhs3390*clhs882 + clhs3402;
            lhs(15,18)=clhs1774*clhs890 + clhs1776*clhs887 - clhs3392*clhs897 + clhs3403;
            lhs(15,19)=-clhs300*(pow(DN(3,0), 2) + pow(DN(3,1), 2) + pow(DN(3,2), 2));
            lhs(16,0)=-clhs1066*clhs3412 - clhs1147*clhs3413 - clhs1293*(-clhs102*clhs1603 + clhs1663 + clhs810*clhs939) - clhs164*clhs2566 + clhs1641 - clhs201*clhs2580 + clhs3404*clhs902 + clhs3405*clhs905 + clhs3406*clhs910 + clhs3407*clhs913 + clhs3408*clhs915 - clhs3409*clhs926 - clhs3410*clhs929 - clhs3411*clhs930;
            lhs(16,1)=-clhs1061*clhs819 - clhs1171*clhs3419 - clhs1179*clhs3411 + clhs1190*clhs3413 + clhs1721 + clhs1958 + clhs2085 - 4.0L/3.0L*clhs3414 + clhs3416 + clhs3418;
            lhs(16,2)=clhs1193*clhs887 + clhs1194*clhs1741 - clhs1197*clhs3411 + clhs1204*clhs3413 + clhs1753 + clhs259*clhs819 + clhs283*clhs3422 + (2.0L/3.0L)*clhs3420 - clhs3421;
            lhs(16,3)=-clhs1207*clhs3429 - clhs1214*clhs3430 + clhs1218*clhs3431 + clhs1225*clhs3413 + clhs1763 - clhs1830*clhs861 + clhs287*clhs819 + (2.0L/3.0L)*clhs3423 - clhs3424;
            lhs(16,4)=clhs67*(-clhs1227*clhs3432 - clhs1741*clhs256 + clhs204*clhs861 + clhs216*clhs887 + clhs3387);
            lhs(16,5)=clhs1279*clhs3406 + clhs1280*clhs3407 - clhs1290*clhs3409 - clhs1291*clhs3410 - clhs1292*clhs3411 - clhs1293*(-clhs102*clhs2763 + clhs2611*clhs939 + clhs2806) - clhs1302*clhs3412 - clhs1342*clhs3413 - clhs2566*clhs383 - clhs2580*clhs411 + clhs2798 + clhs3433*clhs902 + clhs3434*clhs905 + clhs3435*clhs915;
            lhs(16,6)=-clhs1377*clhs819 - clhs1382*clhs3419 - clhs1383*clhs3411 + clhs1388*clhs3413 + clhs2820 + clhs2921 + clhs2984 - 4.0L/3.0L*clhs3436 + clhs3438 + clhs3440;
            lhs(16,7)=clhs1393*clhs887 + clhs1394*clhs1741 - clhs1395*clhs3411 + clhs1399*clhs3413 + clhs2828 + clhs3422*clhs484 + (2.0L/3.0L)*clhs3441 - clhs3442 + clhs463*clhs819;
            lhs(16,8)=-clhs1404*clhs3429 - clhs1407*clhs3430 + clhs1408*clhs3431 + clhs1411*clhs3413 - clhs1877*clhs861 + clhs2831 + (2.0L/3.0L)*clhs3443 - clhs3444 + clhs489*clhs819;
            lhs(16,9)=clhs67*(clhs1412*clhs861 + clhs1413*clhs887 - clhs1414*clhs1741 - clhs1415*clhs3432 + clhs3393);
            lhs(16,10)=-clhs1293*(-clhs102*clhs3160 + clhs3073*clhs939 + clhs3203) + clhs1464*clhs3406 + clhs1465*clhs3407 - clhs1475*clhs3409 - clhs1476*clhs3410 - clhs1477*clhs3411 - clhs1486*clhs3412 - clhs1526*clhs3413 - clhs2566*clhs582 - clhs2580*clhs610 + clhs3195 + clhs3445*clhs902 + clhs3446*clhs905 + clhs3447*clhs915;
            lhs(16,11)=-clhs1560*clhs819 - clhs1565*clhs3419 - clhs1566*clhs3411 + clhs1571*clhs3413 + clhs3217 + clhs3288 + clhs3331 - 4.0L/3.0L*clhs3448 + clhs3450 + clhs3452;
            lhs(16,12)=clhs1575*clhs887 + clhs1576*clhs1741 - clhs1577*clhs3411 + clhs1581*clhs3413 + clhs3225 + clhs3422*clhs683 + (2.0L/3.0L)*clhs3453 - clhs3454 + clhs662*clhs819;
            lhs(16,13)=-clhs1585*clhs3429 - clhs1588*clhs3430 + clhs1589*clhs3431 + clhs1592*clhs3413 - clhs1757*clhs691 + clhs3228 + (2.0L/3.0L)*clhs3455 - clhs3456 + clhs688*clhs819;
            lhs(16,14)=clhs67*(clhs1593*clhs861 + clhs1594*clhs887 - clhs1595*clhs1741 - clhs1596*clhs3432 + clhs3396);
            lhs(16,15)=clhs1645*clhs3406 + clhs1646*clhs3407 - clhs1656*clhs3409 - clhs1657*clhs3410 - clhs1658*clhs3411 - clhs1667*clhs3412 - clhs1707*clhs3413 - clhs2566*clhs781 - clhs2580*clhs809 + clhs3457 + clhs3458*clhs902 + clhs3459*clhs905 + clhs3460*clhs915 - clhs3461*(DN(3,0)*clhs939 + clhs1625*clhs943 + clhs712 + clhs715 + clhs718 + clhs721 - clhs724 - clhs728 - clhs732 - clhs741 + clhs814*clhs84 - clhs821*clhs936 + clhs825*clhs90 - clhs831*clhs937) - clhs3463*clhs3464 - clhs3466*clhs3467 - clhs3468*clhs3469;
            lhs(16,16)=-clhs1017*clhs3481 + clhs1178*clhs3483 - clhs1741*clhs819 - clhs1746*clhs3419 - clhs1747*clhs3411 + clhs1752*clhs3413 + clhs3400 - clhs3468*clhs3477 + clhs3471 + clhs3472 - 4.0L/3.0L*clhs3473 + clhs3475 + clhs3476 - clhs3479*clhs985;
            lhs(16,17)=-clhs1050*clhs3479 + clhs1741*clhs1757 + clhs1756*clhs887 - clhs1758*clhs3411 + clhs1762*clhs3413 + clhs3422*clhs882 - clhs3463*clhs3477 + clhs3482*clhs3486 + (2.0L/3.0L)*clhs3484 - clhs3485 + clhs819*clhs861;
            lhs(16,18)=clhs1215*clhs3481 - clhs1217*clhs3482*clhs3489 - clhs1757*clhs890 - clhs1766*clhs3429 - clhs1769*clhs3430 + clhs1770*clhs3431 + clhs1773*clhs3413 - clhs3466*clhs3477 + (2.0L/3.0L)*clhs3487 - clhs3488 + clhs819*clhs887;
            lhs(16,19)=clhs67*(-clhs1741*clhs1776 + clhs1774*clhs861 + clhs1775*clhs887 - clhs1777*clhs3432 + clhs3401);
            lhs(17,0)=-clhs1066*clhs2545 - clhs1147*clhs3495 - clhs1293*(-clhs158*clhs1599 + clhs1798*clhs857 + clhs1950) + clhs164*clhs3494 + clhs1780*clhs3405 + clhs1783*clhs3407 - clhs1789*clhs3404 - clhs1792*clhs3491 - clhs1793*clhs3410 + clhs1942 - clhs201*clhs2581 + clhs3408*clhs902 + clhs3490*clhs910 + clhs3493*clhs930;
            lhs(17,1)=clhs1154*clhs890 - clhs1167*clhs1957 - clhs1171*clhs3496 + clhs1179*clhs3493 + clhs1190*clhs3495 + clhs1952 - clhs215*clhs819 - clhs3420 + (2.0L/3.0L)*clhs3421;
            lhs(17,2)=clhs1197*clhs3493 + clhs1204*clhs3495 + clhs1711*clhs57 + clhs1717 + clhs1743 - clhs1814*clhs819 - clhs1957*clhs2842 + clhs1960 + clhs2086 - 4.0L/3.0L*clhs3415 + clhs3418 + clhs3497;
            lhs(17,3)=clhs1218*clhs3493 + clhs1225*clhs3495 + clhs1742*clhs287 + clhs1830*clhs1957 + clhs1964 + clhs290*clhs819 + clhs299*clhs3500 + (2.0L/3.0L)*clhs3498 - clhs3499;
            lhs(17,4)=clhs67*(-clhs1227*clhs3504 + clhs204*clhs3503 + clhs216*clhs890 - clhs256*clhs822 + clhs3389);
            lhs(17,5)=clhs1279*clhs3490 + clhs1292*clhs3493 - clhs1293*(-clhs158*clhs2759 + clhs1798*clhs2613 + clhs2918) - clhs1302*clhs2545 - clhs1342*clhs3495 + clhs1780*clhs3434 - clhs1789*clhs3433 + clhs1853*clhs3407 - clhs1857*clhs3491 - clhs1858*clhs3410 - clhs2581*clhs411 + clhs2916 + clhs3435*clhs902 + clhs3494*clhs383;
            lhs(17,6)=-clhs1378*clhs1957 + clhs1380*clhs890 - clhs1382*clhs3496 + clhs1383*clhs3493 + clhs1388*clhs3495 + clhs2919 - clhs3441 + (2.0L/3.0L)*clhs3442 - clhs424*clhs819;
            lhs(17,7)=clhs1395*clhs3493 + clhs1399*clhs3495 - clhs1867*clhs819 - clhs1957*clhs3247 + clhs2810*clhs57 + clhs2816 + clhs2826 + clhs2923 + clhs2985 - 4.0L/3.0L*clhs3437 + clhs3440 + clhs3505;
            lhs(17,8)=clhs1408*clhs3493 + clhs1411*clhs3495 + clhs1742*clhs489 + clhs1877*clhs1957 + clhs2926 + clhs3500*clhs499 + (2.0L/3.0L)*clhs3506 - clhs3507 + clhs492*clhs819;
            lhs(17,9)=clhs67*(clhs1412*clhs3503 + clhs1413*clhs890 - clhs1414*clhs822 - clhs1415*clhs3504 + clhs3394);
            lhs(17,10)=-clhs1293*(-clhs158*clhs3156 + clhs1798*clhs3075 + clhs3285) + clhs1464*clhs3490 + clhs1477*clhs3493 - clhs1486*clhs2545 - clhs1526*clhs3495 + clhs1780*clhs3446 - clhs1789*clhs3445 + clhs1898*clhs3407 - clhs1902*clhs3491 - clhs1903*clhs3410 - clhs2581*clhs610 + clhs3283 + clhs3447*clhs902 + clhs3494*clhs582;
            lhs(17,11)=-clhs1561*clhs1957 + clhs1563*clhs890 - clhs1565*clhs3496 + clhs1566*clhs3493 + clhs1571*clhs3495 + clhs3286 - clhs3453 + (2.0L/3.0L)*clhs3454 - clhs623*clhs819;
            lhs(17,12)=clhs1577*clhs3493 + clhs1581*clhs3495 - clhs1912*clhs819 - clhs1957*clhs203*clhs683*tau2 + clhs3207*clhs57 + clhs3213 + clhs3223 + clhs3290 + clhs3332 - 4.0L/3.0L*clhs3449 + clhs3452 + clhs3508;
            lhs(17,13)=clhs1589*clhs3493 + clhs1592*clhs3495 + clhs1742*clhs688 + clhs1922*clhs1957 + clhs3293 + clhs3500*clhs698 + (2.0L/3.0L)*clhs3509 - clhs3510 + clhs691*clhs819;
            lhs(17,14)=clhs67*(clhs1593*clhs3503 + clhs1594*clhs890 - clhs1595*clhs822 - clhs1596*clhs3504 + clhs3397);
            lhs(17,15)=clhs1645*clhs3490 + clhs1658*clhs3493 - clhs1667*clhs2545 - clhs1707*clhs3495 + clhs1780*clhs3459 - clhs1789*clhs3458 + clhs1943*clhs3407 - clhs1947*clhs3491 - clhs1948*clhs3410 - clhs2581*clhs809 + clhs3460*clhs902 - clhs3461*(DN(3,1)*clhs1798 - clhs1184*clhs820 + clhs1633*clhs1799 + clhs756 + clhs759 + clhs761 + clhs763 - clhs765 - clhs768 - clhs771 - clhs777 + clhs84*clhs848 + clhs864*clhs90 - clhs870*clhs937) + clhs3464*clhs3517 - clhs3467*clhs3515 + clhs3469*clhs3513 + clhs3494*clhs781 + clhs3511;
            lhs(17,16)=clhs1050*clhs3518 - clhs1742*clhs1957 + clhs1744*clhs890 - clhs1746*clhs3496 + clhs1747*clhs3493 + clhs1752*clhs3495 + clhs3477*clhs3513 - clhs3483*clhs985 - clhs3484 + (2.0L/3.0L)*clhs3485 - clhs819*clhs822;
            lhs(17,17)=clhs1758*clhs3493 + clhs1762*clhs3495 - clhs1957*clhs819 - clhs1961*clhs3503*clhs3520 + clhs3400 - 4.0L/3.0L*clhs3470 + clhs3475 + clhs3476 + clhs3477*clhs3517 + clhs3519 + clhs3521 - clhs3522*clhs57;
            lhs(17,18)=clhs1017*clhs3518 + clhs1742*clhs887 + clhs1770*clhs3493 + clhs1773*clhs3495 + clhs1957*clhs1967 - clhs3477*clhs3515 - clhs3481*clhs985 + clhs3500*clhs897 + (2.0L/3.0L)*clhs3523 - clhs3524 + clhs819*clhs890;
            lhs(17,19)=clhs67*(clhs1774*clhs3503 + clhs1775*clhs890 - clhs1776*clhs822 - clhs1777*clhs3504 + clhs3402);
            lhs(18,0)=-clhs1066*clhs2546 - clhs1147*clhs3529 - clhs1293*(-clhs1611*clhs196 + clhs1977*clhs883 + clhs2078) + clhs164*clhs2565 + clhs1780*clhs3404 + clhs1783*clhs3406 - clhs1969*clhs3405 - clhs1972*clhs3525 - clhs1973*clhs3410 + clhs201*clhs3528 + clhs2072 + clhs3408*clhs905 + clhs3490*clhs913 + clhs3527*clhs930;
            lhs(18,1)=-clhs1154*clhs2084 - clhs1167*clhs871 - clhs1171*clhs3530 + clhs1179*clhs3527 + clhs1190*clhs3529 + clhs2079 - clhs226*clhs819 - clhs3423 + (2.0L/3.0L)*clhs3424;
            lhs(18,2)=-clhs1193*clhs2084 + clhs1197*clhs3527 + clhs1204*clhs3529 + clhs1744*clhs259 + clhs2080 - clhs269*clhs819 - clhs283*clhs3531 - clhs3498 + (2.0L/3.0L)*clhs3499;
            lhs(18,3)=clhs1218*clhs3527 + clhs1225*clhs3529 + clhs1712 + clhs1716*clhs57 + clhs1745 + clhs1959 - clhs1992*clhs819 - clhs2084*clhs2936 + clhs2087 + clhs3416 - 4.0L/3.0L*clhs3417 + clhs3497;
            lhs(18,4)=clhs67*(-clhs1227*clhs3533 - clhs204*clhs871 + clhs216*clhs3532 - clhs256*clhs832 + clhs3391);
            lhs(18,5)=clhs1280*clhs3490 + clhs1292*clhs3527 - clhs1293*(-clhs196*clhs2771 + clhs1977*clhs2615 + clhs2981) - clhs1302*clhs2546 - clhs1342*clhs3529 + clhs1780*clhs3433 + clhs1853*clhs3406 - clhs1969*clhs3434 - clhs2017*clhs3525 - clhs2018*clhs3410 + clhs2565*clhs383 + clhs2980 + clhs3435*clhs905 + clhs3528*clhs411;
            lhs(18,6)=-clhs1378*clhs871 - clhs1380*clhs2084 - clhs1382*clhs3530 + clhs1383*clhs3527 + clhs1388*clhs3529 + clhs2982 - clhs3443 + (2.0L/3.0L)*clhs3444 - clhs434*clhs819;
            lhs(18,7)=-clhs1393*clhs2084 + clhs1395*clhs3527 + clhs1399*clhs3529 + clhs1744*clhs463 + clhs2983 - clhs3506 + (2.0L/3.0L)*clhs3507 - clhs3531*clhs484 - clhs473*clhs819;
            lhs(18,8)=clhs1408*clhs3527 + clhs1411*clhs3529 - clhs2026*clhs819 - clhs2084*clhs3305 + clhs2811 + clhs2815*clhs57 + clhs2827 + clhs2922 + clhs2986 + clhs3438 - 4.0L/3.0L*clhs3439 + clhs3505;
            lhs(18,9)=clhs67*(-clhs1412*clhs871 + clhs1413*clhs3532 - clhs1414*clhs832 - clhs1415*clhs3533 + clhs3395);
            lhs(18,10)=-clhs1293*(-clhs196*clhs3168 + clhs1977*clhs3077 + clhs3328) + clhs1465*clhs3490 + clhs1477*clhs3527 - clhs1486*clhs2546 - clhs1526*clhs3529 + clhs1780*clhs3445 + clhs1898*clhs3406 - clhs1969*clhs3446 - clhs2047*clhs3525 - clhs2048*clhs3410 + clhs2565*clhs582 + clhs3327 + clhs3447*clhs905 + clhs3528*clhs610;
            lhs(18,11)=-clhs1561*clhs871 - clhs1563*clhs2084 - clhs1565*clhs3530 + clhs1566*clhs3527 + clhs1571*clhs3529 + clhs3329 - clhs3455 + (2.0L/3.0L)*clhs3456 - clhs633*clhs819;
            lhs(18,12)=-clhs1575*clhs2084 + clhs1577*clhs3527 + clhs1581*clhs3529 + clhs1744*clhs662 + clhs3330 - clhs3509 + (2.0L/3.0L)*clhs3510 - clhs3531*clhs683 - clhs672*clhs819;
            lhs(18,13)=clhs1589*clhs3527 + clhs1592*clhs3529 - clhs203*clhs2084*clhs698*tau2 - clhs2055*clhs819 + clhs3208 + clhs3212*clhs57 + clhs3224 + clhs3289 + clhs3333 + clhs3450 - 4.0L/3.0L*clhs3451 + clhs3508;
            lhs(18,14)=clhs67*(-clhs1593*clhs871 + clhs1594*clhs3532 - clhs1595*clhs832 - clhs1596*clhs3533 + clhs3398);
            lhs(18,15)=clhs1646*clhs3490 + clhs1658*clhs3527 - clhs1667*clhs2546 - clhs1707*clhs3529 + clhs1780*clhs3458 + clhs1943*clhs3406 - clhs1969*clhs3459 - clhs2076*clhs3525 - clhs2077*clhs3410 + clhs2565*clhs781 + clhs3460*clhs905 - clhs3461*(DN(3,2)*clhs1977 - clhs1184*clhs830 + clhs1636*clhs1978 + clhs788 + clhs790 + clhs792 + clhs794 - clhs796 - clhs798 - clhs801 - clhs806 + clhs842*clhs90 + clhs848*clhs90 - clhs869*clhs936) + clhs3464*clhs3538 + clhs3467*clhs3540 + clhs3469*clhs3536 + clhs3528*clhs809 + clhs3534;
            lhs(18,16)=-clhs1017*clhs3483 + clhs1050*clhs3480*clhs3489 - clhs1742*clhs871 - clhs1744*clhs2084 - clhs1746*clhs3530 + clhs1747*clhs3527 + clhs1752*clhs3529 + clhs3477*clhs3536 - clhs3487 + (2.0L/3.0L)*clhs3488 - clhs819*clhs832;
            lhs(18,17)=-clhs1017*clhs3479 + clhs1744*clhs861 - clhs1756*clhs2084 + clhs1758*clhs3527 + clhs1762*clhs3529 + clhs3477*clhs3538 + clhs3480*clhs3486 - clhs3523 + (2.0L/3.0L)*clhs3524 - clhs3531*clhs882 - clhs819*clhs871;
            lhs(18,18)=-clhs1769*clhs3520*clhs3532 + clhs1770*clhs3527 + clhs1773*clhs3529 - clhs2084*clhs819 + clhs3400 + clhs3471 + clhs3472 - 4.0L/3.0L*clhs3474 + clhs3477*clhs3540 + clhs3519 - clhs3521*clhs57 + clhs3522;
            lhs(18,19)=clhs67*(-clhs1774*clhs871 + clhs1775*clhs3532 - clhs1776*clhs832 - clhs1777*clhs3533 + clhs3403);
            lhs(19,0)=-clhs2112*clhs3541 - clhs2115*clhs3542 - clhs2118*clhs3543 - clhs2149*clhs3544 - clhs2164*clhs3545 - clhs2173*clhs3546 - clhs2175*clhs3410 + clhs2184*clhs3548 + clhs2499 + clhs3547*clhs930;
            lhs(19,1)=clhs1179*clhs3547 + clhs1190*clhs3548 + clhs1811*clhs3542 + clhs1989*clhs3543 - clhs2191*clhs3549 - clhs2196*clhs3550 - clhs2201*clhs3551 - clhs2209*clhs3552 - clhs2210*clhs3541 + clhs2526;
            lhs(19,2)=-clhs1052*clhs3541 + clhs1197*clhs3547 + clhs1204*clhs3548 + clhs1988*clhs3543 - clhs2213*clhs3551 - clhs2216*clhs3550 - clhs2219*clhs3549 - clhs2223*clhs3552 - clhs2224*clhs3542 + clhs2549;
            lhs(19,3)=-clhs1053*clhs3541 + clhs1218*clhs3547 + clhs1225*clhs3548 - clhs1810*clhs3542 - clhs2225*clhs3551 - clhs2226*clhs3549 - clhs2227*clhs3550 - clhs2229*clhs3552 - clhs2230*clhs3543 + clhs2568;
            lhs(19,4)=-clhs1150*clhs3554 - clhs1158*clhs3553 - clhs1163*clhs3555 - clhs1227*clhs3557 - clhs2182*clhs3556 + clhs2237*clhs3541 + clhs2238*clhs3542 + clhs2239*clhs3543 + clhs2583;
            lhs(19,5)=clhs1292*clhs3547 - clhs2276*clhs3541 - clhs2277*clhs3542 - clhs2278*clhs3543 - clhs2288*clhs3544 - clhs2293*clhs3545 - clhs2295*clhs3546 - clhs2297*clhs3410 + clhs2299*clhs3548 + clhs3047;
            lhs(19,6)=clhs1383*clhs3547 + clhs1388*clhs3548 - clhs2311*clhs3549 - clhs2316*clhs3550 - clhs2319*clhs3551 - clhs2323*clhs3552 + clhs2324*clhs3542 + clhs2325*clhs3543 - clhs2326*clhs3541 + clhs3050;
            lhs(19,7)=clhs1395*clhs3547 + clhs1399*clhs3548 - clhs2333*clhs3551 - clhs2337*clhs3550 - clhs2340*clhs3549 - clhs2344*clhs3552 + clhs2345*clhs3543 - clhs2346*clhs3541 - clhs2347*clhs3542 + clhs3052;
            lhs(19,8)=clhs1408*clhs3547 + clhs1411*clhs3548 - clhs2352*clhs3551 - clhs2355*clhs3549 - clhs2356*clhs3550 - clhs2360*clhs3552 - clhs2361*clhs3541 - clhs2362*clhs3542 - clhs2363*clhs3543 + clhs3053;
            lhs(19,9)=-clhs1359*clhs3553 - clhs1363*clhs3554 - clhs1368*clhs3555 - clhs1415*clhs3557 - clhs2366*clhs3556 + clhs2367*clhs3541 + clhs2368*clhs3542 + clhs2369*clhs3543 + clhs3054;
            lhs(19,10)=clhs1477*clhs3547 - clhs2391*clhs3541 - clhs2392*clhs3542 - clhs2393*clhs3543 - clhs2403*clhs3544 - clhs2408*clhs3545 - clhs2410*clhs3546 - clhs2412*clhs3410 + clhs2414*clhs3548 + clhs3376;
            lhs(19,11)=clhs1566*clhs3547 + clhs1571*clhs3548 - clhs2423*clhs3549 - clhs2428*clhs3550 - clhs2431*clhs3551 - clhs2435*clhs3552 + clhs2436*clhs3542 + clhs2437*clhs3543 - clhs2438*clhs3541 + clhs3379;
            lhs(19,12)=clhs1577*clhs3547 + clhs1581*clhs3548 - clhs2444*clhs3551 - clhs2448*clhs3550 - clhs2451*clhs3549 - clhs2455*clhs3552 + clhs2456*clhs3543 - clhs2457*clhs3541 - clhs2458*clhs3542 + clhs3381;
            lhs(19,13)=clhs1589*clhs3547 + clhs1592*clhs3548 - clhs2462*clhs3551 - clhs2465*clhs3549 - clhs2466*clhs3550 - clhs2470*clhs3552 - clhs2471*clhs3541 - clhs2472*clhs3542 - clhs2473*clhs3543 + clhs3382;
            lhs(19,14)=-clhs1542*clhs3553 - clhs1546*clhs3554 - clhs1551*clhs3555 - clhs1596*clhs3557 - clhs2475*clhs3556 + clhs2476*clhs3541 + clhs2477*clhs3542 + clhs2478*clhs3543 + clhs3383;
            lhs(19,15)=-1.0L/2.0L*N[3]*clhs2127*clhs3563*clhs7*gamma + clhs1068*clhs3399 - clhs1215*clhs3558*clhs3559 - clhs1217*clhs3558*clhs3561 + clhs1658*clhs3547 - clhs1826*clhs3558*clhs3560 - clhs2500*clhs3541 - clhs2501*clhs3542 - clhs2502*clhs3543 - clhs2512*clhs3544 - clhs2517*clhs3545 - clhs2519*clhs3546 - clhs2521*clhs3410 + clhs2523*clhs3548 + clhs3461*(-clhs1140*clhs2074 - clhs1140*clhs750 - clhs1140*clhs785 - clhs1652*clhs2179 - clhs1736*clhs2179 - clhs1738*clhs2179 - clhs2098*clhs851 - clhs2104*clhs845 - clhs2110*clhs835 + clhs2176*clhs817 + clhs2176*clhs859 + clhs2177*clhs828 + clhs2177*clhs885 + clhs2178*clhs867 + clhs2178*clhs888 + clhs2520);
            lhs(19,16)=-clhs1215*clhs3566*(-6*clhs1652 + clhs1655 + clhs3562 + clhs785 + clhs845 + 6*clhs848 + 3*clhs851) - clhs1217*clhs3565 + clhs1747*clhs3547 + clhs1752*clhs3548 - clhs1826*clhs3564 - clhs2532*clhs3549 - clhs2537*clhs3550 - clhs2540*clhs3551 - clhs2544*clhs3552 + clhs2545*clhs3542 + clhs2546*clhs3543 - clhs2547*clhs3541 + clhs3457 + clhs3477*clhs3559 + clhs3482*clhs3567;
            lhs(19,17)=-clhs1215*clhs3564 - clhs1217*clhs3568 + clhs1758*clhs3547 + clhs1762*clhs3548 - clhs1826*clhs3566*(-6*clhs1738 + clhs3516 + 6*clhs842 + 3*clhs845) - clhs2553*clhs3551 - clhs2557*clhs3550 - clhs2560*clhs3549 - clhs2564*clhs3552 + clhs2565*clhs3543 - clhs2566*clhs3541 - clhs2567*clhs3542 + clhs3477*clhs3560 + clhs3478*clhs3567 + clhs3511;
            lhs(19,18)=-clhs1215*clhs3565 - clhs1217*clhs3566*(-6*clhs1736 + clhs3539 + 6*clhs833 + 3*clhs835) + clhs1770*clhs3547 + clhs1773*clhs3548 - clhs1826*clhs3568 - clhs2571*clhs3551 - clhs2574*clhs3549 - clhs2575*clhs3550 - clhs2579*clhs3552 - clhs2580*clhs3541 - clhs2581*clhs3542 - clhs2582*clhs3543 + clhs3477*clhs3561 + clhs3480*clhs3567 + clhs3534;
            lhs(19,19)=-clhs1017*clhs3480*clhs3569 - clhs1050*clhs3482*clhs3569 - clhs1723*clhs3553 - clhs1727*clhs3554 - clhs1732*clhs3555 - clhs1777*clhs3557 - clhs2584*clhs3556 + clhs2585*clhs3541 + clhs2586*clhs3542 + clhs2587*clhs3543 + clhs3400 + clhs3477*clhs3563*gamma - clhs3478*clhs3569*clhs985;


}

template<>
void CompressibleNavierStokes<3>::ComputeGaussPointRHSContribution(array_1d<double,20>& rhs, const ElementDataStruct& data,double data_v_sc, double data_k_sc)
{
    const int nnodes = 4;
    const int dim = 3;
    const int BlockSize = dim+2;
    const double h = data.h; 

    const double& bdf0 = data.bdf0;
    const double& bdf1 = data.bdf1;
    const double& bdf2 = data.bdf2;
    
    const BoundedMatrix<double,nnodes,BlockSize>& U = data.U;
    const BoundedMatrix<double,nnodes,BlockSize>& Un = data.Un;
    const BoundedMatrix<double,nnodes,BlockSize>& Unn = data.Unn;
    const BoundedMatrix<double,nnodes,dim>& f_ext = data.f_ext;
    const array_1d<double,nnodes>& r = data.r;
    const double mu = data.mu;
    const double nu = data.nu;
    const double lambda = data.lambda;
    const double c_v = data.c_v;
    const double gamma = data.gamma;
    const double cp = c_v*gamma;
    const double v_sc = data_v_sc;
    const double k_sc = data_k_sc;
    
    // Get shape function values
    const array_1d<double,nnodes>& N = data.N;
    const BoundedMatrix<double,nnodes,dim>& DN = data.DN_DX;

    // Auxiliary variables used in the calculation of the RHS
    const array_1d<double,BlockSize> U_gauss = prod(trans(U), N);
    const array_1d<double,dim> f_gauss = prod(trans(f_ext), N);
    const BoundedMatrix<double,dim,BlockSize> grad_U = prod(trans(U), DN); 
    const array_1d<double,BlockSize> accel_gauss = bdf0*U_gauss+bdf1*prod(trans(Un), N)+bdf2*prod(trans(Unn), N);
    
    // Stabilization parameters
    const double stab_c1 = 4.0;
    const double stab_c2 = 2.0;
    double tmp = U_gauss(dim+1)/U_gauss(0);
    for(unsigned int ll=0; ll<dim; ll++)
        tmp -=(U_gauss(ll+1)*U_gauss(ll+1))/(2*U_gauss(0)*U_gauss(0));
    double c = sqrt(gamma*(gamma-1)*tmp);

    double tau1inv = 0.0;
    for(unsigned int ll=0; ll<dim; ll++)
        tau1inv += (U_gauss(ll+1)/U_gauss(0))*(U_gauss(ll+1)/U_gauss(0));
    tau1inv = (sqrt(tau1inv)+c)*stab_c2/h;
    double tau2inv = stab_c1*nu/(h*h)+tau1inv;
    double tau3inv = stab_c1*lambda/(U_gauss(0)*cp*h*h)+tau1inv;
        
    const double tau1 = 1/tau1inv;
    const double tau2 = 1/tau2inv;
    const double tau3 = 1/tau3inv;

    const double crhs0 =             DN(0,2)*U(0,3);
const double crhs1 =             DN(1,2)*U(1,3);
const double crhs2 =             DN(2,2)*U(2,3);
const double crhs3 =             DN(3,2)*U(3,3);
const double crhs4 =             crhs0 + crhs1 + crhs2 + crhs3;
const double crhs5 =             DN(0,0)*U(0,1);
const double crhs6 =             DN(0,1)*U(0,2);
const double crhs7 =             DN(1,0)*U(1,1);
const double crhs8 =             DN(1,1)*U(1,2);
const double crhs9 =             DN(2,0)*U(2,1);
const double crhs10 =             DN(2,1)*U(2,2);
const double crhs11 =             DN(3,0)*U(3,1);
const double crhs12 =             DN(3,1)*U(3,2);
const double crhs13 =             crhs10 + crhs11 + crhs12 + crhs4 + crhs5 + crhs6 + crhs7 + crhs8 + crhs9;
const double crhs14 =             N[0]*(U(0,0)*bdf0 + Un(0,0)*bdf1 + Unn(0,0)*bdf2) + N[1]*(U(1,0)*bdf0 + Un(1,0)*bdf1 + Unn(1,0)*bdf2) + N[2]*(U(2,0)*bdf0 + Un(2,0)*bdf1 + Unn(2,0)*bdf2) + N[3]*(U(3,0)*bdf0 + Un(3,0)*bdf1 + Unn(3,0)*bdf2);
const double crhs15 =             N[0]*(U(0,1)*bdf0 + Un(0,1)*bdf1 + Unn(0,1)*bdf2);
const double crhs16 =             2*crhs15;
const double crhs17 =             N[1]*(U(1,1)*bdf0 + Un(1,1)*bdf1 + Unn(1,1)*bdf2);
const double crhs18 =             2*crhs17;
const double crhs19 =             N[2]*(U(2,1)*bdf0 + Un(2,1)*bdf1 + Unn(2,1)*bdf2);
const double crhs20 =             2*crhs19;
const double crhs21 =             N[3]*(U(3,1)*bdf0 + Un(3,1)*bdf1 + Unn(3,1)*bdf2);
const double crhs22 =             2*crhs21;
const double crhs23 =             2*gamma;
const double crhs24 =             crhs23 - 2;
const double crhs25 =             DN(0,0)*U(0,4);
const double crhs26 =             DN(1,0)*U(1,4);
const double crhs27 =             DN(2,0)*U(2,4);
const double crhs28 =             DN(3,0)*U(3,4);
const double crhs29 =             crhs25 + crhs26 + crhs27 + crhs28;
const double crhs30 =             crhs24*crhs29;
const double crhs31 =             N[0]*f_ext(0,0) + N[1]*f_ext(1,0) + N[2]*f_ext(2,0) + N[3]*f_ext(3,0);
const double crhs32 =             2*N[0]*U(0,0) + 2*N[1]*U(1,0) + 2*N[2]*U(2,0) + 2*N[3]*U(3,0);
const double crhs33 =             crhs31*crhs32;
const double crhs34 =             DN(0,1)*U(0,1);
const double crhs35 =             DN(1,1)*U(1,1);
const double crhs36 =             DN(2,1)*U(2,1);
const double crhs37 =             DN(3,1)*U(3,1);
const double crhs38 =             crhs34 + crhs35 + crhs36 + crhs37;
const double crhs39 =             N[0]*U(0,2) + N[1]*U(1,2) + N[2]*U(2,2) + N[3]*U(3,2);
const double crhs40 =             N[0]*U(0,0) + N[1]*U(1,0) + N[2]*U(2,0) + N[3]*U(3,0);
const double crhs41 =             1.0/crhs40;
const double crhs42 =             2*crhs39*crhs41;
const double crhs43 =             crhs38*crhs42;
const double crhs44 =             crhs10 + crhs12 + crhs6 + crhs8;
const double crhs45 =             N[0]*U(0,1) + N[1]*U(1,1) + N[2]*U(2,1) + N[3]*U(3,1);
const double crhs46 =             2*crhs41*crhs45;
const double crhs47 =             crhs44*crhs46;
const double crhs48 =             DN(0,2)*U(0,1);
const double crhs49 =             DN(1,2)*U(1,1);
const double crhs50 =             DN(2,2)*U(2,1);
const double crhs51 =             DN(3,2)*U(3,1);
const double crhs52 =             crhs48 + crhs49 + crhs50 + crhs51;
const double crhs53 =             N[0]*U(0,3) + N[1]*U(1,3) + N[2]*U(2,3) + N[3]*U(3,3);
const double crhs54 =             2*crhs41*crhs53;
const double crhs55 =             crhs52*crhs54;
const double crhs56 =             crhs4*crhs46;
const double crhs57 =             crhs23 - 6;
const double crhs58 =             crhs11 + crhs5 + crhs7 + crhs9;
const double crhs59 =             crhs41*crhs45;
const double crhs60 =             crhs58*crhs59;
const double crhs61 =             crhs57*crhs60;
const double crhs62 =             DN(0,0)*U(0,2);
const double crhs63 =             DN(1,0)*U(1,2);
const double crhs64 =             DN(2,0)*U(2,2);
const double crhs65 =             DN(3,0)*U(3,2);
const double crhs66 =             crhs62 + crhs63 + crhs64 + crhs65;
const double crhs67 =             crhs39*crhs41;
const double crhs68 =             crhs66*crhs67;
const double crhs69 =             crhs24*crhs68;
const double crhs70 =             DN(0,0)*U(0,3);
const double crhs71 =             DN(1,0)*U(1,3);
const double crhs72 =             DN(2,0)*U(2,3);
const double crhs73 =             DN(3,0)*U(3,3);
const double crhs74 =             crhs70 + crhs71 + crhs72 + crhs73;
const double crhs75 =             crhs41*crhs53;
const double crhs76 =             crhs74*crhs75;
const double crhs77 =             crhs24*crhs76;
const double crhs78 =             DN(0,1)*U(0,0) + DN(1,1)*U(1,0) + DN(2,1)*U(2,0) + DN(3,1)*U(3,0);
const double crhs79 =             crhs45*crhs78;
const double crhs80 =             pow(crhs40, -2);
const double crhs81 =             2*crhs39*crhs80;
const double crhs82 =             crhs79*crhs81;
const double crhs83 =             DN(0,2)*U(0,0) + DN(1,2)*U(1,0) + DN(2,2)*U(2,0) + DN(3,2)*U(3,0);
const double crhs84 =             crhs45*crhs83;
const double crhs85 =             2*crhs53*crhs80;
const double crhs86 =             crhs84*crhs85;
const double crhs87 =             DN(0,0)*U(0,0) + DN(1,0)*U(1,0) + DN(2,0)*U(2,0) + DN(3,0)*U(3,0);
const double crhs88 =             crhs80*crhs87;
const double crhs89 =             pow(crhs45, 2);
const double crhs90 =             gamma - 1;
const double crhs91 =             crhs89*crhs90;
const double crhs92 =             pow(crhs39, 2);
const double crhs93 =             crhs90*crhs92;
const double crhs94 =             pow(crhs53, 2);
const double crhs95 =             crhs90*crhs94;
const double crhs96 =             crhs91 + crhs93 + crhs95;
const double crhs97 =             -2*crhs89 + crhs96;
const double crhs98 =             crhs88*crhs97;
const double crhs99 =             -crhs16 - crhs18 - crhs20 - crhs22 - crhs30 + crhs33 - crhs43 - crhs47 - crhs55 - crhs56 + crhs61 + crhs69 + crhs77 + crhs82 + crhs86 - crhs98;
const double crhs100 =             (1.0L/2.0L)*crhs99*tau2;
const double crhs101 =             N[0]*(U(0,2)*bdf0 + Un(0,2)*bdf1 + Unn(0,2)*bdf2);
const double crhs102 =             2*crhs101;
const double crhs103 =             N[1]*(U(1,2)*bdf0 + Un(1,2)*bdf1 + Unn(1,2)*bdf2);
const double crhs104 =             2*crhs103;
const double crhs105 =             N[2]*(U(2,2)*bdf0 + Un(2,2)*bdf1 + Unn(2,2)*bdf2);
const double crhs106 =             2*crhs105;
const double crhs107 =             N[3]*(U(3,2)*bdf0 + Un(3,2)*bdf1 + Unn(3,2)*bdf2);
const double crhs108 =             2*crhs107;
const double crhs109 =             DN(0,1)*U(0,4);
const double crhs110 =             DN(1,1)*U(1,4);
const double crhs111 =             DN(2,1)*U(2,4);
const double crhs112 =             DN(3,1)*U(3,4);
const double crhs113 =             crhs109 + crhs110 + crhs111 + crhs112;
const double crhs114 =             crhs113*crhs24;
const double crhs115 =             N[0]*f_ext(0,1) + N[1]*f_ext(1,1) + N[2]*f_ext(2,1) + N[3]*f_ext(3,1);
const double crhs116 =             crhs115*crhs32;
const double crhs117 =             2*crhs5;
const double crhs118 =             2*crhs7;
const double crhs119 =             2*crhs9;
const double crhs120 =             2*crhs11;
const double crhs121 =             crhs117 + crhs118 + crhs119 + crhs120;
const double crhs122 =             crhs121*crhs67;
const double crhs123 =             crhs46*crhs66;
const double crhs124 =             DN(0,2)*U(0,2);
const double crhs125 =             DN(1,2)*U(1,2);
const double crhs126 =             DN(2,2)*U(2,2);
const double crhs127 =             DN(3,2)*U(3,2);
const double crhs128 =             crhs124 + crhs125 + crhs126 + crhs127;
const double crhs129 =             crhs128*crhs54;
const double crhs130 =             crhs4*crhs42;
const double crhs131 =             crhs44*crhs67;
const double crhs132 =             crhs131*crhs57;
const double crhs133 =             crhs38*crhs59;
const double crhs134 =             crhs133*crhs24;
const double crhs135 =             DN(0,1)*U(0,3);
const double crhs136 =             DN(1,1)*U(1,3);
const double crhs137 =             DN(2,1)*U(2,3);
const double crhs138 =             DN(3,1)*U(3,3);
const double crhs139 =             crhs135 + crhs136 + crhs137 + crhs138;
const double crhs140 =             crhs139*crhs75;
const double crhs141 =             crhs140*crhs24;
const double crhs142 =             crhs39*crhs87;
const double crhs143 =             2*crhs45*crhs80;
const double crhs144 =             crhs142*crhs143;
const double crhs145 =             crhs39*crhs83;
const double crhs146 =             crhs145*crhs85;
const double crhs147 =             crhs78*crhs80;
const double crhs148 =             -2*crhs92 + crhs96;
const double crhs149 =             crhs147*crhs148;
const double crhs150 =             -crhs102 - crhs104 - crhs106 - crhs108 - crhs114 + crhs116 - crhs122 - crhs123 - crhs129 - crhs130 + crhs132 + crhs134 + crhs141 + crhs144 + crhs146 - crhs149;
const double crhs151 =             (1.0L/2.0L)*crhs150*tau2;
const double crhs152 =             N[0]*(U(0,3)*bdf0 + Un(0,3)*bdf1 + Unn(0,3)*bdf2);
const double crhs153 =             2*crhs152;
const double crhs154 =             N[1]*(U(1,3)*bdf0 + Un(1,3)*bdf1 + Unn(1,3)*bdf2);
const double crhs155 =             2*crhs154;
const double crhs156 =             N[2]*(U(2,3)*bdf0 + Un(2,3)*bdf1 + Unn(2,3)*bdf2);
const double crhs157 =             2*crhs156;
const double crhs158 =             N[3]*(U(3,3)*bdf0 + Un(3,3)*bdf1 + Unn(3,3)*bdf2);
const double crhs159 =             2*crhs158;
const double crhs160 =             DN(0,2)*U(0,4);
const double crhs161 =             DN(1,2)*U(1,4);
const double crhs162 =             DN(2,2)*U(2,4);
const double crhs163 =             DN(3,2)*U(3,4);
const double crhs164 =             crhs160 + crhs161 + crhs162 + crhs163;
const double crhs165 =             crhs164*crhs24;
const double crhs166 =             N[0]*f_ext(0,2) + N[1]*f_ext(1,2) + N[2]*f_ext(2,2) + N[3]*f_ext(3,2);
const double crhs167 =             crhs166*crhs32;
const double crhs168 =             crhs121*crhs75;
const double crhs169 =             crhs46*crhs74;
const double crhs170 =             crhs44*crhs54;
const double crhs171 =             crhs139*crhs42;
const double crhs172 =             crhs4*crhs75;
const double crhs173 =             crhs172*crhs57;
const double crhs174 =             crhs52*crhs59;
const double crhs175 =             crhs174*crhs24;
const double crhs176 =             crhs128*crhs67;
const double crhs177 =             crhs176*crhs24;
const double crhs178 =             crhs53*crhs87;
const double crhs179 =             crhs143*crhs178;
const double crhs180 =             crhs53*crhs78;
const double crhs181 =             crhs180*crhs81;
const double crhs182 =             crhs80*crhs83;
const double crhs183 =             -2*crhs94 + crhs96;
const double crhs184 =             crhs182*crhs183;
const double crhs185 =             -crhs153 - crhs155 - crhs157 - crhs159 - crhs165 + crhs167 - crhs168 - crhs169 - crhs170 - crhs171 + crhs173 + crhs175 + crhs177 + crhs179 + crhs181 - crhs184;
const double crhs186 =             (1.0L/2.0L)*crhs185*tau2;
const double crhs187 =             N[0]*crhs40;
const double crhs188 =             crhs15 + crhs17 + crhs19 + crhs21;
const double crhs189 =             crhs40*v_sc/mu + 1;
const double crhs190 =             DN(0,1)*crhs189*crhs41*mu;
const double crhs191 =             -crhs34 - crhs35 - crhs36 - crhs37 + crhs41*(crhs142 + crhs79) - crhs62 - crhs63 - crhs64 - crhs65;
const double crhs192 =             DN(0,2)*crhs189*crhs41*mu;
const double crhs193 =             crhs41*(crhs178 + crhs84) - crhs48 - crhs49 - crhs50 - crhs51 - crhs70 - crhs71 - crhs72 - crhs73;
const double crhs194 =             DN(0,0)*crhs189*crhs41*mu;
const double crhs195 =             -4.0L/3.0L*DN(0,0)*U(0,1) + (2.0L/3.0L)*DN(0,1)*U(0,2) + (2.0L/3.0L)*DN(0,2)*U(0,3) - 4.0L/3.0L*DN(1,0)*U(1,1) + (2.0L/3.0L)*DN(1,1)*U(1,2) + (2.0L/3.0L)*DN(1,2)*U(1,3) - 4.0L/3.0L*DN(2,0)*U(2,1) + (2.0L/3.0L)*DN(2,1)*U(2,2) + (2.0L/3.0L)*DN(2,2)*U(2,3) - 4.0L/3.0L*DN(3,0)*U(3,1) + (2.0L/3.0L)*DN(3,1)*U(3,2) + (2.0L/3.0L)*DN(3,2)*U(3,3) - 2.0L/3.0L*crhs39*crhs41*crhs78 + (4.0L/3.0L)*crhs41*crhs45*crhs87 - 2.0L/3.0L*crhs41*crhs53*crhs83;
const double crhs196 =             (1.0L/2.0L)*N[0];
const double crhs197 =             crhs30 + crhs43 + crhs47 + crhs55 + crhs56 - crhs61 - crhs69 - crhs77 - crhs82 - crhs86 + crhs98;
const double crhs198 =             DN(0,1)*crhs45;
const double crhs199 =             N[0]*crhs38;
const double crhs200 =             DN(0,0)*crhs39;
const double crhs201 =             N[0]*crhs66;
const double crhs202 =             crhs201*crhs90;
const double crhs203 =             N[0]*crhs41;
const double crhs204 =             crhs203*crhs79;
const double crhs205 =             crhs142*crhs203;
const double crhs206 =             (1.0L/2.0L)*crhs150*crhs41*tau2;
const double crhs207 =             DN(0,2)*crhs45;
const double crhs208 =             N[0]*crhs52;
const double crhs209 =             DN(0,0)*crhs53;
const double crhs210 =             N[0]*crhs74;
const double crhs211 =             crhs210*crhs90;
const double crhs212 =             crhs203*crhs84;
const double crhs213 =             crhs178*crhs203;
const double crhs214 =             (1.0L/2.0L)*crhs185*crhs41*tau2;
const double crhs215 =             (1.0L/2.0L)*tau1*(crhs13 + crhs14);
const double crhs216 =             2*N[0];
const double crhs217 =             -crhs216*crhs31;
const double crhs218 =             N[0]*crhs44;
const double crhs219 =             N[0]*crhs4;
const double crhs220 =             gamma - 3;
const double crhs221 =             N[0]*crhs58;
const double crhs222 =             crhs220*crhs221;
const double crhs223 =             crhs24*crhs39*crhs80;
const double crhs224 =             crhs201*crhs223;
const double crhs225 =             crhs24*crhs53*crhs80;
const double crhs226 =             crhs210*crhs225;
const double crhs227 =             pow(crhs40, -3);
const double crhs228 =             4*N[0]*crhs227*crhs39;
const double crhs229 =             crhs228*crhs79;
const double crhs230 =             4*N[0]*crhs227*crhs53;
const double crhs231 =             crhs230*crhs84;
const double crhs232 =             crhs80*crhs97;
const double crhs233 =             2*N[0]*crhs227;
const double crhs234 =             crhs87*crhs97;
const double crhs235 =             DN(0,2)*crhs53;
const double crhs236 =             crhs75*crhs83;
const double crhs237 =             N[0]*crhs236;
const double crhs238 =             -crhs219 - crhs235 + crhs237;
const double crhs239 =             DN(0,1)*crhs39;
const double crhs240 =             -crhs239;
const double crhs241 =             -crhs218;
const double crhs242 =             DN(0,0)*crhs45;
const double crhs243 =             crhs67*crhs78;
const double crhs244 =             N[0]*crhs243;
const double crhs245 =             crhs41*crhs87;
const double crhs246 =             crhs245*crhs45;
const double crhs247 =             N[0]*crhs246;
const double crhs248 =             (1.0L/2.0L)*crhs41*crhs99*tau2;
const double crhs249 =             N[0]*(U(0,4)*bdf0 + Un(0,4)*bdf1 + Unn(0,4)*bdf2);
const double crhs250 =             N[1]*(U(1,4)*bdf0 + Un(1,4)*bdf1 + Unn(1,4)*bdf2);
const double crhs251 =             N[2]*(U(2,4)*bdf0 + Un(2,4)*bdf1 + Unn(2,4)*bdf2);
const double crhs252 =             N[3]*(U(3,4)*bdf0 + Un(3,4)*bdf1 + Unn(3,4)*bdf2);
const double crhs253 =             N[0]*r[0] + N[1]*r[1] + N[2]*r[2] + N[3]*r[3];
const double crhs254 =             crhs253*crhs40;
const double crhs255 =             crhs31*crhs45;
const double crhs256 =             crhs115*crhs39;
const double crhs257 =             crhs166*crhs53;
const double crhs258 =             crhs29*crhs45;
const double crhs259 =             crhs113*crhs39;
const double crhs260 =             crhs164*crhs53;
const double crhs261 =             crhs24*crhs39*crhs45*crhs80;
const double crhs262 =             crhs24*crhs45*crhs53*crhs80;
const double crhs263 =             crhs41*crhs91;
const double crhs264 =             N[0]*U(0,4);
const double crhs265 =             2*crhs264;
const double crhs266 =             -crhs265;
const double crhs267 =             N[1]*U(1,4);
const double crhs268 =             2*crhs267;
const double crhs269 =             -crhs268;
const double crhs270 =             N[2]*U(2,4);
const double crhs271 =             2*crhs270;
const double crhs272 =             -crhs271;
const double crhs273 =             N[3]*U(3,4);
const double crhs274 =             2*crhs273;
const double crhs275 =             -crhs274;
const double crhs276 =             crhs264 + crhs267 + crhs270 + crhs273;
const double crhs277 =             crhs24*crhs276;
const double crhs278 =             -crhs277;
const double crhs279 =             crhs41*crhs93;
const double crhs280 =             crhs41*crhs95;
const double crhs281 =             crhs266 + crhs269 + crhs272 + crhs275 + crhs278 + crhs279 + crhs280;
const double crhs282 =             3*crhs263 + crhs281;
const double crhs283 =             crhs263 + crhs266 + crhs269 + crhs272 + crhs275 + crhs278 + crhs280;
const double crhs284 =             3*crhs279 + crhs283;
const double crhs285 =             crhs263 + crhs266 + crhs269 + crhs272 + crhs275 + crhs278 + crhs279 + 3*crhs280;
const double crhs286 =             crhs41*(crhs89 + crhs92 + crhs94);
const double crhs287 =             crhs286*crhs90;
const double crhs288 =             crhs263 + crhs266 + crhs269 + crhs272 + crhs275 + crhs278 + crhs279 + crhs287;
const double crhs289 =             crhs280 + crhs288;
const double crhs290 =             crhs128*crhs223*crhs53 + crhs139*crhs223*crhs53 - crhs147*crhs289*crhs39 - crhs182*crhs289*crhs53 - crhs23*crhs258*crhs41 - crhs23*crhs259*crhs41 - crhs23*crhs260*crhs41 - 2*crhs249 - 2*crhs250 - 2*crhs251 - 2*crhs252 + 2*crhs254 + 2*crhs255 + 2*crhs256 + 2*crhs257 + crhs261*crhs38 + crhs261*crhs66 + crhs262*crhs52 + crhs262*crhs74 + crhs282*crhs41*crhs58 + crhs284*crhs41*crhs44 + crhs285*crhs4*crhs41 - crhs289*crhs45*crhs88;
const double crhs291 =             (1.0L/2.0L)*crhs290*crhs90*tau3;
const double crhs292 =             crhs101 + crhs103 + crhs105 + crhs107;
const double crhs293 =             -crhs124 - crhs125 - crhs126 - crhs127 - crhs135 - crhs136 - crhs137 - crhs138 + crhs41*(crhs145 + crhs180);
const double crhs294 =             -2.0L/3.0L*DN(0,0)*U(0,1);
const double crhs295 =             -2.0L/3.0L*DN(1,0)*U(1,1);
const double crhs296 =             -2.0L/3.0L*DN(2,0)*U(2,1);
const double crhs297 =             -2.0L/3.0L*DN(3,0)*U(3,1);
const double crhs298 =             (2.0L/3.0L)*crhs41*crhs45*crhs87;
const double crhs299 =             (4.0L/3.0L)*DN(0,1)*U(0,2) - 2.0L/3.0L*DN(0,2)*U(0,3) + (4.0L/3.0L)*DN(1,1)*U(1,2) - 2.0L/3.0L*DN(1,2)*U(1,3) + (4.0L/3.0L)*DN(2,1)*U(2,2) - 2.0L/3.0L*DN(2,2)*U(2,3) + (4.0L/3.0L)*DN(3,1)*U(3,2) - 2.0L/3.0L*DN(3,2)*U(3,3) + crhs294 + crhs295 + crhs296 + crhs297 + crhs298 - 4.0L/3.0L*crhs39*crhs41*crhs78 + (2.0L/3.0L)*crhs41*crhs53*crhs83;
const double crhs300 =             crhs114 + crhs122 + crhs123 + crhs129 + crhs130 - crhs132 - crhs134 - crhs141 - crhs144 - crhs146 + crhs149;
const double crhs301 =             crhs199*crhs90;
const double crhs302 =             DN(0,2)*crhs39;
const double crhs303 =             N[0]*crhs128;
const double crhs304 =             DN(0,1)*crhs53;
const double crhs305 =             N[0]*crhs139;
const double crhs306 =             crhs305*crhs90;
const double crhs307 =             crhs145*crhs203;
const double crhs308 =             crhs180*crhs203;
const double crhs309 =             -crhs115*crhs216;
const double crhs310 =             2*DN(0,0)*crhs45*crhs80;
const double crhs311 =             crhs218*crhs220;
const double crhs312 =             crhs24*crhs45*crhs80;
const double crhs313 =             crhs199*crhs312;
const double crhs314 =             crhs225*crhs305;
const double crhs315 =             4*N[0]*crhs227*crhs45;
const double crhs316 =             crhs142*crhs315;
const double crhs317 =             crhs145*crhs230;
const double crhs318 =             crhs148*crhs80;
const double crhs319 =             crhs148*crhs78;
const double crhs320 =             -crhs242;
const double crhs321 =             -crhs221;
const double crhs322 =             crhs152 + crhs154 + crhs156 + crhs158;
const double crhs323 =             -2.0L/3.0L*DN(0,1)*U(0,2) + (4.0L/3.0L)*DN(0,2)*U(0,3) - 2.0L/3.0L*DN(1,1)*U(1,2) + (4.0L/3.0L)*DN(1,2)*U(1,3) - 2.0L/3.0L*DN(2,1)*U(2,2) + (4.0L/3.0L)*DN(2,2)*U(2,3) - 2.0L/3.0L*DN(3,1)*U(3,2) + (4.0L/3.0L)*DN(3,2)*U(3,3) + crhs294 + crhs295 + crhs296 + crhs297 + crhs298 + (2.0L/3.0L)*crhs39*crhs41*crhs78 - 4.0L/3.0L*crhs41*crhs53*crhs83;
const double crhs324 =             crhs165 + crhs168 + crhs169 + crhs170 + crhs171 - crhs173 - crhs175 - crhs177 - crhs179 - crhs181 + crhs184;
const double crhs325 =             crhs208*crhs90;
const double crhs326 =             crhs303*crhs90;
const double crhs327 =             -crhs166*crhs216;
const double crhs328 =             crhs219*crhs220;
const double crhs329 =             crhs208*crhs312;
const double crhs330 =             crhs223*crhs303;
const double crhs331 =             crhs178*crhs315;
const double crhs332 =             crhs180*crhs228;
const double crhs333 =             crhs183*crhs80;
const double crhs334 =             crhs183*crhs83;
const double crhs335 =             crhs249 + crhs250 + crhs251 + crhs252;
const double crhs336 =             crhs254 + crhs255 + crhs256 + crhs257;
const double crhs337 =             3*crhs189*crhs39*crhs41*mu;
const double crhs338 =             3*crhs189*crhs41*crhs53*mu;
const double crhs339 =             crhs189*mu;
const double crhs340 =             3*lambda*(c_v*crhs40*k_sc/lambda + 1)/c_v;
const double crhs341 =             (1.0L/3.0L)*crhs41*(crhs191*crhs337 + crhs193*crhs338 + 2*crhs339*crhs41*crhs45*(crhs10 - crhs117 - crhs118 - crhs119 + crhs12 - crhs120 - crhs236 - crhs243 + 2*crhs246 + crhs4 + crhs6 + crhs8) + crhs340*(crhs245*crhs276 - crhs25 - crhs26 - crhs27 - crhs28 + crhs60 + crhs68 + crhs76 - crhs88*crhs89 - crhs88*crhs92 - crhs88*crhs94));
const double crhs342 =             3*crhs189*crhs41*crhs45*mu;
const double crhs343 =             -crhs11 + crhs246 - crhs5 - crhs7 - crhs9;
const double crhs344 =             crhs276*crhs41;
const double crhs345 =             (1.0L/3.0L)*crhs41*(crhs191*crhs342 + crhs293*crhs338 - 2*crhs339*crhs39*crhs41*(-crhs0 - crhs1 + 2*crhs10 + 2*crhs12 - crhs2 + crhs236 - 2*crhs243 - crhs3 + crhs343 + 2*crhs6 + 2*crhs8) + crhs340*(-crhs109 - crhs110 - crhs111 - crhs112 + crhs131 + crhs133 + crhs140 - crhs147*crhs89 - crhs147*crhs92 - crhs147*crhs94 + crhs344*crhs78));
const double crhs346 =             (1.0L/3.0L)*crhs41*(crhs193*crhs342 + crhs293*crhs337 - 2*crhs339*crhs41*crhs53*(2*crhs0 + 2*crhs1 - crhs10 - crhs12 + 2*crhs2 - 2*crhs236 + crhs243 + 2*crhs3 + crhs343 - crhs6 - crhs8) + crhs340*(-crhs160 - crhs161 - crhs162 - crhs163 + crhs172 + crhs174 + crhs176 - crhs182*crhs89 - crhs182*crhs92 - crhs182*crhs94 + crhs344*crhs83));
const double crhs347 =             (1.0L/4.0L)*tau2*(crhs16 + crhs18 + crhs197 + crhs20 + crhs22 - crhs33);
const double crhs348 =             2*N[0]*crhs41*gamma;
const double crhs349 =             6*crhs45*crhs80*crhs90;
const double crhs350 =             crhs282*crhs41;
const double crhs351 =             5*crhs263 + crhs281 + crhs287;
const double crhs352 =             crhs351*crhs80*crhs87;
const double crhs353 =             (1.0L/4.0L)*tau2*(crhs102 + crhs104 + crhs106 + crhs108 - crhs116 + crhs300);
const double crhs354 =             DN(0,0)*crhs24*crhs45*crhs80;
const double crhs355 =             6*crhs39*crhs80*crhs90;
const double crhs356 =             crhs284*crhs41;
const double crhs357 =             5*crhs279 + crhs283 + crhs287;
const double crhs358 =             crhs357*crhs78*crhs80;
const double crhs359 =             (1.0L/4.0L)*tau2*(crhs153 + crhs155 + crhs157 + crhs159 - crhs167 + crhs324);
const double crhs360 =             6*crhs53*crhs80*crhs90;
const double crhs361 =             crhs285*crhs41;
const double crhs362 =             5*crhs280 + crhs288;
const double crhs363 =             crhs362*crhs80*crhs83;
const double crhs364 =             -crhs113*crhs39*gamma + crhs128*crhs39*crhs41*crhs53*crhs90 + crhs139*crhs39*crhs41*crhs53*crhs90 - crhs164*crhs53*gamma + (1.0L/2.0L)*crhs282*crhs58 + (1.0L/2.0L)*crhs284*crhs44 + (1.0L/2.0L)*crhs285*crhs4 - 1.0L/2.0L*crhs289*crhs39*crhs41*crhs78 - 1.0L/2.0L*crhs289*crhs41*crhs45*crhs87 - 1.0L/2.0L*crhs289*crhs41*crhs53*crhs83 - crhs29*crhs45*gamma + crhs38*crhs39*crhs41*crhs45*crhs90 + crhs39*crhs41*crhs45*crhs66*crhs90 + crhs41*crhs45*crhs52*crhs53*crhs90 + crhs41*crhs45*crhs53*crhs74*crhs90;
const double crhs365 =             (1.0L/2.0L)*crhs290*crhs41*gamma*tau3;
const double crhs366 =             2*N[0]*crhs80*gamma;
const double crhs367 =             4*crhs227*crhs39*crhs45;
const double crhs368 =             4*crhs227*crhs45*crhs53;
const double crhs369 =             4*crhs227*crhs39*crhs53;
const double crhs370 =             crhs289*crhs80;
const double crhs371 =             crhs351*crhs80;
const double crhs372 =             crhs357*crhs80;
const double crhs373 =             crhs362*crhs80;
const double crhs374 =             crhs45*crhs87;
const double crhs375 =             -crhs24*crhs286 - crhs263 + crhs265 + crhs268 + crhs271 + crhs274 + crhs277 - crhs279 - crhs280;
const double crhs376 =             2*N[0]*crhs227*crhs375;
const double crhs377 =             crhs39*crhs78;
const double crhs378 =             crhs53*crhs83;
const double crhs379 =             N[1]*crhs40;
const double crhs380 =             DN(1,1)*crhs189*crhs41*mu;
const double crhs381 =             DN(1,2)*crhs189*crhs41*mu;
const double crhs382 =             DN(1,0)*crhs189*crhs41*mu;
const double crhs383 =             (1.0L/2.0L)*N[1];
const double crhs384 =             DN(1,1)*crhs45;
const double crhs385 =             N[1]*crhs38;
const double crhs386 =             DN(1,0)*crhs39;
const double crhs387 =             N[1]*crhs66;
const double crhs388 =             crhs387*crhs90;
const double crhs389 =             N[1]*crhs41;
const double crhs390 =             crhs389*crhs79;
const double crhs391 =             crhs142*crhs389;
const double crhs392 =             DN(1,2)*crhs45;
const double crhs393 =             N[1]*crhs52;
const double crhs394 =             DN(1,0)*crhs53;
const double crhs395 =             N[1]*crhs74;
const double crhs396 =             crhs395*crhs90;
const double crhs397 =             crhs389*crhs84;
const double crhs398 =             crhs178*crhs389;
const double crhs399 =             2*N[1];
const double crhs400 =             -crhs31*crhs399;
const double crhs401 =             N[1]*crhs44;
const double crhs402 =             N[1]*crhs4;
const double crhs403 =             N[1]*crhs58;
const double crhs404 =             crhs220*crhs403;
const double crhs405 =             crhs223*crhs387;
const double crhs406 =             crhs225*crhs395;
const double crhs407 =             4*N[1]*crhs227*crhs39;
const double crhs408 =             crhs407*crhs79;
const double crhs409 =             4*N[1]*crhs227*crhs53;
const double crhs410 =             crhs409*crhs84;
const double crhs411 =             2*N[1]*crhs227;
const double crhs412 =             DN(1,2)*crhs53;
const double crhs413 =             N[1]*crhs236;
const double crhs414 =             -crhs402 - crhs412 + crhs413;
const double crhs415 =             DN(1,1)*crhs39;
const double crhs416 =             -crhs415;
const double crhs417 =             -crhs401;
const double crhs418 =             DN(1,0)*crhs45;
const double crhs419 =             N[1]*crhs243;
const double crhs420 =             N[1]*crhs246;
const double crhs421 =             crhs385*crhs90;
const double crhs422 =             DN(1,2)*crhs39;
const double crhs423 =             N[1]*crhs128;
const double crhs424 =             DN(1,1)*crhs53;
const double crhs425 =             N[1]*crhs139;
const double crhs426 =             crhs425*crhs90;
const double crhs427 =             crhs145*crhs389;
const double crhs428 =             crhs180*crhs389;
const double crhs429 =             -crhs115*crhs399;
const double crhs430 =             crhs220*crhs401;
const double crhs431 =             crhs312*crhs385;
const double crhs432 =             crhs225*crhs425;
const double crhs433 =             4*N[1]*crhs227*crhs45;
const double crhs434 =             crhs142*crhs433;
const double crhs435 =             crhs145*crhs409;
const double crhs436 =             -crhs418;
const double crhs437 =             -crhs403;
const double crhs438 =             crhs393*crhs90;
const double crhs439 =             crhs423*crhs90;
const double crhs440 =             -crhs166*crhs399;
const double crhs441 =             crhs220*crhs402;
const double crhs442 =             crhs312*crhs393;
const double crhs443 =             crhs223*crhs423;
const double crhs444 =             crhs178*crhs433;
const double crhs445 =             crhs180*crhs407;
const double crhs446 =             2*N[1]*crhs41*gamma;
const double crhs447 =             2*N[1]*crhs80*gamma;
const double crhs448 =             2*N[1]*crhs227*crhs375;
const double crhs449 =             N[2]*crhs40;
const double crhs450 =             DN(2,1)*crhs189*crhs41*mu;
const double crhs451 =             DN(2,2)*crhs189*crhs41*mu;
const double crhs452 =             DN(2,0)*crhs189*crhs41*mu;
const double crhs453 =             (1.0L/2.0L)*N[2];
const double crhs454 =             DN(2,1)*crhs45;
const double crhs455 =             N[2]*crhs38;
const double crhs456 =             DN(2,0)*crhs39;
const double crhs457 =             N[2]*crhs66;
const double crhs458 =             crhs457*crhs90;
const double crhs459 =             N[2]*crhs41;
const double crhs460 =             crhs459*crhs79;
const double crhs461 =             crhs142*crhs459;
const double crhs462 =             DN(2,2)*crhs45;
const double crhs463 =             N[2]*crhs52;
const double crhs464 =             DN(2,0)*crhs53;
const double crhs465 =             N[2]*crhs74;
const double crhs466 =             crhs465*crhs90;
const double crhs467 =             crhs459*crhs84;
const double crhs468 =             crhs178*crhs459;
const double crhs469 =             2*N[2];
const double crhs470 =             -crhs31*crhs469;
const double crhs471 =             N[2]*crhs44;
const double crhs472 =             N[2]*crhs4;
const double crhs473 =             N[2]*crhs58;
const double crhs474 =             crhs220*crhs473;
const double crhs475 =             crhs223*crhs457;
const double crhs476 =             crhs225*crhs465;
const double crhs477 =             4*N[2]*crhs227*crhs39;
const double crhs478 =             crhs477*crhs79;
const double crhs479 =             4*N[2]*crhs227*crhs53;
const double crhs480 =             crhs479*crhs84;
const double crhs481 =             2*N[2]*crhs227;
const double crhs482 =             DN(2,2)*crhs53;
const double crhs483 =             N[2]*crhs236;
const double crhs484 =             -crhs472 - crhs482 + crhs483;
const double crhs485 =             DN(2,1)*crhs39;
const double crhs486 =             -crhs485;
const double crhs487 =             -crhs471;
const double crhs488 =             DN(2,0)*crhs45;
const double crhs489 =             N[2]*crhs243;
const double crhs490 =             N[2]*crhs246;
const double crhs491 =             crhs455*crhs90;
const double crhs492 =             DN(2,2)*crhs39;
const double crhs493 =             N[2]*crhs128;
const double crhs494 =             DN(2,1)*crhs53;
const double crhs495 =             N[2]*crhs139;
const double crhs496 =             crhs495*crhs90;
const double crhs497 =             crhs145*crhs459;
const double crhs498 =             crhs180*crhs459;
const double crhs499 =             -crhs115*crhs469;
const double crhs500 =             crhs220*crhs471;
const double crhs501 =             crhs312*crhs455;
const double crhs502 =             crhs225*crhs495;
const double crhs503 =             4*N[2]*crhs227*crhs45;
const double crhs504 =             crhs142*crhs503;
const double crhs505 =             crhs145*crhs479;
const double crhs506 =             -crhs488;
const double crhs507 =             -crhs473;
const double crhs508 =             crhs463*crhs90;
const double crhs509 =             crhs493*crhs90;
const double crhs510 =             -crhs166*crhs469;
const double crhs511 =             crhs220*crhs472;
const double crhs512 =             crhs312*crhs463;
const double crhs513 =             crhs223*crhs493;
const double crhs514 =             crhs178*crhs503;
const double crhs515 =             crhs180*crhs477;
const double crhs516 =             2*N[2]*crhs41*gamma;
const double crhs517 =             2*N[2]*crhs80*gamma;
const double crhs518 =             2*N[2]*crhs227*crhs375;
const double crhs519 =             N[3]*crhs40;
const double crhs520 =             DN(3,1)*crhs189*crhs41*mu;
const double crhs521 =             DN(3,2)*crhs189*crhs41*mu;
const double crhs522 =             DN(3,0)*crhs189*crhs41*mu;
const double crhs523 =             (1.0L/2.0L)*N[3];
const double crhs524 =             DN(3,1)*crhs45;
const double crhs525 =             N[3]*crhs38;
const double crhs526 =             DN(3,0)*crhs39;
const double crhs527 =             N[3]*crhs66;
const double crhs528 =             crhs527*crhs90;
const double crhs529 =             N[3]*crhs41;
const double crhs530 =             crhs529*crhs79;
const double crhs531 =             crhs142*crhs529;
const double crhs532 =             DN(3,2)*crhs45;
const double crhs533 =             N[3]*crhs52;
const double crhs534 =             DN(3,0)*crhs53;
const double crhs535 =             N[3]*crhs74;
const double crhs536 =             crhs535*crhs90;
const double crhs537 =             crhs529*crhs84;
const double crhs538 =             crhs178*crhs529;
const double crhs539 =             2*N[3];
const double crhs540 =             -crhs31*crhs539;
const double crhs541 =             N[3]*crhs44;
const double crhs542 =             N[3]*crhs4;
const double crhs543 =             N[3]*crhs58;
const double crhs544 =             crhs220*crhs543;
const double crhs545 =             crhs223*crhs527;
const double crhs546 =             crhs225*crhs535;
const double crhs547 =             4*N[3]*crhs227*crhs39;
const double crhs548 =             crhs547*crhs79;
const double crhs549 =             4*N[3]*crhs227*crhs53;
const double crhs550 =             crhs549*crhs84;
const double crhs551 =             2*N[3]*crhs227;
const double crhs552 =             DN(3,2)*crhs53;
const double crhs553 =             N[3]*crhs236;
const double crhs554 =             -crhs542 - crhs552 + crhs553;
const double crhs555 =             DN(3,1)*crhs39;
const double crhs556 =             -crhs555;
const double crhs557 =             -crhs541;
const double crhs558 =             DN(3,0)*crhs45;
const double crhs559 =             N[3]*crhs243;
const double crhs560 =             N[3]*crhs246;
const double crhs561 =             crhs525*crhs90;
const double crhs562 =             DN(3,2)*crhs39;
const double crhs563 =             N[3]*crhs128;
const double crhs564 =             DN(3,1)*crhs53;
const double crhs565 =             N[3]*crhs139;
const double crhs566 =             crhs565*crhs90;
const double crhs567 =             crhs145*crhs529;
const double crhs568 =             crhs180*crhs529;
const double crhs569 =             -crhs115*crhs539;
const double crhs570 =             crhs220*crhs541;
const double crhs571 =             crhs312*crhs525;
const double crhs572 =             crhs225*crhs565;
const double crhs573 =             4*N[3]*crhs227*crhs45;
const double crhs574 =             crhs142*crhs573;
const double crhs575 =             crhs145*crhs549;
const double crhs576 =             -crhs558;
const double crhs577 =             -crhs543;
const double crhs578 =             crhs533*crhs90;
const double crhs579 =             crhs563*crhs90;
const double crhs580 =             -crhs166*crhs539;
const double crhs581 =             crhs220*crhs542;
const double crhs582 =             crhs312*crhs533;
const double crhs583 =             crhs223*crhs563;
const double crhs584 =             crhs178*crhs573;
const double crhs585 =             crhs180*crhs547;
const double crhs586 =             2*N[3]*crhs41*gamma;
const double crhs587 =             2*N[3]*crhs80*gamma;
const double crhs588 =             2*N[3]*crhs227*crhs375;
            rhs[0]=-DN(0,0)*crhs100 - DN(0,1)*crhs151 - DN(0,2)*crhs186 + N[0]*crhs13 + N[0]*crhs14;
            rhs[1]=-DN(0,0)*crhs291 + N[0]*crhs188 - crhs187*crhs31 - crhs190*crhs191 - crhs192*crhs193 - crhs194*crhs195 + crhs196*crhs197 + crhs206*(-crhs198 - crhs199 + crhs200*crhs90 + crhs202 + crhs204 - crhs205*crhs90) + crhs214*(-crhs207 - crhs208 + crhs209*crhs90 + crhs211 + crhs212 - crhs213*crhs90) - crhs215*(-DN(0,0)*crhs232 + crhs143*crhs218 + crhs143*crhs219 - crhs143*crhs222 + crhs198*crhs81 + crhs199*crhs81 + crhs207*crhs85 + crhs208*crhs85 + crhs217 - crhs224 - crhs226 - crhs229 - crhs231 + crhs233*crhs234) + crhs248*(crhs220*crhs242 - crhs220*crhs247 + crhs222 + crhs238 + crhs240 + crhs241 + crhs244);
            rhs[2]=-DN(0,1)*crhs291 + N[0]*crhs292 - crhs115*crhs187 + crhs190*crhs299 - crhs191*crhs194 - crhs192*crhs293 + crhs196*crhs300 + crhs206*(crhs220*crhs239 - crhs220*crhs244 + crhs238 + crhs247 + crhs311 + crhs320 + crhs321) + crhs214*(-crhs302 - crhs303 + crhs304*crhs90 + crhs306 + crhs307 - crhs308*crhs90) - crhs215*(-DN(0,1)*crhs318 + crhs143*crhs201 + crhs219*crhs81 + crhs221*crhs81 + crhs233*crhs319 + crhs302*crhs85 + crhs303*crhs85 + crhs309 + crhs310*crhs39 - crhs311*crhs81 - crhs313 - crhs314 - crhs316 - crhs317) - crhs248*(-crhs198*crhs90 + crhs200 + crhs201 + crhs204*crhs90 - crhs205 - crhs301);
            rhs[3]=-DN(0,2)*crhs291 + N[0]*crhs322 - crhs166*crhs187 - crhs190*crhs293 + crhs192*crhs323 - crhs193*crhs194 + crhs196*crhs324 - crhs206*(-crhs302*crhs90 + crhs304 + crhs305 + crhs307*crhs90 - crhs308 - crhs326) + crhs214*(crhs220*crhs235 - crhs220*crhs237 + crhs240 + crhs241 + crhs244 + crhs247 + crhs320 + crhs321 + crhs328) - crhs215*(-DN(0,2)*crhs333 + crhs143*crhs210 + crhs218*crhs85 + crhs221*crhs85 + crhs233*crhs334 + crhs239*crhs85 + crhs305*crhs81 + crhs310*crhs53 + crhs327 - crhs328*crhs85 - crhs329 - crhs330 - crhs331 - crhs332) - crhs248*(-crhs207*crhs90 + crhs209 + crhs210 + crhs212*crhs90 - crhs213 - crhs325);
            rhs[4]=-DN(0,0)*crhs341 - DN(0,1)*crhs345 - DN(0,2)*crhs346 + N[0]*crhs335 - N[0]*crhs336 - crhs203*crhs364 + crhs215*(crhs202*crhs367 + crhs211*crhs368 + crhs216*crhs253 + crhs218*crhs372 + crhs219*crhs373 + crhs221*crhs371 + crhs235*crhs370 + crhs239*crhs370 + crhs242*crhs370 - crhs258*crhs366 - crhs259*crhs366 - crhs260*crhs366 + crhs301*crhs367 + crhs306*crhs369 + crhs325*crhs368 + crhs326*crhs369 + crhs374*crhs376 + crhs376*crhs377 + crhs376*crhs378) - crhs347*(DN(0,0)*crhs350 - N[0]*crhs352 + crhs198*crhs223 + crhs199*crhs223 + crhs207*crhs225 + crhs208*crhs225 + crhs217 + crhs218*crhs312 + crhs219*crhs312 + crhs221*crhs349 + crhs224 + crhs226 - crhs229*crhs90 - crhs231*crhs90 - crhs29*crhs348) - crhs353*(DN(0,1)*crhs356 - N[0]*crhs358 - crhs113*crhs348 + crhs201*crhs312 + crhs218*crhs355 + crhs219*crhs223 + crhs221*crhs223 + crhs225*crhs302 + crhs225*crhs303 + crhs309 + crhs313 + crhs314 - crhs316*crhs90 - crhs317*crhs90 + crhs354*crhs39) - crhs359*(DN(0,2)*crhs361 - N[0]*crhs363 - crhs164*crhs348 + crhs210*crhs312 + crhs218*crhs225 + crhs219*crhs360 + crhs221*crhs225 + crhs223*crhs305 + crhs225*crhs239 + crhs327 + crhs329 + crhs330 - crhs331*crhs90 - crhs332*crhs90 + crhs354*crhs53) - crhs365*(crhs218 + crhs219 + crhs221 + crhs235 - crhs237 + crhs239 + crhs242 - crhs244 - crhs247);
            rhs[5]=-DN(1,0)*crhs100 - DN(1,1)*crhs151 - DN(1,2)*crhs186 + N[1]*crhs13 + N[1]*crhs14;
            rhs[6]=-DN(1,0)*crhs291 + N[1]*crhs188 - crhs191*crhs380 - crhs193*crhs381 - crhs195*crhs382 + crhs197*crhs383 + crhs206*(-crhs384 - crhs385 + crhs386*crhs90 + crhs388 + crhs390 - crhs391*crhs90) + crhs214*(-crhs392 - crhs393 + crhs394*crhs90 + crhs396 + crhs397 - crhs398*crhs90) - crhs215*(-DN(1,0)*crhs232 + crhs143*crhs401 + crhs143*crhs402 - crhs143*crhs404 + crhs234*crhs411 + crhs384*crhs81 + crhs385*crhs81 + crhs392*crhs85 + crhs393*crhs85 + crhs400 - crhs405 - crhs406 - crhs408 - crhs410) + crhs248*(crhs220*crhs418 - crhs220*crhs420 + crhs404 + crhs414 + crhs416 + crhs417 + crhs419) - crhs31*crhs379;
            rhs[7]=-DN(1,1)*crhs291 + N[1]*crhs292 - crhs115*crhs379 - crhs191*crhs382 + crhs206*(crhs220*crhs415 - crhs220*crhs419 + crhs414 + crhs420 + crhs430 + crhs436 + crhs437) + crhs214*(-crhs422 - crhs423 + crhs424*crhs90 + crhs426 + crhs427 - crhs428*crhs90) - crhs215*(-DN(1,1)*crhs318 + crhs143*crhs387 + crhs319*crhs411 + crhs402*crhs81 + crhs403*crhs81 + crhs418*crhs81 + crhs422*crhs85 + crhs423*crhs85 + crhs429 - crhs430*crhs81 - crhs431 - crhs432 - crhs434 - crhs435) - crhs248*(-crhs384*crhs90 + crhs386 + crhs387 + crhs390*crhs90 - crhs391 - crhs421) - crhs293*crhs381 + crhs299*crhs380 + crhs300*crhs383;
            rhs[8]=-DN(1,2)*crhs291 + N[1]*crhs322 - crhs166*crhs379 - crhs193*crhs382 - crhs206*(-crhs422*crhs90 + crhs424 + crhs425 + crhs427*crhs90 - crhs428 - crhs439) + crhs214*(crhs220*crhs412 - crhs220*crhs413 + crhs416 + crhs417 + crhs419 + crhs420 + crhs436 + crhs437 + crhs441) - crhs215*(-DN(1,2)*crhs333 + crhs143*crhs395 + crhs334*crhs411 + crhs401*crhs85 + crhs403*crhs85 + crhs415*crhs85 + crhs418*crhs85 + crhs425*crhs81 + crhs440 - crhs441*crhs85 - crhs442 - crhs443 - crhs444 - crhs445) - crhs248*(-crhs392*crhs90 + crhs394 + crhs395 + crhs397*crhs90 - crhs398 - crhs438) - crhs293*crhs380 + crhs323*crhs381 + crhs324*crhs383;
            rhs[9]=-DN(1,0)*crhs341 - DN(1,1)*crhs345 - DN(1,2)*crhs346 + N[1]*crhs335 - N[1]*crhs336 + crhs215*(crhs253*crhs399 - crhs258*crhs447 - crhs259*crhs447 - crhs260*crhs447 + crhs367*crhs388 + crhs367*crhs421 + crhs368*crhs396 + crhs368*crhs438 + crhs369*crhs426 + crhs369*crhs439 + crhs370*crhs412 + crhs370*crhs415 + crhs370*crhs418 + crhs371*crhs403 + crhs372*crhs401 + crhs373*crhs402 + crhs374*crhs448 + crhs377*crhs448 + crhs378*crhs448) - crhs347*(DN(1,0)*crhs350 - N[1]*crhs352 + crhs223*crhs384 + crhs223*crhs385 + crhs225*crhs392 + crhs225*crhs393 - crhs29*crhs446 + crhs312*crhs401 + crhs312*crhs402 + crhs349*crhs403 + crhs400 + crhs405 + crhs406 - crhs408*crhs90 - crhs410*crhs90) - crhs353*(DN(1,1)*crhs356 - N[1]*crhs358 - crhs113*crhs446 + crhs223*crhs402 + crhs223*crhs403 + crhs223*crhs418 + crhs225*crhs422 + crhs225*crhs423 + crhs312*crhs387 + crhs355*crhs401 + crhs429 + crhs431 + crhs432 - crhs434*crhs90 - crhs435*crhs90) - crhs359*(DN(1,2)*crhs361 - N[1]*crhs363 - crhs164*crhs446 + crhs223*crhs425 + crhs225*crhs401 + crhs225*crhs403 + crhs225*crhs415 + crhs225*crhs418 + crhs312*crhs395 + crhs360*crhs402 + crhs440 + crhs442 + crhs443 - crhs444*crhs90 - crhs445*crhs90) - crhs364*crhs389 - crhs365*(crhs401 + crhs402 + crhs403 + crhs412 - crhs413 + crhs415 + crhs418 - crhs419 - crhs420);
            rhs[10]=-DN(2,0)*crhs100 - DN(2,1)*crhs151 - DN(2,2)*crhs186 + N[2]*crhs13 + N[2]*crhs14;
            rhs[11]=-DN(2,0)*crhs291 + N[2]*crhs188 - crhs191*crhs450 - crhs193*crhs451 - crhs195*crhs452 + crhs197*crhs453 + crhs206*(-crhs454 - crhs455 + crhs456*crhs90 + crhs458 + crhs460 - crhs461*crhs90) + crhs214*(-crhs462 - crhs463 + crhs464*crhs90 + crhs466 + crhs467 - crhs468*crhs90) - crhs215*(-DN(2,0)*crhs232 + crhs143*crhs471 + crhs143*crhs472 - crhs143*crhs474 + crhs234*crhs481 + crhs454*crhs81 + crhs455*crhs81 + crhs462*crhs85 + crhs463*crhs85 + crhs470 - crhs475 - crhs476 - crhs478 - crhs480) + crhs248*(crhs220*crhs488 - crhs220*crhs490 + crhs474 + crhs484 + crhs486 + crhs487 + crhs489) - crhs31*crhs449;
            rhs[12]=-DN(2,1)*crhs291 + N[2]*crhs292 - crhs115*crhs449 - crhs191*crhs452 + crhs206*(crhs220*crhs485 - crhs220*crhs489 + crhs484 + crhs490 + crhs500 + crhs506 + crhs507) + crhs214*(-crhs492 - crhs493 + crhs494*crhs90 + crhs496 + crhs497 - crhs498*crhs90) - crhs215*(-DN(2,1)*crhs318 + crhs143*crhs457 + crhs319*crhs481 + crhs472*crhs81 + crhs473*crhs81 + crhs488*crhs81 + crhs492*crhs85 + crhs493*crhs85 + crhs499 - crhs500*crhs81 - crhs501 - crhs502 - crhs504 - crhs505) - crhs248*(-crhs454*crhs90 + crhs456 + crhs457 + crhs460*crhs90 - crhs461 - crhs491) - crhs293*crhs451 + crhs299*crhs450 + crhs300*crhs453;
            rhs[13]=-DN(2,2)*crhs291 + N[2]*crhs322 - crhs166*crhs449 - crhs193*crhs452 - crhs206*(-crhs492*crhs90 + crhs494 + crhs495 + crhs497*crhs90 - crhs498 - crhs509) + crhs214*(crhs220*crhs482 - crhs220*crhs483 + crhs486 + crhs487 + crhs489 + crhs490 + crhs506 + crhs507 + crhs511) - crhs215*(-DN(2,2)*crhs333 + crhs143*crhs465 + crhs334*crhs481 + crhs471*crhs85 + crhs473*crhs85 + crhs485*crhs85 + crhs488*crhs85 + crhs495*crhs81 + crhs510 - crhs511*crhs85 - crhs512 - crhs513 - crhs514 - crhs515) - crhs248*(-crhs462*crhs90 + crhs464 + crhs465 + crhs467*crhs90 - crhs468 - crhs508) - crhs293*crhs450 + crhs323*crhs451 + crhs324*crhs453;
            rhs[14]=-DN(2,0)*crhs341 - DN(2,1)*crhs345 - DN(2,2)*crhs346 + N[2]*crhs335 - N[2]*crhs336 + crhs215*(crhs253*crhs469 - crhs258*crhs517 - crhs259*crhs517 - crhs260*crhs517 + crhs367*crhs458 + crhs367*crhs491 + crhs368*crhs466 + crhs368*crhs508 + crhs369*crhs496 + crhs369*crhs509 + crhs370*crhs482 + crhs370*crhs485 + crhs370*crhs488 + crhs371*crhs473 + crhs372*crhs471 + crhs373*crhs472 + crhs374*crhs518 + crhs377*crhs518 + crhs378*crhs518) - crhs347*(DN(2,0)*crhs350 - N[2]*crhs352 + crhs223*crhs454 + crhs223*crhs455 + crhs225*crhs462 + crhs225*crhs463 - crhs29*crhs516 + crhs312*crhs471 + crhs312*crhs472 + crhs349*crhs473 + crhs470 + crhs475 + crhs476 - crhs478*crhs90 - crhs480*crhs90) - crhs353*(DN(2,1)*crhs356 - N[2]*crhs358 - crhs113*crhs516 + crhs223*crhs472 + crhs223*crhs473 + crhs223*crhs488 + crhs225*crhs492 + crhs225*crhs493 + crhs312*crhs457 + crhs355*crhs471 + crhs499 + crhs501 + crhs502 - crhs504*crhs90 - crhs505*crhs90) - crhs359*(DN(2,2)*crhs361 - N[2]*crhs363 - crhs164*crhs516 + crhs223*crhs495 + crhs225*crhs471 + crhs225*crhs473 + crhs225*crhs485 + crhs225*crhs488 + crhs312*crhs465 + crhs360*crhs472 + crhs510 + crhs512 + crhs513 - crhs514*crhs90 - crhs515*crhs90) - crhs364*crhs459 - crhs365*(crhs471 + crhs472 + crhs473 + crhs482 - crhs483 + crhs485 + crhs488 - crhs489 - crhs490);
            rhs[15]=-DN(3,0)*crhs100 - DN(3,1)*crhs151 - DN(3,2)*crhs186 + N[3]*crhs13 + N[3]*crhs14;
            rhs[16]=-DN(3,0)*crhs291 + N[3]*crhs188 - crhs191*crhs520 - crhs193*crhs521 - crhs195*crhs522 + crhs197*crhs523 + crhs206*(-crhs524 - crhs525 + crhs526*crhs90 + crhs528 + crhs530 - crhs531*crhs90) + crhs214*(-crhs532 - crhs533 + crhs534*crhs90 + crhs536 + crhs537 - crhs538*crhs90) - crhs215*(-DN(3,0)*crhs232 + crhs143*crhs541 + crhs143*crhs542 - crhs143*crhs544 + crhs234*crhs551 + crhs524*crhs81 + crhs525*crhs81 + crhs532*crhs85 + crhs533*crhs85 + crhs540 - crhs545 - crhs546 - crhs548 - crhs550) + crhs248*(crhs220*crhs558 - crhs220*crhs560 + crhs544 + crhs554 + crhs556 + crhs557 + crhs559) - crhs31*crhs519;
            rhs[17]=-DN(3,1)*crhs291 + N[3]*crhs292 - crhs115*crhs519 - crhs191*crhs522 + crhs206*(crhs220*crhs555 - crhs220*crhs559 + crhs554 + crhs560 + crhs570 + crhs576 + crhs577) + crhs214*(-crhs562 - crhs563 + crhs564*crhs90 + crhs566 + crhs567 - crhs568*crhs90) - crhs215*(-DN(3,1)*crhs318 + crhs143*crhs527 + crhs319*crhs551 + crhs542*crhs81 + crhs543*crhs81 + crhs558*crhs81 + crhs562*crhs85 + crhs563*crhs85 + crhs569 - crhs570*crhs81 - crhs571 - crhs572 - crhs574 - crhs575) - crhs248*(-crhs524*crhs90 + crhs526 + crhs527 + crhs530*crhs90 - crhs531 - crhs561) - crhs293*crhs521 + crhs299*crhs520 + crhs300*crhs523;
            rhs[18]=-DN(3,2)*crhs291 + N[3]*crhs322 - crhs166*crhs519 - crhs193*crhs522 - crhs206*(-crhs562*crhs90 + crhs564 + crhs565 + crhs567*crhs90 - crhs568 - crhs579) + crhs214*(crhs220*crhs552 - crhs220*crhs553 + crhs556 + crhs557 + crhs559 + crhs560 + crhs576 + crhs577 + crhs581) - crhs215*(-DN(3,2)*crhs333 + crhs143*crhs535 + crhs334*crhs551 + crhs541*crhs85 + crhs543*crhs85 + crhs555*crhs85 + crhs558*crhs85 + crhs565*crhs81 + crhs580 - crhs581*crhs85 - crhs582 - crhs583 - crhs584 - crhs585) - crhs248*(-crhs532*crhs90 + crhs534 + crhs535 + crhs537*crhs90 - crhs538 - crhs578) - crhs293*crhs520 + crhs323*crhs521 + crhs324*crhs523;
            rhs[19]=-DN(3,0)*crhs341 - DN(3,1)*crhs345 - DN(3,2)*crhs346 + N[3]*crhs335 - N[3]*crhs336 + crhs215*(crhs253*crhs539 - crhs258*crhs587 - crhs259*crhs587 - crhs260*crhs587 + crhs367*crhs528 + crhs367*crhs561 + crhs368*crhs536 + crhs368*crhs578 + crhs369*crhs566 + crhs369*crhs579 + crhs370*crhs552 + crhs370*crhs555 + crhs370*crhs558 + crhs371*crhs543 + crhs372*crhs541 + crhs373*crhs542 + crhs374*crhs588 + crhs377*crhs588 + crhs378*crhs588) - crhs347*(DN(3,0)*crhs350 - N[3]*crhs352 + crhs223*crhs524 + crhs223*crhs525 + crhs225*crhs532 + crhs225*crhs533 - crhs29*crhs586 + crhs312*crhs541 + crhs312*crhs542 + crhs349*crhs543 + crhs540 + crhs545 + crhs546 - crhs548*crhs90 - crhs550*crhs90) - crhs353*(DN(3,1)*crhs356 - N[3]*crhs358 - crhs113*crhs586 + crhs223*crhs542 + crhs223*crhs543 + crhs223*crhs558 + crhs225*crhs562 + crhs225*crhs563 + crhs312*crhs527 + crhs355*crhs541 + crhs569 + crhs571 + crhs572 - crhs574*crhs90 - crhs575*crhs90) - crhs359*(DN(3,2)*crhs361 - N[3]*crhs363 - crhs164*crhs586 + crhs223*crhs565 + crhs225*crhs541 + crhs225*crhs543 + crhs225*crhs555 + crhs225*crhs558 + crhs312*crhs535 + crhs360*crhs542 + crhs580 + crhs582 + crhs583 - crhs584*crhs90 - crhs585*crhs90) - crhs364*crhs529 - crhs365*(crhs541 + crhs542 + crhs543 + crhs552 - crhs553 + crhs555 + crhs558 - crhs559 - crhs560);

}

template<>
double CompressibleNavierStokes<3>::ShockCapturingViscosity(const ElementDataStruct& data)
{
    const int nnodes = 4;
    const int dim = 3;
    const int BlockSize = dim+2;

   const double h = data.h;                                // Characteristic element size
   const double alpha = 0.8;                               // Algorithm constant
   const double tol = 0.001;                               

    const double& bdf0 = data.bdf0;
    const double& bdf1 = data.bdf1;
    const double& bdf2 = data.bdf2;

    const BoundedMatrix<double,nnodes,BlockSize>& U = data.U;
    const BoundedMatrix<double,nnodes,BlockSize>& Un = data.Un;
    const BoundedMatrix<double,nnodes,BlockSize>& Unn = data.Unn;
    const BoundedMatrix<double,nnodes,dim>& f_ext = data.f_ext;
    const double gamma = data.gamma;
    double v_sc = 0.0;                                      //Shock capturing viscosity
    BoundedMatrix<double,dim,1> res_m; 
    res_m(0,0)= 0; res_m(1,0)= 0; res_m(2,0)= 0;

    // Get shape function values
    const array_1d<double,nnodes>& N = data.N;
    const BoundedMatrix<double,nnodes,dim>& DN = data.DN_DX;

    // Auxiliary variables used in the calculation of the RHS
    const array_1d<double,BlockSize> U_gauss = prod(trans(U), N);
    const array_1d<double,dim> f_gauss = prod(trans(f_ext), N);
    const BoundedMatrix<double,BlockSize,dim> grad_U = prod(trans(U), DN);     // Dfi/Dxj
    const array_1d<double,BlockSize> accel_gauss = bdf0*U_gauss+bdf1*prod(trans(Un), N)+bdf2*prod(trans(Unn), N);
    
    const double cres_m0 =             gamma - 1;
const double cres_m1 =             N[0]*U(0,0) + N[1]*U(1,0) + N[2]*U(2,0) + N[3]*U(3,0);
const double cres_m2 =             DN(0,1)*U(0,1) + DN(1,1)*U(1,1) + DN(2,1)*U(2,1) + DN(3,1)*U(3,1);
const double cres_m3 =             N[0]*U(0,2) + N[1]*U(1,2) + N[2]*U(2,2) + N[3]*U(3,2);
const double cres_m4 =             1.0/cres_m1;
const double cres_m5 =             cres_m3*cres_m4;
const double cres_m6 =             DN(0,1)*U(0,2) + DN(1,1)*U(1,2) + DN(2,1)*U(2,2) + DN(3,1)*U(3,2);
const double cres_m7 =             N[0]*U(0,1) + N[1]*U(1,1) + N[2]*U(2,1) + N[3]*U(3,1);
const double cres_m8 =             cres_m4*cres_m7;
const double cres_m9 =             DN(0,2)*U(0,1) + DN(1,2)*U(1,1) + DN(2,2)*U(2,1) + DN(3,2)*U(3,1);
const double cres_m10 =             N[0]*U(0,3) + N[1]*U(1,3) + N[2]*U(2,3) + N[3]*U(3,3);
const double cres_m11 =             cres_m10*cres_m4;
const double cres_m12 =             DN(0,2)*U(0,3) + DN(1,2)*U(1,3) + DN(2,2)*U(2,3) + DN(3,2)*U(3,3);
const double cres_m13 =             gamma - 3;
const double cres_m14 =             cres_m4*(DN(0,0)*U(0,1) + DN(1,0)*U(1,1) + DN(2,0)*U(2,1) + DN(3,0)*U(3,1));
const double cres_m15 =             DN(0,0)*U(0,2) + DN(1,0)*U(1,2) + DN(2,0)*U(2,2) + DN(3,0)*U(3,2);
const double cres_m16 =             cres_m0*cres_m3*cres_m4;
const double cres_m17 =             DN(0,0)*U(0,3) + DN(1,0)*U(1,3) + DN(2,0)*U(2,3) + DN(3,0)*U(3,3);
const double cres_m18 =             cres_m0*cres_m10*cres_m4;
const double cres_m19 =             DN(0,1)*U(0,0) + DN(1,1)*U(1,0) + DN(2,1)*U(2,0) + DN(3,1)*U(3,0);
const double cres_m20 =             pow(cres_m1, -2);
const double cres_m21 =             cres_m19*cres_m20*cres_m3;
const double cres_m22 =             DN(0,2)*U(0,0) + DN(1,2)*U(1,0) + DN(2,2)*U(2,0) + DN(3,2)*U(3,0);
const double cres_m23 =             cres_m10*cres_m20*cres_m22;
const double cres_m24 =             (1.0L/2.0L)*cres_m20;
const double cres_m25 =             DN(0,0)*U(0,0) + DN(1,0)*U(1,0) + DN(2,0)*U(2,0) + DN(3,0)*U(3,0);
const double cres_m26 =             pow(cres_m7, 2);
const double cres_m27 =             pow(cres_m3, 2);
const double cres_m28 =             pow(cres_m10, 2);
const double cres_m29 =             cres_m0*cres_m26 + cres_m0*cres_m27 + cres_m0*cres_m28;
const double cres_m30 =             DN(0,2)*U(0,2) + DN(1,2)*U(1,2) + DN(2,2)*U(2,2) + DN(3,2)*U(3,2);
const double cres_m31 =             cres_m0*cres_m4*cres_m7;
const double cres_m32 =             DN(0,1)*U(0,3) + DN(1,1)*U(1,3) + DN(2,1)*U(2,3) + DN(3,1)*U(3,3);
const double cres_m33 =             cres_m20*cres_m25*cres_m7;
            res_m(0,0)=-N[0]*(U(0,1)*bdf0 + Un(0,1)*bdf1 + Unn(0,1)*bdf2) - N[1]*(U(1,1)*bdf0 + Un(1,1)*bdf1 + Unn(1,1)*bdf2) - N[2]*(U(2,1)*bdf0 + Un(2,1)*bdf1 + Unn(2,1)*bdf2) - N[3]*(U(3,1)*bdf0 + Un(3,1)*bdf1 + Unn(3,1)*bdf2) - cres_m0*(DN(0,0)*U(0,4) + DN(1,0)*U(1,4) + DN(2,0)*U(2,4) + DN(3,0)*U(3,4)) + cres_m1*(N[0]*f_ext(0,0) + N[1]*f_ext(1,0) + N[2]*f_ext(2,0) + N[3]*f_ext(3,0)) - cres_m11*cres_m9 - cres_m12*cres_m8 + cres_m13*cres_m14*cres_m7 + cres_m15*cres_m16 + cres_m17*cres_m18 - cres_m2*cres_m5 + cres_m21*cres_m7 + cres_m23*cres_m7 - cres_m24*cres_m25*(-2*cres_m26 + cres_m29) - cres_m6*cres_m8;
            res_m(1,0)=-N[0]*(U(0,2)*bdf0 + Un(0,2)*bdf1 + Unn(0,2)*bdf2) - N[1]*(U(1,2)*bdf0 + Un(1,2)*bdf1 + Unn(1,2)*bdf2) - N[2]*(U(2,2)*bdf0 + Un(2,2)*bdf1 + Unn(2,2)*bdf2) - N[3]*(U(3,2)*bdf0 + Un(3,2)*bdf1 + Unn(3,2)*bdf2) - cres_m0*(DN(0,1)*U(0,4) + DN(1,1)*U(1,4) + DN(2,1)*U(2,4) + DN(3,1)*U(3,4)) + cres_m1*(N[0]*f_ext(0,1) + N[1]*f_ext(1,1) + N[2]*f_ext(2,1) + N[3]*f_ext(3,1)) - cres_m11*cres_m30 - cres_m12*cres_m5 + cres_m13*cres_m5*cres_m6 - cres_m14*cres_m3 - cres_m15*cres_m8 + cres_m18*cres_m32 - cres_m19*cres_m24*(-2*cres_m27 + cres_m29) + cres_m2*cres_m31 + cres_m23*cres_m3 + cres_m3*cres_m33;
            res_m(2,0)=-N[0]*(U(0,3)*bdf0 + Un(0,3)*bdf1 + Unn(0,3)*bdf2) - N[1]*(U(1,3)*bdf0 + Un(1,3)*bdf1 + Unn(1,3)*bdf2) - N[2]*(U(2,3)*bdf0 + Un(2,3)*bdf1 + Unn(2,3)*bdf2) - N[3]*(U(3,3)*bdf0 + Un(3,3)*bdf1 + Unn(3,3)*bdf2) - cres_m0*(DN(0,2)*U(0,4) + DN(1,2)*U(1,4) + DN(2,2)*U(2,4) + DN(3,2)*U(3,4)) + cres_m1*(N[0]*f_ext(0,2) + N[1]*f_ext(1,2) + N[2]*f_ext(2,2) + N[3]*f_ext(3,2)) - cres_m10*cres_m14 + cres_m10*cres_m21 + cres_m10*cres_m33 + cres_m11*cres_m12*cres_m13 - cres_m11*cres_m6 + cres_m16*cres_m30 - cres_m17*cres_m8 - cres_m22*cres_m24*(-2*cres_m28 + cres_m29) + cres_m31*cres_m9 - cres_m32*cres_m5;


    double norm_res_m;
    norm_res_m = sqrt(res_m(0,0)*res_m(0,0)+res_m(1,0)*res_m(1,0)+res_m(2,0)*res_m(2,0));

    double norm_gradm = 0.0;                                    // Frobenius norm of momentum gradient
    for (unsigned int i=1; i<dim+1; i++){
        for (unsigned int j=0; j<dim; j++)
            norm_gradm += grad_U(i,j)*grad_U(i,j);
    }
    norm_gradm = sqrt(norm_gradm);
    
    if (norm_gradm>tol)
        v_sc = 0.5*h*alpha*(norm_res_m/norm_gradm);
    
    return v_sc;
}

template<>
double CompressibleNavierStokes<3>::ShockCapturingConductivity(const ElementDataStruct& data)
{
    const int nnodes = 4;
    const int dim = 3;
    const int BlockSize = dim+2;

   const double h = data.h;                                // Characteristic element size
   const double alpha = 0.8;                               // Algorithm constant
   const double tol = 0.001;                               

    const double& bdf0 = data.bdf0;
    const double& bdf1 = data.bdf1;
    const double& bdf2 = data.bdf2;

    const BoundedMatrix<double,nnodes,BlockSize>& U = data.U;
    const BoundedMatrix<double,nnodes,BlockSize>& Un = data.Un;
    const BoundedMatrix<double,nnodes,BlockSize>& Unn = data.Unn;
    const BoundedMatrix<double,nnodes,dim>& f_ext = data.f_ext;
    const array_1d<double,nnodes>& r = data.r;
    const double gamma = data.gamma;
    double k_sc = 0.0;          // Shock Capturing Conductivity
    BoundedMatrix<double,dim,1> res_e;
    res_e(0,0) = 0;
    // Get shape function values
    const array_1d<double,nnodes>& N = data.N;
    const BoundedMatrix<double,nnodes,dim>& DN = data.DN_DX;

    // Auxiliary variables used in the calculation of the RHS
    const array_1d<double,BlockSize> U_gauss = prod(trans(U), N);
    const array_1d<double,dim> f_gauss = prod(trans(f_ext), N);
    const BoundedMatrix<double,BlockSize,dim> grad_U = prod(trans(U), DN);     // Dfi/Dxj
    const array_1d<double,BlockSize> accel_gauss = bdf0*U_gauss+bdf1*prod(trans(Un), N)+bdf2*prod(trans(Unn), N);
    
    const double cres_e0 =             N[0]*U(0,0) + N[1]*U(1,0) + N[2]*U(2,0) + N[3]*U(3,0);
const double cres_e1 =             N[0]*U(0,1) + N[1]*U(1,1) + N[2]*U(2,1) + N[3]*U(3,1);
const double cres_e2 =             N[0]*U(0,2) + N[1]*U(1,2) + N[2]*U(2,2) + N[3]*U(3,2);
const double cres_e3 =             N[0]*U(0,3) + N[1]*U(1,3) + N[2]*U(2,3) + N[3]*U(3,3);
const double cres_e4 =             1.0/cres_e0;
const double cres_e5 =             cres_e4*gamma;
const double cres_e6 =             gamma - 1;
const double cres_e7 =             pow(cres_e0, -2);
const double cres_e8 =             cres_e1*cres_e2*cres_e6*cres_e7;
const double cres_e9 =             cres_e1*cres_e3*cres_e6*cres_e7;
const double cres_e10 =             cres_e2*cres_e3*cres_e6*cres_e7;
const double cres_e11 =             (1.0L/2.0L)*cres_e4;
const double cres_e12 =             N[0]*U(0,4);
const double cres_e13 =             -2*cres_e12;
const double cres_e14 =             N[1]*U(1,4);
const double cres_e15 =             -2*cres_e14;
const double cres_e16 =             N[2]*U(2,4);
const double cres_e17 =             -2*cres_e16;
const double cres_e18 =             N[3]*U(3,4);
const double cres_e19 =             -2*cres_e18;
const double cres_e20 =             -2*cres_e6*(cres_e12 + cres_e14 + cres_e16 + cres_e18);
const double cres_e21 =             pow(cres_e2, 2);
const double cres_e22 =             cres_e4*cres_e6;
const double cres_e23 =             cres_e21*cres_e22;
const double cres_e24 =             pow(cres_e3, 2);
const double cres_e25 =             cres_e22*cres_e24;
const double cres_e26 =             pow(cres_e1, 2);
const double cres_e27 =             cres_e22*cres_e26;
const double cres_e28 =             cres_e13 + cres_e15 + cres_e17 + cres_e19 + cres_e20 + cres_e23 + cres_e27;
const double cres_e29 =             (1.0L/2.0L)*cres_e7*(cres_e22*(cres_e21 + cres_e24 + cres_e26) + cres_e25 + cres_e28);
            res_e(0,0)=-N[0]*(U(0,4)*bdf0 + Un(0,4)*bdf1 + Unn(0,4)*bdf2) - N[1]*(U(1,4)*bdf0 + Un(1,4)*bdf1 + Unn(1,4)*bdf2) - N[2]*(U(2,4)*bdf0 + Un(2,4)*bdf1 + Unn(2,4)*bdf2) - N[3]*(U(3,4)*bdf0 + Un(3,4)*bdf1 + Unn(3,4)*bdf2) + cres_e0*(N[0]*r[0] + N[1]*r[1] + N[2]*r[2] + N[3]*r[3]) - cres_e1*cres_e29*(DN(0,0)*U(0,0) + DN(1,0)*U(1,0) + DN(2,0)*U(2,0) + DN(3,0)*U(3,0)) - cres_e1*cres_e5*(DN(0,0)*U(0,4) + DN(1,0)*U(1,4) + DN(2,0)*U(2,4) + DN(3,0)*U(3,4)) + cres_e1*(N[0]*f_ext(0,0) + N[1]*f_ext(1,0) + N[2]*f_ext(2,0) + N[3]*f_ext(3,0)) + cres_e10*(DN(0,1)*U(0,3) + DN(1,1)*U(1,3) + DN(2,1)*U(2,3) + DN(3,1)*U(3,3)) + cres_e10*(DN(0,2)*U(0,2) + DN(1,2)*U(1,2) + DN(2,2)*U(2,2) + DN(3,2)*U(3,2)) + cres_e11*(3*cres_e25 + cres_e28)*(DN(0,2)*U(0,3) + DN(1,2)*U(1,3) + DN(2,2)*U(2,3) + DN(3,2)*U(3,3)) + cres_e11*(DN(0,0)*U(0,1) + DN(1,0)*U(1,1) + DN(2,0)*U(2,1) + DN(3,0)*U(3,1))*(cres_e13 + cres_e15 + cres_e17 + cres_e19 + cres_e20 + cres_e23 + cres_e25 + 3*cres_e27) + cres_e11*(DN(0,1)*U(0,2) + DN(1,1)*U(1,2) + DN(2,1)*U(2,2) + DN(3,1)*U(3,2))*(cres_e13 + cres_e15 + cres_e17 + cres_e19 + cres_e20 + 3*cres_e23 + cres_e25 + cres_e27) - cres_e2*cres_e29*(DN(0,1)*U(0,0) + DN(1,1)*U(1,0) + DN(2,1)*U(2,0) + DN(3,1)*U(3,0)) - cres_e2*cres_e5*(DN(0,1)*U(0,4) + DN(1,1)*U(1,4) + DN(2,1)*U(2,4) + DN(3,1)*U(3,4)) + cres_e2*(N[0]*f_ext(0,1) + N[1]*f_ext(1,1) + N[2]*f_ext(2,1) + N[3]*f_ext(3,1)) - cres_e29*cres_e3*(DN(0,2)*U(0,0) + DN(1,2)*U(1,0) + DN(2,2)*U(2,0) + DN(3,2)*U(3,0)) - cres_e3*cres_e5*(DN(0,2)*U(0,4) + DN(1,2)*U(1,4) + DN(2,2)*U(2,4) + DN(3,2)*U(3,4)) + cres_e3*(N[0]*f_ext(0,2) + N[1]*f_ext(1,2) + N[2]*f_ext(2,2) + N[3]*f_ext(3,2)) + cres_e8*(DN(0,0)*U(0,2) + DN(1,0)*U(1,2) + DN(2,0)*U(2,2) + DN(3,0)*U(3,2)) + cres_e8*(DN(0,1)*U(0,1) + DN(1,1)*U(1,1) + DN(2,1)*U(2,1) + DN(3,1)*U(3,1)) + cres_e9*(DN(0,0)*U(0,3) + DN(1,0)*U(1,3) + DN(2,0)*U(2,3) + DN(3,0)*U(3,3)) + cres_e9*(DN(0,2)*U(0,1) + DN(1,2)*U(1,1) + DN(2,2)*U(2,1) + DN(3,2)*U(3,1));


    double norm_res_e;
    norm_res_e = sqrt(res_e(0,0)*res_e(0,0));

    double norm_grade = 0.0;              // Frobenius norm of total energy gradient
    for (unsigned int i=0; i<dim; i++)
        norm_grade += grad_U(dim+1,i)*grad_U(dim+1,i);
    norm_grade = sqrt(norm_grade);
    
 
    if (norm_grade>tol)
        k_sc = 0.5*h*alpha*(norm_res_e/norm_grade);

    return k_sc;
}

}
